/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 6/4/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __hep_binary_tree_h__
#define __hep_binary_tree_h__ 1

//   std includes
#include <iterator>
#include <cstddef>

//   project headers


namespace duct {
  
  template<typename _Tp>
  class binary_tree
  {
    /*   Structure for a node with the data and the
     *   pointers to link the nodes.
     */
    struct node {
      //----- creating a node with valid parent -----
      node(node *u, const _Tp& d) : up(u), data(d) {}
      node(node *u,  _Tp&& d) : up(u), data(std::move(d)) {}
      
      //----- assignments -----
      node& operator=(const _Tp& d) { data = d; return *this;}
      node& operator=(_Tp&& d) { data = std::move(d); return *this;}
      
      node *up = nullptr, *left = nullptr, *right = nullptr;
      _Tp data;
    };
    
  public:
    /*    Standatd typedef   */
    typedef _Tp value_type;
    typedef value_type* pointer;
    typedef value_type& reference;
    typedef const value_type* const_pointer;
    typedef const value_type& const_reference;
    typedef std::size_t size_type;
    
    /*  The tree is not linear object thus there are several way
     *  iterate through its elements. We define the three standard
     *  traversal operators, pre-order, in-order and post-order.
     */
    class iterator_base
    {
    public:
      /*   some types to be compatible with the standard   */
      typedef binary_tree::value_type value_type;
      typedef value_type* pointer;
      typedef value_type& reference;
      typedef std::ptrdiff_t difference_type;
      typedef std::forward_iterator_tag iterator_category;
      
      /*   Constructors   */
      iterator_base() = default;
      iterator_base(node *t, node *c) : _M_tree(t), _M_curr(c) {}
     
      /*    Comparison operators   */
      bool operator==(const iterator_base& i) const { return _M_curr == i._M_curr && _M_tree == i._M_tree;}
      bool operator!=(const iterator_base& i) const { return _M_curr != i._M_curr || _M_tree != i._M_tree;}
      
      /*   Dereference operators   */
      reference operator*() const { return _M_curr->data;}
      pointer operator->() const { return &(_M_curr->data);}
      
      /*   Binary tree specific functions   */
      iterator_base up() const { return {_M_tree, _M_curr->up};}
      iterator_base left() const { return {_M_tree, _M_curr->left};}
      iterator_base right() const { return {_M_tree, _M_curr->right};}
      
      /*   Information about the current node   */
      bool is_root() const { return _M_curr == _M_tree;}
      bool is_leaf() const { return _M_curr->left == nullptr && _M_curr->right == nullptr;}
      
    protected:
      node *_M_tree, *_M_curr;
      
      /*   Friend declaration  */
      friend class binary_tree;
      friend class const_iterator_base;
    };
    
    class const_iterator_base
    {
    public:
      /*   some types to be compatible with the standard   */
      typedef binary_tree::value_type value_type;
      typedef const value_type* pointer;
      typedef const value_type& reference;
      typedef std::ptrdiff_t difference_type;
      typedef std::forward_iterator_tag iterator_category;
      
      /*   Constructors   */
      const_iterator_base() = default;
      const_iterator_base(const node *t, const node *c) : _M_tree(t), _M_curr(c) {}
      const_iterator_base(const iterator_base& c) : _M_tree(c._M_tree), _M_curr(c._M_curr) {}
      
      /*    Comparison operators   */
      bool operator==(const const_iterator_base& i) const { return _M_curr == i._M_curr && _M_tree == i._M_tree;}
      bool operator!=(const const_iterator_base& i) const { return _M_curr != i._M_curr || _M_tree != i._M_tree;}
    
      /*   Dereference operators   */
      reference operator*() const { return _M_curr->data;}
      pointer operator->() const { return &(_M_curr->data);}
      
      /*   Binary tree specific functions   */
      const_iterator_base up() const { return {_M_tree, _M_curr->up};}
      const_iterator_base left() const { return {_M_tree, _M_curr->left};}
      const_iterator_base right() const { return {_M_tree, _M_curr->right};}
      
      /*   Information about the current node   */
      bool is_root() const { return _M_curr == _M_tree;}
      bool is_leaf() const { return _M_curr->left == nullptr && _M_curr->right == nullptr;}
      
    protected:
      const node *_M_tree, *_M_curr;
      
      /*   Friend declaration  */
      friend class binary_tree;
    };
    
    /*   In-order iterators   */
    class inorder_iterator : public iterator_base
    {
    public:
      /*   Constructors   */
      inorder_iterator() = default;
      inorder_iterator(node *t, node *c) : iterator_base(t, c) {}
      inorder_iterator(const iterator_base& c) : iterator_base(c) {}
      explicit inorder_iterator(const const_iterator_base& c)
      : iterator_base(const_cast<node*>(c._M_tree), const_cast<node*>(c._M_curr)) {}
      
      /*   Increment operators  */
      inorder_iterator& operator++() {
        node *c = this->_M_curr;
        
        if(c->right) {
          c = c->right;
          while(c->left) c = c->left;
        }
        else {
          while(c->up && c->up->right == c) c = c->up;
          c = c->up;
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      inorder_iterator operator++(int) {
        inorder_iterator tmp = *this;
        operator++();
        return tmp;
      }
    };
    
    class const_inorder_iterator : public const_iterator_base
    {
    public:
      /*   Constructors   */
      const_inorder_iterator() = default;
      const_inorder_iterator(const node *t, const node *c) : const_iterator_base(t, c) {}
      const_inorder_iterator(const const_iterator_base& c) : const_iterator_base(c) {}
      const_inorder_iterator(const iterator_base& c) : const_iterator_base(c) {}
      
      /*   Increment operators  */
      const_inorder_iterator& operator++() {
        const node *c = this->_M_curr;
        
        if(c->right) {
          c = c->right;
          while(c->left) c = c->left;
        }
        else {
          while(c->up && c->up->right == c) c = c->up;
          c = c->up;
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      const_inorder_iterator operator++(int) {
        const_inorder_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    /*   Pre-order iterators   */
    class preorder_iterator : public iterator_base
    {
    public:
      /*   Constructors   */
      preorder_iterator() = default;
      preorder_iterator(node *t, node *c) : iterator_base(t, c) {}
      preorder_iterator(const iterator_base& c) : iterator_base(c) {}
      explicit preorder_iterator(const const_iterator_base& c)
      : iterator_base(const_cast<node*>(c._M_tree), const_cast<node*>(c._M_curr)) {}
      
      /*   Increment operators  */
      preorder_iterator& operator++() {
        node *c = this->_M_curr;
        
        if(c->left) c = c->left;
        else if(c->right) c = c->right;
        else {
          while(c->up && (c->up->right == c || c->up->right == nullptr)) c = c->up;
          c = c->up ? c->up->right : nullptr;
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      preorder_iterator operator++(int) {
        preorder_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    class const_preorder_iterator : public const_iterator_base
    {
    public:
      /*   Constructors   */
      const_preorder_iterator() = default;
      const_preorder_iterator(const node *t, const node *c) : const_iterator_base(t, c) {}
      const_preorder_iterator(const const_iterator_base& c) : const_iterator_base(c) {}
      const_preorder_iterator(const iterator_base& c) : const_iterator_base(c) {}
      
      /*   Increment operators  */
      const_preorder_iterator& operator++() {
        const node *c = this->_M_curr;
        
        if(c->left) c = c->left;
        else if(c->right) c = c->right;
        else {
          while(c->up && (c->up->right == c || c->up->right == nullptr)) c = c->up;
          c = c->up ? c->up->right : nullptr;
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      const_preorder_iterator operator++(int) {
        const_preorder_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    /*   Pre-order iterators   */
    class postorder_iterator : public iterator_base
    {
    public:
      /*   Constructors   */
      postorder_iterator() = default;
      postorder_iterator(node *t, node *c) : iterator_base(t, c) {}
      postorder_iterator(const iterator_base& c) : iterator_base(c) {}
      explicit postorder_iterator(const const_iterator_base& c)
      : iterator_base(const_cast<node*>(c._M_tree), const_cast<node*>(c._M_curr)) {}
    
      /*   Increment operators  */
      postorder_iterator& operator++() {
        node *c = this->_M_curr;
        
        if(c->up == nullptr || c->up->right == c) c = c->up;
        else {
          c = c->up;
          if(c->right) {
            c = c->right;
            while(c->left || c->right)
              c = c->left ? c->left : c->right;
          }
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      postorder_iterator operator++(int) {
        postorder_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    class const_postorder_iterator : public const_iterator_base
    {
    public:
      /*   Constructors   */
      const_postorder_iterator() = default;
      const_postorder_iterator(const node *t, const node *c) : const_iterator_base(t, c) {}
      const_postorder_iterator(const const_iterator_base& c) : const_iterator_base(c) {}
      const_postorder_iterator(const iterator_base& c) : const_iterator_base(c) {}
      
      /*   Increment operators  */
      const_postorder_iterator& operator++() {
        const node *c = this->_M_curr;
        
        if(c->up == nullptr || c->up->right == c) c = c->up;
        else {
          c = c->up;
          if(c->right) {
            c = c->right;
            while(c->left || c->right)
              c = c->left ? c->left : c->right;
          }
        }
        
        this->_M_curr = c;
        return *this;
      }
      
      const_postorder_iterator operator++(int) {
        const_postorder_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    /*   In-order iterators   */
    class leaf_iterator : public iterator_base
    {
    public:
      /*   Constructors   */
      leaf_iterator() = default;
      leaf_iterator(node *t, node *c) : iterator_base(t, c) {}
      leaf_iterator(const iterator_base& c) : iterator_base(c) {}
      explicit leaf_iterator(const const_iterator_base& c)
      : iterator_base(const_cast<node*>(c._M_tree), const_cast<node*>(c._M_curr)) {}
    
      /*   Increment operators  */
      leaf_iterator& operator++() {
        node *c = this->_M_curr;
        
        do {
          if(c->right) {
            c = c->right;
            while(c->left) c = c->left;
          }
          else {
            while(c->up && c->up->right == c) c = c->up;
            c = c->up;
          }
        } while(c && (c->left || c->right));
        
        this->_M_curr = c;
        return *this;
      }
      
      leaf_iterator operator++(int) {
        leaf_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    class const_leaf_iterator : public const_iterator_base
    {
    public:
      /*   Constructors   */
      const_leaf_iterator() = default;
      const_leaf_iterator(const node *t, const node *c) : const_iterator_base(t, c) {}
      const_leaf_iterator(const const_iterator_base& c) : const_iterator_base(c) {}
      const_leaf_iterator(const iterator_base& c) : const_iterator_base(c) {}
      
      /*   Increment operators  */
      const_leaf_iterator& operator++() {
        const node *c = this->_M_curr;
        
        do {
          if(c->right) {
            c = c->right;
            while(c->left) c = c->left;
          }
          else {
            while(c->up && c->up->right == c) c = c->up;
            c = c->up;
          }
        } while(c && (c->left || c->right));
        
        this->_M_curr = c;
        return *this;
      }
      
      const_leaf_iterator operator++(int) {
        const_leaf_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
    };
    
    
    /*   The standard iterator is the in-order iterator  */
    typedef inorder_iterator iterator;
    typedef const_inorder_iterator const_iterator;
    
    
    /*   default constructors and destructor   */
    binary_tree() : _M_head(nullptr) {}
    ~binary_tree() { destroy(_M_head);}
    
    /*   initialize the head  */
    binary_tree(const value_type& d) : _M_head(allocate(nullptr,d)) {}
    binary_tree(value_type&& d) : _M_head(allocate(nullptr,std::move(d))) {}
    
    /*   copy and move constructors   */
    binary_tree(const binary_tree& x) : _M_head(copy(x._M_head)) {}
    binary_tree(binary_tree&& x) : _M_head(nullptr) { std::swap(_M_head, x._M_head);}
    
    template<typename _Up>
    explicit binary_tree(const binary_tree<_Up>& x)
    : _M_head(copy<_Up>(x._M_head)) {}

    template<typename _Up, class _Conv>
    binary_tree(const binary_tree<_Up>& x, _Conv conv)
    : _M_head(copy<_Up, _Conv>(x._M_head, conv)) {}

    /*   copy and move  assignments   */
    binary_tree& operator=(const binary_tree& x) {
      if(&x != this) {
        this->clear();
        _M_head = copy(x._M_head);
      }
      
      return *this;
    }
    
    template<typename _Up>
    binary_tree& operator=(const binary_tree<_Up>& x) {
      if(&x != this) {
        this->clear();
        _M_head = copy<_Up>(x._M_head);
      }
      
      return *this;
    }
    
    binary_tree& operator=(binary_tree&& x) {
      std::swap(_M_head, x._M_head);
      return *this;
    }
  
    /*    assignment with conversion   */
    template<typename _Up, class _Conv>
    binary_tree& assign(const binary_tree<_Up>& x, _Conv conv) {
      if(&x != this) {
        this->clear();
        _M_head = copy<_Up, _Conv>(x._M_head, conv);
      }
      
      return *this;
    }
    
    /*   points to the root node  */
    inorder_iterator root() { return {_M_head, _M_head};}
    const_inorder_iterator root() const { return {_M_head, _M_head};}
    const_inorder_iterator croot() const { return {_M_head, _M_head};}
    
    
    /*   Inorder begin() and end()  */
    inorder_iterator inorder_begin() {
      node *c = _M_head;
      if(c) while(c->left) c = c->left;
      return {_M_head, c};
    }
    
    const_inorder_iterator inorder_begin() const {
      const node *c = _M_head;
      if(c) while(c->left) c = c->left;
      return {_M_head, c};
    }
    
    inorder_iterator inorder_end() { return {_M_head, nullptr};}
    const_inorder_iterator inorder_end() const { return {_M_head, nullptr};}
    
    const_inorder_iterator inorder_cbegin() const { return inorder_begin();}
    const_inorder_iterator inorder_cend() const { return inorder_end();}
    
    
    /* Pre-order begin() and end()   */
    preorder_iterator preorder_begin() { return {_M_head, _M_head};}
    preorder_iterator preorder_end() { return {_M_head, nullptr};}
    
    const_preorder_iterator preorder_begin() const { return {_M_head, _M_head};}
    const_preorder_iterator preorder_end() const { return {_M_head, nullptr};}
    
    const_preorder_iterator preorder_cbegin() const { return preorder_begin();}
    const_preorder_iterator preorder_cend() const { return preorder_end();}
    
    
    /*   Post-order begin() and end()   */
    postorder_iterator postorder_begin() {
      node *c = _M_head;
      if(c) while(c->left || c->right) c = (c->left ? c->left : c->right);
      return {_M_head, c};
    }
    
    const_postorder_iterator postorder_begin() const {
      const node *c = _M_head;
      if(c) while(c->left || c->right) c = (c->left ? c->left : c->right);
      return {_M_head, c};
    }
    
    postorder_iterator postorder_end() { return {_M_head, nullptr};}
    const_postorder_iterator postorder_end() const { return {_M_head, nullptr};}
    
    const_postorder_iterator postorder_cbegin() const { return postorder_begin();}
    const_postorder_iterator postorder_cend() const { return postorder_end();}
    
    
    /*   Leaf begin() and end()  */
    leaf_iterator leaf_begin() {
      leaf_iterator c = begin();
      if(c._M_curr == nullptr || c.is_leaf()) return c;
      return ++c;
    }
    
    const_leaf_iterator leaf_begin() const {
      const_leaf_iterator c = begin();
      if(c._M_curr == nullptr || c.is_leaf()) return c;
      return ++c;
    }
    
    leaf_iterator leaf_end() { return {_M_head, nullptr};}
    const_leaf_iterator leaf_end() const { return {_M_head, nullptr};}
    
    const_leaf_iterator leaf_cbegin() const { return leaf_begin();}
    const_leaf_iterator leaf_cend() const { return leaf_end();}
    
    
    /*   Standard begin() and end() operators   */
    iterator begin() { return inorder_begin();}
    iterator end() { return inorder_end();}
    
    const_iterator begin() const { return inorder_begin();}
    const_iterator end() const { return inorder_end();}
    
    const_iterator cbegin() const { return inorder_begin();}
    const_iterator cend() const { return inorder_end();}
    
    
    /*   append a node to the right if it is a free slot  */
    template<class iter>
    iter append_right(iter pos, const value_type& d) {
      return {_M_head, append_right(pos._M_curr, d)};
    }
    
    template<class iter>
    iter append_right(iter pos, value_type&& d) {
      return {_M_head, append_right(pos._M_curr, std::move(d))};
    }
    
    /*   append a node to the left if it is a free slot  */
    template<class iter>
    iter append_left(iter pos, const value_type& d) {
      return {_M_head, append_left(pos._M_curr, d)};
    }
    
    template<class iter>
    iter append_left(iter pos, value_type&& d) {
      return {_M_head, append_left(pos._M_curr, std::move(d))};
    }
    
    
    /*   append a subtree to the right if it is a free splot  */
    template<class iter>
    iter append_right(iter pos, const binary_tree& d) {
      node *n = pos._M_curr;
      
      if(n->right == nullptr) {
        n->right = copy(d._M_head);
        n->right->up = n;
        return {_M_head, n->right};
      }
      
      return pos;
    }
    
    template<class iter>
    iter append_right(iter pos, binary_tree&& d) {
      node *n = pos._M_curr;
      
      if(n->right == nullptr) {
        std::swap(n->right, d._M_head);
        n->right->up = n;
        return {_M_head, n->right};
      }
      
      return pos;
    }
    
    
    /*   append a subtree to the left if it is a free splot  */
    template<class iter>
    iter append_left(iter pos, const binary_tree& d) {
      node *n = pos._M_curr;
      
      if(n->left == nullptr) {
        n->left = copy(d._M_head);
        n->left->up = n;
        return {_M_head, n->left};
      }
      
      return pos;
    }
    
    template<class iter>
    iter append_left(iter pos, binary_tree&& d) {
      node *n = pos._M_curr;
      
      if(n->left == nullptr) {
        std::swap(n->left, d._M_head);
        n->left->up = n;
        return {_M_head, n->left};
      }
      
      return pos;
    }
    
    
    /*  split the binary tree at a given position   */
    binary_tree split(iterator pos) {
      binary_tree retval;
      node *n = pos._M_curr;
      node *u = n->up;
      
      if(u) {
        if(u->left == n) u->left = nullptr;
        else u->right = nullptr;
        n->up = nullptr;
        retval._M_head = n;
      } else std::swap(retval._M_head, _M_head);
      
      return retval;
    }
    
    /*   erase a subtree (every node starts a subtree) */
    template<class iter>
    iter erase(iter pos) {
      node *n = pos._M_curr;
      node *u = n->up;
      
      if(u) {
        if(u->left == n) u->left = nullptr;
        else u->right = nullptr;
        n->up = nullptr;
      }
      
      destroy(n);
      
      return {_M_head, u};
    }
    
    /*   delete all the nodes   */
    void clear() { _M_head = destroy(_M_head);}
    
    /*   swapping trees  */
    void swap(binary_tree& a) {
      std::swap(a._M_head, _M_head);
    }
    
    /*   structural information  */
    size_type size() const { return std::distance(begin(), end()); }
    bool empty() const { return _M_head == nullptr;}
    
    /*   I/O manipulations   */
    template<class _OStream, class _WriteFunc> _OStream& write(_OStream&, _WriteFunc) const;
    template<class _IStream, class _ReadFunc> _IStream& read(_IStream&, _ReadFunc);
    
  private:
    /*   Some helper function allocate and deallocate a node.
     *   Here we use the system new/delete operators but
     *   it should be done with allocator class.
     */
    static node * allocate(node *u, const _Tp& d) {
      node *ret = new node(u,d);
      if(!ret) throw "__binary_tree::allocate : Unable to allocate memory";
      return ret;
    }
    
    static node * allocate(node *u, _Tp&& d) {
      node *ret = new node(u,std::move(d));
      if(!ret) throw "__binary_tree::allocate : Unable to allocate memory";
      return ret;
    }
    
    static void deallocate(node *n) {
      if(n) delete n;
    }
    
    /*   copy the whole tree  */
    static node * copy(const node *);
    
    template<typename _Up>
    static node * copy(const typename binary_tree<_Up>::node *);

    /*   copy with conversion  */
    template<typename _Up, class _Conv>
    static node * copy(const typename binary_tree<_Up>::node *, _Conv);

    /*   copy the whole tree  */
    static node * destroy(node *);
    
    /*   append a node to the right if it is a free slot  */
    static node * append_right(node *n, const value_type& d) {
      if(n->right == nullptr) return n->right = allocate(n, d);
      return n;
    }
    
    static node * append_right(node *n, value_type&& d) {
      if(n->right == nullptr) return n->right = allocate(n, std::move(d));
      return n;
    }
    
    /*   append a node to the left if it is a free slot  */
    static node * append_left(node *n, const value_type& d) {
      if(n->left == nullptr) return n->left = allocate(n, d);
      return n;
    }
    
    static node * append_left(node *n, value_type&& d) {
      if(n->left == nullptr) return n->left = allocate(n, std::move(d));
      return n;
    }
    
    /*   points to the head of the tree   */
    node *_M_head;
  };
  

  template<typename _Tp> 
  typename binary_tree<_Tp>::node * binary_tree<_Tp>::copy(const node *xc)
  {
    //---- if x is empty, do nothing ----
    if(xc == nullptr) return nullptr;
    
    //---- copy the nodes -----
    node *head, *u = nullptr;
    node **c = &head;
    
    while(xc) {
      u = *c = allocate(u, xc->data);
      
      if(xc->left) {
        xc = xc->left;
        c = &(u->left);
      }
      else if(xc->right) {
        xc = xc->right;
        c = &(u->right);
      }
      else {
        while(xc->up && (xc->up->right == xc || xc->up->right == nullptr)) {
          xc = xc->up;
          u = u->up;
        }
        
        if(xc->up) {
          xc = xc->up->right;
          u = u->up;
          c = &(u->right);
        } else break;
      }
    }
    
    return head;
  }
  
  template<typename _Tp> template<typename _Up>
  typename binary_tree<_Tp>::node * binary_tree<_Tp>::copy(const typename binary_tree<_Up>::node *xc)
  {
    //---- if x is empty, do nothing ----
    if(xc == nullptr) return nullptr;
    
    //---- copy the nodes -----
    node *head, *u = nullptr;
    node **c = &head;
    
    while(xc) {
      u = *c = allocate(u, std::move(_Tp(xc->data)));
      
      if(xc->left) {
        xc = xc->left;
        c = &(u->left);
      }
      else if(xc->right) {
        xc = xc->right;
        c = &(u->right);
      }
      else {
        while(xc->up && (xc->up->right == xc || xc->up->right == nullptr)) {
          xc = xc->up;
          u = u->up;
        }
        
        if(xc->up) {
          xc = xc->up->right;
          u = u->up;
          c = &(u->right);
        } else break;
      }
    }
    
    return head;
  }
  
  template<typename _Tp> template<typename _Up, class _Conv>
  typename binary_tree<_Tp>::node * binary_tree<_Tp>::copy(const typename binary_tree<_Up>::node *xc, _Conv conv)
  {
    //---- if x is empty, do nothing ----
    if(xc == nullptr) return nullptr;
    
    //---- copy the nodes -----
    node *head, *u = nullptr;
    node **c = &head;

    while(xc) {
      u = *c = allocate(u, conv(xc->data));

      if(xc->left) {
        xc = xc->left;
        c = &(u->left);
      }
      else if(xc->right) {
        xc = xc->right;
        c = &(u->right);
      }
      else {
        while(xc->up && (xc->up->right == xc || xc->up->right == nullptr)) {
          xc = xc->up;
          u = u->up;
        }
        
        if(xc->up) {
          xc = xc->up->right;
          u = u->up;
          c = &(u->right);
        } else break;
      }
    }

    return head;
  }
  
  
  template<typename _Tp>
  typename binary_tree<_Tp>::node * binary_tree<_Tp>::destroy(node *c)
  {
    //----- find the post-order begin -----
    if(c) while(c->left || c->right) c = (c->left ? c->left : c->right);
    else return nullptr;
    
    //----- destroying all the nodes -----
    while(c) {
      node *todel = c;

      if(c->up == nullptr || c->up->right == c) c = c->up;
      else {
        c = c->up;
        if(c->right) {
          c = c->right;
          while(c->left || c->right)
            c = c->left ? c->left : c->right;
        }
      }
      
      deallocate(todel);
    }
    
    return nullptr;
  }
  
  
  template<typename _Tp> inline
  void swap(binary_tree<_Tp>& a, binary_tree<_Tp>& b) {
    a.swap(b);
  }
  
  
  template<typename _Tp>
  template<class _OStream, class _WriteFunc>
  _OStream& binary_tree<_Tp>::write(_OStream& os, _WriteFunc wfunc) const
  {
    const node *xc = _M_head;
    //---- if the tree is empty, do nothing ----
    if(xc == nullptr) {
      os<<"[N]";
      return os;
    }
    
    //----- no instruction for the head ----
    os<<"[H]";
    
    while(xc) {
      //---- write the data to the stream -----
      wfunc(os, xc->data);
      
      //---- writing the instruction for construction the tree -----
      os<<"[";
      if(xc->left) {
        xc = xc->left;
        os<<(xc ? "L]" : "N]");
      }
      else if(xc->right) {
        xc = xc->right;
        os<<(xc ? "R]" : "N]");
      }
      else {
        while(xc->up && (xc->up->right == xc || xc->up->right == nullptr)) {
          xc = xc->up;
          os<<"u";
        }
        
        if(xc->up) {
          xc = xc->up->right;
          os<<(xc ? "uR]" : "uN]");
        } else {
          os<<"N]";
          break;
        }
      }
    }
    
    return os;
  }
  
  
  
  template<typename _Tp>
  template<class _IStream, class _ReadFunc>
  _IStream& binary_tree<_Tp>::read(_IStream& is, _ReadFunc rfunc)
  {
    char ch, cmd = 'N';
    value_type data;
    node *xc = nullptr, *head = nullptr;
    
    do {
      //---- parsing the instruction ----
      while(is.get() != '[');
      while((ch = is.get()) != ']')
        if(ch == 'u') xc = xc->up;
        else cmd = ch;
      
      //---- nothing to do ----
      if(cmd == 'N') break;
      
      //---- reading the data -----
      rfunc(is, data);
      
      //---- inserting it into the tree ----
      switch(cmd) {
        case 'H' : xc = head = allocate(nullptr, std::move(data)); break;
        case 'R' : xc = append_right(xc, std::move(data)); break;
        case 'L' : xc = append_left(xc, std::move(data)); break;
        default : throw "binary_tree<_Tp>::read(_IStream&) : unable the parse the traversal instructions";
      }
    } while(true);
    
    //----- make the move ----
    destroy(_M_head);
    _M_head = head;
    
    return is;
  }
  
  
}

#endif
