/*  Copyright (C) 2005-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_diracalgebra_h__
#define __deductor_hep_diracalgebra_h__ 1

//   Standard includes
#include <complex>

//   nlo includes
#include <bits/hep-lorentzvector.h>


namespace duct {
  
  template<typename _Tp>
  class diracalgebra;
  
  //   Helper functions
  namespace __diracalgebra_helper {
	
	template<typename _Float>
	void unit(std::complex<_Float> *g, const _Float& m)
	{
	  g[0] = g[5] = g[10] = g[15] = m;
	}
	
	template<typename _Float>
	void gamma(std::complex<_Float> *g, unsigned int gam)
	{
	  static const std::complex<_Float> I = std::complex<_Float>(0.0,1.0);
	  static const std::complex<_Float> mI = std::complex<_Float>(0.0,-1.0);
	  
	  switch(gam) {
		case 0: g[2] = g[7] = g[8] = g[13] = (_Float) 1.0; break;
		case 1: g[3] = g[6] = 1.0; g[9] = g[12] = (_Float) -1.0; break;
		case 2: g[3] = g[12] = mI; g[9] = g[6] = I; break;
		case 3: g[2] = g[13] = 1.0; g[7] = g[8] = (_Float)-1.0; break;
		case 5: g[0] = g[5] = -1.0; g[10] = g[15] = (_Float)1.0; break;
	  }
	}
		
	template<typename _Float, typename _Tp>
	void create(std::complex<_Float> *g, const _Tp& px, const _Tp& py, const _Tp& pz, const _Tp& E)
	{
	  static const std::complex<_Float> I(0.0,1.0);
	  
	  g[2] = g[13] = E-pz;
	  g[7] = g[8] = E+pz;
	  g[3] = -(g[9] = px - I*py);
	  g[6] = -(g[12] = px + I*py);
	}
	
	template<typename _Tp>
	void create(std::complex<_Tp> *g, const _Tp& px, const _Tp& py, const _Tp& pz, const _Tp& E, const _Tp& m)
	{	  
	  g[2] = g[13] = E-pz;
	  g[7] = g[8] = E+pz;
	  g[3] = -(g[9] = std::complex<_Tp>(px,-py));
	  g[6] = -(g[12] = std::complex<_Tp>(px, py));
	  g[0] = g[5] = g[10] = g[15] = m;
	}
	
	template<typename _Float, typename _FloatX> 
	inline void copy(std::complex<_Float> *g, const std::complex<_FloatX> *x) 
	{
	  std::complex<_Float> *last = g+16;
	  for(; g < last; g++, x++) *g = *x;
	}
	
	template<typename _Float, typename _FloatX> 
	void assmul(std::complex<_Float> *g, const std::complex<_FloatX> *x) 
	{
	  std::complex<_Float> r[16];
	  
	  for(unsigned int i = 0; i < 16U; i+=4U)
		for(unsigned int j = 0; j < 4U; j++) {
		  r[i+j] = g[i]*x[j];
		  for(unsigned int k = 1; k < 4U; k++)
			r[i+j] += g[i+k]*x[4*k+j];
		}
	  
	  for(unsigned int i = 0; i < 16U; i++) g[i] = r[i];
  	}
  }
  
  //  Specializations
  template<>
  class diracalgebra<float>
  {
	//   private types
	typedef float _Float;
	
  public:
    //   types
    typedef std::complex<_Float> value_type;
	
    //   constructors
    diracalgebra() {}	
	explicit diracalgebra(unsigned int i) {
	  __diracalgebra_helper::gamma<_Float>(_M_g, i); 
	}
	explicit diracalgebra(_Float m) {
	  __diracalgebra_helper::unit<_Float>(_M_g, m); 
	}
	explicit diracalgebra(const lorentzvector<_Float>& p, const _Float& m = 0.0f) {
	  __diracalgebra_helper::create<_Float>(_M_g, p.X(), p.Y(), p.Z(), p.T(), m);
	}
    explicit diracalgebra(const lorentzvector<value_type>& p) {
	  __diracalgebra_helper::create<_Float, value_type>(_M_g, p.X(), p.Y(), p.Z(), p.T());
	}
    diracalgebra(const _Float& x, const _Float& y, const _Float& z, const _Float& t) {
	  __diracalgebra_helper::create<_Float, _Float>(_M_g, x, y, z, t);
	}
    
	//   copy constuctor
	template<typename _Xp> 
	explicit diracalgebra(const diracalgebra<_Xp>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}
    
    //   assignments
    diracalgebra& operator=(const diracalgebra& x) {
	  if(this != &x) __diracalgebra_helper::copy<_Float, _Float>(_M_g, x._M_g);	
	  return *this;
	}
	
    //   computed assignments
    diracalgebra& operator*=(const diracalgebra& x) {
	  __diracalgebra_helper::assmul<_Float, _Float>(_M_g, x._M_g);
	  return *this;
	}
    diracalgebra& operator+=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}	
    diracalgebra& operator-=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
    diracalgebra& operator*=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}	
    diracalgebra& operator/=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
	diracalgebra& operator*=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}
	diracalgebra& operator/=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
    
	//   assignments with conversion
    template<typename _Xp> 
	diracalgebra& operator=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	  return *this;
	}

	template<typename _Xp> 
	diracalgebra& operator+=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}
	
    template<typename _Xp> 
	diracalgebra& operator-=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
	
	template<typename _Xp> 
	diracalgebra& operator*=(const diracalgebra<_Xp>&);
	
	//     element access
	value_type& operator[](unsigned int i) {
	  return _M_g[i];
	}	
	value_type& operator()(unsigned int i, unsigned int j) {
	  return _M_g[4*i+j];
	}	

	const value_type& operator[](unsigned int i) const {
	  return _M_g[i];
	}
	const value_type& operator()(unsigned int i, unsigned int j) const {
	  return _M_g[4*i+j];
	}
	
  private:
	//   data member
	value_type _M_g[16];  
  };

  template<>
  class diracalgebra<double>
  {
	//   private types
	typedef double _Float;
	
  public:
    //   types
    typedef std::complex<_Float> value_type;
	
    //   constructors
    diracalgebra() {}	
	explicit diracalgebra(unsigned int i) {
	  __diracalgebra_helper::gamma<_Float>(_M_g, i); 
	}
	explicit diracalgebra(_Float m) {
	  __diracalgebra_helper::unit<_Float>(_M_g, m); 
	}
	explicit diracalgebra(const lorentzvector<_Float>& p, const _Float& m = 0.0) {
	  __diracalgebra_helper::create<_Float>(_M_g, p.X(), p.Y(), p.Z(), p.T(), m);
	}
    explicit diracalgebra(const lorentzvector<value_type>& p) {
	  __diracalgebra_helper::create<_Float, value_type>(_M_g, p.X(), p.Y(), p.Z(), p.T());
	}
    diracalgebra(const _Float& x, const _Float& y, const _Float& z, const _Float& t) {
	  __diracalgebra_helper::create<_Float, _Float>(_M_g, x, y, z, t);
	}
    
	//   copy constuctor
	diracalgebra(const diracalgebra<float>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}

	template<typename _Xp> 
	explicit diracalgebra(const diracalgebra<_Xp>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}
    
    //   assignments
    diracalgebra& operator=(const diracalgebra& x) {
	  if(this != &x) __diracalgebra_helper::copy<_Float, _Float>(_M_g, x._M_g);	
	  return *this;
	}
	
    //   computed assignments
    diracalgebra& operator*=(const diracalgebra& x) {
	  __diracalgebra_helper::assmul<_Float, _Float>(_M_g, x._M_g);
	  return *this;
	}
    diracalgebra& operator+=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}	
    diracalgebra& operator-=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
    diracalgebra& operator*=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}	
    diracalgebra& operator/=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
	diracalgebra& operator*=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}
	diracalgebra& operator/=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
    
	//   assignments with conversion
    template<typename _Xp> 
	diracalgebra& operator=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	  return *this;
	}
	
	template<typename _Xp> 
	diracalgebra& operator+=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}
	
    template<typename _Xp> 
	diracalgebra& operator-=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
	
	template<typename _Xp> 
	diracalgebra& operator*=(const diracalgebra<_Xp>&);
	
	//     element access
	value_type& operator[](unsigned int i) {
	  return _M_g[i];
	}	
	value_type& operator()(unsigned int i, unsigned int j) {
	  return _M_g[4*i+j];
	}	
	
	const value_type& operator[](unsigned int i) const {
	  return _M_g[i];
	}
	const value_type& operator()(unsigned int i, unsigned int j) const {
	  return _M_g[4*i+j];
	}
	
  private:
	//   data member
	value_type _M_g[16];  
  };
  
  template<>
  class diracalgebra<long double>
  {
	//   private types
	typedef long double _Float;
	
  public:
    //   types
    typedef std::complex<_Float> value_type;
	
    //   constructors
    diracalgebra() {}	
	explicit diracalgebra(unsigned int i) {
	  __diracalgebra_helper::gamma<_Float>(_M_g, i); 
	}
	explicit diracalgebra(_Float m) {
	  __diracalgebra_helper::unit<_Float>(_M_g, m); 
	}
	explicit diracalgebra(const lorentzvector<_Float>& p, const _Float& m = 0.0l) {
	  __diracalgebra_helper::create<_Float>(_M_g, p.X(), p.Y(), p.Z(), p.T(), m);
	}
    explicit diracalgebra(const lorentzvector<value_type>& p) {
	  __diracalgebra_helper::create<_Float, value_type>(_M_g, p.X(), p.Y(), p.Z(), p.T());
	}
    diracalgebra(const _Float& x, const _Float& y, const _Float& z, const _Float& t) {
	  __diracalgebra_helper::create<_Float, _Float>(_M_g, x, y, z, t);
	}
    
	//   copy constuctor
	diracalgebra(const diracalgebra<float>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}

	diracalgebra(const diracalgebra<double>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}
	
	template<typename _Xp> 
	explicit diracalgebra(const diracalgebra<_Xp>& x) { 
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	}
    
    //   assignments
    diracalgebra& operator=(const diracalgebra& x) {
	  if(this != &x) __diracalgebra_helper::copy<_Float, _Float>(_M_g, x._M_g);	
	  return *this;
	}
	
    //   computed assignments
    diracalgebra& operator*=(const diracalgebra& x) {
	  __diracalgebra_helper::assmul<_Float, _Float>(_M_g, x._M_g);
	  return *this;
	}
    diracalgebra& operator+=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}	
    diracalgebra& operator-=(const diracalgebra& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
    diracalgebra& operator*=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}	
    diracalgebra& operator/=(const value_type& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
	diracalgebra& operator*=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] *= x;
	  return *this;
	}
	diracalgebra& operator/=(const _Float& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] /= x;
	  return *this;
	}
    
	//   assignments with conversion
    template<typename _Xp> 
	diracalgebra& operator=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] = x[i];
	  return *this;
	}
	
	template<typename _Xp> 
	diracalgebra& operator+=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] += x[i];
	  return *this;
	}
	
    template<typename _Xp> 
	diracalgebra& operator-=(const diracalgebra<_Xp>& x) {
	  for(unsigned int i = 0; i < 16; i++) _M_g[i] -= x[i];
	  return *this;
	}
	
	template<typename _Xp> 
	diracalgebra& operator*=(const diracalgebra<_Xp>&);
	
	//     element access
	value_type& operator[](unsigned int i) {
	  return _M_g[i];
	}	
	value_type& operator()(unsigned int i, unsigned int j) {
	  return _M_g[4*i+j];
	}	
	
	const value_type& operator[](unsigned int i) const {
	  return _M_g[i];
	}
	const value_type& operator()(unsigned int i, unsigned int j) const {
	  return _M_g[4*i+j];
	}
	
  private:
	//   data member
	value_type _M_g[16];  
  };
  
  //   some member function definitions
  template<typename _Xp> 
  diracalgebra<float>& diracalgebra<float>::operator*=(const diracalgebra<_Xp>& x) 
  {
	std::complex<_Float> r[16];
	
	for(unsigned int i = 0; i < 16U; i+=4U)
	  for(unsigned int j = 0; j < 4U; j++) {
		r[i+j] = _M_g[i]*x[j];
		for(unsigned int k = 1; k < 4U; k++)
		  r[i+j] += _M_g[i+k]*x[4*k+j];
	  }
	
	for(unsigned int i = 0; i < 16U; i++) _M_g[i] = r[i];
	
	return *this;
  }
  
  template<typename _Xp> 
  diracalgebra<double>& diracalgebra<double>::operator*=(const diracalgebra<_Xp>& x) 
  {
	std::complex<_Float> r[16];
	
	for(unsigned int i = 0; i < 16U; i+=4U)
	  for(unsigned int j = 0; j < 4U; j++) {
		r[i+j] = _M_g[i]*x[j];
		for(unsigned int k = 1; k < 4U; k++)
		  r[i+j] += _M_g[i+k]*x[4*k+j];
	  }
	
	for(unsigned int i = 0; i < 16U; i++) _M_g[i] = r[i];
	
	return *this;
  }
  
  template<typename _Xp> 
  diracalgebra<long double>& diracalgebra<long double>::operator*=(const diracalgebra<_Xp>& x) 
  {
	std::complex<_Float> r[16];
	
	for(unsigned int i = 0; i < 16U; i+=4U)
	  for(unsigned int j = 0; j < 4U; j++) {
		r[i+j] = _M_g[i]*x[j];
		for(unsigned int k = 1; k < 4U; k++)
		  r[i+j] += _M_g[i+k]*x[4*k+j];
	  }
	
	for(unsigned int i = 0; i < 16U; i++) _M_g[i] = r[i];
	
	return *this;
  }
  
  //    unary operators
  template<typename _Tp>
  diracalgebra<_Tp> operator+(const diracalgebra<_Tp>& x) {
	return x;
  }
  
  template<typename _Tp>
  diracalgebra<_Tp> operator-(const diracalgebra<_Tp>& x) {
	return diracalgebra<_Tp>(x) *= (_Tp) -1.0;
  }
  
  //    Other operators
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator+(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) {
	return diracalgebra<_Tp>(a) += b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator-(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) {
	return diracalgebra<_Tp>(a) -= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator*(const diracalgebra<_Tp>& a, const _Tp & b) {
	return diracalgebra<_Tp>(a) *= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator*(const _Tp & b, const diracalgebra<_Tp>& a) {
	return diracalgebra<_Tp>(a) *= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator*(const diracalgebra<_Tp>& a, const std::complex<_Tp>& b) {
	return diracalgebra<_Tp>(a) *= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator*(const std::complex<_Tp>& b, const diracalgebra<_Tp>& a) {
	return diracalgebra<_Tp>(a) *= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator/(const diracalgebra<_Tp>& a, const _Tp & b) {
	return diracalgebra<_Tp>(a) /= b;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp> 
  operator/(const diracalgebra<_Tp>& a, const std::complex<_Tp> & b) {
	return diracalgebra<_Tp>(a) /= b;
  }
  
  template<typename _Tp> diracalgebra<_Tp>
  operator*(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) 
  {
	diracalgebra<_Tp> r;
	
	for(unsigned int i = 0; i < 4U; i++)
	  for(unsigned int j = 0; j < 4U; j++) {
		r(i,j) = a(i,0)*b(0,j);
		for(unsigned int k = 1; k < 4U; k++)
		  r(i,j) += a(i,k)*b(k,j);
	  }

	return r;
  }
  
  template<typename _Tp> bool 
  operator==(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) 
  {
	for(unsigned int i = 0; i < 4U; i++)
	  for(unsigned int j = 0; j < 4U; j++)
		if(a(i,j) != b(i,j)) return false;
	
	return true;
  }
  
  template<typename _Tp> inline bool 
  operator!=(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) {
	return !(a == b);
  }
  
  template<typename _Tp> inline 
  std::complex<_Tp> Trace(const diracalgebra<_Tp>& a) {
	return a(0,0) + a(1,1) + a(2,2) + a(3,3);
  }
  
  template<typename _Tp> inline diracalgebra<_Tp>
  commutator(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) 
  {
	return a*b-b*a;
  }
  
  template<typename _Tp> inline diracalgebra<_Tp>
  anticommutator(const diracalgebra<_Tp>& a, const diracalgebra<_Tp>& b) 
  {
	return a*b+b*a;
  }
  
  template<typename _Tp>
  std::ostream& operator<<(std::ostream& os, const diracalgebra<_Tp>& d)
  {
	for(unsigned int i = 0; i < 4U; i++) {
	  for(unsigned int j = 0; j < 4U; j++)
		os<<d(i,j)<<"  ";
	  os<<std::endl;
	}
	
	return os;
  }
  
  
}

#endif

