/*  Copyright (C) 2005-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __NLO_HEP_DIRACSPINOR_H__
#define __NLO_HEP_DIRACSPINOR_H__ 1

//   Standard includes
#include <complex>

//   nlo includes
#include <bits/hep-lorentzvector.h>
#include <bits/hep-diracalgebra.h>


namespace duct {
  
  //  empty template declaration
  template<typename _Tp, class _Or>
  class diracspinor;

  struct bra{};
  struct ket{};

  
  namespace __diracspinor_helper {
  
	//----- diracspinors with standard reference momentum -----
	template<typename _Tp>
	void create_weyl_ket(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, int spin)
	{
	  _Tp Ex = p.T() - p.X(), py = p.Y(), pz = p.Z();
	  std::complex<_Tp> norm = std::sqrt(std::complex<_Tp>(2.0*Ex, 0.0));
	  
	  switch(spin) {
		case 1: 
		  s[0] = std::complex<_Tp>(-Ex-pz,-py)/norm;
		  s[1] = std::complex<_Tp>( Ex-pz,-py)/norm;
		  break;
		case -1:
		  s[0] = std::complex<_Tp>( pz-Ex,-py)/norm;
		  s[1] = std::complex<_Tp>(-Ex-pz, py)/norm;
		  break;
	  }		
	}
	
	template<typename _Tp>
	void create_ket(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, char type, int spin)
	{
	  _Tp E = std::abs(p.T()), pv = p.threevector<_Tp>::mag();
	  
	  //----- if it is a V diracspinor we have to flip the spin ----
	  if(type == 'V') spin *= -1;
	  
	  if(std::abs(E-pv) < 1.0e-13*(E+pv)) {
		//    massless diracspinor
		create_weyl_ket(spin == 1 ? s+2 : s, p, spin);
	  } else { 
		//----- lightlike refence momentum n -----
		lorentzvector<_Tp> n(-threevector<_Tp>(p), p.T() < 0.0 ? -pv : pv);
		n /= E+pv;
		
		//----- lighlike momentum k -----
		_Tp m2 = (E+pv)*(E-pv); 
		lorentzvector<_Tp> k = p - _Tp(0.5*m2/pv)*n;
		
		//----- massless diracspinor -----
		std::complex<_Tp> w[2];
		create_weyl_ket(w, k, spin);
		
		//----- calculate the massive diracspinor -----
		_Tp nP = n.plus(), nM = n.minus();
		std::complex<_Tp> nT = std::complex<_Tp>(n.X(), n.Y());
		std::complex<_Tp> fac = std::sqrt(std::complex<_Tp>(m2))/_Tp((type == 'V' ? -2.0 : 2.0)*pv);
		
		switch(spin) {
		  case -1:
			s[0] = w[0];
			s[1] = w[1];
			s[2] = (nP*w[0] + std::conj(nT)*w[1])*fac;
			s[3] = (nM*w[1] + nT*w[0])*fac;
			break;
		  case 1:
			s[0] = (nM*w[0] - std::conj(nT)*w[1])*fac;
			s[1] = (nP*w[1] - nT*w[0])*fac;
			s[2] = w[0];
			s[3] = w[1];
			break;
		}
	  }
	}

	template<typename _Tp>
	void create_weyl_bra(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, int spin)
	{
	  _Tp Ex = p.T() - p.X(), py = p.Y(), pz = p.Z();
	  std::complex<_Tp> norm = std::sqrt(std::complex<_Tp>(2.0*Ex, 0.0));
	  
	  switch(spin) {
		case 1: 
		  s[0] = std::complex<_Tp>(-Ex-pz, py)/norm;
		  s[1] = std::complex<_Tp>( Ex-pz, py)/norm;
		  break;
		case -1:
		  s[0] = std::complex<_Tp>( pz-Ex, py)/norm;
		  s[1] = std::complex<_Tp>(-Ex-pz,-py)/norm;
		  break;
	  }		
	}
	
	template<typename _Tp>
	void create_bra(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, char type, int spin)
	{
	  _Tp E = std::abs(p.T()), pv = p.threevector<_Tp>::mag();
	  
	  //----- if it is a V diracspinor we have to flip the spin ----
	  if(type == 'V') spin *= -1;
	  
	  if(std::abs(E-pv) < 1.0e-13*(E+pv)) {
		//    massless diracspinor
		create_weyl_bra(spin == -1 ? s+2 : s, p, spin);
	  } else { 
		//----- lightlike refence momentum n -----
		lorentzvector<_Tp> n(-threevector<_Tp>(p), p.T() < 0.0 ? -pv : pv);
		n /= E+pv;
		
		//----- lighlike momentum k -----
		_Tp m2 = (E+pv)*(E-pv); 
		lorentzvector<_Tp> k = p - _Tp(0.5*m2/pv)*n;
		
		//----- massless diracspinor -----
		std::complex<_Tp> w[2];
		create_weyl_bra(w, k, spin);
		
		//----- calculate the massive diracspinor -----
		_Tp nP = n.plus(), nM = n.minus();
		std::complex<_Tp> nT = std::complex<_Tp>(n.X(), n.Y());
		std::complex<_Tp> fac = std::sqrt(std::complex<_Tp>(m2))/_Tp((type == 'V' ? -2.0 : 2.0)*pv);
		
		switch(spin) {
		  case -1:
			s[0] = (nP*w[0] + nT*w[1])*fac;
			s[1] = (nM*w[1] + std::conj(nT)*w[0])*fac;
			s[2] = w[0];
			s[3] = w[1];
			break;
		  case 1:
			s[0] = w[0];
			s[1] = w[1];
			s[2] = (nM*w[0] - nT*w[1])*fac;
			s[3] = (nP*w[1] - std::conj(nT)*w[0])*fac;
			break;
		}
	  }
	}
	
	//----- diracspinors with customized reference momentum -----
	/*****
	template<typename _Tp>
	void create_weyl_ket(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, const lorentzvector<_Tp>& n, int spin)
	{
	  std::complex<_Tp> w[2];
	  create_weyl_ket(w, n, -spin);
	  
	  _Tp pP = p.plus(), pM = p.minus();
	  std::complex<_Tp> pT = std::complex<_Tp>(p.X(), p.Y());
	  std::complex<_Tp> norm = std::sqrt(complex<_Tp>(2.0*(p*n), 0.0));
	  
	  switch(spin) {
		case 1:
		  s[0] = (pP*w[0] + std::conj(pT)*w[1])/norm;
		  s[1] = (pM*w[1] + pT*w[0])/norm;
		  break;
		case -1:
		  s[0] = (std::conj(pT)*w[1] - pM*w[0])/norm;
		  s[1] = (pT*w[0] - pP*w[1])/norm;		
		  break;
	  }
	}
	
	template<typename _Tp>
	void create_ket(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, 
					const lorentzvector<_Tp>& Q, char type, int spin)
	{
	  _Tp m2 = p.mag2(), E = p.T();

	  //----- if it is a V diracspinor we have to flip the spin ----
	  if(type == 'V') spin *= -1;
	  
	  if(std::abs(m2) < 1.0e-13*E*E) {
		//    massless diracspinor
		create_weyl_ket(spin == 1 ? s+2 : s, p, spin);
	  } else { 
		//----- lightlike refence momentum n -----
		_Tp pQ = p*Q, Q2 = Q.mag2();
		lorentzvector<_Tp> n = (pQ < 0.0 ? -Q : Q) - (Q2/(std::abs(pQ) + std::sqrt(pQ*pQ-m2*Q2)))*p;
		
		//----- lighlike momentum k -----
		_Tp pn = p*n;
		lorentzvector<_Tp> k = p - (0.5*m2/pn)*n;

		//----- massless diracspinor -----
		std::complex<_Tp> w[2];
		create_weyl_ket(w, k, spin);
		
		//----- calculate the massive diracspinor -----
		_Tp nP = n.plus(), nM = n.minus();
		std::complex<_Tp> nT = std::complex<_Tp>(n.X(), n.Y());
		std::complex<_Tp> fac = std::sqrt(std::complex<_Tp>(m2))/((type == 'V' ? -2.0 : 2.0)*pn);
		
		switch(spin) {
		  case -1:
			s[0] = w[0];
			s[1] = w[1];
			s[2] = (pP*w[0] + std::conj(pT)*w[1])*fac;
			s[3] = (pM*w[1] + pT*w[0])*fac;
			break;
		  case 1:
			s[0] = (pM*w[0] - std::conj(pT)*w[1])*fac;
			s[1] = (pP*w[1] - pT*w[0])*fac;
			s[2] = w[0];
			s[3] = w[1];
			break;
		}
	  }
	}
	
	template<typename _Tp>
	void create_ket(std::complex<_Tp> *s, const lorentzvector<_Tp>& p, const lorentzvector<_Tp>& r, 
					const lorentzvector<_Tp>& Q, char type, int spin)
	{
	  _Tp m2 = p.mag2(), E = p.T();
	  
	  //----- if it is a V diracspinor we have to flip the spin ----
	  if(type == 'V') spin *= -1;
	  
	  if(std::abs(m2) < 1.0e-13*E*E) {
		//    massless diracspinor
		create_weyl_ket(spin == 1 ? s+2 : s, p, r, spin);
	  } else { 
		//----- lightlike refence momentum n -----
		_Tp pQ = p*Q, Q2 = Q.mag2();
		lorentzvector<_Tp> n = (pQ < 0.0 ? -Q : Q) - (Q2/(std::abs(pQ) + std::sqrt(pQ*pQ-m2*Q2)))*p;
		
		//----- lighlike momentum k -----
		_Tp pn = p*n;
		lorentzvector<_Tp> k = p - (0.5*m2/pn)*n;
		
		//----- massless diracspinor -----
		std::complex<_Tp> w[2];
		create_weyl_ket(w, k, r, spin);
		
		//----- calculate the massive diracspinor -----
		_Tp nP = n.plus(), nM = n.minus();
		std::complex<_Tp> nT = std::complex<_Tp>(n.X(), n.Y());
		std::complex<_Tp> fac = std::sqrt(std::complex<_Tp>(m2))/((type == 'V' ? -2.0 : 2.0)*pn);
		
		switch(spin) {
		  case -1:
			s[0] = w[0];
			s[1] = w[1];
			s[2] = (pP*w[0] + std::conj(pT)*w[1])*fac;
			s[3] = (pM*w[1] + pT*w[0])*fac;
			break;
		  case 1:
			s[0] = (pM*w[0] - std::conj(pT)*w[1])*fac;
			s[1] = (pP*w[1] - pT*w[0])*fac;
			s[2] = w[0];
			s[3] = w[1];
			break;
		}
	  }
	}
	******/	
  }
    
  //    ket diracspinors e.g.: |p+>
  template<>
  class diracspinor<float, ket>
  {
  public:
    //  types
	typedef float float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0f;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_ket(_M_s, p, type, spin);
	}
	
	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
 	  return *this;
	}
	
	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }
 
	diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
	
    diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }

	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }

    diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
    
  template<>
  class diracspinor<double, ket>
  {
  public:
    //  types
	typedef double float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_ket(_M_s, p, type, spin);
	}
	
	diracspinor(const diracspinor<float, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }

	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
	  return *this;
    }

	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }

    diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }

    diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
  
  template<>
  class diracspinor<long double, ket>
  {
  public:
    //  types
	typedef long double float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0L;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_ket(_M_s, p, type, spin);
	}
	
	diracspinor(const diracspinor<float, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
	diracspinor(const diracspinor<double, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
	  return *this;
    }

	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }
    
	diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, ket>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
  
  
  //    bra diracspinors e.g.: <p+|
  template<>
  class diracspinor<float, bra>
  {
  public:
    //  types
	typedef float float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0f;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_bra(_M_s, p, type, spin);
	}
	
	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
	  return *this;
    }

	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }
    
	diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
  
  template<>
  class diracspinor<double, bra>
  {
  public:
    //  types
	typedef double float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_bra(_M_s, p, type, spin);
	}
	
	diracspinor(const diracspinor<float, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
	  return *this;
    }

	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }
 
	diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
 
	diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
  
  template<>
  class diracspinor<long double, bra>
  {
  public:
    //  types
	typedef long double float_type;
    typedef std::complex<float_type> value_type;
	
    //  default constructors
    diracspinor() { 
	  _M_s[0] = _M_s[1] = _M_s[2] = _M_s[3] = 0.0L;
	}
	
	//   constructs from lorentz vector (standard reference momenta)
    diracspinor(const lorentzvector<float_type>& p, int spin, char type = 'U') {
	  if(type == 'U' || type == 'V') __diracspinor_helper::create_bra(_M_s, p, type, spin);
	}
	
	diracspinor(const diracspinor<float, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
	diracspinor(const diracspinor<double, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
	//   simple conversions
	template<typename _Xp> 
	explicit diracspinor(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
    }
	
    //   assignments
    template<typename _Xp> 
	diracspinor& operator=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] = x[0]; _M_s[1] = x[1]; _M_s[2] = x[2]; _M_s[3] = x[3];
	  return *this;
    }
	
    //   computed assignments
    diracspinor& operator+=(const diracspinor& x) {
	  _M_s[0] += x._M_s[0]; _M_s[1] += x._M_s[1]; 
	  _M_s[2] += x._M_s[2]; _M_s[3] += x._M_s[3];
	  return *this;
    }

	diracspinor& operator-=(const diracspinor& x) {
	  _M_s[0] -= x._M_s[0]; _M_s[1] -= x._M_s[1]; 
	  _M_s[2] -= x._M_s[2]; _M_s[3] -= x._M_s[3];
	  return *this;
    }
    
	diracspinor& operator*=(const value_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const value_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
	diracspinor& operator*=(const float_type& x) {
	  _M_s[0] *= x; _M_s[1] *= x; _M_s[2] *= x; _M_s[3] *= x;
	  return *this;
    }
    
	diracspinor& operator/=(const float_type& x) {
	  _M_s[0] /= x; _M_s[1] /= x; _M_s[2] /= x; _M_s[3] /= x;
	  return *this;
    }
	
    template<typename _Xp> 
	diracspinor& operator+=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] += x[0]; _M_s[1] += x[1]; _M_s[2] += x[2]; _M_s[3] += x[3];
	  return *this;
    }
    
    template<typename _Xp> 
	diracspinor& operator-=(const diracspinor<_Xp, bra>& x) {
	  _M_s[0] -= x[0]; _M_s[1] -= x[1]; _M_s[2] -= x[2]; _M_s[3] -= x[3];
	  return *this;
    }
	
	//  element access
	value_type& operator[](unsigned int i) { return _M_s[i];}
	const value_type& operator[](unsigned int i) const { return _M_s[i];}
	
  private:
	value_type _M_s[4];
  };
  
  //   unary operators
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator+(const diracspinor<_Tp, _Or>& x) {
	return x;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator-(const diracspinor<_Tp, _Or>& x) {
	return diracspinor<_Tp, _Or>(x) *= -1.0;
  }
  
  //   binary operators
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator*(const diracspinor<_Tp, _Or>& x, const _Tp& y) {
	return diracspinor<_Tp, _Or>(x) *= y;
  }

  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator*(const diracspinor<_Tp, _Or>& x, const std::complex<_Tp>& y) {
	return diracspinor<_Tp, _Or>(x) *= y;
  }

  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator*(const _Tp& y, const diracspinor<_Tp, _Or>& x) {
	return diracspinor<_Tp, _Or>(x) *= y;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator*(const std::complex<_Tp>& y, const diracspinor<_Tp, _Or>& x) {
	return diracspinor<_Tp, _Or>(x) *= y;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator/(const diracspinor<_Tp, _Or>& x, const _Tp& y) {
	return diracspinor<_Tp, _Or>(x) /= y;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator/(const diracspinor<_Tp, _Or>& x, const std::complex<_Tp>& y) {
	return diracspinor<_Tp, _Or>(x) /= y;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator+(const diracspinor<_Tp, _Or>& x, const diracspinor<_Tp, _Or>& y) {
	return diracspinor<_Tp, _Or>(x) += y;
  }
  
  template<typename _Tp, class _Or>
  diracspinor<_Tp, _Or> operator-(const diracspinor<_Tp, _Or>& x, const diracspinor<_Tp, _Or>& y) {
	return diracspinor<_Tp, _Or>(x) -= y;
  }
  
  
  //    special multiplications
  template<typename _Tp> inline 
  std::complex<_Tp> operator*(const diracspinor<_Tp, bra>& a, const diracspinor<_Tp, ket>& b) {
	return a[0]*b[0]+a[1]*b[1]+a[2]*b[2]+a[3]*b[3];
  }
  
  template<typename _Tp> 
  diracspinor<_Tp, ket> operator*(const diracalgebra<_Tp>& a, const diracspinor<_Tp, ket>& b) 
  {
	diracspinor<_Tp, ket> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  for(unsigned int j = 0; j < 4; j++) 	  
		r[i] += a(i,j)*b[j];
	
	return r;
  }

  template<typename _Tp> 
  diracspinor<_Tp, bra> operator*(const diracspinor<_Tp, bra>& b, const diracalgebra<_Tp>& a) 
  {
	diracspinor<_Tp, bra> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  for(unsigned int j = 0; j < 4; j++) 	  
		r[i] += b[j]*a(j,i);
	
	return r;
  }
  
  template<typename _Tp> 
  diracalgebra<_Tp> operator*(const diracspinor<_Tp, ket>& a, const diracspinor<_Tp, bra>& b) 
  {
	diracalgebra<_Tp> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  for(unsigned int j = 0; j < 4; j++) 	  
		r(i,j) = a[i]*b[j];
	
	return r;
  }

  template<typename _Tp>
  lorentzvector<std::complex<_Tp> > make_vector(const diracspinor<_Tp, bra>& a, const diracspinor<_Tp, ket>& b)
  {
	static const std::complex<_Tp> I(0.0, 1.0);
	return lorentzvector<std::complex<_Tp> >(a[0]*b[3]+a[1]*b[2]-a[2]*b[1]-a[3]*b[0],
											 I*(-a[0]*b[3]+a[1]*b[2]+a[2]*b[1]-a[3]*b[0]),
											 a[0]*b[2]-a[1]*b[3]-a[2]*b[0]+a[3]*b[1],
											 a[0]*b[2]+a[1]*b[3]+a[2]*b[0]+a[3]*b[1]);
  }
  
  
  //   complex conjugations
  template<typename _Tp, class _Or> 
  diracspinor<_Tp, _Or> conj(const diracspinor<_Tp, _Or>& x)
  {
	diracspinor<_Tp, _Or> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  r[i] = std::conj(x[i]);
	
	return r;
  }
  
  //   transponse
  template<typename _Tp> 
  diracspinor<_Tp, bra> trans(const diracspinor<_Tp, ket>& x)
  {
	diracspinor<_Tp, bra> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  r[i] = x[i];
	
	return r;
  }
  
  template<typename _Tp> 
  diracspinor<_Tp, ket> trans(const diracspinor<_Tp, bra>& x)
  {
	diracspinor<_Tp, ket> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  r[i] = x[i];
	
	return r;
  }

  //    adjoint diracspinor = complex conj. + transposed
  template<typename _Tp> 
  diracspinor<_Tp, bra> adj(const diracspinor<_Tp, ket>& x)
  {
	diracspinor<_Tp, bra> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  r[i] = std::conj(x[i]);
	
	return r;
  }

  template<typename _Tp> 
  diracspinor<_Tp, ket> adj(const diracspinor<_Tp, bra>& x)
  {
	diracspinor<_Tp, ket> r;
	
	for(unsigned int i = 0; i < 4; i++) 
	  r[i] = std::conj(x[i]);
	
	return r;
  }

  //   dirac adjoint = adjont * gamma^0
  template<typename _Tp> 
  diracspinor<_Tp, bra> dirac_adj(const diracspinor<_Tp, ket>& x)
  {
	diracspinor<_Tp, bra> r;
	
	r[0] = std::conj(x[2]);
	r[1] = std::conj(x[3]);
	r[2] = std::conj(x[0]);
	r[3] = std::conj(x[1]);
	
	return r;
  }
  
  template<typename _Tp> 
  diracspinor<_Tp, ket> dirac_adj(const diracspinor<_Tp, bra>& x)
  {
	diracspinor<_Tp, ket> r;
	
	r[0] = std::conj(x[2]);
	r[1] = std::conj(x[3]);
	r[2] = std::conj(x[0]);
	r[3] = std::conj(x[1]);
	
	return r;
  }

  
  //    stream operations
  template<typename _Tp> 
  std::ostream& operator<<(std::ostream& os, const diracspinor<_Tp, bra>& x)
  {
    os<<"<";
    for(unsigned int i = 0; i < 3; i++)
      os<<x[i]<<", ";
    os<<x[3]<<"|";
    
    return os;
  }
  
  template<typename _Tp> 
  std::ostream& operator<<(std::ostream& os, const diracspinor<_Tp, ket>& x)
  {
    os<<"|";
    for(unsigned int i = 0; i < 3; i++)
      os<<x[i]<<", ";
    os<<x[3]<<">";
    
    return os;
  }
  
}

#endif
