/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_distbook_h__
#define __deductor_hep_distbook_h__ 1


//   Standard includes
#include <map>
#include <string>
#include <tuple>
#include <fstream>

//   deductor includes
#include <bits/hep-dist.h>
#include <bits/nlo-cxx-support.h>


namespace duct {
 
  
  template<typename _Tp, class _Point = void,
    class _Traits = sample_traits<_Tp>,
    class _PointTraits = distpoint::traits<_Point>>
  class distbook
  {
  public:
    /*   public typedefs   */
    typedef _Tp sample_type;
    typedef _Point point_type;
    typedef _Traits sample_traits;
    typedef _PointTraits point_traits;
    typedef distribution<_Tp, _Point, _Traits, _PointTraits> distribution_type;
    typedef std::tuple<std::string, distribution_type, char, ltxplot> entry_type;

    typedef typename std::map<int, entry_type>::value_type        value_type;
    typedef typename std::map<int, entry_type>::reference         reference;
    typedef typename std::map<int, entry_type>::const_reference   const_reference;
    typedef typename std::map<int, entry_type>::iterator          iterator;
    typedef typename std::map<int, entry_type>::const_iterator    const_iterator;

    /*  computed assignment   */
    distbook& operator+=(const distbook& x) {
      auto iter = _M_book.begin();
      auto xiter = x.begin();
      while(iter != _M_book.end())
        std::get<1>((iter++)->second) += std::get<1>((xiter++)->second);
      return *this;
    }

    /*    create an entry in the book   */
    void create(int key, const char *title, unsigned n, const _Point *base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(n, base, norm), 'D', ltxplot());
    }
 
    template<class _InputIterator>
    void create(int key, const char *title, _InputIterator first, _InputIterator last, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(first, last, norm), 'D', ltxplot());
    }
    
    void create(int key, const char *title, std::initializer_list<_Point> base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(base, norm), 'D', ltxplot());
    }

    template<class _Container>
    void create(int key, const char *title, const _Container& base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(base, norm), 'D', ltxplot());
    }

    /*   collecting the weights    */
    template<class _Func>
    void collect(int key, _Func func) {
      std::get<1>(_M_book.find(key)->second).collect(func);
    }
    
    template<class _Func, typename ..._Arg>
    void collect(int key, const _Tp& w, _Func func, _Arg... x) {
      std::get<1>(_M_book.find(key)->second).collect(w, func, x...);
    }
    
    /*   accumulate the collected weights   */
    void accumulate(int key) {
      std::get<1>(_M_book.find(key)->second).accumulate();
    }
    
    template<class _Func>
    void accumulate(int key, _Func func) {
      std::get<1>(_M_book.find(key)->second).accumulate(func);
    }
    
    template<class _Func, typename ..._Arg>
    void accumulate(int key, const _Tp& w, _Func func, _Arg... x) {
      std::get<1>(_M_book.find(key)->second).accumulate(w, func, x...);
    }
    
    
    /*   collecting and accumulating weights for the normalization   */
    void collect_norm(int key, const _Tp& w) { std::get<1>(_M_book.find(key)->second).collect_norm(w);}
    void accumulate_norm(int key, const _Tp& w) { std::get<1>(_M_book.find(key)->second).accumulate_norm(w);}
    void accumulate_norm(int key) { std::get<1>(_M_book.find(key)->second).accumulate_norm();}
    
    
    /*   accumulate the weights in the whole book   */
    void accumulate() {
      for(auto& d : _M_book) {
        std::get<1>(d.second).accumulate();
        std::get<1>(d.second).accumulate_norm();
      }
    }
    
    /*  element access (this can also create an entry) */
    entry_type& operator[](int key) { return _M_book[key];}
    
    /*    iterators  */
    iterator begin() { return _M_book.begin();}
    iterator end() { return _M_book.end();}
    
    const_iterator begin() const { return _M_book.begin();}
    const_iterator end() const { return _M_book.end();}
    
    const_iterator cbegin() const { return _M_book.cbegin();}
    const_iterator cend() const { return _M_book.cend();}

    /*   structural information   */
    unsigned int size() const { return _M_book.size();}
    
    /*   clear the whole book   */
    void clear() { _M_book.clear();}
    
    /*   enable plot  */
    void enable_plot(int key, const char *title = nullptr)
    {
      auto& d = _M_book.find(key)->second;

      std::get<2>(d) = 'E';
      if(title) std::get<3>(d).head.title = title;
      else std::get<3>(d).head.title = std::get<0>(d);
    }
    
    void add_figure(int key, const ltxfigure& fig)
    {
      auto& d = _M_book.find(key)->second;
      
      /*   it has to be explicitly enabled  */
      if('D' == std::get<2>(d)) return;
      
      std::get<3>(d).figures.push_back(fig);
    }
   
    /*   check if the plot is enabled on any distribution  */
    bool is_plot_enabled() const {
      for(const auto& d : _M_book)
        if(std::get<2>(d.second) == 'E') return true;
      
      return false;
    }
    
  private:
    /*    This is a map of distributions   */
    std::map<int, entry_type> _M_book;
  };
  
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  inline distbook<_Tp, _Point, _Traits, _PointTraits>
  operator+(const distbook<_Tp, _Point, _Traits, _PointTraits>& x, const distbook<_Tp, _Point, _Traits, _PointTraits>& y)
  {
    return distbook<_Tp, _Point, _Traits, _PointTraits>(x) += y;
  }
  
  //   I/O operations
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::ostream& write_txt(std::ostream& os, const distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    os<<x.size();
    
    for(const auto& vi : x) {
      os<<" "<<vi.first;
      os<<" "; write_txt(os, std::get<0>(vi.second));
      os<<" "; write_txt(os, std::get<1>(vi.second));
      os<<" "<<std::get<2>(vi.second);
      os<<" "; write_txt(os, std::get<3>(vi.second));
    }
    
    return os;
  }
    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::ostream& write_bin(std::ostream& os, const distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    unsigned sx = x.size();
    os.write(reinterpret_cast<const char *>(&sx), sizeof(unsigned));
      
    for(auto& vi : x) {
      os.write(reinterpret_cast<const char *>(&(vi.first)), sizeof(int));
      write_bin(os, std::get<0>(vi.second));
      write_bin(os, std::get<1>(vi.second));
      os.put(std::get<2>(vi.second));
      write_bin(os, std::get<3>(vi.second));
    }
    
    return os;
  }
    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::ostream& print(std::ostream& os, const distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    for(auto& vi : x) {
      os<<"#  "<<std::get<0>(vi.second)<<"\n";
      print(os, std::get<1>(vi.second))<<"\n\n";
    }
    
    return os;
  }
    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::istream& read_txt(std::istream& is, distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    unsigned sx;
    is>>sx;
    
    for(unsigned int i = 0; i < sx; i++) {
      int key;
      is>>key;
      auto& tmp = x[key];
      
      read_txt(is, std::get<0>(tmp));
      read_txt(is, std::get<1>(tmp));
      is>>std::get<2>(tmp);
      read_txt(is, std::get<3>(tmp));
    }
    
    return is;
  }
    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::istream& read_bin(std::istream& is, distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    unsigned sx;
    
    x.clear();
    is.read(reinterpret_cast<char *>(&sx), sizeof(unsigned));
    
    for(unsigned int i = 0; i < sx; i++)
    {
      int key;
      is.read(reinterpret_cast<char *>(&key), sizeof(int));
      auto& tmp = x[key];
      
      read_bin(is, std::get<0>(tmp));
      read_bin(is, std::get<1>(tmp));
      std::get<2>(tmp) = is.get();
      read_bin(is, std::get<3>(tmp));
    }
    
    return is;
  }
    

  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  void plot(std::string dir, const distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    if(dir.back() != '/') dir += '/';
    
    for(auto i = x.begin(); i != x.end(); ++i)
      if(std::get<2>(i->second) == 'E') {
        std::fstream file((dir+"figure-"+std::to_string(i->first)+".tex").c_str(), std::ios::out);
        distplot<_Tp, _Point, _Traits, _PointTraits>::plot(file, std::get<1>(i->second), std::get<3>(i->second));
        file.flush();
        file.close();
      }
  }
  
  
  template<typename _Tp, class _Traits>
  class distbook<_Tp, void, _Traits, distpoint::traits<void>>
  {
  public:
    /*   public typedefs   */
    typedef _Tp sample_type;
    typedef void point_type;
    typedef _Traits sample_traits;
    typedef distpoint::traits<void> point_traits;
    typedef distribution<_Tp, void, _Traits, distpoint::traits<void>> distribution_type;
    typedef std::tuple<std::string, distribution_type, char, ltxplot> entry_type;
    
    typedef typename std::map<int, entry_type>::value_type        value_type;
    typedef typename std::map<int, entry_type>::reference         reference;
    typedef typename std::map<int, entry_type>::const_reference   const_reference;
    typedef typename std::map<int, entry_type>::iterator          iterator;
    typedef typename std::map<int, entry_type>::const_iterator    const_iterator;
    
    /*  computed assignment   */
    distbook& operator+=(const distbook& x) {
      auto iter = _M_book.begin();
      auto xiter = x.begin();
      while(iter != _M_book.end())
        std::get<1>((iter++)->second) += std::get<1>((xiter++)->second);
      return *this;
    }
    
    /*    create an entry in the book   */
    void create(int key, const char *title) {
      _M_book[key] = entry_type(title, distribution_type(), 'D', ltxplot());
    }
    
    /*   collecting the weights    */
    void collect(int key, const _Tp& w) {
      std::get<1>(_M_book.find(key)->second).collect(w);
    }
    
    /*   accumulate the collected weights   */
    void accumulate(int key) {
      std::get<1>(_M_book.find(key)->second).accumulate();
    }
    
    void accumulate(int key, const _Tp& w) {
      std::get<1>(_M_book.find(key)->second).accumulate(w);
    }
    
    /*   accumulate the weights in the whole book   */
    void accumulate() { for(auto& d : _M_book) std::get<1>(d.second).accumulate();}
    
    /*  element access (this can also create an entry) */
    entry_type& operator[](int key) { return _M_book[key];}
    
    /*    iterators  */
    iterator begin() { return _M_book.begin();}
    iterator end() { return _M_book.end();}
    
    const_iterator begin() const { return _M_book.begin();}
    const_iterator end() const { return _M_book.end();}
    
    const_iterator cbegin() const { return _M_book.cbegin();}
    const_iterator cend() const { return _M_book.cend();}
    
    /*   structural information   */
    unsigned int size() const { return _M_book.size();}
    
    /*   clear the whole book   */
    void clear() { _M_book.clear();}
    
    /*   enable plot  */
    void enable_plot(int key, const char *title = nullptr)
    {
      auto& d = _M_book.find(key)->second;
      
      std::get<2>(d) = 'E';
      if(title) std::get<3>(d).head.title = title;
      else std::get<3>(d).head.title = std::get<0>(d);
    }
    
    void add_figure(int key, const ltxfigure& fig)
    {
      auto& d = _M_book.find(key)->second;
      
      /*   it has to be explicitly enabled  */
      if('D' == std::get<2>(d)) return;
      
      std::get<3>(d).figures.push_back(fig);
    }
    
    /*   check if the plot is enabled on any distribution  */
    bool is_plot_enabled() const {
      for(const auto& d : _M_book)
        if(std::get<2>(d.second) == 'E') return true;
      
      return false;
    }

  private:
    /*    This is a map of distributions   */
    std::map<int, entry_type> _M_book;
  };
  


}
    
#endif
