/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 5/15/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __hep_dynmultiarray_h__
#define __hep_dynmultiarray_h__ 1

//   std includes
#include <vector>
#include <iterator>

//   project headers


namespace duct {
  
  
  template<typename _Tp, unsigned _Size, unsigned...>
  class dynmultiarray;
  
  
  template<typename _Tp, unsigned _Size>
  class dynmultiarray<_Tp, _Size>
  {
    typedef std::vector<_Tp> _Base;
    
  public:
    /*   Public types. These are usual in vectors   */
    typedef typename _Base::value_type             value_type;
    typedef typename _Base::reference              reference;
    typedef typename _Base::const_reference        const_reference;
    typedef typename _Base::pointer                pointer;
    typedef typename _Base::const_pointer          const_pointer;
    typedef typename _Base::size_type              size_type;
    typedef typename _Base::difference_type        difference_type;
    typedef typename _Base::iterator               iterator;
    typedef typename _Base::const_iterator         const_iterator;
    typedef typename _Base::const_reverse_iterator const_reverse_iterator;
    typedef typename _Base::reverse_iterator       reverse_iterator;
    
    /*   Default constructor  */
    dynmultiarray() : _M_data(_Size) {}
    
    /*   Destructor  */
    ~dynmultiarray() = default;
    
    /*   Create with uniform initialization   */
    explicit dynmultiarray(const _Tp& c)
    : _M_data(_Size, c) {}
    
    /*   Construct using initializer list   */
    dynmultiarray(std::initializer_list<_Tp> il)
    : _M_data(_Size) {
      unsigned int n = std::min((unsigned)(il.end()-il.begin()), _Size);
      std::move(il.begin(), il.begin()+n, _M_data.begin());
    }
    
    /*   Copy   */
    dynmultiarray(const dynmultiarray&) = default;
    dynmultiarray& operator=(const dynmultiarray&) = default;
    dynmultiarray& operator=(std::initializer_list<_Tp> il) { _M_data = il; return *this;}
    
    /*   Move   */
    dynmultiarray(dynmultiarray&&) = default;
    dynmultiarray& operator=(dynmultiarray&&) = default;
    
    
    /*   Iterators   */
    iterator begin() { return _M_data.begin();}
    iterator end() { return _M_data.end(); }
    
    const_iterator begin() const { return _M_data.begin();}
    const_iterator end() const { return _M_data.end();}
    
    const_iterator cbegin() const { return _M_data.begin();}
    const_iterator cend() const { return _M_data.end();}
    
    reverse_iterator rbegin() { return reverse_iterator(end());}
    reverse_iterator rend() { return reverse_iterator(begin());}
    
    const_reverse_iterator rbegin()  const { return reverse_iterator(end());}
    const_reverse_iterator rend() const { return reverse_iterator(begin());}
    
    /*  Capacity: Standard operators   */
    size_type size() const { return _Size;}
    bool empty() const { return _M_data.empty();}
    
    /*  Number of the rows and the element.  */
    static const size_type nrow  = _Size;
    static const size_type nelem = _Size;
    
    /*   Element access   */
    reference       operator[](size_type i) { return _M_data[i];}
    const_reference operator[](size_type i) const { return _M_data[i];}
    
    reference       front()       { return _M_data.front(); }
    const_reference front() const { return _M_data.front(); }
    reference       back()        { return _M_data.back(); }
    const_reference back()  const { return _M_data.back(); }
    
    const_reference at(size_type n) const { return _M_data.at(n); }
    reference       at(size_type n)       { return _M_data.at(n); }
    
  private:
    _Base _M_data;
  };
  
  
  template<typename _Tp, unsigned _Size, unsigned..._Rest>
  class dynmultiarray
  {
    /*   Private types  */
    typedef dynmultiarray<_Tp, _Rest...> _Row;
    typedef dynmultiarray<_Row, _Size>   _Base;
    
  public:
    /*   Public types  */
    typedef typename _Row::value_type             value_type;
    typedef typename _Row::reference              reference;
    typedef typename _Row::const_reference        const_reference;
    typedef typename _Row::pointer                pointer;
    typedef typename _Row::const_pointer          const_pointer;
    typedef typename _Row::size_type              size_type;
    typedef typename _Row::difference_type        difference_type;
    
    /*   Iterators   */
    struct iterator
    {
      /*   some types to be compatible with the standard   */
      typedef typename _Row::value_type         value_type;
      typedef typename _Row::reference          reference;
      typedef typename _Row::pointer            pointer;
      typedef typename _Row::difference_type    difference_type;
      typedef std::random_access_iterator_tag   iterator_category;
      
      /*   Increment operators  */
      iterator& operator++() { ++idx; return *this;}
      iterator operator++(int) { return {array, idx++};}
      
      /*  Decrement operators   */
      iterator& operator--() { --idx; return *this;}
      iterator operator--(int) { return {array, idx--};}
      
      /*  Shift operators  */
      iterator& operator+=(difference_type n) {idx += n; return *this;}
      iterator& operator-=(difference_type n) {idx -= n; return *this;}
      iterator  operator+ (difference_type n) const { return {array, idx+n};}
      iterator  operator- (difference_type n) const { return {array, idx-n};}
      
      /*   Distance   */
      difference_type operator-(const iterator& a) const { return idx-a.idx;}
      
      /*   Dereference operators   */
      reference operator*() const { return *((array+idx/_Row::nelem)->begin() + idx%_Row::nelem);}
      reference operator[](difference_type n) const { return *((*this)+n);}
      pointer operator->() const { return &(*(*this));}
      
      /*    Comparison operators   */
      bool operator==(const iterator& x) const { return (array == x.array && idx == x.idx);}
      bool operator!=(const iterator& x) const { return !(this -> operator==(x));}
      bool operator> (const iterator& x) const { return (idx > x.idx);}
      bool operator< (const iterator& x) const { return (idx < x.idx);}
      bool operator>=(const iterator& x) const { return (idx >= x.idx);}
      bool operator<=(const iterator& x) const { return (idx <= x.idx);}
      
      typename _Base::iterator array;
      difference_type idx;
    };
    
    struct const_iterator
    {
      /*   some types to be compatible with the standard   */
      typedef typename _Row::value_type         value_type;
      typedef typename _Row::const_reference    reference;
      typedef typename _Row::const_pointer      pointer;
      typedef typename _Row::difference_type    difference_type;
      typedef std::random_access_iterator_tag   iterator_category;
      
      /*   Increment operators  */
      const_iterator& operator++() { ++idx; return *this;}
      const_iterator operator++(int) { return {array, idx++};}
      
      /*  Decrement operators   */
      const_iterator& operator--() { --idx; return *this;}
      const_iterator operator--(int) { return {array, idx--};}
      
      /*  Shift operators  */
      const_iterator& operator+=(difference_type n) {idx += n; return *this;}
      const_iterator& operator-=(difference_type n) {idx -= n; return *this;}
      const_iterator  operator+ (difference_type n) const { return {array, idx+n};}
      const_iterator  operator- (difference_type n) const { return {array, idx-n};}
      
      /*   Distance   */
      difference_type operator-(const const_iterator& a) const { return idx-a.idx;}
      
      /*   Dereference operators   */
      const_reference operator*() const { return *((array+idx/_Row::nelem)->begin() + idx%_Row::nelem);}
      const_reference operator[](difference_type n) const { return *((*this)+n);}
      const_pointer operator->() const { return &(*(*this));}
      
      /*    Comparison operators   */
      bool operator==(const const_iterator& x) const { return (array == x.array && idx == x.idx);}
      bool operator!=(const const_iterator& x) const { return !(this->operator==(x));}
      bool operator> (const const_iterator& x) const { return (idx > x.idx);}
      bool operator< (const const_iterator& x) const { return (idx < x.idx);}
      bool operator>=(const const_iterator& x) const { return (idx >= x.idx);}
      bool operator<=(const const_iterator& x) const { return (idx <= x.idx);}
      
      typename _Base::const_iterator array;
      difference_type idx;
    };
    
    /*   Reverse iterators  */
    typedef std::reverse_iterator<iterator>       reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    
  public:
    /*   Default constructor  */
    dynmultiarray() = default;
    
    /*   Destructor  */
    ~dynmultiarray() = default;
    
    /*   Create with uniform initialization   */
    explicit dynmultiarray(const _Tp& c)
    : _M_data(_Row(c)) {}
    
    /*   Construct using initializer list   */
    dynmultiarray(std::initializer_list<_Row> il)
    : _M_data() {
      unsigned int n = std::min((unsigned)(il.end()-il.begin()), _Size);
      std::move(il.begin(), il.begin()+n, _M_data.begin());
    }
    
    /*   Copy   */
    dynmultiarray(const dynmultiarray&) = default;
    dynmultiarray& operator=(const dynmultiarray&) = default;
    dynmultiarray& operator=(std::initializer_list<_Tp> il) { _M_data = il; return *this;}
    
    /*   Move   */
    dynmultiarray(dynmultiarray&&) = default;
    dynmultiarray& operator=(dynmultiarray&&) = default;
    
    /*   Iterators   */
    iterator begin() { return {_M_data.begin(), 0U};}
    iterator end() { return {_M_data.begin(), nelem};}
    
    const_iterator begin() const { return {_M_data.begin(), 0U};}
    const_iterator end() const { return {_M_data.begin(), nelem};}
    
    const_iterator cbegin() const { return {_M_data.begin(), 0U};}
    const_iterator cend() const { return {_M_data.begin(), nelem};}
    
    reverse_iterator rbegin() { return reverse_iterator(end());}
    reverse_iterator rend() { return reverse_iterator(begin());}
    
    const_reverse_iterator rbegin()  const { return reverse_iterator(end());}
    const_reverse_iterator rend() const { return reverse_iterator(begin());}
    
    /*  Capacity: Standard operators  */
    size_type size() const { return nelem;}
    bool empty() const { return _M_data.empty();}
    
    /*  Number of the rows and the element.  */
    static const size_type nrow  = _Size;
    static const size_type nelem = _Size*_Row::nelem;
    
    /*   Element access   */
    _Row&       operator[](size_type i) { return _M_data[i];}
    const _Row& operator[](size_type i) const { return _M_data[i];}
    
    reference       front()       { return *begin();}
    reference       back()        { return *(end()-1);}
    const_reference front() const { return *begin();}
    const_reference back()  const { return *(end()-1);}
    
    const_reference at(size_type n) const { return *(begin()+n);}
    reference       at(size_type n)       { return *(begin()+n);}
    
    /*   We might want more features to handle rows....   */
    
  private:
    _Base _M_data;
  };
  
  
  template<typename _Tp, unsigned _Size, unsigned..._Rest>
  inline typename dynmultiarray<_Tp, _Size, _Rest...>::iterator
  operator+(typename dynmultiarray<_Tp, _Size, _Rest...>::difference_type n,
            const typename dynmultiarray<_Tp, _Size, _Rest...>::iterator& a) { return a+n;}
  
  
  template<typename _Tp, unsigned _Size, unsigned..._Rest>
  inline typename dynmultiarray<_Tp, _Size, _Rest...>::const_iterator
  operator+(typename dynmultiarray<_Tp, _Size, _Rest...>::difference_type n,
            const typename dynmultiarray<_Tp, _Size, _Rest...>::const_iterator& a) { return a+n;}
  
}

#endif
