/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_sample_statistics_h__
#define __deductor_hep_sample_statistics_h__ 1

//   std includes
#include <cmath>
#include <array>
#include <iostream>

//   project includes


namespace duct {
  
  //    template class sample_statistic
  template<typename _Tp>
  struct sample_statistic
  {
    //    types
    typedef _Tp sample_type;
    
    /*   computed assignments   */
    sample_statistic& operator+=(const sample_statistic& x) {
      number_of_samples += x.number_of_samples;
      sum_samples += x.sum_samples;
      sum_samples_sqr += x.sum_samples_sqr;
      return *this;
    }
    
    /*   accumulate weights   */
    sample_statistic& accumulate(const sample_type&);
    
    /*    add zero sample   */
    sample_statistic& accumulate() { ++number_of_samples; return *this;}
    
    /*    re-initialize    */
    void reset() { number_of_samples = 0ULL; sum_samples = sum_samples_sqr = 0.0;}
    
    /*   mean of values   */
    _Tp mean() const { return number_of_samples ? sum_samples/number_of_samples : 0.0;}
    _Tp mean2() const { return number_of_samples ? sum_samples_sqr/number_of_samples : 0.0;}
    
    /*    variance    */
    _Tp variance() const {
      _Tp m = mean();
      return mean2() - m*m;
    }
    
    /*   estimated error   */
    _Tp error() const {
      using namespace std;
      return sqrt(variance()/number_of_samples);
    }
    
    /*   Public data memebers, it might be not that safe but the code is
     *   much simpler this way.
     */
    unsigned long long int number_of_samples = 0ULL;
    _Tp sum_samples = 0.0, sum_samples_sqr = 0.0;

    /*   Static members  (this is useful in the class distribition)  */
    static void zero(sample_type& x) { x = 0.0;}
    static void assadd(sample_type& ths, const sample_type& x) { ths += x;}
    
    template<typename _Xp>
    static sample_type mul(const sample_type& ths, const _Xp& x) { return ths*x;}
  };
  
  
  template<typename _Tp>
  sample_statistic<_Tp>& sample_statistic<_Tp>::accumulate(const sample_type& x)
  {
    using namespace std;
    if(isfinite(x) == false) return *this;
    
    sum_samples += x;
    sum_samples_sqr += x*x;
    ++number_of_samples;
    
    return *this;
  }
  
  
  template<typename _Tp>
  inline sample_statistic<_Tp>
  operator+(const sample_statistic<_Tp> a, const sample_statistic<_Tp> b) {
    return sample_statistic<_Tp>(a) += b;
  }
  
  
  //        I/O operations
  template<typename _Tp> inline
  std::ostream& write_txt(std::ostream& os, const sample_statistic<_Tp>& x) {
    os<<x.number_of_samples<<"  "<<x.sum_samples<<"  "<<x.sum_samples_sqr;
    return os;
  }
  
  template<typename _Tp> inline
  std::istream& read_txt(std::istream& is, sample_statistic<_Tp>& x) {
    is>>x.number_of_samples>>x.sum_samples>>x.sum_samples_sqr;
    return is;
  }
  
  template<typename _Tp>
  std::ostream& write_bin(std::ostream& os, const sample_statistic<_Tp>& x)
  {
    os.write((const char *) &(x.number_of_samples), sizeof(unsigned long long int));
    os.write((const char *) &(x.sum_samples), sizeof(_Tp));
    os.write((const char *) &(x.sum_samples_sqr), sizeof(_Tp));
    return os;
  }
  
  template<typename _Tp>
  std::istream& read_bin(std::istream& is, sample_statistic<_Tp>& x)
  {
    is.read((char *) &(x.number_of_samples), sizeof(unsigned long long int));
    is.read((char *) &(x.sum_samples), sizeof(_Tp));
    is.read((char *) &(x.sum_samples_sqr), sizeof(_Tp));
    return is;
  }
 
}   //  namespace duct



namespace duct {
  
  //    template class sample_statistic
  template<typename _Tp, std::size_t _Dim>
  struct sample_statistic<std::array<_Tp, _Dim>>
  {
    //    types
    typedef std::array<_Tp, _Dim> sample_type;
    
    /*   default constructor   */
    sample_statistic() { this->reset();}
    
    /*   computed assignments   */
    sample_statistic& operator+=(const sample_statistic&);
    
    /*   accumulate weights   */
    sample_statistic& accumulate(const sample_type&);
    
    /*    add zero sample   */
    sample_statistic& accumulate() { ++number_of_samples; return *this;}
    
    /*    re-initialize    */
    void reset();
    
    /*   mean of values   */
    _Tp mean(unsigned int i) const {
      return number_of_samples ? sum_samples[i]/number_of_samples : 0.0;
    }
  
    _Tp mean2(unsigned int r, unsigned int c) const {
      return number_of_samples ? sum_samples_sqr[r >= c ? (r*(r+1))/2 + c : (c*(c+1))/2 + r]/number_of_samples : 0.0;
    }
    
    /*    variance    */
    _Tp variance(unsigned int r) const {
      _Tp m = mean(r);
      return mean2(r,r) - m*m;
    }
    
   _Tp cov(unsigned int r, unsigned int c) const {
      return mean2(r,c) - mean(r)*mean(c);
    }
    
    /*   estimated error   */
    _Tp error(unsigned int r) const {
      using namespace std;
      return sqrt(variance(r)/number_of_samples);
    }
    
    /*   Public data memebers, it might be not that safe but the code is
     *   much simpler this way.
     */
    unsigned long long int number_of_samples;
    
    /*   This is the sum of samples   */
    std::array<_Tp, _Dim> sum_samples;
    
    /*   This is sum of samples products (Ai * Aj). 
     *   Symmetric matrix stored in an array.
     */
    std::array<_Tp, (_Dim*(_Dim+1))/2> sum_samples_sqr;
    
    /*   Static members  (this is useful in the class distribition)  */
    static void zero(sample_type& x) {
      for(std::size_t r = 0; r < _Dim; r++) x[r] = 0.0;
    }
    
    static void assadd(sample_type& ths, const sample_type& x) {
      for(std::size_t r = 0; r < _Dim; r++) ths[r] += x[r];
    }

    template<typename _Xp>
    static sample_type mul(const sample_type& ths, const _Xp& x) {
      sample_type ret;
      for(std::size_t r = 0; r < _Dim; r++) ret[r] = ths[r]*x;
      return ret;
    }
  };
  
  
  template<typename _Tp, std::size_t _Dim>
  void sample_statistic<std::array<_Tp, _Dim>>::reset()
  {
    number_of_samples = 0ULL;
    
    for(std::size_t k = 0; k < _Dim; k++)
      sum_samples[k] = 0.0;
    
    for(std::size_t k = 0; k < (_Dim*(_Dim+1))/2; k++)
      sum_samples_sqr[k] = 0.0;
  }
  

  template<typename _Tp, std::size_t _Dim>
  sample_statistic<std::array<_Tp, _Dim>>&
  sample_statistic<std::array<_Tp, _Dim>>::operator+=(const sample_statistic<std::array<_Tp, _Dim>>& x)
  {
    number_of_samples += x.number_of_samples;
    
    for(std::size_t k = 0; k < _Dim; k++)
      sum_samples[k] += x.sum_samples[k];
    
    for(std::size_t k = 0; k < (_Dim*(_Dim+1))/2; k++)
      sum_samples_sqr[k] += x.sum_samples_sqr[k];
    
    return *this;
  }

  
  template<typename _Tp, std::size_t _Dim>
  sample_statistic<std::array<_Tp, _Dim>>&
  sample_statistic<std::array<_Tp, _Dim>>::accumulate(const sample_type& x)
  {
    using namespace std;
    for(std::size_t k = 0; k < _Dim; k++)
      if(isfinite(x[k]) == false) return *this;
    
    for(std::size_t r = 0; r < _Dim; r++) {
      sum_samples[r] += x[r];
      
      for(std::size_t c = 0; c <= r; c++)
        sum_samples_sqr[(r*(r+1))/2 + c] += x[r]*x[c];
    }
    
    ++number_of_samples;
    
    return *this;
  }
  
  
  //        I/O operations
  template<typename _Tp, std::size_t _Dim>
  std::ostream& write_txt(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x)
  {
    os<<x.number_of_samples;
    for(const _Tp& e : x.sum_samples) os<<"  "<<e;
    for(const _Tp& e : x.sum_samples_sqr) os<<"  "<<e;
    return os;
  }
  
  template<typename _Tp, std::size_t _Dim>
  std::istream& read_txt(std::istream& is, sample_statistic<std::array<_Tp, _Dim>>& x)
  {
    is>>x.number_of_samples;
    for(_Tp& e : x.sum_samples) is>>e;
    for(_Tp& e : x.sum_samples_sqr) is>>e;
    return is;
  }
  
  template<typename _Tp, std::size_t _Dim>
  std::ostream& write_bin(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x)
  {
    os.write((const char *) &(x.number_of_samples), sizeof(unsigned long long int));
    os.write((const char *) x.sum_samples.data(), _Dim*sizeof(_Tp));
    os.write((const char *) x.sum_samples_sqr.data(), ((_Dim*(_Dim+1))/2)*sizeof(_Tp));
    return os;
  }
  
  template<typename _Tp, std::size_t _Dim>
  std::istream& read_bin(std::istream& is, sample_statistic<std::array<_Tp, _Dim>>& x)
  {
    is.read((char *) &(x.number_of_samples), sizeof(unsigned long long int));
    is.read((char *) x.sum_samples.data(), _Dim*sizeof(_Tp));
    is.read((char *) x.sum_samples_sqr.data(), ((_Dim*(_Dim+1))/2)*sizeof(_Tp));
    return is;
  }
  
  
}   //  namespace duct



/*   Same helper function to do fancy printing   */
#include <string>
#include <sstream>

namespace duct {
  
  template<typename _Tp>
  struct sample_traits
  {
    /*   printing the result   */
    static void print(std::ostream& os, const sample_statistic<_Tp>& x) {
      os<<x.mean()<<"  "<<x.error();
    }
    
    /*   printing the result with normalization   */
    static void print(std::ostream& os, const sample_statistic<_Tp>& x, const sample_statistic<_Tp>& n) {
      _Tp norm = n.mean(), m = x.mean(), e = x.error();
      os<<m/norm<<"  "<<e/norm<<"   "<<m<<"  "<<e;
    }
    
    /*   meaning of printed result   */
    static std::string descrption(bool isnorm = false) {
      return (isnorm ? "val1  err1   val2  err2" : "val1  err1");
    }


    /*   printing the val+err   */
    static void print_plus(std::ostream& os, const sample_statistic<_Tp>& x) {
      os<<x.mean() + x.error();
    }
    
    /*   printing the val-err   */
    static void print_minus(std::ostream& os, const sample_statistic<_Tp>& x) {
      os<<x.mean() - x.error();
    }

    /*   printing the result with normalization, val+err  */
    static void print_plus(std::ostream& os, const sample_statistic<_Tp>& x, const sample_statistic<_Tp>& n) {
      _Tp norm = n.mean(), m = x.mean(), e = x.error();
      os<<(m + e)/norm<<"   "<<m + e;
    }

    /*   printing the result with normalization, val+err  */
    static void print_minus(std::ostream& os, const sample_statistic<_Tp>& x, const sample_statistic<_Tp>& n) {
      _Tp norm = n.mean(), m = x.mean(), e = x.error();
      os<<(m - e)/norm<<"   "<<m - e;
    }
    
    /*   meaning of printed result   */
    static std::string descrption_err(bool isnorm = false) {
      return (isnorm ? "val1  val2" : "val1");
    }
  };
  
  
  template<typename _Tp, std::size_t _Dim>
  struct sample_traits<std::array<_Tp, _Dim>>
  {
    /*   printing the result   */
    static void print(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x) {
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k)<<" "<<x.error(k);
    }
    
    /*   printing the result with normalization */
    static void print(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm<<" "<<x.error(k)/norm;
      }
      
      /*   non normalized print  */
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k)<<" "<<x.error(k);
    }
    
    /*   meaning of printed result   */
    static std::string descrption(bool isnorm = false)
    {
      std::ostringstream ss("val1  err1", std::ios_base::ate);
      
      for(std::size_t k = 2; k <= _Dim; k++)
        ss<<"   val"<<k<<"  err"<<k;
      
      if(isnorm) {
        for(std::size_t k = _Dim+1; k <= 2*_Dim; k++)
          ss<<"   val"<<k<<"  err"<<k;
      }
      
      return ss.str();
    }
    
    /*   printing the val + err   */
    static void print_plus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x) {
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k) + x.error(k);
    }
    
    /*   printing the result with normalization, val+err */
    static void print_plus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm + x.error(k)/norm;
      }
      
      /*   non normalized print  */
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k) + x.error(k);
    }

    /*   printing the val - err   */
    static void print_minus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x) {
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k) - x.error(k);
    }
    
    /*   printing the result with normalization, val-err */
    static void print_minus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm - x.error(k)/norm;
      }
      
      /*   non normalized print  */
      for(std::size_t k = 0; k < _Dim; k++) os<<"   "<<x.mean(k) - x.error(k);
    }

    /*   meaning of printed result   */
    static std::string descrption_err(bool isnorm = false)
    {
      std::ostringstream ss("val1", std::ios_base::ate);
      
      for(std::size_t k = 2; k <= _Dim; k++)
        ss<<"   val"<<k;
      
      if(isnorm) {
        for(std::size_t k = _Dim+1; k <= 2*_Dim; k++)
          ss<<"   val"<<k;
      }
      
      return ss.str();
    }
  };
  
  
  template<typename _Tp>
  struct sample_ratio_traits;
  
  template<typename _Tp, std::size_t _Dim>
  struct sample_ratio_traits<std::array<_Tp, _Dim>>
  {
    /*   printing the result   */
    static void print(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean<<"  "<<mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k)<<" "<<x.error(k);
    }

    /*   printing the result with normalization */
    static void print(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean<<"  "<<mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm<<" "<<x.error(k)/norm;
      }
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k)<<" "<<x.error(k);
    }
    
    /*   meaning of printed result   */
    static std::string descrption(bool isnorm = false) {
      std::ostringstream ss("val1  err1",std::ios_base::ate);
      
      for(std::size_t k = 2; k <= _Dim+1; k++)
        ss<<"   val"<<k<<"  err"<<k;
        
        if(isnorm) {
          for(std::size_t k = _Dim+2; k <= 2*_Dim+1; k++)
            ss<<"   val"<<k<<"  err"<<k;
        }
      
      return ss.str();
    }
  
    
    /*   printing the val + err   */
    static void print_plus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean + mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k) + x.error(k);
    }

    /*   printing the result with normalization, val+err */
    static void print_plus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean + mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm + x.error(k)/norm;
      }
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k) + x.error(k);
    }
    
    /*   printing the val-err   */
    static void print_minus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean - mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k) - x.error(k);
    }

    /*   printing the result with normalization, val+err */
    static void print_minus(std::ostream& os, const sample_statistic<std::array<_Tp, _Dim>>& x, const sample_statistic<std::array<_Tp, _Dim>>& n)
    {
      _Tp m1 = x.mean(0), m2 = x.mean(1);
      _Tp v1 = x.variance(0), v2 = x.variance(1);
      _Tp c12 = x.cov(0,1);
      
      _Tp var = v1/(m1*m1) + v2/(m2*m2)-2.0*c12/(m1*m2);
      _Tp mean = m1/m2;
      
      os<<"  "<<mean - mean*sqrt(var/x.number_of_samples);
      
      for(std::size_t k = 0; k < _Dim; k++) {
        _Tp norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm - x.error(k)/norm;
      }
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k) - x.error(k);
    }
 
    /*   meaning of printed result   */
    static std::string descrption_err(bool isnorm = false) {
      std::ostringstream ss("val1",std::ios_base::ate);
      
      for(std::size_t k = 2; k <= _Dim+1; k++)
        ss<<"  val"<<k;
      
      if(isnorm) {
        for(std::size_t k = _Dim+2; k <= 2*_Dim+1; k++)
          ss<<"  val"<<k;
      }
      
      return ss.str();
    }
  };
  
}

#endif
