/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 6/10/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __hep_void_ptr_h__
#define __hep_void_ptr_h__ 1

//   std includes

//   project headers


namespace duct {
  
  
  template<typename _Tp>
  struct void_ptr_deleter {
    typedef void (*deleter_func_t)(void*);
    
    /*   convertable to void (*)(void*)   */
    operator deleter_func_t (){
      return deleter;
    }
    
    /*   this is the deleter function  */
    static void deleter(void *p) {
      if(p) delete reinterpret_cast<_Tp*>(p);
    }
  };
  
  
  template<typename _Tp>
  struct void_ptr_deleter<_Tp[]> {
    typedef void (*deleter_func_t)(void*);
    
    /*   convertable to void (*)(void*)   */
    operator deleter_func_t (){
      return deleter;
    }
    
    /*   this is the deleter function  */
    static void deleter(void *p) {
      if(p) delete [] reinterpret_cast<_Tp*>(p);
    }
  };
  
  
  
  class void_ptr
  {
  public:
    /*   default constructor   */
    constexpr void_ptr() : _M_ptr(nullptr), _M_del(nullptr) {}
    void_ptr(std::nullptr_t) : void_ptr() { }
    
    /*   creating from pointer   */
    template<typename _Tp>
    explicit void_ptr(_Tp *p) : _M_ptr(p), _M_del(void_ptr_deleter<_Tp>()) {}
    
    /*   creating from pointer with specialized deleter function   */
    template<typename _Tp>
    explicit void_ptr(_Tp *p, void (*del)(void*)) : _M_ptr(p), _M_del(del) {}
    
    /*   move constructor  */
    void_ptr(void_ptr&& p) : _M_ptr(p._M_ptr), _M_del(p._M_del) { p._M_ptr = nullptr;}
    
    /*   destructor  */
    ~void_ptr() { if(_M_ptr) (*_M_del)(_M_ptr); _M_ptr = nullptr;}
    
    /*   disable copy  */
    void_ptr(const void_ptr&) = delete;
    void_ptr& operator=(const void_ptr&) = delete;
    
    /*   assignments   */
    void_ptr& operator=(void_ptr&& u) {
      _M_ptr = u._M_ptr; _M_del = u._M_del; u._M_ptr = nullptr;
      return *this;
    }
    
    void_ptr& operator=(std::nullptr_t) {
      this->~void_ptr();
      return *this;
    }
    
    /*   Release the pointer, void_ptr is not responsible for it anymore   */
    void * release() {
      void *ret = _M_ptr;
      _M_ptr = nullptr;
      return ret;
    }
    
    /*   resetting the pointer   */
    void reset(void *p = nullptr) {
      if(_M_ptr) (*_M_del)(_M_ptr);
      _M_ptr = p;
    }
    
    template<typename _Tp>
    void reset(_Tp *p = nullptr, void (*del)(void*) = void_ptr_deleter<_Tp>()) {
      if(_M_ptr) (*_M_del)(_M_ptr);
      _M_ptr = reinterpret_cast<void*>(p);
      _M_del = del;
    }
    
    /*   swapping  pointers   */
    void swap(void_ptr& u) {
      std::swap(_M_ptr, u._M_ptr);
      std::swap(_M_del, u._M_del);
    }
    
    /*   consversions   */
    void * get() const { return _M_ptr;}
    explicit operator void *() const { return _M_ptr;}
    explicit operator bool() const { return _M_ptr == nullptr ? false : true;}
    
    template<typename _Tp>
    explicit operator _Tp* () const { return reinterpret_cast<_Tp *>(_M_ptr);}
    
  private:
    /*   pointer and it is deleter  */
    void *_M_ptr;
    void (*_M_del)(void *);
  };
  

  inline void swap(void_ptr& a, void_ptr& b) {
    a.swap(b);
  }
  
  
}

#endif
