/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 5/22/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_imp_sampling_h__
#define __math_imp_sampling_h__ 1

//   std includes

//   project headers


namespace duct {
 
  
  template<unsigned _Np>
  class importance_sampling
  {
  public:
    /*   This consturctor initializes the grid   */
    explicit importance_sampling(double xl = 0.0, double xh = 1.0)
    : _M_dx((xh-xl)/_Np)
    {
      for(unsigned k = 0; k <= _Np; k++)
        _M_grid[k].x  = xl + k*_M_dx;
    }
    
    /*  constructing using functor  */
    template<class _Func>
    importance_sampling(const _Func& func, double xl, double xh) {
      this->assign(func, xl, xh);
    }
    
    /*    Assignments   */
    template<class _Func>
    importance_sampling& assign(const _Func&);
    
    template<class _Func>
    importance_sampling& assign(const _Func&, double, double);
    
    template<class _Func>
    importance_sampling& operator=(const _Func& func) {
      return this->assign(func);
    }
    
    /*   mapping form the old (x) to the new variable (chi)  */
    double mapping_x2chi(double x) const {
      return this->integral(x)/_M_grid[_Np].ifunc;
    }
    
    /*   mapping form the new variable to the old variable   */
    double mapping_chi2x(double chi) const {
      return this->integral_inverse(chi*_M_grid[_Np].ifunc);
    }
    
    /*   Jacobian of this mapping   */
    double jacobian(double x) const {
      return _M_grid[_Np].ifunc/this->operator()(x);
    }
    
    /*   Interpolating the distribution function   */
    double operator()(double) const;
    
    /*   Commulative integral   */
    double integral(double) const;
    
    /*   Inverse of the commulative function   */
    double integral_inverse(double) const;
    
    /*   Integral of the distribution function  */
    double integral() const {
      return _M_grid[_Np].ifunc;
    }
    
  private:
    /*   distance between two grid points   */
    double _M_dx;
    
    /*   this defines the grid   */
    struct __gridpoint {
      double x, func, ifunc;
    } _M_grid[_Np+1];
  };
  
  
  template<unsigned int _Np> template<class _Func>
  importance_sampling<_Np>& importance_sampling<_Np>::assign(const _Func& func)
  {
    /*   calculate the function in the gridpoints  */
    for(unsigned k = 0; k <= _Np; k++) {
      _M_grid[k].func = func(_M_grid[k].x);
    }
    
    /*    calculating the integral function in the gridpoints   */
    _M_grid[0].ifunc = 0.0;
    for(unsigned k = 0; k < _Np; k++)
      _M_grid[k+1].ifunc = _M_grid[k].ifunc + 0.5*_M_dx*(_M_grid[k].func+_M_grid[k+1].func);
    
    return *this;
  }
  
  
  template<unsigned int _Np> template<class _Func>
  importance_sampling<_Np>& importance_sampling<_Np>::assign(const _Func& func, double xmin, double xmax)
  {
    /*   calculate the function in the gridpoints  */
    _M_dx = (xmax-xmin)/_Np;
    
    for(unsigned k = 0; k <= _Np; k++) {
      _M_grid[k].x = xmin + k*_M_dx;
      _M_grid[k].func = func(_M_grid[k].x);
    }
    
    /*    calculating the integral function in the gridpoints   */
    _M_grid[0].ifunc = 0.0;
    for(unsigned k = 0; k < _Np; k++)
      _M_grid[k+1].ifunc = _M_grid[k].ifunc + 0.5*_M_dx*(_M_grid[k].func+_M_grid[k+1].func);
    
    return *this;
  }
  
  
  template<unsigned int _Np>
  double importance_sampling<_Np>::operator()(double x) const
  {
    int k = static_cast<int>((x-_M_grid[0].x)/_M_dx);
        
    /*   extrapolating by freezing the basepoints  */
    if(k < 0) return _M_grid[0].func;
    if(k >= (int) _Np) return _M_grid[_Np].func;

    /*   interpolating the function  */
    const __gridpoint& g1 = _M_grid[k];
    const __gridpoint& g2 = _M_grid[k+1];
      
    return (g1.func*(g2.x-x) + g2.func*(x-g1.x))/_M_dx;
  }
  
  
  template<unsigned int _Np>
  double importance_sampling<_Np>::integral(double x) const
  {
    int k = static_cast<int>((x-_M_grid[0].x)/_M_dx);
    
    /*  extrapolating by freezing the basepoints  */
    if(k < 0) return 0.0;
    if(k >= (int) _Np) return _M_grid[_Np].ifunc + _M_grid[_Np].func*(x-_M_grid[_Np].x);
    
    /*   interpolating the function  */
    const __gridpoint& g1 = _M_grid[k];
    const __gridpoint& g2 = _M_grid[k+1];
    
    double da = (g2.func-g1.func);
    double ca = (g1.func*g2.x-g2.func*g1.x);
    
    return g1.ifunc + (x-g1.x)*(0.5*da*(x+g1.x) + ca)/_M_dx;
  }
  
  
  template<unsigned int _Np>
  double importance_sampling<_Np>::integral_inverse(double F) const
  {
    /*   if it is out of range return something reasonable  */
    if(F <= 0.0) return _M_grid[0].x;
    if(F >= _M_grid[_Np].ifunc) return _M_grid[_Np].x + (F-_M_grid[_Np].ifunc)/_M_grid[_Np].func;
    
    /*   findig the gridpoint  */
    int lo = 0, hi = _Np, mi;
    
    while(lo != (mi = (lo+hi)/2)) {
      if(_M_grid[mi].ifunc <= F) lo = mi;
      else hi = mi;
    }
    
    /*   interpolation the inverse function   */
    const __gridpoint& g1 = _M_grid[lo];
    double d = (_M_grid[lo+1].func-g1.func)/_M_dx;
    
    return g1.x + (std::sqrt(g1.func*g1.func+2.0*d*(F-g1.ifunc))-g1.func)/d;
  }
  
}

#endif
