/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/29/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_mc_basic_user_h__
#define __deductor_mc_basic_user_h__ 1

//   std includes
#include <typeinfo>
#include <iostream>
#include <fstream>
#include <functional>
#include <ctime>
#include <chrono>
#include <iomanip>
#include <tuple>

//   project headers
#include <bits/mc-user.h>
#include <bits/hep-distbook.h>
#include "bits/nlo-c-support.h"




namespace duct {
  
  //
  //   Abstract class for defining basic_user types
  //
  template<class _Tag>
  class basic_user_base : public user<_Tag>
  {
  public:
    //   destructor
    virtual ~basic_user_base() {}
    
    //   set the output (file name saving mode)
    void param(const char *dir, const char *rname, const char *uname,
               unsigned long nsave = 10000UL, unsigned long long nevent = 0UL, bool txt = false)
    {
      _M_fname = dir;
      if(_M_fname.back() != '/') _M_fname += '/';
      _M_fname += rname;
      _M_fname += "-";
      _M_fname += uname;
      _M_fname += ".dat";
      
      _M_rname = rname;
      _M_uname = uname;
      
      _M_txt = txt;
      _M_nsave = nsave;
      user<_Tag>::max_number_of_events = nevent;
    }
  
    /*   finish function   */
    void finish() {
      std::cout<<"\nAnalyser \'"<<_M_uname<<"\' has finished its job and saved the results into file \'"<<_M_fname<<"\'.\n";
      std::cout<<"This file has "<<(_M_txt ? "text fomat but" : "binary format and")
      <<" contains raw data, so you should produce the (human readable) results by using function\n";
      std::cout<<"\'__basic_user_make_result<...>::collect_and_print(const char *, const char **)\'\n";
      std::cout<<"or a program based on this function, like \'deductor --add ...\'.\n"<<std::endl;
    }

    /*    Returns a function that can read the output of the user<...> class
     *    and sum them up and print the result into file(s) in a nice way.
     */
    virtual std::function<unsigned long long(const char *, const char **)>
    collect_and_print_result() const = 0;

    /*   retruns the name of the output file   */
    const std::string& name() const { return _M_uname;}
    
  protected:
    /*    default constructor is protected  */
    basic_user_base()
    : _M_txt(false), _M_nsave(10000UL), _M_fname("run-user.dat"), _M_rname("run"), _M_uname("user"), _M_start_time(std::time(0)) {}
    
  private:
    // operations at the end of the event and saving the results
    void operations_at_the_end_of_event();

    //   do any operetaions at the end of the event
    virtual void basic_user_end_of_event() = 0;
    
    //   write some information about the type of the sample
    virtual std::string get_typeinfo() const = 0;
    
    //   save the results in txt and binary mode
    virtual void write_txt(std::ostream&) = 0;
    virtual void write_bin(std::ostream&) = 0;
    
  private:
    /*    data members   */
    bool _M_txt;
    unsigned long _M_nsave;
    std::basic_string<char> _M_fname, _M_rname, _M_uname;
    std::time_t _M_start_time;
  };
  

  template<class _Tag>
  void basic_user_base<_Tag>::operations_at_the_end_of_event()
  {
    this -> basic_user_end_of_event();
    
    if(user<_Tag>::number_of_events() % _M_nsave == 0 || user<_Tag>::is_done()) {
      std::fstream ofile(_M_fname.c_str(), std::ios::out);
      
      /*   number of events   */
      unsigned long long int nevent = user<_Tag>::number_of_events();
      
      /*   first byte of the file tells if the stream is txt or binary   */
      ofile.put(_M_txt ? 'T' : 'B');
      
      if(_M_txt) {
        ofile<<"  "<<nevent<<"  ";
        ::duct::write_txt(ofile, get_typeinfo());
        
        ofile.setf(std::ios::scientific, std::ios::floatfield);
		ofile.precision(14);
        
		this -> write_txt(ofile);
      } else {
        ofile.write(reinterpret_cast<const char *>(&nevent), sizeof(unsigned long long int));
        ::duct::write_bin(ofile, get_typeinfo());
        this -> write_bin(ofile);
      }
      
      ofile.flush();
      ofile.close();
      
      std::time_t hh, mm, ss = std::time(0) - _M_start_time;
      
      hh  = ss/3600L;
      ss -= hh*3600L;
      mm  = ss/60L;
      ss -= mm*60L;
      
      std::cout<<"    Analyser \'"<<_M_uname<<"\' of run \'"<<_M_rname<<"\' : "<<user<_Tag>::number_of_events()<<" events in "
      <<(hh < 10 ? "0" : "")<<hh
      <<(mm < 10 ? ":0" : ":")<<mm
      <<(ss < 10 ? ":0" : ":")<<ss<<std::endl;
    }
  }
  

  
  /*
   *   Helper class to produce results 
   */
  template<class ..._Distbook>
  class __basic_user_make_result
  {
    template<class _Tuple, std::size_t N>
    struct unpacker
    {
      /*   ass_add (+=) operator */
      static void ass_add(_Tuple& r, const _Tuple& x) {
        unpacker<_Tuple, N-1>::ass_add(r, x);
        std::get<N-1>(r) += std::get<N-1>(x);
      }
      
      /*   formated read from a stream   */
      static void read_txt(std::istream& is, _Tuple& x) {
        unpacker<_Tuple, N-1>::read_txt(is, x);
        ::duct::read_txt(is, std::get<N-1>(x));
      }
      
      /*   binary read from a stream   */
      static void read_bin(std::istream& is, _Tuple& x) {
        unpacker<_Tuple, N-1>::read_bin(is, x);
        ::duct::read_bin(is, std::get<N-1>(x));
      }
      
      /*   print the reults into files   */
      static void print(unsigned long long nsaved, const char * dir, const _Tuple& x)
      {
        unpacker<_Tuple, N-1>::print(nsaved, dir, x);
        
        /*   printing every distbook into to separate files   */
        std::basic_string<char> fn(dir);
        
        /*   filename  */
        if(fn.back() != '/') fn += '/';
        fn += "result-";
        fn += std::to_string(N-1);
        
        /*  printing   */
        std::fstream f(fn.c_str(), std::ios::out);
        std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        
        f<<"#  This result was calculated by Deductor-1.0.0.\n"
        << "#  The number of collected MC events is "<<nsaved<<".\n"
        << "#  It was created on ";
        char mbstr[100];
       
        if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
          f<<mbstr;
      
        
        f.setf(std::ios::scientific, std::ios::floatfield);
        ::duct::print(f, std::get<N-1>(x));
        f.close();
      }
      
      /*   plot the result into directories  */
      static void plot(unsigned long long nsaved, const char * dir, const _Tuple& x)
      {
        unpacker<_Tuple, N-1>::plot(nsaved, dir, x);
        
        /*   check if there is any plot enabled  */
        if(std::get<N-1>(x).is_plot_enabled() == false) return;
        
        /*   plotting every distbook into to separate directory   */
        std::string fn(dir);
        
        /*   filename  */
        if(fn.back() != '/') fn += '/';
        fn += "plot-";
        fn += std::to_string(N-1);
        
        /*   make directory  */
        if(::__nlo_helper_mkdir(fn.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
          std::cerr<<"Unable to make directory : "<<fn<<std::endl;
        }

        /*  printing the info */
        std::fstream f((fn+"/info").c_str(), std::ios::out);
        std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        char mbstr[100];
       
        f<<"#  The results in this directory were calculated by Deductor-1.0.0.\n"
        << "#  The number of collected MC events is "<<nsaved<<".\n"
        << "#  It was created on ";
        
        if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n", std::localtime(&now)))
          f<<mbstr;

        f.close();
      
        /*   plotting the results   */
        ::duct::plot(std::move(fn), std::get<N-1>(x));
      }
    };
    
  public:
    /*   result type is a tuple   */
    typedef std::tuple<_Distbook...> result_type;
    
    /*   reading from stream   */
    static unsigned long long read(std::istream& is, result_type& x);
    
    /*   collect the files and print the result   */
    static unsigned long long collect_and_print(const char *, const char **);
  };
  
  
  template<class ..._Distbook> template<class _Tuple>
  struct __basic_user_make_result<_Distbook...>::unpacker<_Tuple, 1>
  {
    /*   ass_add (+=) operator */
    static void ass_add(_Tuple& r, const _Tuple& x) {
      std::get<0>(r) += std::get<0>(x);
    }
    
    /*   formated read from a stream   */
    static void read_txt(std::istream& is, _Tuple& x) {
      ::duct::read_txt(is, std::get<0>(x));
    }
    
    /*   binary read from a stream   */
    static void read_bin(std::istream& is, _Tuple& x) {
      ::duct::read_bin(is, std::get<0>(x));
    }
    
    /*   print the reults into files   */
    static void print(unsigned long long nsaved, const char * dir, const _Tuple& x)
    {
      /*   printing every distbook into to separate files   */
      std::basic_string<char> fn(dir);
      
      /*   filename  */
      if(fn.back() != '/') fn += '/';
      fn += "result-0";
      
      /*  printing   */
      std::fstream f(fn.c_str(), std::ios::out);
      std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
      char mbstr[100];
     
      f<<"#  This result was calculated by Deductor-1.0.0.\n"
      << "#  The number of collected MC events is "<<nsaved<<".\n"
      << "#  It was created on ";

      if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
        f<<mbstr;
      
      f.setf(std::ios::scientific, std::ios::floatfield);
      ::duct::print(f, std::get<0>(x));
      f.close();
    }
    
    /*   plot the result into directories  */
    static void plot(unsigned long long nsaved, const char * dir, const _Tuple& x)
    {
      /*   check if there is any plot enabled  */
      if(std::get<0>(x).is_plot_enabled() == false) return;
      
      /*   plotting every distbook into to separate directory   */
      std::string fn(dir);
      
      /*   filename  */
      if(fn.back() != '/') fn += '/';
      fn += "plot-0";
      
      /*   make directory  */
      if(::__nlo_helper_mkdir(fn.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
        std::cerr<<"Unable to make directory : "<<fn<<std::endl;
      }
      
      /*  printing the info */
      std::fstream f((fn+"/info").c_str(), std::ios::out);
      std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
      char mbstr[100];
      
      f<<"#  The results in this directory were calculated by Deductor-1.0.0.\n"
      << "#  The number of collected MC events is "<<nsaved<<".\n"
      << "#  It was created on ";
      
      if (std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n", std::localtime(&now)))
        f<<mbstr;
      
      f.close();
      
      /*   plotting the results   */
      ::duct::plot(std::move(fn), std::get<0>(x));
    }
  };
  
  
  template<class ..._Distbook>
  unsigned long long __basic_user_make_result<_Distbook...>::read(std::istream& is, result_type& x)
  {
    unsigned long long nsaved;
    
    /*   type of the saved file   */
    bool txt = (is.get() == 'T' ? true : false);
    std::string __typeinfo_name;
    
    /*   reading number of events and typeinfo   */
    if(txt) {
      is>>nsaved;
      read_txt(is, __typeinfo_name);
    }
    else {
      is.read(reinterpret_cast<char *>(&nsaved), sizeof(unsigned long long int));
      read_bin(is, __typeinfo_name);
    }
    
    /*   If there is any mismatch it gives a warning.    */
    if(typeid(__basic_user_make_result<_Distbook...>).name() != __typeinfo_name) {
      std::cerr
      <<"**** The type of the distbook(s) in the current stream is different than in the user function! "
      <<"Probably you are using different version or variant of the C++ complier.\n"
      <<"Please check your result! ****"<<std::endl;
    }
    
    /*   Reading the distbook  */
    if(txt) unpacker<result_type, sizeof...(_Distbook)>::read_txt(is, x);
    else unpacker<result_type, sizeof...(_Distbook)>::read_bin(is, x);
    
    return nsaved;
  }
  
  
  template<class ..._Distbook> unsigned long long int
  __basic_user_make_result<_Distbook...>::collect_and_print(const char *dir, const char **files)
  {
    std::fstream file;
    unsigned long int nsaved = 0UL;
    result_type tmp, *res = nullptr;
    
    for(const char **fi = files; *fi != nullptr; fi++) {
      //----- open the file ----
      file.open(*fi, std::ios::in);
      
      //----- read the file if it is exist -----
      if(file.is_open()) {
        nsaved += read(file, tmp);
        file.flush();
        
        if(!res) res = new result_type(tmp);
        else unpacker<result_type, sizeof...(_Distbook)>::ass_add(*res, tmp);
      }
      
      //----- close the file ----
      file.close();
    }
    
    if(!res) return 0;
    
    //----- save the results into the directory dir  -----
    unpacker<result_type, sizeof...(_Distbook)>::print(nsaved, dir, *res);
    unpacker<result_type, sizeof...(_Distbook)>::plot(nsaved, dir, *res);
    
    //---- cleaning up -----
    if(res) delete res;
    
    return nsaved;
  } 
  

  
  /*
   *      General basic_user class.
   */
  template<class _Tag, typename _Sample, typename _Point,
  class _Traits = sample_traits<_Sample>, class _PointTraits = distpoint::traits<_Point>>
  class basic_user : public virtual basic_user_base<_Tag>
  {
  public:
    //   public types
    typedef _Sample sample_type;
    typedef _Point  point_type;
    typedef _Traits sample_traits;
    typedef _PointTraits point_traits;
    typedef distbook<_Sample, _Point, _Traits, _PointTraits> distbook_type;
    
    //   destructor
    virtual ~basic_user() {}
 
    /*   result generator   */
    std::function<unsigned long long(const char *, const char **)> collect_and_print_result() const {
      return __basic_user_make_result<distbook_type>::collect_and_print;
    }
   
    /*   static version of result generator   */
    static unsigned long long collect_and_print(const char *dir, const char **files) {
      return __basic_user_make_result<distbook_type>::collect_and_print(dir, files);
    }
    
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title, unsigned n, const _Point *base, bool norm = false) {
      _M_book.create(key, title, n, base, norm);
    }
    
    template<class _InputIterator>
    void phys(int key, const char *title, _InputIterator first, _InputIterator last, bool norm = false) {
      _M_book.create(key, title, first, last, norm);
    }
    
    void phys(int key, const char *title, std::initializer_list<_Point> base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }

    template<class _Container>
    void phys(int key, const char *title, const _Container& base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }

    /*   collecting weight   */
    template<class _Func>
    void physfill(int key, _Func func) {
      _M_book.collect(key, func);
    }
    
    template<class _Func, typename ..._Arg>
    void physfill(int key, const sample_type& w, _Func func, _Arg... x) {
      _M_book.collect(key, w, func, x...);
    }
  
    /*   collects weights for the normalization   */
    void normfill(int key, const sample_type& w) {
      _M_book.collect_norm(key, w);
    }
    
    /*   enable plot  */
    void enable_plot(int key, const char *title = nullptr) { _M_book.enable_plot(key, title);}
    void add_figure(int key, const ltxfigure& fig) { _M_book.add_figure(key, fig);}
    
  private:
    /*   defining some virtual functions   */
    void basic_user_end_of_event() { _M_book.accumulate();}
    void write_txt(std::ostream& os) { os<<"  "; ::duct::write_txt(os, _M_book);}
    void write_bin(std::ostream& os) { ::duct::write_bin(os, _M_book);}
    std::string get_typeinfo() const { return typeid(__basic_user_make_result<distbook_type>).name();}

  private:
    //    data members
    distbook_type _M_book;

    /*   friend declaration   */
    template<class ..._Users> friend class basic_user_set;
  };

  
  //  Specialization:
  //         Scalar distribution (In this case the _Point = void)
  //
  template<class _Tag, typename _Sample, class _Traits>
  class basic_user<_Tag, _Sample, void, _Traits, distpoint::traits<void>>
  : public virtual basic_user_base<_Tag>
  {
  public:
    //   public types
    typedef _Sample sample_type;
    typedef void  point_type;
    typedef _Traits sample_traits;
    typedef distpoint::traits<void> point_traits;
    typedef distbook<_Sample, void, _Traits, point_traits> distbook_type;
   
    //   destructor
    virtual ~basic_user() {}
    
    /*   result generator   */
    std::function<unsigned long long(const char *, const char **)> collect_and_print_result() const {
      return __basic_user_make_result<distbook_type>::collect_and_print;
    }

    /*   static version of result generator   */
    static unsigned long long collect_and_print(const char *dir, const char **files) {
      return __basic_user_make_result<distbook_type>::collect_and_print(dir, files);
    }
    
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title) { _M_book.create(key, title);}
    
    /*   collecting weight   */
    void physfill(int key, const sample_type& w) { _M_book.collect(key, w);}
    
    /*   enable plot  */
    void enable_plot(int key) { _M_book.enable_plot(key);}
    void enable_plot(int key, const ltxplot& ltx) { _M_book.enable_plot(key, ltx);}
    
  private:
    /*   defining some virtual functions   */
    void basic_user_end_of_event() { _M_book.accumulate();}
    void write_txt(std::ostream& os) { os<<"  "; ::duct::write_txt(os, _M_book);}
    void write_bin(std::ostream& os) { ::duct::write_bin(os, _M_book);}
    std::string get_typeinfo() const { return typeid(__basic_user_make_result<distbook_type>).name();}
    
  private:
    /*    data members  */
    distbook_type _M_book;

    /*   friend declaration   */
    template<class ..._Users> friend class basic_user_set;
  };
  
    
  template<class ..._Users>
  class basic_user_set : public _Users...
  {
    /*   helper class to unpack the template variable   */
    template<class _Arg, class ..._Args>
    struct __basic_user {
      static void end_of_event(basic_user_set *bs) {
        bs->_Arg::basic_user_end_of_event();
        __basic_user<_Args...>::end_of_event(bs);
      }
      
      static void write_txt(basic_user_set *bs, std::ostream& os) {
        bs->_Arg::write_txt(os);
        __basic_user<_Args...>::write_txt(bs, os);
      }
      
      static void write_bin(basic_user_set *bs, std::ostream& os) {
        bs->_Arg::write_bin(os);
        __basic_user<_Args...>::write_bin(bs, os);
      }
    };
    
  public:
    /*   result generator   */
    std::function<unsigned long long(const char *, const char **)> collect_and_print_result() const {
      return __basic_user_make_result<typename _Users::distbook_type...>::collect_and_print;
    }
    
    /*   static version of result generator   */
    static unsigned long long collect_and_print(const char *dir, const char **files) {
      return __basic_user_make_result<typename _Users::distbook_type...>::collect_and_print(dir, files);
    }
    
  protected:
    /*   Aliases to the basic_user<...> types in the template arguments.   */
    template<std::size_t _Nth>
    using user_type = typename std::tuple_element<_Nth, std::tuple<_Users...>>::type;
 
  private:
    /*   overriding the virtual function   */
    void basic_user_end_of_event() { __basic_user<_Users...>::end_of_event(this);}
    void write_txt(std::ostream& os) { os<<"  "; __basic_user<_Users...>::write_txt(this, os);}
    void write_bin(std::ostream& os) { __basic_user<_Users...>::write_bin(this, os);}
    std::string get_typeinfo() const { return typeid(__basic_user_make_result<typename _Users::distbook_type...>).name();}
   };
 
  
  template<class ..._Users> template<class _Arg>
  struct basic_user_set<_Users...>::__basic_user<_Arg> {
    static void end_of_event(basic_user_set<_Users...> *bs) {
      bs->_Arg::basic_user_end_of_event();
    }
    
    static void write_txt(basic_user_set<_Users...> *bs, std::ostream& os) {
      bs->_Arg::write_txt(os);
    }
    
    static void write_bin(basic_user_set<_Users...> *bs, std::ostream& os) {
      bs->_Arg::write_bin(os);
    }
  };
  
 }  //   namespace nlo

#endif
