/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/9/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_bounding_function_h__
#define __mc_bounding_function_h__ 1

//   std includes
#include <cstring>
#include <utility>
#include <limits>
#include <initializer_list>
#include <memory>
#include <thread>
#include <mutex>
#include <iostream>

//   project headers
#include <bits/hep-dynmultiarray.h>
#include <bits/hep-binary-tree.h>
#include <bits/hep-void_ptr.h>


namespace duct {
  
  /* Empty template declaration of the bounding functions class.   */
  template<unsigned...> class bounding_function;


  namespace __bounding_function_helper {

    /*    interval type is a pair of doubles   */
    typedef std::pair<double,double> interval;
    
    
    /*  This class is used for scanning the function and collecting information.
     *  It can offer some possibilities for making split in a cell.
     */
    template<unsigned _Nv>
    struct proposals {
      /*   This class represents the cell.  */
      struct cell {
        //----- check if the point is in the cell -----
        bool is_in(const double *v) const {
          for(unsigned int i = 0; i < _Nv; i++)
            if(v[i] < box[i].first || v[i] > box[i].second)
              return false;
          return true;
        }
        
        //----- volume of the variables -----
        double volume() const {
          double retval = 1.0;
          for(unsigned i = 0; i < _Nv; i++)
            retval *= box[i].second-box[i].first;
          return retval;
        }
        
        //----- coordinates of the hypercube -----
        interval box[_Nv];
        
        double vol, vmin[_Nv], fmin, vmax[_Nv], fmax;
      };
      
      /*   Constructors   */
      explicit proposals(const interval *);
      
      explicit proposals(const cell& hc) : proposals(hc.box) {
        this->accumulate(hc.vmax, hc.fmax);
        this->accumulate(hc.vmin, hc.fmin);
      }

      /*   Accumulates weights in the scan phase   */
      void accumulate(const double *, double);
      
      /*   integral of the maximum function with split  */
      double ifmax(unsigned int i, unsigned int k) const {
        const cell &hL = histL[i][k], &hR = histR[i][k];
        return hL.vol*hL.fmax + hR.vol*hR.fmax;
      }
      
      /*   integral of the minimum function with split  */
      double ifmin(unsigned int i, unsigned int k) const {
        const cell &hL = histL[i][k], &hR = histR[i][k];
        return hL.vol*hL.fmin + hR.vol*hR.fmin;
      }
      
      /*   gain of the maximum function with split  */
      double gfmax(unsigned int i, unsigned int k) const {
        const cell &hL = histL[i][k], &hR = histR[i][k];
        return (hL.vol+hR.vol)*std::max(hL.fmax, hR.fmax) - hL.vol*hL.fmax - hR.vol*hR.fmax;
      }
      
      /*   integral of the minimum and maximum function */
      double ifmin() const { return vol*fmin;}
      double ifmax() const { return vol*fmax;}
      
      /*   The mother cell volume, fmin and fmax  */
      double vol, fmin, fmax;
      
      /*   We have two proposalss for the left and the right side.
       *   These are the possible daughter cells.
       */
      dynmultiarray<cell,_Nv,9> histL;
      dynmultiarray<cell,_Nv,9> histR;
    };
 
    
    template<unsigned _Nv>
    proposals<_Nv>::proposals(const interval *box)
    {
      /*   Initializing the daughter cells   */
      vol = 1.0;
      
      for(unsigned i = 0; i < _Nv; i++) {
        double dpi = (box[i].second-box[i].first)/10;
        double xik = box[i].first + dpi;
        
        for(unsigned k = 0; k < 9; k++, xik += dpi) {
          cell &hL = histL[i][k], &hR = histR[i][k];
          
          std::memcpy(hL.box, box, _Nv*sizeof(interval));
          std::memcpy(hR.box, box, _Nv*sizeof(interval));
          hL.box[i].second = hR.box[i].first = xik;
          
          hL.fmax = hR.fmax = -std::numeric_limits<double>::infinity();
          hL.fmin = hR.fmin =  std::numeric_limits<double>::infinity();
          hL.vol = hL.volume(); hR.vol = hR.volume();
        }
        
        /*   Volume of the mother cell   */
        vol *= box[i].second-box[i].first;
      }
      
      /*   Mother cell fmin and fmax   */
      fmax = -std::numeric_limits<double>::infinity();
      fmin =  std::numeric_limits<double>::infinity();
    }
    
    
    template<unsigned _Nv>
    void proposals<_Nv>::accumulate(const double *v, double fval)
    {
      /*   Accumulating the weight in the possible daughter cells   */
      for(unsigned i = 0; i < _Nv; i++)
        for(unsigned k = 0; k < 9; k++) {
          cell &h = histL[i][k].is_in(v) ? histL[i][k] : histR[i][k];
          
          if(h.fmax < fval) {
            h.fmax = fval;
            std::memcpy(h.vmax, v, _Nv*sizeof(double));
          }
          
          if(h.fmin > fval) {
            h.fmin = fval;
            std::memcpy(h.vmin, v, _Nv*sizeof(double));
          }
        }
    
      /*   Mother cell fmin and fmax   */
      if(fmax < fval) fmax = fval;
      if(fmin > fval) fmin = fval;
    }
  
  
  }   //   namespace __bounding_function_helper
}  //  namespace deductor



/*  Specialization of the bounding_function:
 *  This class defines the bounding function with
 *  certain number of variables.
 */
namespace duct {
  
  
  template<unsigned int _Nv>
  class bounding_function<_Nv>
  {
  public:
    /*  The interval is a pair of doubles. This help me to avoid defining
     *  a new trivial class for this simple stucture.
     */
    typedef __bounding_function_helper::interval interval;
    
   private:
    /*  This class helps us to find the best split. It monitors the function
     *  over the cell and make some proposals for the best split.
     */
    typedef __bounding_function_helper::proposals<_Nv> proposals;

    /*  In priciple we have to build a tree of cells but 
     *  The leaf cells need some auxiliary data during the
     *  scanning phase. This auxiliary data can be also used 
     *  for monitoring the function in the main calculation.
     */
    struct cell {
      /*   it is rather a trivial object, but copy is not allowed  */
      cell() = default;
      ~cell() = default;
      cell(cell&&) = default;
      cell(const cell&) = delete;
      cell& operator=(cell&&) = default;
      cell& operator=(const cell&) = delete;
  
      /*   non-standard constructor  */
      explicit cell(const interval *bx, double fmx = -std::numeric_limits<double>::infinity()) {
        for(unsigned i = 0; i < _Nv; ++i) box[i] = bx[i];
        fmax = fmx;
      }
      
      /*  scanning the cell and try to find the maximum  */
      template<class _Func> void scan(const _Func&);
      
      /*   check if the point is in the box   */
      bool is_in(const double *v) const {
        for(unsigned int i = 0; i < _Nv; i++)
          if(v[i] < box[i].first || v[i] > box[i].second)
            return false;
        return true;
      }
      
      /*    the volume of the box   */
      double volume() const {
        double retval = 1.0;
        for(unsigned i = 0; i < _Nv; i++)
          retval *= box[i].second-box[i].first;
        return retval;
      }
      
      /*   integral of the cell over an interval  */
      double integral(const interval *) const;

      /*   integral of the function over a given interval with weight function  */
      template<class _Func> double integral(const interval *, _Func) const;

      /*   average of the function over a given interval   */
      double average(const interval *) const;
      
      /*   average of the function over a given interval with weight function  */
      template<class _Func> double average(const interval *, _Func, double=0.0001) const;

      /*   coordinates of the hypercube   */
      interval box[_Nv];
      
      /*  maximum of the function in this box   */
      double fmax = -std::numeric_limits<double>::infinity();
      
      /*  it points to the proposals if it is a leaf cell  */
      void_ptr aux;
    };
    
    /*    some useful typdefs   */
    typedef typename binary_tree<cell>::leaf_iterator leaf_iterator;
    
  public:
    /*  default constructor and destructor  */
    bounding_function() = default;
    ~bounding_function() = default;
    
    /*   Intializing the main box   */
    explicit bounding_function(const interval *bx) : _M_tree(cell(bx)) {}
    
    /*   move constructor and assignment   */
    bounding_function(bounding_function&& x) = default;
    bounding_function& operator=(bounding_function&&) = default;
    
    /*   copy constructor   */
    bounding_function(const bounding_function& x)
    : _M_tree(x._M_tree, [=](const cell& c) -> cell { return cell(c.box, c.fmax);}) {}
    
    /*   copy assinment  */
    bounding_function& operator=(const bounding_function& x) {
      if(&x != this)
        _M_tree.assign(x._M_tree, [=](const cell& c) -> cell { return cell(c.box, c.fmax);});
      return *this;
    }
    
    /*   scanning and making boxes   */
    template<class _Func> void scan(const _Func&, unsigned int);
    
    /*   calculate the function in a given point   */
    double operator()(const double *v) const {
      auto c = this->_M_tree.root();
      while(!c.is_leaf()) { c = c.left()->is_in(v) ? c.left() : c.right();}
      return c->fmax;
    }

    double operator()(std::initializer_list<double> v) const {
      return this->operator()(v.begin());
    }

    
    /*   calculate the integral over a given interval   */
    double integral(const interval *v) const {
      double retval = 0.0;
      for(auto ic = _M_tree.leaf_cbegin(); ic != _M_tree.leaf_cend(); ++ic)
        retval += ic->integral(v);
      return retval;
    }

    double integral(std::initializer_list<interval> v) const {
      return this->integral(v.begin());
    }

    
    /*   calculate the integral over a given interval with weight function   */
    template<class _Func>
    double integral(const interval *v, _Func weight) const {
      double retval = 0.0;
      for(auto ic = _M_tree.leaf_cbegin(); ic != _M_tree.leaf_cend(); ++ic)
        retval += ic->integral(v, weight);
      return retval;
    }
    
    template<class _Func>
    double integral(std::initializer_list<interval> v, _Func weight) const {
      return this->integral(v.begin(), weight);
    }

    
    /*   calculate the average over a given interval   */
    double average(const interval *v) const {
      double retval = 0.0;
      for(auto ic = _M_tree.leaf_cbegin(); ic != _M_tree.leaf_cend(); ++ic)
        retval += ic->average(v);
      return retval;
    }
    
    double average(std::initializer_list<interval> v) const {
      return this->average(v.begin());
    }
 
    
    /*   calculate the average over a given interval with weight function   */
    template<class _Func>
    double average(const interval *v, _Func weight) const {
      double retval = 0.0;
      for(auto ic = _M_tree.leaf_cbegin(); ic != _M_tree.leaf_cend(); ++ic)
        retval += ic->average(v, weight);
      return retval;
    }
    
    template<class _Func>
    double average(std::initializer_list<interval> v, _Func weight) const {
      return this->average(v.begin(), weight);
    }
    
   
    /*   I/O manipulations   */
    template<class _OStream> _OStream& write(_OStream& ) const;
    template<class _IStream> _IStream& read(_IStream&);
    
  private:
    /*   We have a binary tree to store the cells   */
    binary_tree<cell> _M_tree;

    /*   make an equidistant grid   */
    void _M_make_grid(unsigned);

    /*   try to find the best cut   */
    double _M_find_best_split(leaf_iterator&, unsigned&, unsigned&) const;
    double _M_find_best_middle_split(leaf_iterator&, unsigned&) const;

    /*   make a split   */
    template<class _Func> void _M_make_a_split(leaf_iterator&, unsigned, unsigned, const _Func&);

    /*   Friend declaration  */
    template<unsigned...Args> friend class bounding_function;
  };
  
  
  template<unsigned int _Nv> template<class _Func> 
  void bounding_function<_Nv>::cell::scan(const _Func& func)
  {
    constexpr unsigned long long int mask = 1;
        
    /*    Scanning the function over the combined box   */
    double fval, r, v[_Nv];
    proposals *hist = static_cast<proposals *>(aux);
    
    /*  Generate points in the corners */
    for(double scale : {1e-7, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.49})
      for(unsigned long long int idx = 0U; idx < (1<<_Nv); idx++) {
        //----- generate a corner point -----
        for(unsigned int i = 0; i < _Nv; i++) {
          r = (mask << i) & idx ? 1.0-scale : scale;
          v[i] = (this->box[i].second - this->box[i].first)*r + this->box[i].first;
        }
        
        //----- calculate the function and check if it is a maximum ----
        fval = func(v);
        if(fval != fval) fval = 0.0;
        if(fval > this->fmax) this->fmax = fval;
        if(hist) hist->accumulate(v, fval);
      }
    
    /*  Generate points along the edges */
    for(unsigned k = 0; k < _Nv; k++) {
      double dx = (this->box[k].second - this->box[k].first)/10;
      
      for(unsigned long long int idx = 0U; idx < (1<<_Nv); idx++)
        if((mask << k) & idx) {
          //----- generate a corner point -----
          for(unsigned int i = 0; i < _Nv; i++) {
            r = (mask << i) & idx ? 0.9999999 : 0.0000001;
            v[i] = (this->box[i].second - this->box[i].first)*r + this->box[i].first;
          }
          
          //----- calculate the function and check if it is a maximum ----
          for(unsigned j = 1; j < 10; j++) {
            v[k] = this->box[k].first + j*dx;
            
            fval = func(v);
            if(fval != fval) fval = 0.0;
            if(fval > this->fmax) this->fmax = fval;
            if(hist) hist->accumulate(v, fval);
          }
        }
    }
  }
  
  
  template<unsigned int _Nv>
  double bounding_function<_Nv>::_M_find_best_split(leaf_iterator& cbest, unsigned& ibest, unsigned& kbest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    ibest = 0u; kbest = 4u;
    cbest = static_cast<leaf_iterator>(_M_tree.leaf_begin());
    
    /*   Calculating  Imax and Imin    */
    double Imax = 0.0, Imin = 0.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      proposals *hist = static_cast<proposals *>(ic->aux);
      Imax += hist->ifmax();
      Imin += hist->ifmin();
    }
    
    /*   Finding the best place to split   */
    double dr = -1.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*   Find the best split. This corresponds to the cut with
       *   the highest change of the function between left and right side.
       */
      proposals *hist = static_cast<proposals *>(ic->aux);
      double ifmin = hist->ifmin(), ifmax = hist->ifmax();
      
      for(unsigned i = 0; i < _Nv; i++)
        for(unsigned k = 0; k < 9; k++) {
          double dvfmax = ifmax - hist->ifmax(i, k);
          double dvfmin = hist->ifmin(i, k) - ifmin;
          double drik = (Imax*dvfmin + Imin*dvfmax)/(Imax-dvfmax);
          
          if(drik > dr) {
            dr = drik;
            cbest = static_cast<leaf_iterator>(ic);
            ibest = i;
            kbest = k;
          }
        }
    }
    
    return dr;
  }
  
  
  template<unsigned int _Nv>
  double bounding_function<_Nv>::_M_find_best_middle_split(leaf_iterator& cbest, unsigned& ibest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    cbest = static_cast<leaf_iterator>(_M_tree.leaf_begin());
    
    /*   Finding the best place to split   */
    double vhi = 0.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*  If there is a diagonal ridge then the previus slplit ususally fails.
       *  In this case we cut the cell with the bigest contribution in the middle.
       */
      proposals *hist = static_cast<proposals *>(ic->aux);
      double ifmin = hist->ifmin(), ifmax = hist->ifmax();

      if(ifmax-ifmin > vhi) {
        vhi = ifmax-ifmin;
        cbest = static_cast<leaf_iterator>(ic);
      }
    }
    
    /*   We cut the bigest box along the longest side in the middle.  */
    double ilen = 0.0;
    
    for(unsigned i = 0; i < _Nv; i++)
      if(cbest->box[i].second - cbest->box[i].first > ilen) {
        ibest = i;
        ilen = cbest->box[i].second - cbest->box[i].first;
      }
    
    return vhi;
  }
  
  
  template<unsigned int _Nv>
  template<class _Func> void bounding_function<_Nv>::
  _M_make_a_split(leaf_iterator& cbest, unsigned ibest, unsigned kbest, const _Func& func)
  {
    /*   make the splitting happen  */
    proposals *hist = static_cast<proposals *>(cbest->aux);
    const typename proposals::cell &hL = hist->histL[ibest][kbest];
    const typename proposals::cell &hR = hist->histR[ibest][kbest];
    
    auto left = _M_tree.append_left(cbest, cell(hL.box, hL.fmax));
    auto right = _M_tree.append_right(cbest, cell(hR.box, hR.fmax));
    
    //---- initializing the proposalss -----
    left->aux.reset(new proposals(hL));
    right->aux.reset(new proposals(hR));
    
    cbest->aux = nullptr;
    
    //---- scanning the new cells ------
    left->scan(func);
    right->scan(func);
  }
  
  
  template<unsigned _Nv> template<class _Func> 
  void bounding_function<_Nv>::scan(const _Func& func, unsigned int nsvar)
  {
    /*   making a small grid   */
    _M_make_grid(2u);
    
    for(auto iv = _M_tree.leaf_begin(); iv != _M_tree.leaf_end(); ++iv) {
      iv->aux.reset(new proposals(iv->box));
      iv->scan(func);
    }
    
    /*   Making boxes according to the best cut   */
    unsigned ibest = 0u, kbest = 4u;
    auto cbest = _M_tree.leaf_begin();
    
    for(unsigned ns = 0; ns < nsvar/2; ++ns) {
      _M_find_best_split(cbest, ibest, kbest);
      _M_make_a_split(cbest, ibest, kbest, func);
    }
    
    /*   Making boxes by alternating the best cut and best middle cut methods  */
    for(unsigned ns = 0; ns < nsvar/4; ++ns) {
      _M_find_best_middle_split(cbest, ibest);
      _M_make_a_split(cbest, ibest, 4u, func);
      
      _M_find_best_split(cbest, ibest, kbest);
      _M_make_a_split(cbest, ibest, kbest, func);
    }

    /*   Removing the auxilary data from the leaf cells  */
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic)
      ic->aux = nullptr;
    
    /*   Consolidating the tree  */
    for(auto ic = _M_tree.postorder_begin(); ic != _M_tree.postorder_end(); ++ic)
      if(ic.is_leaf() == false) ic->fmax = std::max(ic.left()->fmax, ic.right()->fmax);
  }
 
  
  template<unsigned _Nv>
  template<class _OStream>
  _OStream& bounding_function<_Nv>::write(_OStream& os) const
  {
    _M_tree.write(os, [=](_OStream& __os, const cell& c) -> void {
      __os.write(reinterpret_cast<const char *>(c.box), _Nv*sizeof(interval));
      __os.write(reinterpret_cast<const char *>(&c.fmax), sizeof(double));
    });
  
    return os;
  }
  
  template<unsigned _Nv>
  template<class _IStream>
  _IStream& bounding_function<_Nv>::read(_IStream& is)
  {
    _M_tree.read(is, [=](_IStream& __is, cell& c) -> void {
      __is.read(reinterpret_cast<char *>(c.box), _Nv*sizeof(interval));
      __is.read(reinterpret_cast<char *>(&(c.fmax)), sizeof(double));
      c.aux = nullptr;
    });
    
    return is;
  }
  

  template<unsigned _Nv>
  void bounding_function<_Nv>::_M_make_grid(unsigned ndeep)
  {
    /*   make the splitting happen  */
    for(unsigned k = 0; k < _Nv; ++k)
      for(unsigned d = 0; d < ndeep; ++d) {
        auto ic = _M_tree.postorder_begin();
        
        while(ic != _M_tree.postorder_end())
          if(ic.is_leaf()) {
            cell cleft(ic->box), cright(ic->box);
            
            cleft.box[k].second = cright.box[k].first = 0.5*(ic->box[k].first + ic->box[k].second);
            
            _M_tree.append_left(ic, std::move(cleft));
            _M_tree.append_right(ic, std::move(cright));
          } else ++ic;
      }
  }

  
  
  template<unsigned int _Nv>
  double bounding_function<_Nv>::cell::integral(const interval *v) const
  {
    double low, up, vol = 1.0;
    
    for(unsigned i = 0; i < _Nv; i++) {
      low = std::max(this->box[i].first, v[i].first);
      up = std::min(this->box[i].second, v[i].second);
      
      if(low >= up) return 0.0;
      else vol *= up-low;
    }
    
    return this->fmax*vol;
  }

  
  template<unsigned int _Nv> template<class _Func>
  double bounding_function<_Nv>::cell::integral(const interval *v, _Func weight) const
  {
    interval overlap[_Nv];
    
    for(unsigned i = 0; i < _Nv; i++) {
      overlap[i].first = std::max(this->box[i].first, v[i].first);
      overlap[i].second = std::min(this->box[i].second, v[i].second);
      
      if(overlap[i].first >= overlap[i].second) return 0.0;
    }
    
    return this->fmax*weight(overlap);
  }
  

  template<unsigned int _Nv>
  double bounding_function<_Nv>::cell::average(const interval *v) const
  {
    double low, up, vol = 1.0;
    
    for(unsigned i = 0; i < _Nv; i++) {
      if(std::abs(v[i].second-v[i].first) < 1e-14*std::min(1.0, box[i].second-box[i].first)) {
        if(v[i].first < box[i].first || box[i].second <= v[i].first) return 0.0;
      } else {
        low = std::max(this->box[i].first, v[i].first);
        up = std::min(this->box[i].second, v[i].second);
        
        if(low >= up) return 0.0;
        else vol *= (up-low)/(v[i].second-v[i].first);
      }
    }
    
    return this->fmax*vol;
  }
  
  
  template<unsigned int _Nv> template<class _Func>
  double bounding_function<_Nv>::cell::average(const interval *v, _Func weight, double eps) const
  {
    double dx, vol = 1.0;
    interval overlap[_Nv];
    
    for(unsigned i = 0; i < _Nv; i++) {
      dx = std::min(1.0, box[i].second-box[i].first);
      
      if(std::abs(v[i].first-v[i].second) < 1e-14*dx) {
        if(v[i].first < box[i].first || box[i].second <= v[i].first)
          return 0.0;
        
        overlap[i].first = v[i].first-eps*dx;
        overlap[i].second = v[i].first+eps*dx;
        vol /= 2.0*eps*dx;
      } else {
        overlap[i].first = std::max(this->box[i].first, v[i].first);
        overlap[i].second = std::min(this->box[i].second, v[i].second);
        
        if(overlap[i].first >= overlap[i].second) return 0.0;
        else vol /= v[i].second-v[i].first;
      }
    }
    
    return this->fmax*weight(overlap)*vol;
  }
  
}



namespace duct {
  
  
  template<unsigned int _Np, unsigned int _Nv>
  class bounding_function<_Np, _Nv>
  {
  public:
    /*  The interval is a pair of doubles. This help me to avoid defining
     *  a new trivial class for this simple stucture.
     */
    typedef __bounding_function_helper::interval interval;
    
  private:
    /*  This represent a hypercube in the _Np dimensional space.
     *  We store in every leaf cell an object of bounding_function<_Nv>.
     */
    struct cell {
      /*   it is rather a trivial obbject, but copy is not allowed  */
      cell() = default;
      ~cell() = default;
      cell(cell&&) = default;
      cell(const cell&) = delete;
      cell& operator=(cell&&) = default;
      cell& operator=(const cell&) = delete;
      
      /*   creating by initalizing the parameter box  */
      explicit cell(const interval *bp) {
        for(unsigned i = 0; i < _Np; i++) box[i] = bp[i];
        fvar = nullptr;
      }
      
      /*   creating by making the head of the tree   */
      cell(const interval *bp, const interval *bv) {
        for(unsigned i = 0; i < _Np; i++) box[i] = bp[i];
        fvar.reset(new bounding_function<_Nv>(bv));
      }
      
      /*   copy by members  */
      cell(const interval *bp, const bounding_function<_Nv>& fv)
      : cell(bp) { fvar.reset(new bounding_function<_Nv>(fv));}
      
      /*   check if the point is in the cell (parameters only)  */
      bool is_in(const double *p) const {
        for(unsigned int i = 0; i < _Np; i++)
          if(p[i] < box[i].first || p[i] > box[i].second)
            return false;
        return true;
      }
      
      /*   volume of the box (parameters only)   */
      double volume() const {
        double retval = 1.0;
        for(unsigned i = 0; i < _Np; i++)
          retval *= box[i].second-box[i].first;
        return retval;
      }
      
      /*  this defines the box in the parameter space  */
      interval box[_Np];
      
      /*  Bounding function in terms of the integral variables.
       *  This is just a pointer because we don't need a big tree
       *  in every cells. We need them only in the leaf cells.
       */
      typedef bounding_function<_Nv> _FuncNv;
      std::unique_ptr<_FuncNv> fvar;
 
      /*   scanning a cell in bounding_function<_Nv>  */
      template<class _Func> void scan(typename _FuncNv::leaf_iterator, const _Func&);

      /*   calculate the integral of the maximum function with split  */
      double ifmax(unsigned int i, unsigned int k) const {
        double retval = 0.0;
        for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic)
          retval += static_cast<const proposals *>(ic->aux)->ifmax(i, k);
        return retval;
      }
      
      /*   calculate the integral of the minimum function with split  */
      double ifmin(unsigned int i, unsigned int k) const {
        double retval = 0.0;
        for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic)
          retval += static_cast<const proposals *>(ic->aux)->ifmin(i, k);
        return retval;
      }

      /*   calculate the integral of the maximum function with split  */
      double gfmax(unsigned int i, unsigned int k) const {
        double retval = 0.0;
        for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic)
          retval += static_cast<const proposals *>(ic->aux)->gfmax(i, k);
        return retval;
      }
      
      /*   calculate the integral of the maximum function  */
      double ifmax() const {
        double retval = 0.0;
        for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic)
          retval += static_cast<const proposals *>(ic->aux)->ifmax();
        return retval;
      }

      /*   calculate the integral of the minimum function  */
      double ifmin() const {
        double retval = 0.0;
        for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic)
          retval += static_cast<const proposals *>(ic->aux)->ifmin();
        return retval;
      }
      
      /*   scanning the left cells   */
      template<class _Func> static void scan_in_thread(cell *left, const _Func& func) {
        for(auto iv = left->fvar->_M_tree.leaf_begin(); iv != left->fvar->_M_tree.leaf_end(); ++iv)
          left->scan(iv, func);
      }

      /*   trying to find the best place to cut  */
      double find_best_split(typename _FuncNv::leaf_iterator&, unsigned&, unsigned&) const;
      double find_best_middle_split(typename _FuncNv::leaf_iterator&, unsigned&) const;

      /*   make the splitting happen  */
      template<class _Func>
      void make_a_split(typename _FuncNv::leaf_iterator&, unsigned, unsigned, const _Func&);
      
      /*   scanning and making more splits in the variable space    */
      template<class _Func> void scan(const _Func&, unsigned);
    };

    /*  This class helps us to find the best split. It monitores the function
     *  over the cell and make some proposals for the best split.
     */
    typedef __bounding_function_helper::proposals<_Np+_Nv> proposals;

    
  public:
    /*   Default constructor  */
    bounding_function() = default;

    /*   Intializing the main box   */
    bounding_function(const interval *bp, const interval *bv)
    : _M_tree(cell(bp, bv)) {}
    
    /*   move constructor and assignment   */
    bounding_function(bounding_function&& x) = default;
    bounding_function& operator=(bounding_function&&) = default;
    
    /*   copy constructor   */
    bounding_function(const bounding_function& x)
    : _M_tree(x._M_tree, [=](const cell& c) -> cell { return c.fvar ? cell(c.box, *(c.fvar)) : cell(c.box);}) {}
    
    /*   copy assinment  */
    bounding_function& operator=(const bounding_function& x) {
      if(&x != this)
        _M_tree.assign(x._M_tree, [=](const cell& c) -> cell {
          return c.fvar ? cell(c.box, *(c.fvar)) : cell(c.box);
        });
      return *this;
    }

    /*   scanning and making boxes   */
    template<class _Func> void scan(const _Func&, unsigned int, unsigned int);

    /*   multi threaded version of the scan function   */
    template<class _Func> void scan_mt(const _Func&, unsigned int, unsigned int);
    
    /*   returns the bounding function at fixed parameters   */
    const bounding_function<_Nv>& bind(const double *p) const {
      auto c = this->_M_tree.root();
      while(!c.is_leaf()) c = (c.left()->is_in(p) ? c.left() : c.right());
      return *(c->fvar);
    }
    
    const bounding_function<_Nv>& bind(std::initializer_list<double> il) const {
      return bind(il.begin());
    }
    
    /*   I/O manipulations   */
    template<class _OStream> _OStream& write(_OStream&) const;
    template<class _IStream> _IStream& read(_IStream&);
    
  private:
    /*   We have a binary tree to store the cells   */
    binary_tree<cell> _M_tree;
  
    /*   make a split in the parameter space  */
    template<class _Func> void make_a_split(const typename binary_tree<cell>::leaf_iterator&, unsigned, unsigned, const _Func&);
    template<class _Func> void make_a_split_mt(const typename binary_tree<cell>::leaf_iterator&, unsigned, unsigned, const _Func&);
    
    /*   It tries to find the best place to make cut.   */
    double _M_find_best_split(typename binary_tree<cell>::leaf_iterator&, unsigned&, unsigned&) const;
    double _M_find_best_gain_split(typename binary_tree<cell>::leaf_iterator&, unsigned&, unsigned&) const;
    double _M_find_best_middle_split(typename binary_tree<cell>::leaf_iterator&, unsigned&) const;
  };
  
  
  template<unsigned int _Np, unsigned int _Nv> template<class _Func>
  void bounding_function<_Np, _Nv>::cell::scan(typename _FuncNv::leaf_iterator ic, const _Func& func)
  {
    constexpr unsigned int Nvar = _Np+_Nv;
    constexpr unsigned long long int mask = 1;
    
    /*   The combined parameter and variable box   */
    interval pvbox[Nvar];
    
    for(unsigned i = 0; i < _Np; i++) pvbox[i] = this->box[i];
    for(unsigned i = 0; i < _Nv; i++) pvbox[i+_Np] = ic->box[i];
    
    /*    Scanning the function over the combined box   */
    double fval, r, v[Nvar];
    proposals *hist = static_cast<proposals *>(ic->aux);
    
    /*  Generate points in the corners */
    //   for(double scale : {1e-7, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.49})
    for(double scale : {1e-7, 0.1, 0.2, 0.3, 0.4, 0.49})
      for(unsigned long long int idx = 0U; idx < (1<<Nvar); idx++) {
        //----- generate a corner point -----
        for(unsigned int i = 0; i < Nvar; i++) {
          r = (mask << i) & idx ? 1.0-scale : scale;
          v[i] = (pvbox[i].second - pvbox[i].first)*r + pvbox[i].first;
        }
        
        //----- calculate the function and check if it is a maximum ----
        fval = func(v, v+_Np);
        if(fval != fval) fval = 0.0;
        if(fval > ic->fmax) ic->fmax = fval;
        if(hist) hist->accumulate(v, fval);
      }
    
    /*  Generate points along the edges */
    for(unsigned k = 0; k < Nvar; k++) {
      double dx = (pvbox[k].second - pvbox[k].first)/10;
      
      for(unsigned long long int idx = 0U; idx < (1<<Nvar); idx++)
        if((mask << k) & idx) {
          //----- generate a corner point -----
          for(unsigned int i = 0; i < Nvar; i++) {
            r = (mask << i) & idx ? 0.99999 : 0.00001;
            v[i] = (pvbox[i].second - pvbox[i].first)*r + pvbox[i].first;
          }
          
          //----- calculate the function and check if it is a maximum ----
          for(unsigned j = 1; j < 10; j++) {
            v[k] = pvbox[k].first + j*dx;
            
            fval = func(v, v+_Np);
            if(fval != fval) fval = 0.0;
            if(fval > ic->fmax) ic->fmax = fval;
            if(hist) hist->accumulate(v, fval);
          }
        }
    }
  }

  
  template<unsigned int _Np, unsigned int _Nv>
  double bounding_function<_Np, _Nv>::cell::
  find_best_split(typename _FuncNv::leaf_iterator& cbest, unsigned& ibest, unsigned& kbest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    ibest = 0u; kbest = 4u;
    cbest = static_cast<typename _FuncNv::leaf_iterator>(fvar->_M_tree.leaf_begin());

    /*    Finding the best place to split (variables only)   */
    double dr = -1.0, Imin = this->ifmin(), Imax = this->ifmax();
    
    for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic) {
      /*   Find the best split. This corresponds to the cut with
       *   the highest change of the function between left and right side.
       */
      proposals *hist = static_cast<proposals *>(ic->aux);
      double ifmin = hist->ifmin(), ifmax = hist->ifmax();
      
      for(unsigned i = _Np; i < _Np+_Nv; i++)
        for(unsigned k = 0; k < 9; k++) {
          double dvfmin = hist->ifmin(i, k) - ifmin;
          double dvfmax = ifmax - hist->ifmax(i, k);
          double drik = (Imax*dvfmin/(Imin+1e-16) + dvfmax)/(Imax-dvfmax);
          
          if(drik > dr) {
            dr = drik;
            cbest = static_cast<typename _FuncNv::leaf_iterator>(ic);
            ibest = i;
            kbest = k;
          }
        }
    }
    
    return dr;
  }
  

  template<unsigned int _Np, unsigned int _Nv>
  double bounding_function<_Np, _Nv>::cell::
  find_best_middle_split(typename _FuncNv::leaf_iterator& cbest, unsigned& ibest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    cbest = static_cast<typename _FuncNv::leaf_iterator>(fvar->_M_tree.leaf_begin());
    
    /*    Finding the best place to split (variables only)   */
    double vhi = -1.0;
    
    for(auto ic = fvar->_M_tree.leaf_begin(); ic != fvar->_M_tree.leaf_end(); ++ic) {
      /*  If there is a diagonal ridge then the previus slplit ususally fails.
       *  In this case we cut the cell with the bigest contribution in the middle.
       */
      proposals *hist = static_cast<proposals *>(ic->aux);
      double ifmin = hist->ifmin(), ifmax = hist->ifmax();
      
      if(ifmax-ifmin > vhi) {
        vhi = ifmax-ifmin;
        cbest = static_cast<typename _FuncNv::leaf_iterator>(ic);
      }
    }
    
    /*   We cut the bigest box along the longest side in the middle.   */
    double ilen = -1.0;
    
    for(unsigned i = 0; i < _Nv; i++)
      if(cbest->box[i].second - cbest->box[i].first > ilen) {
        ibest = _Np+i;
        ilen = cbest->box[i].second - cbest->box[i].first;
      }
    
    return vhi;
  }
  
  
  template<unsigned int _Np, unsigned int _Nv>
  template<class _Func> void bounding_function<_Np, _Nv>::cell::
  make_a_split(typename _FuncNv::leaf_iterator& cbest, unsigned ibest, unsigned kbest, const _Func& func)
  {
    /*    make the spliting happen (variable space)  */
    typedef typename bounding_function<_Nv>::cell vcell;
    proposals *hist = static_cast<proposals *>(cbest->aux);
    
    /*    inserting the new cells   */
    const typename proposals::cell &hL = hist->histL[ibest][kbest];
    const typename proposals::cell &hR = hist->histR[ibest][kbest];
    
    auto left = fvar->_M_tree.append_left(cbest, vcell(hL.box+_Np, hL.fmax));
    auto right = fvar->_M_tree.append_right(cbest, vcell(hR.box+_Np, hR.fmax));
    
    /*    initializing the auxilary data   */
    left->aux.reset(new proposals(hL));
    right->aux.reset(new proposals(hR));
    
    /*   cleaning up  */
    cbest->aux = nullptr;
    
    /*   scanning the new cells   */
    this->scan(left, func);
    this->scan(right, func);
  }

  template<unsigned int _Np, unsigned int _Nv> template<class _Func>
  void bounding_function<_Np, _Nv>::cell::scan(const _Func& func, unsigned nsvar)
  {
    /*   Making boxes according to the best cut   */
    unsigned ibest = 0u, kbest = 4u;
    auto cbest = fvar->_M_tree.leaf_begin();
    
    for(unsigned ns = 0; ns < nsvar/2; ++ns) {
      find_best_split(cbest, ibest, kbest);
      make_a_split(cbest, ibest, kbest, func);
    }
    
    /*   Making boxes by alternating the best cut and best middle cut methods  */
    for(unsigned ns = 0; ns < nsvar/4; ++ns) {
      find_best_middle_split(cbest, ibest);
      make_a_split(cbest, ibest, 4u, func);
      
      find_best_split(cbest, ibest, kbest);
      make_a_split(cbest, ibest, kbest, func);
    }
  }
  
  
  
  template<unsigned int _Np, unsigned int _Nv> template<class _Func>
  void bounding_function<_Np, _Nv>::
  make_a_split(const typename binary_tree<cell>::leaf_iterator& cbest, unsigned i, unsigned k, const _Func& func)
  {
    /*    make the spliting happen (parameter space)  */
    auto left = _M_tree.append_left(cbest, cell());
    auto right = _M_tree.append_right(cbest, cell());

    /*    copy the boxes  */
    proposals *hist = static_cast<proposals *>(cbest->fvar->_M_tree.leaf_begin()->aux);

    for(unsigned m = 0; m < _Np; m++) {
      left->box[m] = hist->histL[i][k].box[m];
      right->box[m] = hist->histR[i][k].box[m];
    }

    /*  creating bounding_function<_Nv> on the left and right side  */
    left->fvar.reset(new bounding_function<_Nv>);
    right->fvar.reset(new bounding_function<_Nv>);
  
    /*   Conversion function to be able to copy proposals */
    typedef typename bounding_function<_Nv>::cell vcell;
    
    auto convL = [=](const vcell& n) -> vcell {
      if(n.aux) {
        const typename proposals::cell& h = static_cast<proposals *>(n.aux)->histL[i][k];
        vcell ret(h.box+_Np, h.fmax);
        ret.aux.reset(new proposals(h));
        return ret;
      }

      return vcell(n.box, n.fmax);
    };
    
    auto convR = [=](const vcell& n) -> vcell {
      if(n.aux) {
        const typename proposals::cell& h = static_cast<proposals *>(n.aux)->histR[i][k];
        vcell ret(h.box+_Np, h.fmax);
        ret.aux.reset(new proposals(h));
        return ret;
      }
      
      return vcell(n.box, n.fmax);
    };
    
    /*   copying the trees */
    left->fvar->_M_tree.assign(cbest->fvar->_M_tree, convL);
    right->fvar->_M_tree.assign(cbest->fvar->_M_tree, convR);

    /*   scanning the left cells   */
    for(auto iv = left->fvar->_M_tree.leaf_begin(); iv != left->fvar->_M_tree.leaf_end(); ++iv)
      left->scan(iv, func);
    
    /*   scanning the right cells   */
    for(auto iv = right->fvar->_M_tree.leaf_begin(); iv != right->fvar->_M_tree.leaf_end(); ++iv)
      right->scan(iv, func);

    /*   cleaning up  */
    cbest->fvar = nullptr;
  }
  
  
  template<unsigned int _Np, unsigned int _Nv> template<class _Func>
  void bounding_function<_Np, _Nv>::
  make_a_split_mt(const typename binary_tree<cell>::leaf_iterator& cbest, unsigned i, unsigned k, const _Func& func)
  {
    /*    make the spliting happen (parameter space)  */
    auto left = _M_tree.append_left(cbest, cell());
    auto right = _M_tree.append_right(cbest, cell());
    
    /*    copy the boxes  */
    proposals *hist = static_cast<proposals *>(cbest->fvar->_M_tree.leaf_begin()->aux);
    
    for(unsigned m = 0; m < _Np; m++) {
      left->box[m] = hist->histL[i][k].box[m];
      right->box[m] = hist->histR[i][k].box[m];
    }
    
    /*  creating bounding_function<_Nv> on the left and right side  */
    left->fvar.reset(new bounding_function<_Nv>);
    right->fvar.reset(new bounding_function<_Nv>);
    
    /*   Conversion function to be able to copy proposals */
    typedef typename bounding_function<_Nv>::cell vcell;
    
    auto convL = [=](const vcell& n) -> vcell {
      if(n.aux) {
        const typename proposals::cell& h = static_cast<proposals *>(n.aux)->histL[i][k];
        vcell ret(h.box+_Np, h.fmax);
        ret.aux.reset(new proposals(h));
        return ret;
      }
      
      return vcell(n.box, n.fmax);
    };
    
    auto convR = [=](const vcell& n) -> vcell {
      if(n.aux) {
        const typename proposals::cell& h = static_cast<proposals *>(n.aux)->histR[i][k];
        vcell ret(h.box+_Np, h.fmax);
        ret.aux.reset(new proposals(h));
        return ret;
      }
      
      return vcell(n.box, n.fmax);
    };
    
    /*   copying the trees */
    left->fvar->_M_tree.assign(cbest->fvar->_M_tree, convL);
    right->fvar->_M_tree.assign(cbest->fvar->_M_tree, convR);
    
    /*   the new boxes are scanned in paralell  */
    std::thread thleft(cell::template scan_in_thread<_Func>, &(*left), func);
    std::thread thright(cell::template scan_in_thread<_Func>, &(*right), func);
    
    thleft.join();
    thright.join();
    
    /*   cleaning up  */
    cbest->fvar = nullptr;
  }
  
  
  
  template<unsigned int _Np, unsigned int _Nv>
  double bounding_function<_Np, _Nv>::
  _M_find_best_split(typename binary_tree<cell>::leaf_iterator& cbest, unsigned& ibest, unsigned& kbest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    ibest = 0u; kbest = 4u;
    cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(_M_tree.leaf_begin());
    
    /*   Calculate Imax and Imin   */
    double Imin = 0.0, Imax = 0.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      Imin += ic->ifmin();
      Imax += ic->ifmax();
    }
    
    /*   Trying to make the best cut in the parameter space.
     *   We always do split in the parameter space.
     */
    double dr = -1.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*   Find the best split. This corresponds to the cut with
       *   the highest change of the function between left and right side.
       *   Here we consider a split only in the parameter space.
       */
      double ifmin = ic->ifmin(), ifmax = ic->ifmax();
      
      for(unsigned i = 0; i < _Np; i++)
        for(unsigned k = 0; k < 9; k++) {
          double dvfmin = ic->ifmin(i, k)-ifmin;
          double dvfmax = ifmax-ic->ifmax(i, k);
          double drik = (Imax*dvfmin/(Imin+1e-10) + dvfmax)/(Imax-dvfmax);
          
          if(drik > dr) {
            dr = drik;
            cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(ic);
            ibest = i;
            kbest = k;
          }
        }
    }

    return dr;
  }

  template<unsigned int _Np, unsigned int _Nv>
  double bounding_function<_Np, _Nv>::
  _M_find_best_gain_split(typename binary_tree<cell>::leaf_iterator& cbest, unsigned& ibest, unsigned& kbest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    ibest = 0u; kbest = 4u;
    cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(_M_tree.leaf_begin());
    
    /*   Trying to make the best cut in the parameter space.
     *   We always do split in the parameter space.
     */
    double gain = -1.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*   Find the best split. This corresponds to the cut with
       *   the highest gain of the integral.
       *   Here we consider a split only in the parameter space.
       */
      for(unsigned i = 0; i < _Np; i++)
        for(unsigned k = 0; k < 9; k++) {
          double gik = ic->gfmax(i, k);
          
          if(gik > gain) {
            gain = gik;
            cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(ic);
            ibest = i;
            kbest = k;
          }
        }
    }
    
    return gain;
  }
  

  template<unsigned int _Np, unsigned int _Nv>
  double bounding_function<_Np, _Nv>::
  _M_find_best_middle_split(typename binary_tree<cell>::leaf_iterator& cbest, unsigned& ibest) const
  {
    /*   just in case ... it helps to avoid segmentation fault   */
    cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(_M_tree.leaf_begin());
    
    /*   Trying to make the best cut in the parameter space.
     *   We always do split in the parameter space.
     */
    double vhi = -1.0;
    
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*  If there is a diagonal ridge then the previous split ususally fails.
       *  In this case we cut the cell with the bigest contribution in the middle.
       */
      double ifmin = ic->ifmin(), ifmax = ic->ifmax();
      
      if(ifmax-ifmin > vhi) {
        vhi = ifmax-ifmin;
        cbest = static_cast<typename binary_tree<cell>::leaf_iterator>(ic);
      }
    }
    
    /*   We cut the bigest box along the longest side in the middle.   */
    double ilen = -1.0;
    
    for(unsigned i = 0; i < _Np; i++)
      if(cbest->box[i].second - cbest->box[i].first > ilen) {
        ibest = i;
        ilen = cbest->box[i].second - cbest->box[i].first;
      }
  
    return vhi;
  }
  
  
  
  template<unsigned int _Np, unsigned int _Nv> template<class _Func> 
  void bounding_function<_Np, _Nv>::scan(const _Func& func, unsigned nspar, unsigned nsvar)
  {
    auto head = _M_tree.root();
    interval pvbox[_Np+_Nv];
    
    /*   making a small grid in the variable space   */
    head->fvar->_M_make_grid(2u);
  
    /*   make the initial scan    */
    for(unsigned i = 0; i < _Np; i++)
      pvbox[i] = head->box[i];
    
    for(auto iv = head->fvar->_M_tree.leaf_begin(); iv != head->fvar->_M_tree.leaf_end(); ++iv) {
      for(unsigned i = 0; i < _Nv; i++) pvbox[i+_Np] = iv->box[i];
      iv->aux.reset(new proposals(pvbox));
      head->scan(iv, func);
    }
    
    /*   Making many boxes in the parameter space   */
    unsigned ibest = 0u, kbest = 4u;
    auto cbest = _M_tree.leaf_begin();
    
    /*   Making boxes according to the best cut   */
    for(unsigned ns = 0; ns < nspar/4; ++ns) {
      this->_M_find_best_split(cbest, ibest, kbest);
      this->make_a_split(cbest, ibest, kbest, func);

      this->_M_find_best_gain_split(cbest, ibest, kbest);
      this->make_a_split(cbest, ibest, kbest, func);
    }
    
    /*   Making boxes by alternating the best cut and best middle cut methods  */
    for(unsigned ns = 0; ns < nspar/4; ++ns) {
      this->_M_find_best_middle_split(cbest, ibest);
      this->make_a_split(cbest, ibest, 4u, func);
      
      this->_M_find_best_split(cbest, ibest, kbest);
      this->make_a_split(cbest, ibest, kbest, func);
      
      this->_M_find_best_gain_split(cbest, ibest, kbest);
      this->make_a_split_mt(cbest, ibest, kbest, func);
    }
    
    /*   Afterburner, consolidating the tree and cleaning up   */
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*   Afterburner   */
      ic->scan(func, nsvar);
      
      /*   Consolidating the tree  */
      for(auto iv = ic->fvar->_M_tree.postorder_begin(); iv != ic->fvar->_M_tree.postorder_end(); ++iv)
        if(iv.is_leaf() == false) iv->fmax = std::max(iv.left()->fmax, iv.right()->fmax);
      
      /*  Cleaning up  */
      for(auto iv = ic->fvar->_M_tree.leaf_begin(); iv != ic->fvar->_M_tree.leaf_end(); ++iv)
        iv->aux = nullptr;
    }
  }
  

  template<unsigned int _Np, unsigned int _Nv> template<class _Func>
  void bounding_function<_Np, _Nv>::scan_mt(const _Func& func, unsigned nspar, unsigned nsvar)
  {
    auto head = _M_tree.root();
    interval pvbox[_Np+_Nv];
    
    /*   making a small grid in the variable space   */
    head->fvar->_M_make_grid(2u);
    
    /*   make the initial scan    */
    for(unsigned i = 0; i < _Np; i++)
      pvbox[i] = head->box[i];
    
    for(auto iv = head->fvar->_M_tree.leaf_begin(); iv != head->fvar->_M_tree.leaf_end(); ++iv) {
      for(unsigned i = 0; i < _Nv; i++) pvbox[i+_Np] = iv->box[i];
      iv->aux.reset(new proposals(pvbox));
      head->scan(iv, func);
    }
    
    /*   Making many boxes in the parameter space   */
    unsigned ibest = 0u, kbest = 4u;
    auto cbest = _M_tree.leaf_begin();
    
    /*   Making boxes according to the best cut   */
    for(unsigned ns = 0; ns < nspar/4; ++ns) {
      this->_M_find_best_split(cbest, ibest, kbest);
      this->make_a_split_mt(cbest, ibest, kbest, func);
 
      this->_M_find_best_gain_split(cbest, ibest, kbest);
      this->make_a_split_mt(cbest, ibest, kbest, func);
    }
    
   
    /*   Making boxes by alternating the best cut and best middle cut methods  */
    for(unsigned ns = 0; ns < nspar/4; ++ns) {
      this->_M_find_best_middle_split(cbest, ibest);
      this->make_a_split_mt(cbest, ibest, 4u, func);
      
      this->_M_find_best_split(cbest, ibest, kbest);
      this->make_a_split_mt(cbest, ibest, kbest, func);
      
      this->_M_find_best_gain_split(cbest, ibest, kbest);
      this->make_a_split_mt(cbest, ibest, kbest, func);
    }
    
    /*   Afterburner   */
    auto queue = _M_tree.leaf_begin();
    auto iend = _M_tree.leaf_end();
    std::mutex queue_mutex;
    
    auto task = [=, &queue, &queue_mutex]() -> void {
      do {
        queue_mutex.lock();
        auto c = (queue != iend ? queue++ : iend);
        queue_mutex.unlock();
        
        if(c != iend) c->scan(func, nsvar);
        else break;
      } while(true);
    };
    
    unsigned ncpu = std::thread::hardware_concurrency();
    std::thread *worker = new std::thread[ncpu];
    
    for(unsigned i = 0; i < ncpu; i++)
      worker[i] = std::thread(task);
    
    for(unsigned i = 0; i < ncpu; i++)
      worker[i].join();
    
    delete [] worker;
  
    /*   Consolidating the tree and cleaning up   */
    for(auto ic = _M_tree.leaf_begin(); ic != _M_tree.leaf_end(); ++ic) {
      /*   Consolidating the tree  */
      for(auto iv = ic->fvar->_M_tree.postorder_begin(); iv != ic->fvar->_M_tree.postorder_end(); ++iv)
        if(iv.is_leaf() == false) iv->fmax = std::max(iv.left()->fmax, iv.right()->fmax);
      
      /*  Cleaning up  */
      for(auto iv = ic->fvar->_M_tree.leaf_begin(); iv != ic->fvar->_M_tree.leaf_end(); ++iv)
        iv->aux = nullptr;
    }
  }
  
  
  template<unsigned int _Np, unsigned int _Nv>
  template<class _OStream>
  _OStream& bounding_function<_Np, _Nv>::write(_OStream& os) const
  {
    _M_tree.write(os, [=](_OStream& os, const cell& c) -> void {
      os.write(reinterpret_cast<const char *>(c.box), _Np*sizeof(interval));
      
      if(c.fvar) {
        os.put('1');
        c.fvar->write(os);
      } else os.put('0');
    });
    
    return os;
  }
  
  
  template<unsigned int _Np, unsigned int _Nv>
  template<class _IStream>
  _IStream& bounding_function<_Np, _Nv>::read(_IStream& is)
  {
    _M_tree.read(is, [=](_IStream& is, cell& c) -> void {
      is.read(reinterpret_cast<char *>(c.box), _Np*sizeof(interval));
      
      switch(is.get()) {
        case '0': return; break;
        case '1': { c.fvar.reset(new bounding_function<_Nv>); c.fvar->read(is);} break;
        default: throw "_IStream&  bounding_function<_Np, _Nv>::read(_IStream& is) : something wrong with the stream";
      }
    });
    
    return is;
  }
 
}

#endif


