/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/19/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_deductor_h__
#define __mc_deductor_h__ 1

//   std includes
#include <cstdlib>
#include <list>
#include <vector>
#include <queue>
#include <string>
#include <thread>
#include <mutex>
#include <chrono>
#include <atomic>
#include <functional>
#include <type_traits>
#include <condition_variable>

//   project headers
#include <bits/nlo-tags.h>
#include <bits/mc-user.h>
#include <bits/mc-engine.h>
#include <bits/qcd-dglap.h>
#include <bits/mc-default-pdfs.h>




namespace duct {

  
  template<typename _Tag>
  class __deductor_base
  {
  public:
    /*   default destructor  */
    ~__deductor_base() = default;
    
    /*   Create a buffer with n elements   */
    explicit __deductor_base(unsigned int n = 10u) : _M_pool(n), _M_kill(false), _M_newusr(nullptr)
    { for(unsigned k = 0; k < n; k++) _M_consumed.push(k);}
    
    /*   Copy and move are deleted because the mutex and conditional_variable
     *   cannot be copied or moved.   
     */
    __deductor_base(const __deductor_base&) = delete;
    __deductor_base& operator=(const __deductor_base&) = delete;
    
    __deductor_base(__deductor_base&&) = delete;
    __deductor_base& operator=(__deductor_base&&) = delete;
    
    /*  This is the shower generator. It write new events to the buffer.
     *  It stops when it receives kill signal form the analyzer routine.
     */
    void generator(const any_hard_state<_Tag>&, const any_shower_emission<_Tag>&);
    
    /*  This is the analyser routine. It reads elements from a buffer and call the user
     *  routines. It stops and send the kill signal when all the analysises are done.
     */
    void analyser(std::list<std::unique_ptr<user<_Tag>>> *);
    
    /*   Checked if the calculation is finished   */
    bool is_killed() const { return _M_kill;}
    
    /*   Send the kill signal too all of the threads   */
    void kill() { _M_kill = true;}
    
    /*   Attach a new user routine  (the user has to be initialized) */
    void attach_usr(user<_Tag> *usr) {
      _M_newusr = usr;
      _M_user_is_not_empty.notify_one();
    }
    
  private:
    /*   This is the pool where we store the elements of the buffer   */
    std::vector<shower_evolution<_Tag>> _M_pool;
    
    /*    This is the queue of the produced objects.  */
    std::queue<unsigned int> _M_produced;
    
    /*   This is the queue of the consumed objects. They are ready for recycling.  */
    std::queue<unsigned int> _M_consumed;
    
    /*   Signal to kill the generators   */
    std::atomic<bool> _M_kill;
    std::atomic<user<_Tag> *> _M_newusr;
    
    /*    Mutexes for protecting the queues.   */
    std::mutex _M_pmtx, _M_cmtx, _M_umtx;
    
    /*   Conditional variables to syncronize threads    */
    std::condition_variable _M_prod_is_not_empty;
    std::condition_variable _M_cons_is_not_empty;
    std::condition_variable _M_user_is_not_empty;
  };
  

  template<class _Tag>
  void __deductor_base<_Tag>::generator(const any_hard_state<_Tag>& hard, const any_shower_emission<_Tag>& engine)
  {
    /*     initialize the random number generator   */
    std::size_t seed = std::hash<std::thread::id>()(std::this_thread::get_id());
    seed ^= std::hash<void *>()((void *) this) << 1;
    seed ^= std::hash<unsigned long>()(std::chrono::system_clock::to_time_t(std::chrono::system_clock::now())) << 1;
    
    random_engine_type rne(seed);
    shower_emission<_Tag> *eng = engine.get();

    /*   The generator runs until it gets the kill signal.   */
    while(_M_kill == false) {
      /*   pop from the consumed queue   */
      unsigned int current;
      //------
      {
        std::unique_lock<std::mutex> lock(_M_cmtx);
        while(_M_consumed.empty()) {
          if(_M_kill) return;
          _M_cons_is_not_empty.wait(lock);
        }
        
        current = _M_consumed.front(); _M_consumed.pop();
      }
      
      /*   Generate the hard event and do the shower evolution.   */
      _M_pool[current](rne, eng, 1e+120, hard(rne));
      
      /*   puts the product into the produced queue   */
      _M_pmtx.lock(); _M_produced.push(current); _M_pmtx.unlock();
      _M_prod_is_not_empty.notify_one();
    }
  }
  
  
  template<class _Tag>
  void __deductor_base<_Tag>::analyser(std::list<std::unique_ptr<user<_Tag>>> *users)
  {
    /*   This is a fresh calculation, let the generators run...  */
    _M_kill = false;
    
    /*   The analyzer runs until all the user routines are done.   */
    while(_M_kill == false) {
      /*   Check if there is a new user to attachment   */
      user<_Tag> *newusr = _M_newusr.exchange(nullptr);
      if(newusr) users->push_back(std::unique_ptr<user<_Tag>>(newusr));
      
      /*   If all the analyseses are done, wait 20 minutes for new user. 
       *   If there is no new attachment send the kill signal.
       */
      if(users->empty())
      {
        std::unique_lock<std::mutex> lock(_M_umtx);
        user<_Tag> *unew = _M_newusr.exchange(nullptr);

        if(unew) users->push_back(std::unique_ptr<user<_Tag>>(unew));
        else {
          if(_M_user_is_not_empty.wait_for(lock, std::chrono::minutes(10)) == std::cv_status::no_timeout)
            users->push_back(std::unique_ptr<user<_Tag>>(_M_newusr.exchange(nullptr)));
          else {
            _M_kill = true;
            break;
          }
        }
      }
 
      /*   pop from the produced queue   */
      unsigned int current;
      //------
      {
        std::unique_lock<std::mutex> lock(_M_pmtx);
        while(_M_produced.empty()) _M_prod_is_not_empty.wait(lock);
        current = _M_produced.front(); _M_produced.pop();
      }
    
      /*   Analysing the event  */
      const shower_evolution<_Tag>& history = _M_pool[current];
      
      for(auto iu = users->begin(); iu != users->end();) {
        (*iu)->userfunc(history);
        (*iu)->end_of_event();
        
        if((*iu)->is_done()) {
          (*iu)->finish();
          iu = users->erase(iu);
        } else ++iu;
      }
 
      /*   puts the consumed item back to the consumed queue   */
      _M_cmtx.lock(); _M_consumed.push(current); _M_cmtx.unlock();
      _M_cons_is_not_empty.notify_one();
    }
    
    /*   There was a kill signal, notify all blocked generators to quit.   */
    _M_cons_is_not_empty.notify_all();
  }
  
  
  /*    This is a class for representing a deductor process. 
   *    It starts many generator and one analyser thread let them run. 
   *    The instances of this class own the thread and they join in the 
   *    destructor of this class.
   */
  template<class _Tag>
  class process
  {
  public:
    /*   default constructor  */
    process() : _M_usr(nullptr), _M_buffer(nullptr) {}
    
    /*   copy is deleted   */
    process(const process&) = delete;
    process& operator=(const process&) = delete;
    
    /*   move is allowed    */
    process(process&& x)
    : _M_usr(nullptr), _M_buffer(nullptr), _M_threads(std::move(x._M_threads))
    { std::swap(_M_usr, x._M_usr); std::swap(_M_buffer, x._M_buffer);}
    
    process& operator=(process&& x) {
      std::swap(_M_usr, x._M_usr);
      std::swap(_M_buffer, x._M_buffer);
      _M_threads = std::move(x._M_threads);
      return *this;
    }
    
    /*   destructor  */
    ~process() {
      /*   Every thread has to be joined to the main thread before
       *   the data members get destroyed.
       */
      for(auto& gi : _M_threads) gi.join();
      
      /*   deleting the buffer   */
      if(_M_buffer) delete _M_buffer;
      
      /*  deleting the user list  */
      if(_M_usr) delete _M_usr;
    }
    
    /*   Creates and starts the calculation for a given hard process and with the user functions.   */
    process(any_hard_state<_Tag>, any_shower_emission<_Tag>, std::list<std::unique_ptr<user<_Tag>>>&&, unsigned = 0u);
    
    /*   Checked if the calculation is finished   */
    bool is_killed() const { return _M_buffer ? _M_buffer->is_killed() : true;}
    
    /*   Send the kill signal too all of the threads   */
    void kill() { if(_M_buffer) _M_buffer->kill();}
  
    /*   Attach a new user routine  */
    void attach_usr(user<_Tag> *usr) {
      usr->initialize();
      _M_buffer->attach_usr(usr);
    }
    
  private:
    /*    List of user functions and initilizing them.  */
    std::list<std::unique_ptr<user<_Tag>>> *_M_usr;
    
    /*   Buffer to queue the generated events.   */
    __deductor_base<_Tag> *_M_buffer;
    
    /*   Vector of threads   */
    std::vector<std::thread> _M_threads;
  };
  
  
  template<class _Tag>
  process<_Tag>::process(any_hard_state<_Tag> hard, any_shower_emission<_Tag> engine, std::list<std::unique_ptr<user<_Tag>>>&& usr, unsigned ngen)
  {
    /*   move the user list  */
    _M_usr = new std::list<std::unique_ptr<user<_Tag>>>(std::move(usr));
    
    /*   Initialize the user functions   */
    for(const auto& p : *_M_usr) p->initialize();
    
    /*   if it not specified then the number of thread is the number of available CPUs   */
    if(ngen == 0u) ngen = std::thread::hardware_concurrency();
    if(ngen == 0u) ngen = 1u;
    
    /*   Buffer to queue the generated events.   */
    _M_buffer = new __deductor_base<_Tag>(2*ngen+20u);
    
    /*   Vector of threads   */
    _M_threads.resize(ngen+1);
    
    /*   Starting the analyser process   */
    _M_threads[0] = std::thread(&__deductor_base<_Tag>::analyser, _M_buffer, _M_usr);
    
    /*   Starting the generator process(es)   */
    for(unsigned i = 1; i <= ngen; ++i)
      _M_threads[i] = std::thread(&__deductor_base<_Tag>::generator, _M_buffer, hard, engine);
  }
  
  
  
  
  /*   Bsae class for deductor attribute.
   *   It consists of some parameter and the model.
   */
  template<class _Tag>
  struct __deductor_attr
  {
    /*   derieved quantities   */
    double s() const { return Ecm*Ecm;}
    double q2min() const {return minEcm*minEcm;}
    double k2Tmin() const {return minkT*minkT;}
    
    /*   c.m. energy, kT cutoff and minimal c.m. energy   */
    double Ecm, minkT, minEcm;
    
    /*   model   */
    model mdl;
  };
  
  template<class _Tag> inline
  void write(std::ostream& os, const __deductor_attr<_Tag>& attr)
  {
    /*   writting the paramters   */
    os.write(reinterpret_cast<const char *>(&(attr.Ecm)), sizeof(double));
    os.write(reinterpret_cast<const char *>(&(attr.minkT)), sizeof(double));
    os.write(reinterpret_cast<const char *>(&(attr.minEcm)), sizeof(double));
    
    /*   writting model   */
    attr.mdl.write(os);
  }
  
  template<class _Tag> inline
  void read(std::istream& is, __deductor_attr<_Tag>& attr)
  {
    /*   reading parameters   */
    is.read(reinterpret_cast<char *>(&(attr.Ecm)), sizeof(double));
    is.read(reinterpret_cast<char *>(&(attr.minkT)), sizeof(double));
    is.read(reinterpret_cast<char *>(&(attr.minEcm)), sizeof(double));
    
    /*   reading model   */
    attr.mdl.read(is);
  }
  
  
  template<>
  struct __deductor_attr<dis>
  {
    /*   derieved quantities   */
    double s() const { return 4.0*El*Eh;}
    double q2min() const {return minEcm*minEcm;}
    double k2Tmin() const {return minkT*minkT;}
    
    /*   c.m. energy, kT cutoff and minimal c.m. energy   */
    double El, Eh, minkT, minEcm;
    
    /*   model   */
    model mdl;
  };
  
  template<> inline
  void write(std::ostream& os, const __deductor_attr<dis>& attr)
  {
    /*   writting the paramters   */
    os.write(reinterpret_cast<const char *>(&(attr.El)), sizeof(double));
    os.write(reinterpret_cast<const char *>(&(attr.Eh)), sizeof(double));
    os.write(reinterpret_cast<const char *>(&(attr.minkT)), sizeof(double));
    os.write(reinterpret_cast<const char *>(&(attr.minEcm)), sizeof(double));
    
    /*   writting model   */
    attr.mdl.write(os);
  }
  
  template<> inline
  void read(std::istream& is, __deductor_attr<dis>& attr)
  {
    /*   reading parameters   */
    is.read(reinterpret_cast<char *>(&(attr.El)), sizeof(double));
    is.read(reinterpret_cast<char *>(&(attr.Eh)), sizeof(double));
    is.read(reinterpret_cast<char *>(&(attr.minkT)), sizeof(double));
    is.read(reinterpret_cast<char *>(&(attr.minEcm)), sizeof(double));
    
    /*   reading model   */
    attr.mdl.read(is);
  }
  
  
  /*   Run time informartion for the deductor processes. 
   *   This are constant data like model and pdf info.
   *   They have to be in the deductor_kernel and destroyed 
   *   after the processes.
   */
  template<class _Tag>
  struct __deductor_kernel_rtd
  {
    /*   Name of the kernel   */
    std::string _M_name;
    
    /*   Keeps tracking the process ID   */
    unsigned int _M_pid = 1U;
    
    /*   model and other parameters   */
    __deductor_attr<_Tag> _M_attr;
    
    /*   PDF pool   */
    std::list<std::unique_ptr<qcd_hadron>> _M_pdf_pool;
  };
  
  template<>
  struct __deductor_kernel_rtd<epa>
  {
    /*   Name of the kernel   */
    std::string _M_name;
    
    /*   Keeps tracking the process ID   */
    unsigned int _M_pid = 1U;
    
    /*   model and other parameters   */
    __deductor_attr<epa> _M_attr;
  };
  

  /*   Helper class to be able to make the hard states  */
  template<class _Tag, class _Hard>
  struct __make_hard_state {
    static any_hard_state<_Tag> make(const _Hard& mkh, const __deductor_attr<_Tag>& attr) {
      return any_hard_state<_Tag>(mkh(&(attr.mdl), attr.Ecm));
    }
  };
  
  template<class _Hard>
  struct __make_hard_state<dis, _Hard> {
    static any_hard_state<dis> make(const _Hard& mkh, const __deductor_attr<dis>& attr) {
      return any_hard_state<dis>(mkh(&(attr.mdl), attr.El, attr.Eh));
    }
  };
  
  


  /*   This is the abstract deductor base class. It can handle the processes (start/kill)
   *   but doesn't know anything about the shower engine. 
   */
  template<class _Tag>
  class deductor_kernel : protected __deductor_kernel_rtd<_Tag>
  {
  public:
    /*   default constructor and destructor   */
    deductor_kernel() = default;
    virtual ~deductor_kernel() = default;
    
    /*   copy is deleted because the process cannot be copied   */
    deductor_kernel(const deductor_kernel&) = delete;
    deductor_kernel& operator=(const deductor_kernel&) = delete;

    /*   move   */
    deductor_kernel(deductor_kernel&&) = default;
    deductor_kernel& operator=(deductor_kernel&&) = default;

    
    /*   Start a calculation with a given hard process and a user function.   */
    template<class _Hard>
    unsigned int operator()(const char *, _Hard, user<_Tag> *, unsigned = 0u);
    
    /*   Start a calculation with a given hard process and a list of user function.   */
    template<class _Hard>
    unsigned int operator()(const char *, _Hard, std::initializer_list<user<_Tag>*>, unsigned = 0u);
    
    /*   Start a calculation with a given hard process, a user function and adjust the parameters for the process.   */
    template<class _Hard>
    unsigned int operator()(const char *, const shower_engine_attr&, _Hard, user<_Tag> *, unsigned = 0u);
    
    /*   Start a calculation with a given hard process and a list of user function.   */
    template<class _Hard>
    unsigned int operator()(const char *, const shower_engine_attr&, _Hard, std::initializer_list<user<_Tag>*>, unsigned = 0u);
    
    
    /*   Checked if the process is killed   */
    bool is_killed(unsigned int pid) const {
      auto p = _M_proc.find(pid);
      if(p != _M_proc.end())
        return p->second.second.is_killed();
      return true;
    }
    
    /*    Send a kill signal to a process    */
    void kill(unsigned int pid) {
      auto p = _M_proc.find(pid);
      if(p != _M_proc.end()) {
        p->second.second.kill();
        _M_proc.erase(p);
      }
    }
    
    /*   Attach a new user routine   */
    void attach_usr(unsigned int pid, user<_Tag> *usr) {
      auto p = _M_proc.find(pid);
      if(p != _M_proc.end())
        p->second.second.attach_usr(usr);
    }
    
    /*    return  the name of the kernel   */
    const std::string& name() const { return __deductor_kernel_rtd<_Tag>::_M_name;}
    
    /*   returns the name of the process with the given ID    */
    const std::string& name(unsigned pid) const {
      auto p = _M_proc.find(pid);
      if(p != _M_proc.end()) return p->second.first;
      
      return __deductor_kernel_rtd<_Tag>::_M_name;
    }
    
  protected:
    /*    Process table   */
    std::map<unsigned int, std::pair<std::string, process<_Tag>>> _M_proc;
 
  private:
    /*   Cloning the shower engine   */
    virtual any_shower_emission<_Tag> _M_clone_engine(const incoming_hadrons<_Tag>&) = 0;
  };
  
  
  template<class _Tag> template<class _Hard>
  unsigned int deductor_kernel<_Tag>::operator()(const char *name, _Hard mkh, user<_Tag> *usr, unsigned ngen)
  {
    /*  Creating the hard state   */
    any_hard_state<_Tag> hard = __make_hard_state<_Tag, _Hard>::make(mkh, __deductor_kernel_rtd<_Tag>::_M_attr);
    
    /*  Cloning the shower engine   */
    any_shower_emission<_Tag> engine = this->_M_clone_engine(hard);
    
    /*  Putting together the user list   */
    std::list<std::unique_ptr<user<_Tag>>> users;
    users.push_back(std::unique_ptr<user<_Tag>>(usr));
    
    /*   Inserting the new process    */
    unsigned int pid = __deductor_kernel_rtd<_Tag>::_M_pid++;
    
    _M_proc[pid] = {name, process<_Tag>(std::move(hard), std::move(engine), std::move(users), ngen)};
    
    return pid;
  }
  
  
  template<class _Tag> template<class _Hard>
  unsigned int deductor_kernel<_Tag>::
  operator()(const char *name, _Hard mkh, std::initializer_list<user<_Tag>*> usr, unsigned ngen)
  {
    /*  Creating the hard state   */
    any_hard_state<_Tag> hard = __make_hard_state<_Tag, _Hard>::make(mkh, __deductor_kernel_rtd<_Tag>::_M_attr);
    
    /*  Cloning the shower engine   */
    any_shower_emission<_Tag> engine = this->_M_clone_engine(hard);
    
    /*  Putting together the user list   */
    std::list<std::unique_ptr<user<_Tag>>> users;
    for(user<_Tag> *u : usr) users.push_back(std::unique_ptr<user<_Tag>>(u));
   
    /*   Inserting the new process    */
    unsigned int pid = __deductor_kernel_rtd<_Tag>::_M_pid++;
    
    _M_proc[pid] = {name, process<_Tag>(std::move(hard), std::move(engine), std::move(users), ngen)};
    
    return pid;
  }
  

  template<class _Tag> template<class _Hard>
  unsigned int deductor_kernel<_Tag>::
  operator()(const char *name, const shower_engine_attr& pp, _Hard mkh, user<_Tag> *usr, unsigned ngen)
  {
    /*  Creating the hard state   */
    any_hard_state<_Tag> hard = __make_hard_state<_Tag, _Hard>::make(mkh, __deductor_kernel_rtd<_Tag>::_M_attr);
    
    /*  Cloning the shower engine   */
    any_shower_emission<_Tag> engine = this->_M_clone_engine(hard);
    
    /*   Adjusting the parameters only for this process  */
    engine.param(pp);
    
    /*  Putting together the user list   */
    std::list<std::unique_ptr<user<_Tag>>> users;
    users.push_back(std::unique_ptr<user<_Tag>>(usr));
    
    /*   Inserting the new process    */
    unsigned int pid = __deductor_kernel_rtd<_Tag>::_M_pid++;
    
    _M_proc[pid] = {name, process<_Tag>(std::move(hard), std::move(engine), std::move(users), ngen)};
    
    return pid;
  }
  
  
  template<class _Tag> template<class _Hard>
  unsigned int deductor_kernel<_Tag>::
  operator()(const char *name, const shower_engine_attr& pp, _Hard mkh, std::initializer_list<user<_Tag>*> usr, unsigned ngen)
  {
    /*  Creating the hard state   */
    any_hard_state<_Tag> hard = __make_hard_state<_Tag, _Hard>::make(mkh, __deductor_kernel_rtd<_Tag>::_M_attr);
    
    /*  Cloning the shower engine   */
    any_shower_emission<_Tag> engine = this->_M_clone_engine(hard);
    
    /*   Adjusting the parameters only for this process  */
    engine.param(pp);

    /*  Putting together the user list   */
    std::list<std::unique_ptr<user<_Tag>>> users;
    for(user<_Tag> *u : usr) users.push_back(std::unique_ptr<user<_Tag>>(u));
    
    /*   Inserting the new process    */
    unsigned int pid = __deductor_kernel_rtd<_Tag>::_M_pid++;
    
    _M_proc[pid] = {name, process<_Tag>(std::move(hard), std::move(engine), std::move(users), ngen)};
    
    return pid;
  }
  

  
  
  /*   Attribute class that has information of the actual pdf.
   *   PDFs are given at 1GeV.
   */
  template<class _Tag>
  struct deductor_attr : public __deductor_attr<_Tag>
  {
    /*   default constructor   */
    deductor_attr() {
      /*   Default is LHC @ 14TeV  */
      __deductor_attr<_Tag>::Ecm = 14.0_TeV;
      __deductor_attr<_Tag>::minkT = 1.0_GeV;
      __deductor_attr<_Tag>::minEcm = 4.0_GeV;
      
      /*   default running alpha_s and proton pdf  */
      __deductor_attr<_Tag>::mdl.alpha_qcd = qcd_coupling(__herapdflo_as);
      pdf_at_1GeV[2212] = __herapdflo;
    }
    
    /*   constructor   */
    explicit deductor_attr(double ecm, double (*as)(double) = __herapdflo_as,
                           void (*pdf)(double, double *) = __herapdflo)
    {
      /*   Default is LHC @ 14TeV  */
      __deductor_attr<_Tag>::Ecm = ecm;
      __deductor_attr<_Tag>::minkT = 1.0_GeV;
      __deductor_attr<_Tag>::minEcm = 4.0_GeV;
      
      /*   default running alpha_s and proton pdf  */
      __deductor_attr<_Tag>::mdl.alpha_qcd = qcd_coupling(as);
      pdf_at_1GeV[2212] = pdf;
    }
    
    /*   points to the pdf function  */
    std::map<int, void (*)(double, double *)> pdf_at_1GeV;
  };
  
  
  /*   Specialization for DIS    */
  template<>
  struct deductor_attr<dis> : public __deductor_attr<dis>
  {
    /*   default constructor   */
    deductor_attr() {
      /*   Default is HERA @ 27.5GeV, 920GeV  */
      El = 27.5_TeV;
      Eh = 920.0_TeV;
      minkT = 1.0_GeV;
      minEcm = 4.0_GeV;
      
      /*   default running alpha_s and proton pdf  */
      mdl.alpha_qcd = qcd_coupling(__herapdflo_as);
      pdf_at_1GeV[2212] = __herapdflo;
    }
  
    /*   default constructor   */
    deductor_attr(double el, double eh, double (*as)(double) = __herapdflo_as,
                  void (*pdf)(double, double *) = __herapdflo)
    {
      /*   Default is HERA @ 27.5GeV, 920GeV  */
      El = el;
      Eh = eh;
      minkT = 1.0_GeV;
      minEcm = 4.0_GeV;
      
      /*   default running alpha_s and proton pdf  */
      mdl.alpha_qcd = qcd_coupling(as);
      pdf_at_1GeV[2212] = pdf;
    }
    
    /*   points to the pdf function  */
    std::map<int, void (*)(double, double *)> pdf_at_1GeV;
  };
  
  
  /*  In e+e- annihilation there is no pdf */
  template<>
  struct deductor_attr<epa> : public __deductor_attr<epa>
  {
    /*   default constructor   */
    deductor_attr() {
      /*   Default is LEP II @ 209GeV  */
      Ecm = 209.0_GeV;
      minkT = 1.0_GeV;
      minEcm = 0.0_GeV;
    }
    
    /*   constructor   */
    explicit deductor_attr(double ecm, double kT = 1.0) {
      Ecm = ecm;
      minkT = kT;
      minEcm = 0.0_GeV;
    }
  };
  
  
  
  /*   Class deductor to generate and analyse events
   *   for DIS and hadron-hadron collision processes.
   */
  template<class _Tag, class _Ordering = ordering::lambda>
  class deductor : public deductor_kernel<_Tag>
  {
  public:
    /*   This constructor creates a kernel, initializes the engine
     *   and saves the auxialiary data to file.
     */
    deductor(const char *, const deductor_attr<_Tag>&);

    /*   This constructor reconstructs a kernel from the disk.  */
    explicit deductor(const char *);
    
 
    /*   copy is deleted because the process cannot be copied   */
    deductor(const deductor&) = delete;
    deductor& operator=(const deductor&) = delete;
    
    /*   move   */
    deductor(deductor&&) = default;
    deductor& operator=(deductor&&) = default;
    
    /*   destructor   */
    ~deductor() = default;

    /*   Changing the engine parameters   */
    void engine_attr(const shower_engine_attr& pp) {
      _M_engine.attr(pp);
    }
    
    /*   Changing the kT cutoff   */
    void kT_cutoff(double kT) {
      shower_engine_attr p = _M_engine.attr();
      p.kT_cutoff = kT;
      _M_engine.attr(p);
    }
 
    /*   Changing the virtuality cutoff   */
    void virt_cutoff(double virt) {
      shower_engine_attr p = _M_engine.attr();
      p.virt_cutoff = virt;
      _M_engine.attr(p);
    }
  
    /*   Changing the max allowed color suppression   */
    void max_color_suppression(unsigned imax) {
      shower_engine_attr p = _M_engine.attr();
      p.max_color_suppression = imax;
      _M_engine.attr(p);
    }
    
  private:
    /*   Cloning the shower engine   */
    any_shower_emission<_Tag> _M_clone_engine(const incoming_hadrons<_Tag>& hh)
    {
      /*   Initializing the engine   */
      _M_engine.initialize(_M_dir.c_str(), hh,
                           __deductor_kernel_rtd<_Tag>::_M_attr.q2min(),
                           __deductor_kernel_rtd<_Tag>::_M_attr.s(),
                           256, 16);
      
      /*   cloning the engine   */
      return any_shower_emission<_Tag>(_M_engine);
    }
    
    /*   Name of the directory to store runtime data.   */
    std::string _M_dir;
    
    /*   The prototype of the shower engine. This is copied in the multi-threaded version.
     *   Thus the big trees of the splitting kernel approximation are stored once.  */
    shower_engine<_Tag, _Ordering> _M_engine;
  };
  

  template<class _Tag, class _Ordering>
  deductor<_Tag, _Ordering>::deductor(const char *name, const deductor_attr<_Tag>& attr)
  {
    /*    directory name for the auxiliary data   */
    __deductor_kernel_rtd<_Tag>::_M_name = name;
    
    _M_dir = std::getenv("HOME");
    _M_dir += "/.deductor/";
    _M_dir += name;
    _M_dir += '/';
    
    std::string dir = _M_dir;
    dir += _Ordering::name;
    dir += '/';
    dir += _Tag::name;
    dir += '/';

    if(::__nlo_helper_mkdir(dir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
      std::cerr<<"Unable to make directory : "<<dir<<std::endl;
      throw "Unable to make directory";
    }

    /*   copying the attributes (some parametes and the model)   */
    __deductor_kernel_rtd<_Tag>::_M_attr = attr;
    
    /*   saving the model to file   */
    std::fstream file((dir+"kernel-attr.dat").c_str(), std::ios::out);
    
    write(file, __deductor_kernel_rtd<_Tag>::_M_attr);
    file.flush();
    file.close();
    
    
    /*   setting up pdfs   */
    model *mdl = &__deductor_kernel_rtd<_Tag>::_M_attr.mdl;
    
    for(const auto& pdf : attr.pdf_at_1GeV)
      if(pdf.first > 0)
      {
        std::cout<<"Generating the pdf table of hadron("<<pdf.first<<") ...    "<<std::flush;
        
        /*   generate the pdf table   */
        qcd_dglap *tbl = new qcd_dglap(__dglap_mode<_Ordering>::value, 1.0, pdf.second, mdl);
        qcd_antihadron *antitbl = new qcd_antihadron(tbl);
        
        __deductor_kernel_rtd<_Tag>::_M_pdf_pool.push_back(std::unique_ptr<qcd_hadron>(tbl));
        __deductor_kernel_rtd<_Tag>::_M_pdf_pool.push_back(std::unique_ptr<qcd_hadron>(antitbl));
        
        std::cout<<"OK"<<std::endl;
        
        /*   inserting the hadron and its pdf  */
        auto ih = mdl->hadron_table.find(pdf.first);
        
        if(ih == mdl->hadron_table.end()) {
          std::string name = "hadron(";
          name += std::to_string(pdf.first)+")";
          
          mdl->hadron_table[pdf.first] = {pdf.first, name.c_str(), tbl};
        } else ih->second.pdf = tbl;
        
        
        /*   inserting the antihadron and its pdf  */
        ih = mdl->hadron_table.find(-pdf.first);
        
        if(ih == mdl->hadron_table.end()) {
          std::string name = "hadron(";
          name += std::to_string(-pdf.first)+")";
          
          mdl->hadron_table[-pdf.first] = {pdf.first, name.c_str(), antitbl};
        } else ih->second.pdf = antitbl;
      }
    
    /*   saving the pdf tables to files   */
    for(const auto& ih : mdl->hadron_table)
      if(ih.first > 0 && ih.second.pdf != nullptr) {
        /*   opening file   */
        file.open((dir+"hadron-"+std::to_string(ih.first)+".pdg").c_str(), std::ios::out);
        
        /*   writing pdf to file   */
        dynamic_cast<const qcd_dglap *>(ih.second.pdf) -> write(file);
        
        /*  closing file   */
        file.flush();
        file.close();
      }

    
    /*   Initializing the engine   */
    _M_engine = shower_engine<_Tag, _Ordering>(__deductor_kernel_rtd<_Tag>::_M_attr.k2Tmin(), mdl);
    
    for(const auto& ih : mdl->hadron_table)
      if(ih.first > 0 && ih.second.pdf != nullptr)
        _M_engine.initialize(_M_dir.c_str(), incoming_hadrons<_Tag>(ih.first),
                             __deductor_kernel_rtd<_Tag>::_M_attr.q2min(),
                             __deductor_kernel_rtd<_Tag>::_M_attr.s(),
                             256, 16);
  }
  
  
  template<class _Tag, class _Ordering>
  deductor<_Tag, _Ordering>::deductor(const char *name)
  {
    /*    directory name for the auxiliary data   */
    __deductor_kernel_rtd<_Tag>::_M_name = name;
    
    _M_dir = std::getenv("HOME");
    _M_dir += "/.deductor/";
    _M_dir += name;
    _M_dir += '/';
    
    std::string dir = _M_dir;
    dir += _Ordering::name;
    dir += '/';
    dir += _Tag::name;
    dir += '/';

    /*    Checking if the kernel exists   */
    struct ::stat sb;
    
    if(::stat((dir+"kernel-attr.dat").c_str(), &sb)) {
      std::cerr<<"The kernel \'"<<name<<"\' doesn't exist. Please create a kernel before try to use it!"<<std::endl;
      throw "Kernel is not defined.";
    }
    
    
    /*   reading back the model   */
    std::fstream file((dir+"kernel-attr.dat").c_str(), std::ios::in);
    
    read(file, __deductor_kernel_rtd<_Tag>::_M_attr);
    file.flush();
    file.close();

    /*   reading back the pdf tables   */
    model *mdl = &__deductor_kernel_rtd<_Tag>::_M_attr.mdl;

    for(auto& ih : mdl->hadron_table)
      if(ih.first > 0) {
        /*   opening file   */
        file.open((dir+"hadron-"+std::to_string(ih.first)+".pdg").c_str(), std::ios::in);

        if(file.is_open())
        {
          /*   reading the pdf from the file   */
          qcd_dglap *tbl = new qcd_dglap;
          qcd_antihadron *antitbl = new qcd_antihadron(tbl);
          
          tbl->read(file);
          
          __deductor_kernel_rtd<_Tag>::_M_pdf_pool.push_back(std::unique_ptr<qcd_hadron>(tbl));
          __deductor_kernel_rtd<_Tag>::_M_pdf_pool.push_back(std::unique_ptr<qcd_hadron>(antitbl));
          
          ih.second.pdf = tbl;
          mdl->hadron_table[-ih.first].pdf = antitbl;
          
          /*  closing file   */
          file.flush();
          file.close();
        }
      }
    
    /*   Creating the engine   */
    _M_engine = shower_engine<_Tag, _Ordering>(__deductor_kernel_rtd<_Tag>::_M_attr.k2Tmin(), mdl);
  }
  
 
  
  
  /*   Class deductor to generate and analyse events
   *   for e+e- annihilation processes.
   */ 
  template<class _Ordering>
  class deductor<epa, _Ordering> : public deductor_kernel<epa>
  {
  public:
    /*   This constructor creates a kernel, initializes the engine
     *   and saves the auxialiary data to file.
     */
    deductor(const char *, const deductor_attr<epa>&);
    
    /*   This constructor reconstructs a kernel from the disk.  */
    explicit deductor(const char *);
    
    
    /*   copy is deleted because the process cannot be copied   */
    deductor(const deductor&) = delete;
    deductor& operator=(const deductor&) = delete;
    
    /*   move   */
    deductor(deductor&&) = default;
    deductor& operator=(deductor&&) = default;
    
    /*   destructor   */
    ~deductor() = default;
    
    /*   Changing the engine parameters   */
    void engine_attr(const shower_engine_attr& pp) {
      _M_engine.param(pp);
    }
    
    /*   Changing the kT cutoff   */
    void kT_cutoff(double kT) {
      shower_engine_attr p = _M_engine.attr();
      p.kT_cutoff = kT;
      _M_engine.attr(p);
    }
    
    /*   Changing the virtuality cutoff   */
    void virt_cutoff(double virt) {
      shower_engine_attr p = _M_engine.attr();
      p.virt_cutoff = virt;
      _M_engine.attr(p);
    }
    
    /*   Changing the max allowed color suppression   */
    void max_color_suppression(unsigned imax) {
      shower_engine_attr p = _M_engine.attr();
      p.max_color_suppression = imax;
      _M_engine.attr(p);
    }
    
  private:
    /*   Cloning the shower engine   */
    any_shower_emission<epa> _M_clone_engine(const incoming_hadrons<epa>&)
    {
      /*   Initializing the engine   */
      _M_engine.initialize(_M_dir.c_str(), __deductor_kernel_rtd<epa>::_M_attr.s(), 256, 16);
      
      /*   cloning the engine   */
      return any_shower_emission<epa>(_M_engine);
    }
    
    /*   Name of the directory to store runtime data.   */
    std::string _M_dir;
    
    /*   The prototype of the shower engine. This is copied in the multi-threaded version.
     *   Thus the big trees of the splitting kernel approximation are stored once.  */
    shower_engine<epa, _Ordering> _M_engine;
  };
  
  
  template<class _Ordering>
  deductor<epa, _Ordering>::deductor(const char *name, const deductor_attr<epa>& attr)
  {
    /*    directory name for the auxiliary data   */
    __deductor_kernel_rtd<epa>::_M_name = name;
    
    _M_dir = std::getenv("HOME");
    _M_dir += "/.deductor/";
    _M_dir += name;
    _M_dir += '/';
    
    std::string dir = _M_dir;
    dir += _Ordering::name;
    dir += '/';
    dir += epa::name;
    dir += '/';
    
    if(::__nlo_helper_mkdir(dir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
      std::cerr<<"Unable to make directory : "<<dir<<std::endl;
      throw "Unable to make directory";
    }
    
    /*   copying the attributes (some parametes and the model)   */
    __deductor_kernel_rtd<epa>::_M_attr = attr;
    
    /*   saving the model to file   */
    std::fstream file((dir+"kernel-attr.dat").c_str(), std::ios::out);
    
    write(file, __deductor_kernel_rtd<epa>::_M_attr);
    file.flush();
    file.close();
    
    
    /*   Creating the engine   */
    _M_engine = shower_engine<epa, _Ordering>(__deductor_kernel_rtd<epa>::_M_attr.k2Tmin(),
                                              &__deductor_kernel_rtd<epa>::_M_attr.mdl);
    
    /*   Initialize the engine   */
    _M_engine.initialize(_M_dir.c_str(), __deductor_kernel_rtd<epa>::_M_attr.s(), 256, 16);
  }
  
  
  template<class _Ordering>
  deductor<epa, _Ordering>::deductor(const char *name)
  {
    /*    directory name for the auxiliary data   */
    __deductor_kernel_rtd<epa>::_M_name = name;
    
    _M_dir = std::getenv("HOME");
    _M_dir += "/.deductor/";
    _M_dir += name;
    _M_dir += '/';
    
    std::string dir = _M_dir;
    dir += _Ordering::name;
    dir += '/';
    dir += epa::name;
    dir += '/';
    
    /*    Checking if the kernel exists   */
    struct ::stat sb;
    
    if(::stat((dir+"kernel-attr.dat").c_str(), &sb)) {
      std::cerr<<"The kernel \'"<<name<<"\' doesn't exist. Please create a kernel before try to use it!"<<std::endl;
      throw "Kernel is not defined.";
    }
    
    
    /*   reading back the model   */
    std::fstream file((dir+"kernel-attr.dat").c_str(), std::ios::in);
    
    read(file, __deductor_kernel_rtd<epa>::_M_attr);
    file.flush();
    file.close();
    
    
    /*   Creating the engine   */
    _M_engine = shower_engine<epa, _Ordering>(__deductor_kernel_rtd<epa>::_M_attr.k2Tmin(),
                                              &__deductor_kernel_rtd<epa>::_M_attr.mdl);
    
    /*   Initialize the engine   */
    _M_engine.initialize(_M_dir.c_str(), __deductor_kernel_rtd<epa>::_M_attr.s(), 256, 16);
  }

 
}

#endif
