/*  Copyright (C) 2010-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 6/17/10.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __nlo_mc_emission_h__
#define __nlo_mc_emission_h__ 1

// standard c++ includes

// nlo++ includes
#include <bits/nlo-event.h>
#include <bits/nlo-mapping.h>
#include <bits/qcd-color-overlap.h>
#include <bits/qcd-color-overlap-u3.h>


namespace duct {
  
  
  /*  This function selects the color operator. This can be different 
   *  in different shower implementations. This function select the 
   *  color operator in dipole shower. 
   */
  double selecting_color_operator(random_engine_type&, splitting_variables<lo_shower> *, const statistical_basis&);
  double selecting_color_operator_u3(random_engine_type&, splitting_variables<lo_shower> *, const statistical_basis&);
  
  
  
  /*  Class for recording every stage of the shower evolution. 
   *  It consists of
   *   => the event
   *   => spitting variable for the next stage
   *   => shower time of the next splitting
   *   => weight of the current event
   *   => reference momentum for defining shower time
   */
  template<class _Tag> 
  struct shower_stage
  {
    //----- public types -----
    typedef event<_Tag> event_type;	
    
    /*  The event of the state. This contains every informations about the
     *  event, momenta, flavors, color connections, incoming objects,...
     */
    event<_Tag> state;
    
    /*  Splitting varibles: This gives the splitting varibles of the next stage. 
     *  It consists of delta, u, phi, emitter, spectator, color related variables. 
     */
    splitting_variables<lo_shower> vars;
    
    /*  This is the refernce momentum that help us to define shower time.
     *  We have a copy in very stage, since we want to resume shower from 
     *  an intermediate state. 
     */
    lorentzvector<double> qnull;
    
    /*  This is maximum allowed transverse momentum square in the initial
     *  and final state splitting.
     */
    double max_k2T_initial_state;
    double max_k2T_final_state;
    
    /*  This is maximum allowed virtuality in the initial
     *  and final state splitting.
     */
    double max_q2_initial_state;
    double max_q2_final_state;

    /*  The posibble weight of the event. This weight can come from color
     *  factors or from spin correlation (when we add spin correlations).
     */
    double weight;
    
    /*  The shower time. This state is evolved without emission upto this time.
     */
    double time;
    
    /*   It is true if the state is consired as U(Nc) state. 
     */
    bool is_color_u3;
  };
  
  /*    calculate the color weight of shower stage   */
  template<class _Tag>
  double color_weight(color_overlap *overlap, color_overlap_u3 *overlap_u3, const shower_stage<_Tag>& s)
  {
    if(s.is_color_u3) return overlap_u3->operator()(s.state);
    return overlap->operator()(s.state);
  }
  
  /*   This is the overlap calculator function. Since the thread_local is not really implemented in 
   *   g++ or clang++, I had to find a workaround. It looks to me OK. We should use thread_local 
   *   instead of __thread.
   */
  template<class _Tag>
  double color_weight(const shower_stage<_Tag>& s)
  {
    static __thread color_overlap *overlap = new color_overlap;
    static __thread color_overlap_u3 *overlap_u3 = new color_overlap_u3;
    
    return color_weight(overlap, overlap_u3, s);
  }

  
  /*  Some general parameters of the engine   */
  struct shower_engine_attr
  {
    /*   constructor to be able to do brace construction   */
    shower_engine_attr() : kT_cutoff(0.0), virt_cutoff(0.0), max_color_suppression(0u) {}
    shower_engine_attr(double kT, double virt, unsigned imax) : kT_cutoff(kT), max_color_suppression(imax) {}
    
    double kT_cutoff;
    double virt_cutoff;
    unsigned max_color_suppression;
  };

  /* Abstract class for makeing trial emission.
   * It is defined for the various events. This 
   * is the engine of the shower....
   */
  template<class _Tag> 
  class shower_emission
  {
  public:
    //----- public types -----
    typedef shower_stage<_Tag> stage;	
    
    //----- destructor -----
    virtual ~shower_emission() = default;
    
    /*  This function makes a trial emission. It returns the splitting variables 
     *  and the new shower time. If the emission is not possible the shower time 
     *  is unchaged and returns zero weight otherwise returns the weight of the 
     *  splitting.
     */
    virtual double operator()(random_engine_type&, stage&, double, double) = 0;

    /*   function to clone the objects   */
    virtual shower_emission * clone() const = 0;
    
    /*   setting the parameters   */
    void attr(const shower_engine_attr& pp) {
      _M_attr = pp;
    }
  
    void attr(double kT_cutoff, double virt_cutoff, unsigned imax) {
      _M_attr = {kT_cutoff, virt_cutoff, imax};
    }
    
    /*   getting the parameters  */
    const shower_engine_attr& attr() const {
      return _M_attr;
    }
    
  private:
    shower_engine_attr _M_attr;
  };
  
 
  /*   This class can store any shower_emission object.
   *   It is copyable, moveable, so, it can be used
   *   in multi-threaded enviroment safely.
   */
  template<class _Tag>
  class any_shower_emission
  {
  public:
    /*   no default constructor   */
    any_shower_emission() : _M_ptr(nullptr) {}
    
    /*    constructors  */
    explicit any_shower_emission(const shower_emission<_Tag>& r)
    : _M_ptr(r.clone()) {}
    
    /*    copy   */
    any_shower_emission(const any_shower_emission& h)
    : _M_ptr(h._M_ptr->clone()) {}
    
    any_shower_emission& operator=(const any_shower_emission& h)
    {
      if(this != &h) {
        this->~any_hard_state();
        _M_ptr = h._M_ptr->clone();
      }
      
      return *this;
    }
    
    /*   move   */
    any_shower_emission(any_shower_emission&& h) {
      _M_ptr = nullptr;
      std::swap(_M_ptr, h._M_ptr);
    }
    
    any_shower_emission& operator=(any_shower_emission&& h) {
      std::swap(_M_ptr, h._M_ptr);
      return *this;
    }
    
    /*    destructor   */
    ~any_shower_emission() {
      if(_M_ptr) delete _M_ptr;
    }
    
    /*  this operator generates an emission   */
    double operator()(random_engine_type& rne, shower_stage<_Tag>& s, double tmU, double tmL) const {
      return _M_ptr->operator()(rne, s, tmU, tmL);
    }
    
    /*   covert to shower_emission reference and pointer   */
    operator shower_emission<_Tag>& () const { return *_M_ptr;}
    
    /*    get the address of the shower_emission object   */
    shower_emission<_Tag>* get() const { return  _M_ptr;}

    
    /*   setting the parameters   */
    void attr(const shower_engine_attr& pp) { _M_ptr->param(pp);}
    void attr(double kT_cutoff, unsigned imax) { _M_ptr->param(kT_cutoff, imax);}
    
    /*   getting the parameters  */
    const shower_engine_attr& attr() const { return _M_ptr->param();}
    
  private:
    /*    pointer to the hard state   */
    shower_emission<_Tag> *_M_ptr;
  };

  
  
  
  /*  Abstract class for interfacing the hard part of the cross section.
   *  It is a callable object and it return a new hard event. These are 
   *  usually the simple 2->2 like processes. 
   */
  template<class _Tag>
  struct hard_state : public shower_stage<_Tag>, public incoming_hadrons<_Tag>
  {
    /*   virtual destructor   */
    virtual ~hard_state() = default;
    
    /*   function to clone the objects   */
    virtual hard_state * clone() const = 0;
    
    /*  this operator generates a new hard event   */
    virtual shower_stage<_Tag>& operator()(random_engine_type&) = 0;
  };


  /*   This class can store any hard_state object. 
   *   It is copyable, moveable, so, it can be used 
   *   in multi-threaded enviroment safely.
   */
  template<class _Tag>
  class any_hard_state
  {
  public:
    /*   no default constructor   */
    any_hard_state() : _M_ptr(nullptr) {}
    
    /*    constructors  */
    explicit any_hard_state(const hard_state<_Tag>& r)
    : _M_ptr(r.clone()) {}
        
    /*    copy   */
    any_hard_state(const any_hard_state& h)
    : _M_ptr(h._M_ptr->clone()) {}
    
    any_hard_state& operator=(const any_hard_state& h)
    {
      if(this != &h) {
        this->~any_hard_state();
        _M_ptr = h._M_ptr->clone();
      }
      
      return *this;
    }
    
    /*   move   */
    any_hard_state(any_hard_state&& h) {
      _M_ptr = nullptr;
      std::swap(_M_ptr, h._M_ptr);
    }
    
    any_hard_state& operator=(any_hard_state&& h) {
      std::swap(_M_ptr, h._M_ptr);
      return *this;
    }
    
    /*    destructor   */
    ~any_hard_state() {
      if(_M_ptr) delete _M_ptr;
    }
    
    /*  this operator generates a new hard event   */
    shower_stage<_Tag>& operator()(random_engine_type& rne) const {
      return _M_ptr->operator()(rne);
    }
    
    /*   covert to hard_state reference  */
    operator hard_state<_Tag>& () const { return *_M_ptr;}
    
    /*    get the address of the hard_state object   */
    hard_state<_Tag>* get() const { return  _M_ptr;}

  private:
    /*    pointer to the hard state   */
    hard_state<_Tag> *_M_ptr;
  };
  
}    //   namespace nlo

#endif

