/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_virt_h__
#define __mc_engine_virt_h__ 1

//   standard c++ includes
#include <memory>
#include <limits>

//   deductor includes
#include "bits/mc-engine.h"
#include "bits/mc-splitting-functions.h"
#include "bits/math-imp-sampling.h"


namespace duct {
  
  /*   Exact splitting kernel for the initial state splittings.  */
  template<>
  class splitting_kernel<ini, ordering::lambda>
  {
  public:
    /*     constructor   */
    explicit splitting_kernel(const model *mdl) : _M_mdl(mdl) {}
    
    /*    initialize the function by fixing the parameters (final state spectator)  */
    double bindF(int hdr, int fl, double m2T, double m2b, double eta, double Q2, double xi, double xT) {
      _M_pars = {false, hdr, fl, 1, xi, xT, m2T/Q2, _M_mdl->mass2(fl)/Q2, m2b/Q2, eta, Q2};
      return _M_y_mapping.refresh(this);
    }
    
    /*   initialize the function by fixing the parameters (initial state spectator)  */
    double bindI(int hdr, int fl, double m2T, double m2b, double eta, double Q2) {
      _M_pars = {false, hdr, fl, 0, m2b/(m2b+Q2), 0.0, m2T/Q2, _M_mdl->mass2(fl)/Q2, m2b/Q2, eta, Q2};
      return _M_y_mapping.refresh(this);
    }
    
    /*    initialize the function by fixing the parameters (final state spectator, heavy quark)  */
    double bindFH(int hdr, int fl, double m2T, double m2b, double eta, double Q2, double xi, double xT) {
      _M_pars = {true, hdr, fl, 1, xi, xT, m2T/Q2, _M_mdl->mass2(fl)/Q2, m2b/Q2, eta, Q2};
      return _M_y_mapping.refresh_hq(this);
    }
    
    /*   initialize the function by fixing the parameters (initial state spectator, heavy quark)  */
    double bindIH(int hdr, int fl, double m2T, double m2b, double eta, double Q2) {
      _M_pars = {true, hdr, fl, 0, m2b/(m2b+Q2), 0.0, m2T/Q2, _M_mdl->mass2(fl)/Q2, m2b/Q2, eta, Q2};
      return _M_y_mapping.refresh_hq(this);
    }
    
    /*   calculate the function  (with transverse momentum veto)  */
    double operator()(double kappa, double chi, double k2Tmin = 0.0, double k2Tmax = std::numeric_limits<double>::max()) const {
      if(_M_pars.hq) return _M_quark_hq(kappa, chi, k2Tmin, k2Tmax);
      if(_M_pars.fl) return _M_quark(kappa, chi, k2Tmin, k2Tmax);
      return _M_gluon(kappa, chi, k2Tmin, k2Tmax);
    }

    /*    vertex as function of a random variable   */
    splitting_variables<lo_shower>::zeta_flavor vertex(double) const;
    
    /*   azimuthal angle as function of a random variable  */
    double azimuthal_angle(int emitter, int spectator, double rphi) const {
      if(emitter != spectator) return 2.0*M_PI*__mc_helper::newton_root_finder_phi(this->_M_func_emit, rphi);
      else return 2.0*M_PI*rphi;
    }
    
    /*   mappings:  kappa <--> y   */
    double mapping_y2k(double y) const { return _M_y_mapping.mapping_tau2k(_M_mdl->alpha_qcd.tauk(y*_M_pars.Q2));}
    double mapping_k2y(double kappa) const { return _M_mdl->alpha_qcd.tauk_inverse(_M_y_mapping.mapping_k2tau(kappa))/_M_pars.Q2;}
    
    void mapping_k2y(double kappa, double& y, double &yjac) const {
      double tau = _M_y_mapping.mapping_k2tau(kappa);
      y = _M_mdl->alpha_qcd.tauk_inverse(tau)/_M_pars.Q2;
      yjac = _M_y_mapping.jacobian(tau);
    }
    
    /*   mappings:  chi <--> z   */
    double mapping_z2chi(double y, double z) const {
      z_mapping zmap; zmap.refresh(this, y);
      return zmap.mapping_z2chi(z);
    }
    
    double mapping_chi2z(double y, double chi) const {
      z_mapping zmap; zmap.refresh(this, y);
      return zmap.mapping_chi2z(chi);
    }
    
    void mapping_chi2z(double y, double chi, double& z, double& zjac) const {
      z_mapping zmap; zmap.refresh(this, y);
      zmap.mapping_chi2z(chi, z, zjac);
    }
    
    /*   returns the threshold and freezing scale in kappa variable  */
    double kappaT() const { return _M_y_mapping.kappaT;}
    double kappaF() const { return _M_y_mapping.kappaF;}
    
    /*   returns the flavor of the emitter  */
    int emitter_flavor() const { return _M_pars.fl;}
    
    /*   returns the pointer to the model  */
    const model * get_model() const { return _M_mdl;}
    
  private:
    /*   pointer to the model   */
    const model *_M_mdl;
    
    /*   Exact splitting function:
     *   These variables are mutable because they are kind of return values.
     */
    mutable __mc_splitting_functions::emitting_gluon_ini _M_func_emit;
    mutable double _M_prob_proc[11];
    
    /*   splitting functions for gluon and quark emitters   */
    double _M_gluon(double, double, double, double) const;
    double _M_quark(double, double, double, double) const;
    double _M_quark_hq(double, double, double, double) const;
    
    /*  Class for making the z mapping faster and simpler.   */
    struct z_mapping
    {
      //----- refreshing the y dependence ----
      void refresh(const splitting_kernel *, double);
      void refresh_hq(const splitting_kernel *, double);
      
      //----- mapping from z to chi ------
      double mapping_z2chi(double z) const {
        double oz = 1.0-z;
        if(is_gluon) oz /= z;
        return std::log((1.0+as*std::log(oz))/Lmax)/J;
      }
      
      //----- mapping from chi to z ------
      double mapping_chi2z(double chi) const {
        double oz = std::exp((Lmax*std::exp(chi*J)-1.0)/as);
        return is_gluon ? 1.0/(1.0+oz) : 1.0-oz;
      }
      
      //----- mapping from chi to z with jacobian ------
      void mapping_chi2z(double chi, double& z, double& jac) const {
        double oz = std::exp((Lmax*std::exp(chi*J)-1.0)/as);
        z = is_gluon ? 1.0/(1.0+oz) : 1.0-oz;
        jac = J/as*(1.0-z);
        if(is_gluon) jac *= z*(1.0+as*std::log(1.0/z-1.0))/(1.0+as*std::log(1.0-z));
      }
      
      //----- cached variables ------
      bool is_gluon;
      double J, Lmax, as;
    };
    
    /*  Class for mapping out some y dependence  */
    struct y_mapping
    {
      //----- refreshing the spline ----
      double refresh(const splitting_kernel *);
      double refresh_hq(const splitting_kernel *);
      
      /*   making the transformation and its jacobian   */
      double mapping_k2tau(double kappa) const { return spline.mapping_chi2x(kappa);}
      double mapping_tau2k(double tau) const { return spline.mapping_x2chi(tau);}
      double jacobian(double tau) const { return spline.jacobian(tau)/spline.integral();}
      
      //---- data members ----
      importance_sampling<20U> spline;
      double kappaT, kappaF;
    } _M_y_mapping;
    
    /*   storage for bindding the parameters   */
    struct {
      bool hq;
      int hdr, fl, spec;
      double xi, xT, mT, ma, mb, eta, Q2;
    } _M_pars;
  };

  
  /*   Exact splitting kernel for the final state splittings   */
  template<>
  class splitting_kernel<fin, ordering::lambda>
  {
  public:
    //    constructors
    explicit splitting_kernel(const model *mdl) : _M_mdl(mdl) {}
    
    /*    initialize the function by fixing the parameters (initial state spectator)  */
    double bindI(int fl, double m2T, double rl, double ymax, double pQ, double xi, double mu) {
      _M_pars = {fl, 'I', rl, ymax, pQ, _M_mdl->mass2(fl)/pQ, m2T/pQ, xi, mu};
      return _M_y_mapping.refresh(this);
    }
    
    /*    initialize the function by fixing the parameters (final state spectator)  */
    double bindF(int fl, double m2T, double rl, double ymax, double pQ, double xi, double mu) {
      _M_pars = {fl, 'F', rl, ymax, pQ, _M_mdl->mass2(fl)/pQ, m2T/pQ, xi, mu};
      return _M_y_mapping.refresh(this);
    }
   
    /*    initialize the function by fixing the parameters  (decay shower, initial state spectator)   */
    double bindD(int fl, double m2T, double rl, double ymax, double pQ) {
      _M_pars = {fl, 'D', rl, ymax, pQ, _M_mdl->mass2(fl)/pQ, m2T/pQ, -1.0, -1.0};
      return _M_y_mapping.refresh(this);
    }

    /*    initialize the function by fixing the parameters  (spectator set is massless, final state spectator)   */
    double bindS(int fl, double m2T, double ymax, double pQ) {
      _M_pars = {fl, 'S', ymax, ymax, pQ, _M_mdl->mass2(fl)/pQ, m2T/pQ, 1.0, 0.0};
      return _M_y_mapping.refresh(this);
    }
    
    /*   calculate the function  (with transverse momentum veto)    */
    double operator()(double kappa, double chi, double k2Tmin = 0.0, double k2Tmax = std::numeric_limits<double>::max()) const {
      switch(_M_pars.spec) {
        case 'I': case 'F': return _M_funcFI(kappa, chi, k2Tmin, k2Tmax); break;
        case 'D': return _M_funcD(kappa, chi, k2Tmin, k2Tmax); break;
        case 'S': return _M_funcS(kappa, chi, k2Tmin, k2Tmax); break;
        default : return 0.0;
      }
    }
    
    /*   azimuthal angle as function of a random variable  */
    double azimuthal_angle(int emitter, int spectator, double rphi) const {
      if(emitter != spectator) return 2.0*M_PI*__mc_helper::newton_root_finder_phi(this->_M_func_emit, rphi);
      else return 2.0*M_PI*rphi;
    }
    
    /*    vertex as function of a random variable   */
    splitting_variables<lo_shower>::zeta_flavor vertex(double) const;

    /*   mappings:  kappa <--> y   */
    double mapping_y2k(double y) const { return _M_y_mapping.mapping_tau2k(_M_mdl->alpha_qcd.tauk(y*_M_pars.pQ));}
    double mapping_k2y(double kappa) const { return _M_mdl->alpha_qcd.tauk_inverse(_M_y_mapping.mapping_k2tau(kappa))/_M_pars.pQ;}
    
    /*   mappings:  chi <--> z   */
    double mapping_u2chi(double y, double u) const {
      z_mapping zmap; zmap.refresh(this, y);
      return zmap.mapping_u2chi(u);
    }
    
    double mapping_chi2u(double y, double chi) const {
      z_mapping zmap; zmap.refresh(this, y);
      return zmap.mapping_chi2u(chi);
    }

    void mapping_chi2u(double y, double chi, double& u, double& jac) const {
      z_mapping zmap; zmap.refresh(this, y);
      zmap.mapping_chi2u(chi, u, jac);
    }

    /*   returns the flavor of the emitter  */
    int emitter_flavor() const { return _M_pars.fl;}
    
    /*   returns the pointer to the model  */
    const model * get_model() const { return _M_mdl;}

  private:
    /*   pointer to the model   */
    const model *_M_mdl;
    
    /*    exact splitting function   */
    mutable __mc_splitting_functions::emitting_gluon_fin _M_func_emit;
    mutable double _M_prob_proc[8];
    
    /*   splitting function in terms of splitting variables   */
    double _M_funcFI(double, double, double, double) const;
    double _M_funcD(double, double, double, double) const;
    double _M_funcS(double, double, double, double) const;
    
    /*  Class for making the z mapping faster and simpler.   */
    struct z_mapping
    {
      //----- refreshing the y dependence ----
      void refresh(const splitting_kernel<fin, ordering::lambda> *, double);
      
      //----- mapping from u to chi ------
      double mapping_u2chi(double u) const {
        return std::log(Lm/(1.0+as*std::log(1.0-u*ox0)))/J;
      }
      
      //----- mapping from chi to u ------
      double mapping_chi2u(double chi) const {
        return (1.0-std::exp((Lm*std::exp(-chi*J)-1.0)/as))/ox0;
      }
      
      //----- mapping from chi to u with jacobian ------
      void mapping_chi2u(double chi, double& u, double& jac) const {
        double ecJ = Lm*std::exp(-chi*J);
        u = (1.0-std::exp((ecJ-1.0)/as))/ox0;
        jac = J1*(1.0-u*ox0);
      }
      
      //----- data member -----
      double ox0, as, Lm, J, J1;
    };
    
    /*  Class for mapping out some y dependence  */
    struct y_mapping
    {
      //----- refreshing the spline ----
      double refresh(const splitting_kernel *);
      
      /*   making the transformation and its jacobian   */
      double mapping_k2tau(double kappa) const { return spline.mapping_chi2x(kappa);}
      double mapping_tau2k(double tau) const { return spline.mapping_x2chi(tau);}
      double jacobian(double tau) const { return spline.jacobian(tau)/spline.integral();}
      
      //---- data members ----
      importance_sampling<20U> spline;
    } _M_y_mapping;
    
    /*   storage for bindding the parameters   */
    struct {
      int fl;
      char spec;
      double rl, ymax, pQ, bl, mT, xi, mu;
    } _M_pars;
  };
  

  /*  Splitting probabilities for the initial state splitting. 
   *  It can make trial emissions.
   */
  template<>
  class splitting_prob<ini, ordering::lambda>
  {
  public:
    //    constructors
    splitting_prob(double q2np, const model *mdl)
    : _M_q2np(q2np), _M_k2T_veto_min(0.0), _M_k2T_veto_max(std::numeric_limits<double>::max()),
    _M_q2_veto_min(0.0), _M_q2_veto_max(std::numeric_limits<double>::max()), _M_split(mdl) {}
    
    /*  Initialize the bounding functions  */
    void initialize(const char *, int, double, double, unsigned int, unsigned int);
    
    /*  Making a trial emission considering all the possible final state partons and their color
     *  connections. It returns with the splitting variables if the splitting is possible.
     */
    template<class _Tag>
    bool make_trial(random_engine_type&, splitting_variables<lo_shower> *, double&,
                    double&, double, const event<_Tag>&, const lorentzvector<double>&) const;
    
    /*   set and get transverse momentum veto   */
    void k2T_veto(double kTmin, double kTmax) {
      _M_k2T_veto_min = kTmin;
      _M_k2T_veto_max = kTmax;
    }
    
    std::pair<double,double> k2T_veto() const {
      return {_M_k2T_veto_min, _M_k2T_veto_max};
    }
    
    /*   set and get the virtuality veto   */
    void q2_veto(double q2min, double q2max) {
      _M_q2_veto_min = q2min;
      _M_q2_veto_max = q2max;
    }
    
    std::pair<double,double> q2_veto() const {
      return {_M_q2_veto_min, _M_q2_veto_max};
    }
    
  private:
    /*   absolute transverse momentum cutoff   */
    double _M_q2np;
    
    /*   transverse momentum veto   */
    double _M_k2T_veto_min, _M_k2T_veto_max;
    
    /*   virtuality veto   */
    double _M_q2_veto_min, _M_q2_veto_max;
    
    /*   exact splitting function   */
    mutable splitting_kernel<ini, ordering::lambda> _M_split;
    
    /*   This is the splitting kernel approximation. We have a map since
     *   we have to introduce different approximation for different hadrons
     */
    std::map<int, splitting_kernel_approx<ini>> _M_split_approx;
    
    /*  The boundig function at fixed parameters.  */
    mutable bool _M_isheavy;
    mutable double _M_yL, _M_yH, _M_ynull, _M_bind_jac, _M_funcF_hq;
    mutable const bounding_function<2U> *_M_bounding_kappa_chi;
       
    /*  Setting up the distribution to genrate a trial emission.  */
    bool _M_parameters(const event<dis>&, bool, int, const lorentzvector<double>&) const;
    bool _M_parameters(const event<hhc>&, bool, int, int, const lorentzvector<double>&) const;
    
    /*  Making a trial emission   */
    bool _M_make_trial(random_engine_type&, splitting_variables<lo_shower> *,
                       double&, double&, int, int, double, double, double) const;

    /*  Function for generating the variable kappa.   */
    double _M_primitive_kappa(double, double) const;
    
    double _M_bounding_kappa(double kappa) const {
      double kappaF = kappa, thrs = 1.0;
      
      if(!_M_isheavy && std::abs(_M_split.emitter_flavor()) > 3) {
        kappaF = std::min(kappaF, _M_split.kappaF());
        thrs = _M_split.kappaT()-kappa;
      }

      return _M_bounding_kappa_chi->average({{kappaF,kappaF},{0.0,1.0}})/thrs;
    }
    
    /*  Functions for generating variable chi.   */
    double _M_bounding_chi(double kappa, double chi) const {
      return (*_M_bounding_kappa_chi)({kappa,chi});
    }
    
    double _M_primitive_chi(double kappa, double chi) const {
      return _M_bounding_kappa_chi->average({{kappa,kappa},{0.0,chi}})*chi;
    }
    
    double _M_function_veto(double kappa, double chi) const {
      double thrs = 1.0;
      
      if(!_M_isheavy && std::abs(_M_split.emitter_flavor()) > 3) {
        kappa = std::min(kappa, _M_split.kappaF());
        thrs = _M_split.kappaT()-kappa;
      }
      
      return _M_split(kappa, chi, _M_k2T_veto_min, _M_k2T_veto_max)*thrs;
    }
    
    /*  The root finder has to be friend  */
    friend double __mc_helper::newton_root_finder_chi<>(const splitting_prob *, double, double);
    friend double __mc_helper::newton_root_finder_kappa<>(const splitting_prob *, double, double);
  };
  
  
  /*  Splitting probabilities for the final state splitting.
   *  It can make trial emissions.
   */
  template<>
  class splitting_prob<fin, ordering::lambda>
  {
  public:
    //    constructors
    splitting_prob(double q2np, const model *mdl)
    : _M_q2np(q2np), _M_k2T_veto_min(0.0), _M_k2T_veto_max(std::numeric_limits<double>::max()),
    _M_q2_veto_min(0.0), _M_q2_veto_max(std::numeric_limits<double>::max()), _M_split(mdl), _M_is_init(false) {}
    
    /*  Initialize the bounding functions  */
    void initialize(const char *filename, double s, unsigned int nspar, unsigned int nsvar) {
      if(!_M_is_init) {
        _M_split_approx = splitting_kernel_approx<fin>(_M_q2np, s);
        _M_split_approx.initialize(filename, _M_split, nspar, nsvar);
        _M_is_init = true;
      }
    }
    
    /*  Making a trial emission considering all the possible final state partons and their color
     *  connections. It returs with the splitting variables if the splitting is possible.
     */
    template<class _Tag>
    bool make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
                    double& tmU, double tmL, const event<_Tag>& p, const lorentzvector<double>& qnull) const {
      return _M_make_trial(rne, vars, ratio, tmU, tmL, p, qnull);
    }
    
    /*   set and get transverse momentum veto   */
    void k2T_veto(double kTmin, double kTmax) {
      _M_k2T_veto_min = kTmin;
      _M_k2T_veto_max = kTmax;
    }
    
    std::pair<double,double> k2T_veto() const {
      return {_M_k2T_veto_min, _M_k2T_veto_max};
    }
    
    /*   set and get the virtuality veto   */
    void q2_veto(double q2min, double q2max) {
      _M_q2_veto_min = q2min;
      _M_q2_veto_max = q2max;
    }
    
    std::pair<double,double> q2_veto() const {
      return {_M_q2_veto_min, _M_q2_veto_max};
    }
    
  private:
    /*   absolute transverse momentum cutoff   */
    double _M_q2np;
    
    /*   transverse momentum veto   */
    double _M_k2T_veto_min, _M_k2T_veto_max;
 
    /*   virtuality veto   */
    double _M_q2_veto_min, _M_q2_veto_max;
    
    /*   exact splitting function   */
    mutable splitting_kernel<fin, ordering::lambda> _M_split;
    
    /*   This is the splitting kernel approximation.
     */
    bool _M_is_init;
    splitting_kernel_approx<fin> _M_split_approx;
    
    /*  The boundig function at fixed parameters.  */
    mutable double _M_ynull, _M_yL, _M_yH, _M_bind_jac;
    mutable const bounding_function<2U> *_M_bounding_kappa_chi;

    /*  Setting up the distribution to genrate a trial emission.  */
    bool _M_parameters(const statistical_basis&, int, int, const lorentzvector<double>&) const;
    
    /*  Making a trial emission   */
    bool _M_make_trial(random_engine_type&, splitting_variables<lo_shower> *,
                       double&, double&, int, int, double, double, double) const;
    
    bool _M_make_trial(random_engine_type&, splitting_variables<lo_shower> *, double&, double&,
                       double, const statistical_basis&, const lorentzvector<double>&) const;

    /*  Function for generating the variable kappa.   */
    double _M_bounding_kappa(double kappa) const {
      return _M_bounding_kappa_chi->average({{kappa,kappa},{0.0,1.0}});
    }
    
    double _M_primitive_kappa(double kappaU, double kappaL) const {
      if(kappaL > kappaU) return 0.0;
      return _M_bounding_kappa_chi->integral({{kappaL,kappaU},{0.0,1.0}});
    }
    
    /*  Functions for generatung variable chi.   */
    double _M_bounding_chi(double kappa, double chi) const {
      return (*_M_bounding_kappa_chi)({kappa,chi});
    }
    
    double _M_primitive_chi(double kappa, double chi) const {
      return _M_bounding_kappa_chi->average({{kappa,kappa},{0.0,chi}})*chi;
    }
    
    /*  The root finder has to be friend  */
    friend double __mc_helper::newton_root_finder_chi<>(const splitting_prob *, double, double);
    friend double __mc_helper::newton_root_finder_kappa<>(const splitting_prob *, double, double);
  };


}


#endif

