/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1

//   std includes
#include <cmath>
#include <utility>
#include <fstream>
#include <string>

//   project headers
#include <bits/nlo-tags.h>
#include <bits/nlo-event.h>
#include <bits/nlo-mapping.h>
#include <bits/mc-emission.h>
#include <bits/mc-bounding-function.h>
#include <bits/nlo-c-support.h>
#include <bits/nlo-cxx-support.h>


namespace duct {
 
  /*  This block defines some routines and functions
   *  those are used commonly. E.g.: root finder algorithm.
   */
  namespace __mc_helper {
    
    template<class _Func>
    double newton_root_finder_chi(const _Func *xInt, double kappa, double rho)
    {
      //----- optimized newton method ----
      double lo = 0.0, hi = 1.0, mi = rho;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= xInt->_M_primitive_chi(kappa, 1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = xInt->_M_primitive_chi(kappa, mi)))/xInt->_M_bounding_chi(kappa, mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
      
      return mi;
    }
    
    template<class _Func>
    double newton_root_finder_kappa(const _Func *xInt, double low, double rho)
    {
      //----- optimized newton method ----
      double lo = low, mi = low, hi = 1.0;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = xInt->_M_primitive_kappa(mi, low)))/xInt->_M_bounding_kappa(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < low || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(1.0-mi/mi1) > 1e-8 && ++iter < 30U);
      
      
      return mi;
    }
    
    template<class _Func>
    double newton_root_finder_phi(const _Func& xInt, double rho)
    {
      //----- optimized newton method ----
      double lo = 0.0, hi = 1.0, mi = rho;
      double mi1=-10.0, mi2=-10.0, u, v, prim;
      unsigned int iter = 0U;
      
      //----- normalization ----
      rho *= xInt.primitive_phi(1.0);
      
      do {
        //----- start with Newton method -----
        mi2 = mi1; mi1 = mi;
        mi += (rho - (prim = xInt.primitive_phi(mi)))/xInt.function_phi(mi);
        
        if(mi2 > 0.0) {
          u = mi1-mi2;
          v = mi-2.0*mi1+mi2;
          if(v != 0.0) mi = mi2 - u*u/v;
        }
        
        //---- adjusting the boundaries to the old value -----
        if(prim > rho) { if(mi1 < hi) hi = mi1;}
        else { if(mi1 > lo) lo = mi1;}
        
        //----- if it doesn't work we use bisect method-----
        if(mi < 0.0 || mi > 1.0) mi = 0.5*(lo+hi);
      } while(std::abs(mi1-mi) > 1e-8*mi1 && ++iter < 30U);
      
      return mi;
    }


    /*  This routine counts the color connected parton of an emitter. It considers
     *  both the amplitude and conjugate amplitude. This function needs to do the
     *  shower evolution.
     */
    unsigned int counting_color_connections(const statistical_basis&, int, std::pair<int, double> *);
  }
  
  
  template<class _Type, class _Ordering>
  class splitting_kernel;

  template<class _Type>
  class splitting_kernel_approx;

  template<class _Type, class _Ordering>
  class splitting_prob;
  
  
  /*  This class represent the bounding function for the initial state splitting kernels.
   *  Here we create a big trees to be able to approximate the splitting kernel. This
   *  should be a constant and rather static information. The template argument is
   *  the underlying exact splitting kernel, but this class is not inhereted from or there is
   *  no delegated instance of it. We want to use these trees in many threads.
   */
  template<>
  class splitting_kernel_approx<ini>
  {
  public:
    /*     constructor   */
    explicit splitting_kernel_approx(double m2T=1.0, double q2min=16.0, double s=1.96e+8)
    : _M_s(s), _M_m2T(m2T), _M_q2min(q2min), _M_ebm(std::pow(std::sqrt(m2T) + std::sqrt(m2T+q2min),2)/s) {}
    
    
    /*  Initialize the bounding functions  */
    template<class _Kernel>
    void initialize(const char *, const _Kernel&, int, unsigned int, unsigned int);
    
    /*    initialize the function by fixing the parameters (final state spectator)  */
    const bounding_function<2u>& bindF(int fl, double eta, double Q2, double xi, double xT) const {
      double par[4] = {0.0, 0.0, xi, xT};
      this -> _M_mapping_eta2x(par, eta, Q2);
      return _M_bounding[5+fl] -> funcF.bind(par);
    }
    
    /*   initialize the function by fixing the parameters (initial state spectator)  */
    const bounding_function<2u>& bindI(int fl, double eta, double Q2) const {
      double par[2];
      this -> _M_mapping_eta2x(par, eta, Q2);
      return _M_bounding[5+fl] -> funcI.bind(par);
    }

    /*    initialize the function by fixing the parameters (final state spectator, heavy quark)  */
    const bounding_function<2u>& bindFH(int fl, double eta, double Q2, double xi, double xT) const {
      double par[4] = {0.0, 0.0, xi, xT};
      this -> _M_mapping_eta2x(par, eta, Q2);
      return _M_bounding_hq[std::abs(fl)-4] -> funcF.bind(par);
    }
    
    /*   initialize the function by fixing the parameters (initial state spectator, heavy quark)  */
    const bounding_function<2u>& bindIH(int fl, double eta, double Q2) const {
      double par[2];
      this -> _M_mapping_eta2x(par, eta, Q2);
      return _M_bounding_hq[std::abs(fl)-4] -> funcI.bind(par);
    }
    
    /*   Transoform to the anti-hadron version of the splitting kernel.  */
    void anti() {
      for(int i : {1,2,3}) std::swap(_M_bounding[5+i], _M_bounding[5-i]);
    }
        
  private:
    /*   some variables to derermine the limits  */
    double _M_s, _M_m2T, _M_q2min, _M_ebm;
    
    /*   Mapping the eta and Q2 to the internal variables   */
    void _M_mapping_eta2x(double *x, double eta, double Q2) const {
      constexpr double n = 2.0;
      double etb = Q2/(eta*_M_s), mtl = _M_m2T/_M_s, etm = _M_q2min/_M_s;
      double eta_max = std::pow(1.0-2.0*std::sqrt(mtl/etb), 1.0/n);
      double eta_min = std::pow(etm/etb, 1.0/n);
      
      x[0] = (std::pow(eta,1.0/n)-eta_min)/(eta_max-eta_min);
      x[1] = (etb-_M_ebm)/(1.0-_M_ebm);
    }
    
    void _M_mapping_x2eta(double& eta, double& Q2, const double *x) const {
      constexpr double n = 2.0;
      double etb = (1.0-_M_ebm)*x[1]+_M_ebm, mtl = _M_m2T/_M_s, etm = _M_q2min/_M_s;
      eta = std::pow(std::pow(1.0-2.0*std::sqrt(mtl/etb), 1.0/n)*x[0] + std::pow(etm/etb, 1.0/n)*(1.0-x[0]), n);
      Q2 = eta*etb*_M_s;
    }
    
    /*  This is a class to boundle overshooting functions for the
     *  different spectator. It makes the life simpler.....
     */
    struct bounding_set {
      /*   constructor   */
      bounding_set() = default;
      explicit bounding_set(const std::pair<double,double> *ival)
      : funcF(ival, ival+4), funcI(ival, ival+4) {}
      
      /*   I/O manipulations   */
      void write(std::ostream& os) const { funcI.write(os); funcF.write(os);}
      void read(std::istream& is) { funcI.read(is); funcF.read(is);}
      
      /*    bounding functions with final and initial state spectator  */
      bounding_function<4U,2U> funcF;
      bounding_function<2U,2U> funcI;
    };
    
    /*  We have an array of pointers to the actual bounding_set.
     *  The memory is allocated dynamically and we try to avoid
     *  doubling huge objects by using shared pointers.
     */
    std::shared_ptr<bounding_set> _M_bounding[11];
    std::shared_ptr<bounding_set> _M_bounding_hq[2];
    
    /*  Scanning a bounding set (multi threaded)  */
    template<class _Kernel> void _M_scan(const _Kernel&, int, int, unsigned, unsigned);
    template<class _Kernel> void _M_scan_hq(const _Kernel&, int, int, unsigned, unsigned);

    /*   Fuctor for doing the scanning  */
    template<class _Kernel>
    struct __split
    {
      /*   This constructor captures some variables.   */
      __split(const splitting_kernel_approx<ini> *that, const _Kernel& kern, int hdr, int fl, char spec)
      : _this_(that), _kernel_(kern), _hadron_(hdr), _flavor_(fl), _spec_(spec) {}
      
      double operator()(const double *p, const double *v) const {
        /*  momentum fractions of the incomming partons   */
        double eta, Q2, bjac = 0.0;
        _this_->_M_mapping_x2eta(eta, Q2, p);
        
        /*   binding the function to the parameters  */
        switch(_spec_) {
          case 'I': bjac = _kernel_.bindI(_hadron_, _flavor_, _this_->_M_m2T, 0.0, eta, Q2); break;
          case 'F': bjac = _kernel_.bindF(_hadron_, _flavor_, _this_->_M_m2T, 0.0, eta, Q2, p[2], p[3]); break;
 
          case 'i': bjac = _kernel_.bindIH(_hadron_, _flavor_, _this_->_M_m2T, 0.0, eta, Q2); break;
          case 'f': bjac = _kernel_.bindFH(_hadron_, _flavor_, _this_->_M_m2T, 0.0, eta, Q2, p[2], p[3]); break;
       }
        
        if(bjac < 1e-13) return 1e-13;
        
        /*   heavy flavor threshold   */
        double kappa = v[0], thrs = 1.0;
        
        if(std::abs(_flavor_) > 3) {
          if(kappa > _kernel_.kappaF()) kappa = _kernel_.kappaF();
          thrs = _kernel_.kappaT()-kappa;
        }
        
        /*   splitting function  */
        return std::max(_kernel_(kappa, v[1])*thrs, 1e-13);
      }
      
      /*   captured variables  */
      const splitting_kernel_approx<ini> *_this_;
      mutable _Kernel _kernel_;
      int _hadron_, _flavor_;
      char _spec_;
    };
  };
  

  template<class _Kernel>
  void splitting_kernel_approx<ini>::
  initialize(const char *filename, const _Kernel& kernel, int hdr, unsigned int nspar, unsigned int nsvar)
  {
    /*   check if it is already initialized  */
    if(_M_bounding[5] != nullptr) return;
    
    /*   otherwise try to read from file   */
    std::fstream file;
    std::shared_ptr<bounding_set> *__bounding = _M_bounding + 5;
 
    /*  Try to open the cache file   */
    file.open(filename, std::ios::in);
    
    if(file.is_open())
    {
      /*  Reading variables _M_q2np and s variables and rescan if it is needed */
      double __m2T, __q2min, __s;
      
      file.read(reinterpret_cast<char *>(&__m2T), sizeof(double));
      file.read(reinterpret_cast<char *>(&__q2min), sizeof(double));
      file.read(reinterpret_cast<char *>(&__s), sizeof(double));
      
      /*   We have to rescan the splitting kernels
       *   if we want to extend the evolution region.
       */
      if(std::abs(1.0-__s/_M_s) > 1.0e-11 || std::abs(1.0-__m2T/_M_m2T) > 1.0e-11 || std::abs(1.0-__q2min/_M_q2min) > 1.0e-11) {
        file.close();
        goto rescan;
      }
      
      /*  The cache file looks OK for this run
       *
       *  Reading gluon function
       */
      __bounding[0] = std::shared_ptr<bounding_set>(new bounding_set);
      __bounding[0] -> read(file);
      
      /*  Reading the light quark functions  */
      for(int fl : {1,2,3}) {
        //----- light quark -----
        __bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set);
        __bounding[fl] -> read(file);
        
        //----- light anti-quark -----
        __bounding[-fl] = std::shared_ptr<bounding_set>(new bounding_set);
        __bounding[-fl] -> read(file);
      }
      
      
      /*  Reading the heavy quark functions  */
      for(int fl : {4,5}) {
        //----- quark function -----
        __bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set);
        __bounding[fl] -> read(file);
        
        //----- anti-quark function -----
        __bounding[-fl] = __bounding[fl];
      }
      
      /*   Reading the extra gluon emission contribution   */
      char is_heavy[2];
      
      file.read(is_heavy, 2*sizeof(char));
      
      if(is_heavy[0] == 'h') {
        _M_bounding_hq[0] = std::shared_ptr<bounding_set>(new bounding_set);
        _M_bounding_hq[0] -> read(file);
      } else _M_bounding_hq[0] = nullptr;
      
      if(is_heavy[1] == 'h') {
        _M_bounding_hq[1] = std::shared_ptr<bounding_set>(new bounding_set);
        _M_bounding_hq[1] -> read(file);
      } else _M_bounding_hq[1] = nullptr;
      
      
      //---- close the file -----
      file.flush();
      file.close();
      
      return;
    }
    
    
  rescan:
    /*  Defining the scan intervals of the parameters and variables  */
    std::pair<double,double> intervals[] = {
      /*  xa        xb          xi         xT       kappa       chi     */
      {0.0,1.0}, {0.0,1.0}, {0.0,1.0}, {0.0,1.0}, {0.0,1.0}, {0.0,1.0}
    };

    /*    Fancy messages   */
    static const char *qname[] = {"Anti-strange", "Anti-up", "Anti-down", "Gluon", "Down", "Up", "Strange", "Charm", "Bottom"};
    static const char *space[] = {"", "     ", "   ", "       ", "        ", "          ", "     ", "       ", "      "};
    
    /*  creating the bounding function for gluon and scan it */
    mtx_cout.lock();
    std::cout<<"Scanning the initial state splitting kernels for "<<kernel.get_model()->name(hadron(hdr))<<": \n";
    std::cout<<"     "<<qname[3]<<" emitter ... "<<space[3]<<std::flush;
    mtx_cout.unlock();
    
    __bounding[0] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
    this -> _M_scan(kernel, hdr, 0, nspar, nsvar);
    mtx_cout.lock(); std::cout<<"OK"<<std::endl; mtx_cout.unlock();
    
    /*  creating the light quark objects and scan them  */
    for(int fl = 1; fl < 4; fl++) {
      //----- light quark -----
      mtx_cout.lock(); std::cout<<"     "<<qname[ fl+3]<<" emitter ... "<<space[fl+3]<<std::flush; mtx_cout.unlock();
      __bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
      this -> _M_scan(kernel, hdr, fl, nspar, nsvar);
      mtx_cout.lock(); std::cout<<"OK"<<std::endl; mtx_cout.unlock();
      
      //----- light anti-quark -----
      mtx_cout.lock(); std::cout<<"     "<<qname[-fl+3]<<" emitter ... "<<space[-fl+3]<<std::flush; mtx_cout.unlock();
      __bounding[-fl] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
      this -> _M_scan(kernel, hdr, -fl, nspar, nsvar);
      mtx_cout.lock(); std::cout<<"OK"<<std::endl; mtx_cout.unlock();
    }
    
    /*  charm and bottom quark boundig functions   */
    const model *mdl = kernel.get_model();
    
    for(int fl = 4; fl < 6; fl++) {
      //----- quark function -----
      mtx_cout.lock(); std::cout<<"     "<<qname[ fl+3]<<" emitter ... "<<space[fl+3]<<std::flush; mtx_cout.unlock();
      __bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
      this -> _M_scan(kernel, hdr, fl, nspar, nsvar);
      
      //----- anti-quark function -----
      __bounding[-fl] = __bounding[fl];
 
      /*    Extra gluon emission contribution    */
      if(mdl->mass2(fl) > 1e-6) {
        _M_bounding_hq[fl-4] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
        this -> _M_scan_hq(kernel, hdr, fl, nspar, nsvar);
      } else _M_bounding_hq[fl-4] = nullptr;
      
      mtx_cout.lock(); std::cout<<"OK"<<std::endl; mtx_cout.unlock();
    }
    
    
    /** Write the cache file  */
    file.open(filename, std::ios::out);
    
    /*  Writting the variables _M_q2np and s  */
    file.write(reinterpret_cast<const char *>(&_M_m2T), sizeof(double));
    file.write(reinterpret_cast<const char *>(&_M_q2min), sizeof(double));
    file.write(reinterpret_cast<const char *>(&_M_s), sizeof(double));
    
    /*  Writting gluon function  */
    __bounding[0] -> write(file);
    
    /*  Writting the light quark function  */
    for(int fl : {1,2,3}) {
      __bounding[ fl] -> write(file);
      __bounding[-fl] -> write(file);
    }
    
    /*  Writting the heavy quark functions  */
    __bounding[4] -> write(file);
    __bounding[5] -> write(file);
    
    /*   Witting the extra gluon emission contribution   */
    char is_heavy[2] = {'l','l'};
    
    if(_M_bounding_hq[0] != nullptr) is_heavy[0] = 'h';
    if(_M_bounding_hq[1] != nullptr) is_heavy[1] = 'h';
    
    file.write(is_heavy, 2*sizeof(char));
    
    if(is_heavy[0] == 'h') _M_bounding_hq[0] -> write(file);
    if(is_heavy[1] == 'h') _M_bounding_hq[1] -> write(file);


    /*  closing the file  */
    file.flush();
    file.close();
  }
  

  /*  This function scans the spitting functions and sets up the corresponding
   *  bounding functions for all the possible spectator types.
   */
  template<class _Kernel> void splitting_kernel_approx<ini>::
  _M_scan(const _Kernel& kernel, int __hdr, int flavor, unsigned nspar, unsigned nsvar)
  {
    /*   Fuctor in the case when we have initial and final state spectator  */
    __split<_Kernel> splitI{this, kernel, __hdr, flavor, 'I'}, splitF{this, kernel, __hdr, flavor, 'F'};
    
    
    std::thread thI(&bounding_function<2u,2u>::template scan_mt<__split<_Kernel>>,
                    &(_M_bounding[flavor+5]->funcI),
                    std::move(splitI), nspar, nsvar);
    
    std::thread thF(&bounding_function<4u,2u>::template scan_mt<__split<_Kernel>>,
                    &(_M_bounding[flavor+5]->funcF),
                    std::move(splitF), nspar, nsvar);
    
    thI.join();
    thF.join();
  }

  template<class _Kernel> void splitting_kernel_approx<ini>::
  _M_scan_hq(const _Kernel& kernel, int __hdr, int flavor, unsigned nspar, unsigned nsvar)
  {
    /*   Fuctor in the case when we have initial and final state spectator  */
    __split<_Kernel> splitI{this, kernel, __hdr, flavor, 'i'}, splitF{this, kernel, __hdr, flavor, 'f'};
    
    
    std::thread thI(&bounding_function<2u,2u>::template scan_mt<__split<_Kernel>>,
                    &(_M_bounding_hq[std::abs(flavor)-4]->funcI),
                    std::move(splitI), nspar, nsvar);
    
    std::thread thF(&bounding_function<4u,2u>::template scan_mt<__split<_Kernel>>,
                    &(_M_bounding_hq[std::abs(flavor)-4]->funcF),
                    std::move(splitF), nspar, nsvar);
   
    thI.join();
    thF.join();
  }
  
  
  
  /*  This class represent the bounding function for the final state splitting kernels.
   *  Here we create a big trees to be able to approximate the splitting kernel. This
   *  should be a constant and rather static information. The template argument is
   *  the underlying exact splitting kernel, but this class is not inhereted from or there is
   *  no delegated instance of it. We want to use the same trees in many threads.
   */
  template<>
  class splitting_kernel_approx<fin>
  {
  public:
    /*     constructor   */
    explicit splitting_kernel_approx(double m2T=1.0, double s=1.96e+8)
    : _M_s(s), _M_m2T(m2T) {}
    
    /*  Initialize the bounding functions  */
    template<class _Kernel>
    void initialize(const char *, const _Kernel&, unsigned int, unsigned int);
    
    /*    initialize the function by fixing the parameters (initial state spectator)  */
    const bounding_function<2u>& bindI(int fl, double ymax, double pQ, double xi, double mu) const {
      double uk = (xi-0.5)/std::sqrt(std::abs(1.0-2.0*mu))+0.5, m2l = _M_bounding[std::abs(fl)]->mass2;
      double par[4] = {0.0, 0.0, uk, mu};
      
      /*   mapping parameters  */
      this->mapping_pQ2x(par, m2l, ymax, pQ);
      
      return _M_bounding[std::abs(fl)]->funcI.bind(par);
    }
    
    /*    initialize the function by fixing the parameters (final state spectator)  */
    const bounding_function<2u>& bindF(int fl, double ymax, double pQ, double xi, double mu) const {
      double uk = (xi-0.5)/std::sqrt(std::abs(1.0-2.0*mu))+0.5, m2l = _M_bounding[std::abs(fl)]->mass2;
      double par[4] = {0.0, 0.0, uk, mu};
      
      /*   mapping parameters  */
      this->mapping_pQ2x(par, m2l, ymax, pQ);
      
      return _M_bounding[std::abs(fl)]->funcF.bind(par);
    }
    
    /*   initialize the function by fixing the parameters (decay shower)  */
    const bounding_function<2u>& bindD(int fl, double ymax, double pQ) const {
      double par[2], m2l = _M_bounding[std::abs(fl)]->mass2;
      
      /*   mapping parameters  */
      this->mapping_pQ2x(par, m2l, ymax, pQ);
      
      return _M_bounding[std::abs(fl)]->funcD.bind(par);
    }
        
    /*   initialize the function by fixing the parameters (special case)  */
    const bounding_function<2u>& bindS(int fl, double pQ) const {
      double m2l = _M_bounding[std::abs(fl)]->mass2;
      
      /*   mapping parameters  */
      double pQm = 2.0*(m2l+_M_m2T + std::sqrt(_M_m2T*(m2l+_M_m2T)));
      double par = std::log(pQ/pQm)/std::log((_M_s+m2l)/pQm);
            
      return _M_bounding[std::abs(fl)]->funcS.bind(&par);
    }
    
 private:
    /*   some variables to derermine the limits  */
    double _M_s, _M_m2T;
    
    /*   mapping the hard variables to a unit cube and back  */
    void mapping_pQ2x(double *x, double m2l, double ymax, double pQ) const
    {
      double bl = m2l/pQ, pQm = 2.0*(m2l+_M_m2T + std::sqrt(_M_m2T*(m2l+_M_m2T)));
      double yL = 2.0*_M_m2T/pQ*(1.0+std::sqrt(1.0+m2l/_M_m2T));
      
      double dmm = std::sqrt(yL*(yL+bl)/(1.0-2.0*bl-yL*bl));
      double dmax = std::sqrt(ymax*(bl+ymax)/(1.0-bl*(2.0+ymax)));
      
      x[0] = (dmax-dmm)/(1.0-dmm);
      x[1] = std::log(pQ/pQm)/std::log((_M_s+m2l)/pQm);
    }
    
    void mapping_x2pQ(double& rl, double& ymax, double& pQ, double m2l, const double *x) const
    {
      double pQm = 2.0*(m2l+_M_m2T + std::sqrt(_M_m2T*(m2l+_M_m2T)));
      pQ = pQm*std::pow((_M_s+m2l)/pQm, x[1]);
      
      double bl = m2l/pQ, yL = 2.0*_M_m2T/pQ*(1.0+std::sqrt(1.0+m2l/_M_m2T));
      double dmm = std::sqrt(yL*(yL+bl)/(1.0-2.0*bl-yL*bl));
      double dmax = (1.0-dmm)*x[0]+dmm;
      double bdx = 0.5*bl*(1.0/(dmax+1e-16) + dmax);
      
      rl = std::sqrt(bdx*bdx+1.0-2.0*bl)-bdx;
      ymax = rl*dmax;
    }
      
    /*  This is a class to boundle overshooting functions for the
     *  different spectator. It makes the life simpler.....
     */
    struct bounding_set {
      /*   constructors   */
      bounding_set() = default;
      
      explicit bounding_set(const std::pair<double,double> *ival, double m2=0.0)
      : funcF(ival, ival+4), funcI(ival, ival+4), funcD(ival, ival+4), funcS(ival+1, ival+4), mass2(m2) {}
      
      /*   I/O manipulations   */
      void write(std::ostream& os) const {
        funcI.write(os); funcF.write(os); funcD.write(os); funcS.write(os);
        os.write(reinterpret_cast<const char *>(&mass2), sizeof(double));
      }
      
      void read(std::istream& is) {
        funcI.read(is); funcF.read(is); funcD.read(is); funcS.read(is);
        is.read(reinterpret_cast<char *>(&mass2), sizeof(double));
      }
      
      /*    bounding functions with final and initial state spectator  */
      bounding_function<4U,2U> funcF, funcI;

      /*    bounding function for the decay shower  */
      bounding_function<2U,2U> funcD;
      
      /*    bounding function for the special case  */
      bounding_function<1U,2U> funcS;
      
      /*    mass of the emitter   */
      double mass2;
    };
    
    /*  We have an array of pointers to the actual bounding_set.
     *  The memory is allocated dynamically and we try to avoid
     *  doubling huge objects by using shared pointers.
     */
    std::shared_ptr<bounding_set> _M_bounding[7];
    
    /*  Scanning a bounding set (multi threaded)  */
    template<class _Kernel> void _M_scan(const _Kernel&, int, unsigned, unsigned);

    /*   Fuctor for doing the scanning (final, initial and decaying parton spectators) */
    template<class _Kernel>
    struct __splitFID
    {
      /*   This constructor captures some variables.   */
      __splitFID(const splitting_kernel_approx<fin> *that, const _Kernel& kern, int fl, char spec)
      : _this_(that), _kernel_(kern), _flavor_(fl), _spec_(spec) { m2l = _kernel_.get_model()->mass2(_flavor_);}
      
      double operator()(const double *p, const double *v) const
      {
        double bjac = 0.0, xi, rl, ymax, pQ;
        
        /*   mapping the parameters  */
        _this_->mapping_x2pQ(rl, ymax, pQ, m2l, p);
        if(_spec_ != 'D') xi = std::sqrt(std::abs(1.0-2.0*p[3]))*(p[2]-0.5) + 0.5;
        
        /*   binding the function to the parameters  */
        switch(_spec_) {
          case 'F': bjac = _kernel_.bindF(_flavor_, _this_->_M_m2T, rl, ymax, pQ, xi, p[3]); break;
          case 'I': bjac = _kernel_.bindI(_flavor_, _this_->_M_m2T, rl, ymax, pQ, xi, p[3]); break;
          case 'D': bjac = _kernel_.bindD(_flavor_, _this_->_M_m2T, rl, ymax, pQ); break;
        }
        
        if(bjac < 1e-13) return 1e-13;
        
        /*   splitting function  */
        return std::max(_kernel_(v[0], v[1]), 1e-13);
      }
      
      /*   captured variables  */
      const splitting_kernel_approx<fin> *_this_;
      mutable _Kernel _kernel_;
      int _flavor_;
      char _spec_;
      
      /*   derived variables   */
      double m2l;
    };
 
    /*   Fuctor in the case when we have special spectator  */
    template<class _Kernel>
    struct __splitS
    {
      /*   This constructor captures some variables.   */
      __splitS(const splitting_kernel_approx<fin> *that, const _Kernel& kern, int fl)
      : _this_(that), _kernel_(kern), _flavor_(fl)
      {
        m2T = _this_->_M_m2T;
        m2l = _kernel_.get_model()->mass2(_flavor_);
        pQmin = 2.0*(m2l+m2T + std::sqrt(m2T*(m2l+m2T)));
        pQmax = _this_->_M_s+m2l;
      }
      
      double operator()(const double *p, const double *v) const {
        double pQ = pQmin*std::pow(pQmax/pQmin, p[0]);
        
        /*   binding the function to the parameters  */
        if(_kernel_.bindS(_flavor_, m2T, 1.0-2.0*m2l/pQ, pQ) < 1e-13) return 1e-13;
        
        /*   splitting function  */
        return std::max(_kernel_(v[0], v[1]), 1e-13);
      };
      
      /*   captured variables  */
      const splitting_kernel_approx<fin> *_this_;
      mutable _Kernel _kernel_;
      int _flavor_;
      
      /*   derived variables   */
      double m2T, m2l, pQmin, pQmax;
    };
  };

  
  /*   Function to initialize the bounding function. This scans the available variable and parameter
   *   space and find the maximum of the splitting function over cells.
   */
  template<class _Kernel> void splitting_kernel_approx<fin>::
  initialize(const char *filename, const _Kernel& kernel, unsigned int nspar, unsigned int nsvar)
  {
    /*   check if it is already initialized  */
    if(_M_bounding[0] != nullptr) return;
    
    /*   otherwise try to read from file   */ 
    std::fstream file;
    
    /** Try ro open the cache file   */
    file.open(filename, std::ios::in);
    
    if(file.is_open())
    {
      /*  Reading variables _M_q2np and s variables and rescan if it is needed */
      double __s, __m2T;
      
      file.read(reinterpret_cast<char *>(&__m2T), sizeof(double));
      file.read(reinterpret_cast<char *>(&__s), sizeof(double));
      
      /*   We have to rescan the splitting kernels 
       *   if we want to change the evolution region.
       */
      if(std::abs(1.0-__s/_M_s) > 1.0e-11 || std::abs(1.0-__m2T/_M_m2T) > 1.0e-11) {
        file.close();
        goto rescan;
      }
      
      /*  The cache file looks OK for this run
       *
       *  Reading gluon function
       */
      _M_bounding[0] = std::shared_ptr<bounding_set>(new bounding_set);
      _M_bounding[0] -> read(file);
      
      /*  Reading the light quark function  */
      std::shared_ptr<bounding_set> lquark(new bounding_set);
      lquark -> read(file);
      
      /*  down, up and strange quarks are always light  */
      _M_bounding[1] = _M_bounding[2] = _M_bounding[3] = lquark;
      
      /*  the charm, bottom and top quarks usually have non-zero mass   */
      for(unsigned int fl = 4; fl <= 6; fl++) {
        char is_heavy;
        file.read(&is_heavy, sizeof(char));
        
        if(is_heavy == 'h') {
          _M_bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set);
          _M_bounding[fl] -> read(file);
        }
        else if(is_heavy == 'l') _M_bounding[fl] = lquark;
        else _M_bounding[fl] = nullptr;
      }
      
      //---- close the file -----
      file.flush();
      file.close();
      
      return;
    }
    
    
  rescan:
    /*   Pointer to the model descriptor   */
    const model *mdl = kernel.get_model();
    
    /*  Defining the scan intervals of the parameters and variables  */
    std::pair<double,double> intervals[] = {
      /*  ymax       pQ         uk         mu       kappa      chi  */
      {0.0,1.0}, {0.0,1.0}, {0.0,1.0}, {0.0,0.5}, {0.0,1.0}, {0.0,1.0},
    };
    
    /*  creating the bounding function for gluon and scan it */
    mtx_cout.lock(); std::cout<<"Scanning the final state splitting kernels: \n";
    std::cout<<"     Gluon emitter ...       "<<std::flush; mtx_cout.unlock();
    _M_bounding[0] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
    this -> _M_scan(kernel, 0, nspar, nsvar);
    mtx_cout.lock(); std::cout<<" OK\n"; mtx_cout.unlock();
   
    /*  creating a light quark object and scan it  */
    mtx_cout.lock(); std::cout<<"     Light quark emitter ... "<<std::flush; mtx_cout.unlock();
    _M_bounding[1] = std::shared_ptr<bounding_set>(new bounding_set(intervals));
    this -> _M_scan(kernel, 1, nspar, nsvar);
    mtx_cout.lock(); std::cout<<" OK\n"; mtx_cout.unlock();
   
    /*  down, up and strange quarks are always light  */
    _M_bounding[2] = _M_bounding[3] = _M_bounding[1];
    
    /*  the charm, bottom and top quarks usually have non-zero mass   */
    static const char * hqnames[] = {"Charm", "Bottom", "Top"};
    static const char * hqspace[] = {" ", "", "   "};
    
    for(unsigned int fl = 4; fl <= 6; fl++) {
      double m2q = mdl->mass2(fl);
      
      if(m2q > 1e-12) {
        double pQmin = 2.0*(m2q+_M_m2T + std::sqrt(_M_m2T*(m2q+_M_m2T)));
        
        if(pQmin < _M_s+m2q) {
          mtx_cout.lock(); std::cout<<"     "<<hqnames[fl-4]<<" quark emitter ... "<<hqspace[fl-4]<<std::flush; mtx_cout.unlock();
          _M_bounding[fl] = std::shared_ptr<bounding_set>(new bounding_set(intervals, mdl->mass2(fl)));
          this -> _M_scan(kernel, fl, nspar, nsvar);
          mtx_cout.lock(); std::cout<<"OK\n"; mtx_cout.unlock();
        } else _M_bounding[fl] = nullptr;
      } else _M_bounding[fl] = _M_bounding[1];
    }
 
    
    /** Write the cache file  */
    file.open(filename, std::ios::out);
    
    /*  Writting the variables _M_q2np and s  */
    file.write(reinterpret_cast<const char *>(&_M_m2T), sizeof(double));
    file.write(reinterpret_cast<const char *>(&_M_s), sizeof(double));
    
    /*  Writting gluon function  */
    _M_bounding[0]->write(file);
    
    /*  Writting the light quark function  */
    _M_bounding[1]->write(file);
    
    /*  Writting the heavy quark functions (if there is any)  */
    for(unsigned int fl = 4; fl <= 6; fl++) {
      char is_heavy = 'h';
      if(_M_bounding[fl] == _M_bounding[1]) is_heavy = 'l';
      if(_M_bounding[fl] == nullptr) is_heavy = 'n';
      
      file.write(&is_heavy, sizeof(char));
      if(is_heavy != 'n') _M_bounding[fl]->write(file);
    }
    
    /*  closing the file  */
    file.flush();
    file.close();
  }

  
  /*  This function scans the spitting functions and sets up the corresponding
   *  bounding functions for all the possible spectator types.
   */
  template<class _Kernel> void splitting_kernel_approx<fin>::
  _M_scan(const _Kernel& kernel, int flavor, unsigned nspar, unsigned nsvar)
  {
    /*   Fuctor in the case when we have initial and final state spectator  */
    __splitFID<_Kernel> splitF{this, kernel, flavor, 'F'}, splitI{this, kernel, flavor, 'I'};
  
    /*   Fuctor in the case when we have decay shower  */
    __splitFID<_Kernel> splitD{this, kernel, flavor, 'D'};
    
    /*   Fuctor in the case when we have special spectator  */
    __splitS<_Kernel> splitS{this, kernel, flavor};
    

    std::thread thF(&bounding_function<4u,2u>::template scan_mt<__splitFID<_Kernel>>,
                    &(_M_bounding[flavor]->funcF),
                    std::move(splitF), nspar, nsvar);

    std::thread thI(&bounding_function<4u,2u>::template scan_mt<__splitFID<_Kernel>>,
                    &(_M_bounding[flavor]->funcI),
                    std::move(splitI), nspar, nsvar);
    
    std::thread thD(&bounding_function<2u,2u>::template scan_mt<__splitFID<_Kernel>>,
                    &(_M_bounding[flavor]->funcD),
                    std::move(splitD), nspar, nsvar);

    std::thread thS(&bounding_function<1u,2u>::template scan_mt<__splitS<_Kernel>>,
                    &(_M_bounding[flavor]->funcS),
                    std::move(splitS), nspar, nsvar);
    
    
    thF.join();
    thI.join();
    thD.join();
    thS.join();
  }

}


/*   Here we put together the shower engines. 
 */
namespace duct {
  
  /*   Shower engine for DIS and hadron-hadron like processes 
   */
  template<class _Tag, class _Ordering=ordering::lambda>
  class shower_engine : public shower_emission<_Tag>
  {
  public:
    //----- constructor -----
    explicit shower_engine(double q2n = 1.0, const model *mdl = nullptr)
    : _M_trialI(q2n, mdl), _M_trialF(q2n, mdl) {}
    
    /*  Initialize the bounding functions  */
    void initialize(const char *, const incoming_hadrons<_Tag>&, double, double, unsigned int , unsigned int);
    
    /*  Generates the trial emission   */
    double operator()(random_engine_type&, shower_stage<_Tag>&, double, double);
    
    /*   function to clone the objects   */
    shower_emission<_Tag> * clone() const {
      return new shower_engine(*this);
    }
    
  private:
    /*   Initial and final state splitting probabilities   */
    splitting_prob<ini, _Ordering> _M_trialI;
    splitting_prob<fin, _Ordering> _M_trialF;
  };
  
 
  /*   Helper class to initialize the initial state probabilities.   */
  template<class _Tag, class _Ordering>
  struct __init_engine;
  
  
  template<class _Ordering>
  struct __init_engine<dis, _Ordering>
  {
    static void initI(splitting_prob<ini, _Ordering>& trialI, const std::basic_string<char>& dname,
                      const incoming_hadrons<dis>& hdr, double q2min, double s, unsigned nspar, unsigned nsvar)
    {
      int h = hdr.hadron, x = std::abs(hdr.hadron);
      
      /*   first we initialize with hadron then as anti-hadron if it is needed   */
      trialI.initialize((dname+"initial-state-"+std::to_string(x)+".bfn").c_str(), x, q2min, s, nspar, nsvar);
      trialI.initialize((dname+"initial-state-"+std::to_string(h)+".bfn").c_str(), h, q2min, s, nspar, nsvar);
    }
  };
  

  
  template<class _Ordering>
  struct __init_engine<hhc, _Ordering>
  {
    static void initI(splitting_prob<ini, _Ordering>& trialI, const std::basic_string<char>& dname,
                      const incoming_hadrons<hhc>& hdr, double q2min, double s, unsigned nspar, unsigned nsvar)
    {
      int hA = hdr.hadronA, hB = hdr.hadronB;
      int xA = std::abs(hA), xB = std::abs(hB);
      
      /*   first we initialize with hadron   */
      trialI.initialize((dname+"initial-state-"+std::to_string(xA)+".bfn").c_str(), xA, q2min, s, nspar, nsvar);
      trialI.initialize((dname+"initial-state-"+std::to_string(xB)+".bfn").c_str(), xB, q2min, s, nspar, nsvar);
      
      /*   if some of them was an anti-hadron then we initialize them by using the table of hadrons   */
      trialI.initialize((dname+"initial-state-"+std::to_string(hA)+".bfn").c_str(), hA, q2min, s, nspar, nsvar);
      trialI.initialize((dname+"initial-state-"+std::to_string(hB)+".bfn").c_str(), hB, q2min, s, nspar, nsvar);
    }
  };
  
  
  
  template<class _Tag, class _Ordering>
  double shower_engine<_Tag, _Ordering>::
  operator()(random_engine_type& rne, shower_stage<_Tag>& stage, double tmU, double tmL)
  {
    /*   Setting the transverse momentum veto   */
    double kTveto = this->attr().kT_cutoff;
    double virtveto = this->attr().virt_cutoff;
   
    _M_trialI.k2T_veto(kTveto*kTveto, stage.max_k2T_initial_state);
    _M_trialF.k2T_veto(kTveto*kTveto, stage.max_k2T_final_state);
 
    _M_trialI.q2_veto(virtveto*virtveto, stage.max_q2_initial_state);
    _M_trialF.q2_veto(virtveto*virtveto, stage.max_q2_final_state);

    /*  making the trial emission   */
    unsigned int niter = 0u;
    double ratio = 0.999999999, tmT = tmU, weight = 1.0;
    splitting_variables<lo_shower> *vars = &(stage.vars);
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    do {
      /*   making the trial emission   */
      bool ret = _M_trialI.make_trial(rne, vars, ratio, tmT, tmL, stage.state, stage.qnull);
      ret |=  _M_trialF.make_trial(rne, vars, ratio, tmT, tmL, stage.state, stage.qnull);

      /*   the splitting is impossible   */
      if(!ret) return 0.0;

      /*   if the hit-or-miss ration is OK, then break the loop    */
      if(ratio <= 1.0) break;
      
      /*  otherwise accept it with probability 1.0/ratio   */
      weight *= -ratio;
      
      if(rng(rne)*ratio < 1.0) {
        weight *= -ratio;
        break;
      }
      
      /*   it is rejected  */
      tmL = tmT;
      tmT = tmU;
    } while(++niter < 100u);
  
    /*   select the color   */
    stage.time = tmT;
    
    if(stage.is_color_u3) weight *= selecting_color_operator_u3(rne, vars, stage.state);
    else weight *= selecting_color_operator(rne, vars, stage.state);
    
    return weight;
  }
  
  
  template<class _Tag, class _Ordering>
  void shower_engine<_Tag, _Ordering>::
  initialize(const char *dirname, const incoming_hadrons<_Tag>& hdr, double q2min, double s, unsigned int nspar, unsigned int nsvar)
  {
    /*   putting together the directory name   */
    std::basic_string<char> dname = dirname;
    if(dname.back() != '/') dname += '/';
    
    dname += _Ordering::name;
    dname += '/';
    dname += _Tag::name;
    dname += '/';
    
    if(::__nlo_helper_mkdir(dname.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
      std::cerr<<"Unable to make directory : "<<dname<<std::endl;
      throw "Unable to make directory";
    }
    
    _M_trialF.initialize((dname+"final-state.bfn").c_str(), s, nspar, nsvar);
    __init_engine<_Tag, _Ordering>::initI(_M_trialI, dname, hdr, q2min, s, nspar, nsvar);
  }
  
  
  
  /*   Shower engine for e+e- like processes
   */
  template<class _Ordering>
  class shower_engine<epa, _Ordering> : public shower_emission<epa>
  {
  public:
    //----- constructor -----
    explicit shower_engine(double q2n = 1.0, const model *mdl = nullptr)
    : _M_trial(q2n, mdl) {}
    
    /*  Initialize the bounding functions  */
    void initialize(const char *, double, unsigned int, unsigned int);
    
    /*  Generates the trial emission   */
    double operator()(random_engine_type&, shower_stage<epa>&, double, double);
    
    /*   function to clone the objects   */
    shower_emission<epa> * clone() const {
      return new shower_engine<epa, _Ordering>(*this);
    }

  private:
    /*   e+e- annihilation we need only the final state probability   */
    splitting_prob<fin, _Ordering> _M_trial;
  };
  
  
  template<class _Ordering>
  double shower_engine<epa, _Ordering>::
  operator()(random_engine_type& rne, shower_stage<epa>& stage, double tmU, double tmL)
  {
    /*   Setting the transverse momentum veto   */
    double kTveto = this->attr().kT_cutoff;
    double virtveto = this->attr().virt_cutoff;
    
    _M_trial.k2T_veto(kTveto*kTveto, stage.max_k2T_final_state);
    _M_trial.q2_veto(virtveto*virtveto, stage.max_q2_final_state);
    
    /*  making the trial emission   */
    unsigned int niter = 0u;
    double ratio = 0.999999999, tmT = tmU, weight = 1.0;
    splitting_variables<lo_shower> *vars = &(stage.vars);
    std::uniform_real_distribution<double> rng(0.0,1.0);

    do {
      /*   making the trial emission   */
      if(!_M_trial.make_trial(rne, vars, ratio, tmT, tmL, stage.state, stage.qnull)) return 0.0;
      
      /*   if the hit-or-miss ration is OK, then break the loop    */
      if(ratio <= 1.0) break;
  
      /*  otherwise accept it with probability 1.0/ratio   */
      weight *= -ratio;
  
      if(rng(rne)*ratio < 1.0) {
       weight *= -ratio;
        break;
      }
      
      /*   it is rejected  */
      tmL = tmT;
      tmT = tmU;
    } while(++niter < 100u);
    
    /*   select the color   */
    stage.time = tmT;
    
    if(stage.is_color_u3) weight *= selecting_color_operator_u3(rne, vars, stage.state);
    else weight *= selecting_color_operator(rne, vars, stage.state);
    
    return weight;
  }
  
  
  template<class _Ordering>
  void shower_engine<epa, _Ordering>::
  initialize(const char *dirname, double s, unsigned int nspar, unsigned int nsvar)
  {
    /*   putting together the directory name   */
    std::basic_string<char> dname = dirname;
    if(dname.back() != '/') dname += '/';
    
    dname += _Ordering::name;
    dname += '/';
    dname += epa::name;
    dname += '/';
    
    if(::__nlo_helper_mkdir(dname.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
      std::cerr<<"Unable to make directory : "<<dname<<std::endl;
      throw "Unable to make directory";
    }
    
    _M_trial.initialize((dname+"final-state.bfn").c_str(), s, nspar, nsvar);
  }

  
  
  /*   Shower engine for decay processes
   */
  template<class _Ordering>
  class shower_engine<decay, _Ordering> : public shower_emission<decay>
  {
  public:
    //----- constructor -----
    explicit shower_engine(double q2n = 1.0, const model *mdl = nullptr)
    : _M_trial(q2n, mdl) {}
    
    /*  Initialize the bounding functions  */
    void initialize(const char *, double, unsigned int, unsigned int);
    
    /*  Generates the trial emission   */
    double operator()(random_engine_type&, shower_stage<decay>&, double, double);
    
    /*   function to clone the objects   */
    shower_emission<decay> * clone() const {
      return new shower_engine<decay, _Ordering>(*this);
    }

  private:
    /*   in decay process we need only the final state probability   */
    splitting_prob<fin, _Ordering> _M_trial;
  };
  
  
  template<class _Ordering>
  double shower_engine<decay, _Ordering>::
  operator()(random_engine_type& rne, shower_stage<decay>& stage, double tmU, double tmL)
  {
    /*   Setting the transverse momentum veto   */
    double kTveto = this->attr().kT_cutoff;
    double virtveto = this->attr().virt_cutoff;
    
    _M_trial.k2T_veto(kTveto*kTveto, stage.max_k2T_final_state);
    _M_trial.q2_veto(virtveto*virtveto, stage.max_q2_final_state);
    
    /*  making the trial emission   */
    unsigned int niter = 0u;
    double ratio = 0.999999999, tmT = tmU, weight = 1.0;
    splitting_variables<lo_shower> *vars = &(stage.vars);
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    do {
      /*   making the trial emission   */
      if(!_M_trial.make_trial(rne, vars, ratio, tmT, tmL, stage.state, stage.qnull)) return 0.0;
      
      /*   if the hit-or-miss ration is OK, then break the loop    */
      if(ratio <= 1.0) break;
      
      /*  otherwise accept it with probability 1.0/ratio   */
      weight *= -ratio;
      
      if(rng(rne)*ratio < 1.0) {
        weight *= -ratio;
        break;
      }
      
      /*   it is rejected  */
      tmL = tmT;
      tmT = tmU;
    } while(++niter < 100u);
    
    /*   select the color   */
    stage.time = tmT;
    
    if(stage.is_color_u3) weight *= selecting_color_operator_u3(rne, vars, stage.state);
    else weight *= selecting_color_operator(rne, vars, stage.state);
    
    return weight;
  }
 
  
  template<class _Ordering>
  void shower_engine<decay, _Ordering>::
  initialize(const char *dirname, double s, unsigned int nspar, unsigned int nsvar)
  {
    /*   putting together the directory name   */
    std::basic_string<char> dname = dirname;
    if(dname.back() != '/') dname += '/';
    
    dname += _Ordering::name;
    dname += '/';
    dname += decay::name;
    dname += '/';
    
    if(::__nlo_helper_mkdir(dname.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
      std::cerr<<"Unable to make directory : "<<dname<<std::endl;
      throw "Unable to make directory";
    }
    
    _M_trial.initialize((dname+"final-state.bfn").c_str(), s, nspar, nsvar);
  }
  
}

#endif







