/*  Copyright (C) 2010-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 6/17/10.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __nlo_mc_evolution_h__
#define __nlo_mc_evolution_h__ 1


/*  standard c++ includes  */
#include <list>

/*  deductor includes   */
#include <bits/mc-emission.h>
#include <bits/qcd-color-overlap.h>
#include <bits/qcd-color-overlap-u3.h>



namespace duct {
  
  
  
  template<class _Tag>
  class shower_history
  {
  public:
    //----- public types -----
    typedef event<_Tag> event_type;
    typedef shower_stage<_Tag> stage;
    
    typedef typename std::list<stage>::const_iterator const_iterator;
    typedef typename std::list<stage>::const_reference const_reference;
    
    /*   access to the shower history    */
    const_iterator begin() const { return _M_begin_evol;}
    const_iterator end() const { return _M_end_evol;}
    
    const_iterator cbegin() const { return _M_begin_evol;}
    const_iterator cend() const { return _M_end_evol;}
    
    /*  The function front() returns a reference to the hard event and
     *  function back() retunrs a reference to the end of the shower.
     */
    const_reference front() const { return *_M_begin_evol;}
    const_reference back() const { return *(--const_iterator(_M_end_evol));}
    
    const_reference cfront() const { return *_M_begin_evol;}
    const_reference cback() const { return *(--const_iterator(_M_end_evol));}
    
  protected:
    /*   non constant access is restricted   */
    typedef typename std::list<stage>::iterator iterator;
    typedef typename std::list<stage>::reference reference;
    
    /*   access to the shower history    */
    iterator begin() { return _M_begin_evol;}
    iterator end() { return _M_end_evol;}
    
    /*  The function front() returns a reference to the hard event and
     *  function back() retunrs a reference to the end of the shower.
     */
    reference front() { return *_M_begin_evol;}
    reference back() { return *(--iterator(_M_end_evol));}
    
    /*   List to store the shower history.   */
    std::list<shower_stage<_Tag>> _M_history;
    
    /*  These iterators point to the begining and end of the evolution.
     *  They help us to save lots of allocations, deallocations.
     */
    iterator _M_begin_evol, _M_end_evol;
  };

  
  
  template<class _Tag>
  class shower_evolution : public shower_history<_Tag>
  {
    /*   type aliasses  */
    typedef shower_history<_Tag> _Base;
    typedef typename _Base::stage _Stage;

  public:
    /*  Do the shower evolution of the given shower stage.   */
    const _Stage& operator()(random_engine_type&, shower_emission<_Tag> *, double, const _Stage&);
    
  private:
    /*  Color overlap calculators   */
    color_overlap _M_overlap;
    color_overlap_u3 _M_overlap_u3;
    
    /*    Generates one step of the shower    */
    void _M_make_one_step(random_engine_type&, shower_emission<_Tag> *, double, double);
  };
  
  
  template<class _Tag>
  void shower_evolution<_Tag>::_M_make_one_step(random_engine_type& rne, shower_emission<_Tag> *engine, double tmax, double tmin)
  {
    // trivial evolution
    if(tmax <= tmin) return;
    
    //----- current state -----
    typename _Base::iterator current = _Base::_M_end_evol;
    
    //----- make trial shower emissions -----
    double weight = engine->operator()(rne, *(--current), tmax, tmin);
    
    //------ if no emission then the evolution stops ------
    if(weight == 0.0) return;
    
    //----- allocating memory for the next stage (if it is necessary) -----
    typename _Base::iterator next;
    
    if(_Base::_M_end_evol == _Base::_M_history.end()) {
      _Base::_M_history.push_back(_Stage());
      next = _Base::_M_end_evol = _Base::_M_history.end();
      --next;
    } else next = _Base::_M_end_evol++;
    
    //---- initialization of the new stage -----
    next->weight = current->weight*weight;
    next->time = tmax;
    next->qnull = current->qnull;
    next->max_k2T_initial_state = current->max_k2T_initial_state;
    next->max_k2T_final_state = current->max_k2T_final_state;
    next->max_q2_initial_state = current->max_q2_initial_state;
    next->max_q2_final_state = current->max_q2_final_state;
    next->is_color_u3 = current->is_color_u3;
    
    //----- assigning the new momenta and flavors -----
    mapping(next->state, current->state, current->vars);
    
    /*   change color group (SU(Nc) ==> U(Nc))    */
    if(next->is_color_u3 == false && next->state.color_suppression_index >= engine->attr().max_color_suppression) {
      next->weight *= _M_overlap(next->state);
      next->weight /= _M_overlap_u3(next->state);
      next->is_color_u3 = true;
    }
  }
  
  
  template<class _Tag>
  const typename shower_history<_Tag>::stage&
  shower_evolution<_Tag>::operator()(random_engine_type& rne, shower_emission<_Tag> *engine, double tmF, const _Stage& p)
  {
    //----- if the history list is empty then we start a new history -----
    if(_Base::_M_history.empty()) {
      _Base::_M_history.push_back(p);
      _Base::_M_begin_evol = _Base::_M_history.begin();
      _Base::_M_end_evol = _Base::_M_history.end();
    }
    //----- use the previous history -----
    else {
      _Base::_M_end_evol = _Base::_M_begin_evol = _Base::_M_history.begin();
      *(_Base::_M_end_evol++) = p;
    }
    
    /*   change color group of the starting stage, SU(Nc) ==> U(Nc), if it is needed   */
    if(_Base::_M_begin_evol->is_color_u3 == false && _Base::_M_begin_evol->state.color_suppression_index >= engine->attr().max_color_suppression) {
      _Base::_M_begin_evol->weight *= _M_overlap(_Base::_M_begin_evol->state);
      _Base::_M_begin_evol->weight /= _M_overlap_u3(_Base::_M_begin_evol->state);
      _Base::_M_begin_evol->is_color_u3 = true;
    }
    
    //---- evolve if it is needed ------
    if(tmF <  p.time || p.weight == 0.0) return *_Base::_M_begin_evol;
    
    //------ start the evolution ------
    double tmI = _Base::_M_begin_evol->time;
    typename _Base::iterator current = _Base::_M_begin_evol;
    current->time = tmF;
    
    do {
      this->_M_make_one_step(rne, engine, tmF, tmI);
      tmI = current->time;
    } while(_Base::_M_end_evol != ++current);
    
    return *(--current);
  }
  
}

#endif

