/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_user_h__
#define __mc_user_h__ 1


/*   project headers     */
#include <bits/mc-evolution.h>



namespace duct {
  
  /*   Abstract class to define the most general user interface.
   */
  template<class _Tag>
  class user
  {
  public:
    /*   Virtual destructor   */
    virtual ~user() {}
    
    /*   The analyser routine. It analyses the events and fill the histograms.   */
    virtual void userfunc(const shower_history<_Tag>&) = 0;
    
    /*   User defined finish function. It is called once
     *   at the end of the calculation.
     */
    virtual void finish() {}
    
    /*   Make sure that the event counter is set to zero
     *   and the maximum number of events is set correctly.
     */
    void initialize() {
      this->initfunc();
      this->_M_number_of_events = 0;
    }
    
    /*   Increment the event number and do some user defined operations at the
     *   end of the event
     */
    void end_of_event() {
      ++_M_number_of_events;
      this -> operations_at_the_end_of_event();
    }
    
    /*   Returns the number of the analysed events    */
    unsigned long long int number_of_events() const {
      return _M_number_of_events;
    }
    
    /*   returns true if this calculation is finished (no more events are needed)   */
    bool is_done() const {
      return max_number_of_events == 0 ? false : _M_number_of_events >= max_number_of_events;
    }
  
  private:
    /*   User defined init function. It is called once
     *   at the begining of the calculation.
     */
    virtual void initfunc() = 0;
    
    /*   User defined operations at the end of the event. It is virtual but not abstract
     *   function. User might don't have to do anything at the end of the event.
     */
    virtual void operations_at_the_end_of_event() {}

  protected:
    /*   maximum number of events to generate  */
    unsigned long long int max_number_of_events = 0;
    
  private:
    /*   number of the analysed events   */
    unsigned long long int _M_number_of_events = 0;

  };


}

#endif
