/*  Copyright (C) 2002-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 3/12/02.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __NLO_NLO_EVENT_H__
#define __NLO_NLO_EVENT_H__ 1


//   standard c++ includes

//   nlojet++ includes
#include "bits/nlo-tags.h"
#include "bits/nlo-statistical-basis.h"


namespace duct {
  
    
  /* This class helps to decribe the incoming object. 
   * We need more information if the incoming particle 
   * is composite object like a hadron.
   */
  struct colliding_hadron {
    /* The lorentz momentum of the incoming particle*/
    lorentzvector<double> momentum;
    
    /* Type of the hadron (PDG key) */
    int key;
  };


  /* Empty template declaration of the class event.
   * This class is inhereted from the statistical_basis class
   * and add extra, process dependent properties to it.
   */
  template<class _Inh>
  class event; 
  
 
  /*  Definition of the event class for the decay processes. 
   *  In the decay processes we have only one incoming particle.
   *  Here we consider only elementary particle decay.
   */
  template<>
  class event<decay> : public statistical_basis
  {
  public:
    //----- constructors -----
    explicit event(const model *mdl = nullptr)
    : statistical_basis(mdl) {}
    
    event(size_type n, size_type m, const model *mdl = nullptr)
    : statistical_basis(n, 1, m, mdl) {}

    //----- resize the event -----
    void resize(size_type n, size_type m) {
      statistical_basis::resize(n,1U,m);
    }

    //----- merge back the decay subevent -----
    event& merge(int, const event&);
  };
  
  
  /*  Definition of the event class for e+e- like processes 
   *  (the colliding particles are elementary). We have two incomings.
   */
  template<>
  class event<epa> : public statistical_basis
  {
  public:
    //----- constructors -----
    explicit event(const model *mdl = nullptr)
    : statistical_basis(mdl) {}
    
    event(size_type n, size_type m, const model *mdl = nullptr)
    : statistical_basis(n, 2, m, mdl) {}

    //----- resize the event -----
    void resize(size_type n, size_type m) {
      statistical_basis::resize(n,2U,m);
    }

    //----- merge back the decay subevent -----
    event& merge(int, const event<decay>&);
  };
  
  
  /*  Definition of the event class for DIS like processes, one elementary (electron) 
   *  and one hadron in the initial state. Here we have to provide pdf for the hadron.
   */
  template<>
  class event<dis> : public statistical_basis
  {
  public:
    //----- constructors -----
    explicit event(const model *mdl = nullptr)
    : statistical_basis(mdl) {}
    
    event(size_type n, size_type m, const model *mdl = nullptr)
    : statistical_basis(n, 2, m, mdl) {}
    
    //----- acces to the hadron objects -----
    colliding_hadron& operator[](const hadron& i) {
      return _M_hadron;
    }
 
    const colliding_hadron& operator[](const hadron& i) const {
      return _M_hadron;
    }
    
    //----- merge back the decay subevent -----
    event& merge(int, const event<decay>&);

    /* I don't know why it is needed now... could be a gcc-4.6 issue... */
    using statistical_basis::operator[];
     
    //----- resize the event -----
    void resize(size_type n, size_type m) {
      statistical_basis::resize(n,2U,m);
    }

  private:
    //----- hadron object -----
    colliding_hadron _M_hadron;
  };
  
  /*  Definitions of the event class for hadron-hadron collision. We have two hadron 
   *  in the initial state and the corresponding pdf should be provided. 
   */
  template<>
  class event<hhc> : public statistical_basis
  {
  public:
    //----- constructors -----
    explicit event(const model *mdl = nullptr)
    : statistical_basis(mdl) {}
    
    event(size_type n, size_type m, const model *mdl = nullptr) 
    : statistical_basis(n, 2, m, mdl) {}

    //----- acces to the hadron objects -----
    colliding_hadron& operator[](const hadron& i) {
      return _M_hadron[i.idx+1];
    }
 
    const colliding_hadron& operator[](const hadron& i) const {
      return _M_hadron[i.idx+1];
    }
    
    //----- merge back the decay subevent -----
    event& merge(int, const event<decay>&);

    /* I don't know why it is needed now... could be a gcc-4.6 issue... */
    using statistical_basis::operator[];
    
    //----- resize the event -----
    void resize(size_type n, size_type m) {
      statistical_basis::resize(n,2U,m);
    }
    
  private:
    //----- hadron objects -----
    colliding_hadron _M_hadron[2];
  };
  
  
  //  FIXME:  this is very simple output operator
  template<class _Tag> inline
  std::ostream& operator<<(std::ostream& os, const event<_Tag>& a) {
    return os<<static_cast<const statistical_basis&>(a);
  }
 
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<dis>& a) 
  {
    os<<"H : "<<a[hadron(0)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    return os<<static_cast<const statistical_basis&>(a);
  }
 
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<hhc>& a) 
  {
    os<<"A : "<<a[hadron(-1)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    os<<"B : "<<a[hadron(0)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    return os<<static_cast<const statistical_basis&>(a);
  }
}

#endif
