/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/18/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __nlo_model_h__
#define __nlo_model_h__ 1

//   std includes
#include <map>
#include <random>
#include <string>
#include <climits>
#include <initializer_list>

//   project headers
#include <bits/nlo-tags.h>
#include <bits/qcd-coupling.h>


namespace duct {
  
  
  /*   Structure to describe a particle */
  struct particle_info
  {
    /*    default constructor   */
    particle_info()
    : key(INT_MAX), name("---"), mass(0.0), mass2(0.0), color(0), spin(0) {}

    /*    constructor    */
    particle_info(int k, const char *n, double m, int c, int s)
    : key(k), name(n), mass(m), mass2(m*m), color(c), spin(s) {}
    
    //  id of the particle
    int key;
    
    //   name of the particle
    std::string name;
    
    //   mass of the particle
    double mass, mass2;
    
    //   color representation (-3, 0, 3, 8)
    int color;
    
    //   2*spin (0, 1, 2)
    int spin;
  };
  
  
  /*   Structure to describe a hadron. Note that the hadron can play the
   *   role of an incomig hadron, thus we need the pdf information.
   */
  struct hadron_info {
    /*    default constructors    */
    hadron_info() : key(INT_MAX), name("---"), pdf(nullptr) {}
    
    /*    constructor    */
    hadron_info(int k, const char *n, const qcd_hadron *p)
    : key(k), name(n), pdf(p) {}
    
    /*   id of the hadron   */
    int key;
    
    /*   name of the hadron  */
    std::string name;
    
    /*  if it can be an incomming hadron, then
     *  we need the corresponding pdfs.
     */
    const qcd_hadron *pdf;
  };
  
  
  
  /*   Base class to define particles and hadron and basic things
   *   those are important to have parton shower implemented.
   *   All the model extension should be inhereted from this.
   */
  struct model
  {
    /*  types  */
    typedef std::map<int, particle_info> ptbl_type;
    typedef std::map<int, hadron_info>   htbl_type;
    
    /*  default constructor */
    model();
    
    /*  construct from list of particles and hadrons  */
    model(const particle_info *, const hadron_info *);

    /*   adding more particles from initializer list   */
    model(std::initializer_list<particle_info> pl, std::initializer_list<hadron_info> hl) {
      for(const particle_info& p : pl) particle_table[p.key] = p;
      for(const hadron_info& h : hl) hadron_table[h.key] = h;
    }
    
    /*    return the mass and mass2 of a particle   */
    double mass(int key) const { return particle_table.at(key).mass;}
    double mass2(int key) const { return particle_table.at(key).mass2;}
 
    /*    return the color and spin of a particle   */
    int color(int key) const { return particle_table.at(key).color;}
    int spin(int key) const { return particle_table.at(key).spin;}
   
    /*    return the name of a particle or hadron   */
    const std::string& name(int key) const { return particle_table.at(key).name;}
    const std::string& name(const hadron& ih) const { return hadron_table.at(ih.idx).name;}
    
    /*    returns a pointer to the pdfs   */
    const qcd_hadron * pdf(const hadron& ih) const { return hadron_table.at(ih.idx).pdf;}
    
    /*   add new elements to the tables or modify them   */
    particle_info& operator[](int k) { return particle_table[k];}
    hadron_info& operator[](const hadron& ih) { return hadron_table[ih.idx];}
    
    /*   read and write the model from/to stream   */
    void read(std::istream&);
    void write(std::ostream&) const;
    
    /*   This is the particle table. We use a map for this and particles can be added easily.
     */
    std::map<int, particle_info> particle_table;
    
    /*   Similarly we need a hadron table.
     *   At the moment we have to describe only the incomming hadrons.
     */
    std::map<int, hadron_info> hadron_table;
    
    /*   Strong coupling   */
    qcd_coupling alpha_qcd;

    /*   Simple electromagnetic coupling 
     *   It is virtual, thus, it can be overridden. 
     */
    virtual double alpha_qed(double) const;
  };
  
 
  /*  Originaly I wanted to put a thread_local random engine in the model class,
   *  but there are problem in the compilers with thread_local implementation. 
   *  It is better to have separate random engine for every shower engine. 
   *  The type of the random engine is defined here.
   */
  typedef std::mt19937_64 random_engine_type;
  
  
}

#endif
