/*  Copyright (C) 2010-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 8/29/10.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __NLO_NLO_STATISTICAL_BASIS_H__
#define __NLO_NLO_STATISTICAL_BASIS_H__ 1


//   standard c++ includes
#include <iostream>
#include <climits>
#include <initializer_list>

//   nlojet++ includes
#include <bits/hep-lorentzvector.h>
#include <bits/hep-diracspinor.h>
#include <bits/nlo-tags.h>
#include <bits/nlo-model.h>



namespace duct {

  
  class statistical_basis
  {
  public:
    /*  This class is for storing one particle informations.
     *    => flavor        : flavor of the particle
     *    => momentum      : four mementum of the particle
     *    => plus, minus   : momentum in Dirac spinor representation
     *    => color, colorP : color connections 
     */
    struct particle 
    {
      //   default constructor
      particle() : flavor(0) {}
      
      //   flavor information
      int flavor;
      
      //----- momentum of the particle -----
      lorentzvector<double> momentum;
      
      //----- associated spinors -----
      diracspinor<double, ket> plus, minus;
      
      /*  Healthiness of a gluon:
       *    => '-' not applicable
       *    => 'f' frail gluon (or color 8 object)
       *    => 'h' healthy gluon (or color 8 object)
       */
      char health;
      
      struct colorlink 
      {
        //----- make quark a quark connection ----
        static colorlink dead() { 
          return {INT_MAX, INT_MIN};
        } 
     
        //----- make quark a quark connection ----
        static colorlink quark(int n) { 
          return {quarkend, n};
        } 
        
        //----- make quark a antiquark connection ----
        static colorlink antiquark(int p) { 
          return {p, antiquarkend};
        } 
        
        //----- test if it is a quark -----
        bool is_quark() const { return prev == quarkend;}
        bool is_triplet() const { return prev == quarkend;}
         
        //----- test if it is a quark -----
        bool is_antiquark() const { return next == antiquarkend;}
        bool is_antitriplet() const { return next == antiquarkend;}
        
        //----- test if it is a quark -----
        bool is_gluon() const { return next > antiquarkend && prev < quarkend;}
        bool is_octet() const { return next > antiquarkend && prev < quarkend;}
        
        //----- test if it is connection ----
        bool is_qcd() const {
          return next > INT_MIN && prev < INT_MAX && !(prev == quarkend && next == antiquarkend); 
        }
        
        //---- labeling the quark and antiquark ends -----
        static constexpr int quarkend =  INT_MAX-1; 
        static constexpr int antiquarkend = INT_MIN+1; 
        
        /*  The label of the previous and next parton on the color string.
         *  this should follow our labeling convetion (tagged integers). 
         */
        int prev, next;
      } color, colorP;
    };
    
    
    //----- standard public types -----
    typedef size_t size_type;
    typedef particle value_type;
    typedef particle* iterator;
    typedef particle& reference;
    typedef particle* pointer;
    typedef const particle* const_iterator;
    typedef const particle& const_reference;
    typedef const particle* const_pointer;
    typedef particle parton_type;
    
    
    //----- default constructor -----
    explicit statistical_basis(const model *mdl = nullptr)
    : color_suppression_index(0U), _M_start(nullptr), _M_incoming(nullptr), _M_outgoing(nullptr), 
    _M_finish(nullptr), _M_end_of_storage(nullptr), _M_model(mdl) {}
    
    //----- constructor -----
    statistical_basis(size_type k, size_type l, size_type m, const model *mdl)
    : color_suppression_index(0U), _M_model(mdl) { this->_M_allocate(k, l, m);}
    
    //---- copy -----
    statistical_basis(const statistical_basis& p) : _M_model(p._M_model)
    {
      if(p._M_start) {
        this->_M_allocate(p._M_incoming-p._M_start, p._M_outgoing-p._M_incoming, p._M_finish-p._M_outgoing);
        std::copy(p._M_start, p._M_finish, _M_start);
      } else _M_start = _M_incoming = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
      
      color_suppression_index = p.color_suppression_index; 
    }
    
    statistical_basis& operator=(const statistical_basis& p) 
    {
      if(this != &p && p._M_start) {
        this->resize(p._M_incoming-p._M_start, p._M_outgoing-p._M_incoming, p._M_finish-p._M_outgoing);
        std::copy(p._M_start, p._M_finish, _M_start);
        _M_model = p._M_model;
        color_suppression_index = p.color_suppression_index; 
      }
      
      return *this;
    }

    //----- move ----
    statistical_basis(statistical_basis&& p) noexcept {
      _M_start = _M_incoming = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
      _M_model = nullptr;
      color_suppression_index = 0U;
      this->swap(p);
    }
 
    statistical_basis& operator=(statistical_basis&& p) noexcept {
      if(this != &p) this->swap(p);
      return *this;
    }
    
    //----- destructor -----
    ~statistical_basis() {
      if(_M_start) delete [] _M_start;
    }
    
    /* Access to the partons or parton like objects*/ 
    reference operator[](int k) {
      return *(_M_outgoing + k-1);
    }
 
    const_reference operator[](int k) const {
      return *(_M_outgoing + k-1);
    }
    
    //----- iterators ------
    iterator begin() { return _M_start;} 
    iterator end() { return _M_finish;} 
    
    const_iterator begin() const { return _M_start;} 
    const_iterator end() const { return _M_finish;} 

    const_iterator cbegin() const { return _M_start;} 
    const_iterator cend() const { return _M_finish;} 

    template<class _Range>
    std::pair<iterator, iterator> range();

    template<class _Range>
    std::pair<const_iterator, const_iterator> range() const;
  
    template<class _Range>
    std::pair<const_iterator, const_iterator> crange() const {
      return range<_Range>();
    }
    
    //--- structural informations ----
    size_type size() const { return _M_finish-_M_start;}
    size_type capacity() const { return _M_end_of_storage-_M_start;}
    
    template<class _Range>
    size_type number_of() const;
        
    //----- get/set the particle table -----
    const model * get_model() const {
      return _M_model;
    }
    
    void set_model(const model *mdl) {
      _M_model = mdl;
    }

    //----- make the health report of color 8 objects -----
    char health_report(int) const;
    void make_health_report();
    
    //----- momentum sums ----
    template<class _Range>
    lorentzvector<double> momentum_sum() const;
  
    //----- resize the array -----
    void resize(size_type k, size_type l, size_type m) {
      if((size_type)(_M_end_of_storage-_M_start) >= k+l+m) {
        _M_incoming = _M_start + k;
        _M_outgoing = _M_incoming + l; 
        _M_finish = _M_outgoing + m;
      } else {
        this->_M_deallocate();
        this->_M_allocate(k,l,m);
      }
    }
    
    //----- swap two statistical_basis -----
    void swap(statistical_basis& p) noexcept {
      std::swap(_M_start, p._M_start);
      std::swap(_M_incoming, p._M_incoming);
      std::swap(_M_outgoing, p._M_outgoing);
      std::swap(_M_finish, p._M_finish);
      std::swap(_M_end_of_storage, p._M_end_of_storage);
      std::swap(_M_model, p._M_model);
      std::swap(color_suppression_index, p.color_suppression_index); 
    }
    
    /*  This is a public data member. It represents the color supression index 
     *  of the state. During the evolution we are able to keep tracing this 
     *  quantity (I hope). It should be always an even number, unless we start with odd.
     */
    unsigned int color_suppression_index;
 
    //----- functions for inserting open and closed color strings -----
    void insert_closed_string(const int *, const int *);
    void insert_closed_stringP(const int *, const int *);
    
    void insert_open_string(const int *, const int *);
    void insert_open_stringP(const int *, const int *);

    
    //----- inserting open and closed color strings using initializer list -----
    void insert_closed_string(std::initializer_list<int> ll) {
      insert_closed_string(ll.begin(), ll.end());
    }
    
    void insert_closed_stringP(std::initializer_list<int> ll) {
      insert_closed_stringP(ll.begin(), ll.end());
    }

    void insert_open_string(std::initializer_list<int> ll) {
      insert_open_string(ll.begin(), ll.end());
    }

    void insert_open_stringP(std::initializer_list<int> ll) {
      insert_open_stringP(ll.begin(), ll.end());
    }

  private:
    /*  In this class we implement a vector directly 
     *  instead of using the std::vector<...> class. 
     *  With this we can deal with the indexing more efficiently.
     */
    particle *_M_start, *_M_incoming, *_M_outgoing, *_M_finish, *_M_end_of_storage;
    
    //   particle table
    const model *_M_model;
   
    /* Private function to allocate and deallocate memory. 
     * One can use the standard allocate<..> class for this but I 
     * think the new/delete commands are good enough. 
     */
    void _M_allocate(size_type k, size_type l, size_type m) {
      size_type np = k+l+m;
      _M_start = new particle[np];
      if(!_M_start) throw "statistical_basis : Unable to allocate memory.";
      
      _M_finish = _M_end_of_storage = _M_start + np;
      _M_incoming = _M_start + k;
      _M_outgoing = _M_incoming + l; 
    };
    
    void _M_deallocate() {
      if(_M_start) delete [] _M_start;
      _M_start = _M_incoming = _M_outgoing = _M_finish = _M_end_of_storage = 0;
    }
  };
  
  //----- I/O operations -----
  std::ostream& operator<<(std::ostream&, const statistical_basis&); 

  template<> inline 
  statistical_basis::size_type statistical_basis::number_of<qcd_out>() const { 
    return _M_start ? _M_finish-_M_outgoing : 0U;
  }
  
  template<> inline 
  statistical_basis::size_type statistical_basis::number_of<non_qcd>() const { 
    return _M_start ? _M_incoming-_M_start : 0U;
  }
  
  template<> inline 
  statistical_basis::size_type statistical_basis::number_of<outgoings>() const { 
    return _M_start ? (_M_finish-_M_outgoing) + (_M_incoming-_M_start) : 0U;
  }
  
  template<> inline 
  statistical_basis::size_type statistical_basis::number_of<incomings>() const { 
    return _M_start ?_M_outgoing - _M_incoming : 0U;
  }
  
  template<> inline 
  lorentzvector<double> statistical_basis::momentum_sum<qcd_out>() const 
  {
    lorentzvector<double> retval;
    for(auto ri = _M_outgoing; ri < _M_finish; ++ri) 
      retval += ri->momentum;
    return retval;
  }
  
  template<> inline 
  lorentzvector<double> statistical_basis::momentum_sum<non_qcd>() const 
  {
    lorentzvector<double> retval;
    if(_M_start)
      for(auto ri = _M_start; ri < _M_incoming; ++ri) 
        retval += ri->momentum;
    return retval;
  }
  
  template<> inline 
  lorentzvector<double> statistical_basis::momentum_sum<incomings>() const 
  {
    lorentzvector<double> retval;
    if(_M_start)
      for(auto ri = _M_incoming; ri < _M_outgoing; ++ri) 
        retval += ri->momentum;
    return retval;
  }
  
  template<> inline 
  lorentzvector<double> statistical_basis::momentum_sum<outgoings>() const {
    return this->momentum_sum<qcd_out>() + this->momentum_sum<non_qcd>();
  }
  
  template<> inline
  std::pair<statistical_basis::iterator, statistical_basis::iterator> 
  statistical_basis::range<qcd_out>()
  {
    if(_M_start) return std::pair<iterator, iterator>(_M_outgoing, _M_finish);
    return std::pair<iterator, iterator>(nullptr, nullptr);
  }
  
  template<> inline
  std::pair<statistical_basis::const_iterator, statistical_basis::const_iterator> 
  statistical_basis::range<qcd_out>() const
  {
    if(_M_start) return std::pair<const_iterator, const_iterator>(_M_outgoing, _M_finish);
    return std::pair<const_iterator, const_iterator>(nullptr, nullptr);
  }
  

  template<> inline
  std::pair<statistical_basis::iterator, statistical_basis::iterator> 
  statistical_basis::range<non_qcd>()
  {
    if(_M_start) return std::pair<iterator, iterator>(_M_start, _M_incoming);
    return std::pair<iterator, iterator>(nullptr, nullptr);
  }
  
  template<> inline
  std::pair<statistical_basis::const_iterator, statistical_basis::const_iterator> 
  statistical_basis::range<non_qcd>() const
  {
    if(_M_start) return std::pair<const_iterator, const_iterator>(_M_start, _M_incoming);
    return std::pair<const_iterator, const_iterator>(nullptr, nullptr);
  }
  
 
  template<> inline
  std::pair<statistical_basis::iterator, statistical_basis::iterator> 
  statistical_basis::range<incomings>()
  {
    if(_M_start) return std::pair<iterator, iterator>(_M_incoming, _M_outgoing);
    return std::pair<iterator, iterator>(nullptr, nullptr);
  }
  
  template<> inline
  std::pair<statistical_basis::const_iterator, statistical_basis::const_iterator> 
  statistical_basis::range<incomings>() const
  {
    if(_M_start) return std::pair<const_iterator, const_iterator>(_M_incoming, _M_outgoing);
    return std::pair<const_iterator, const_iterator>(nullptr, nullptr);
  }
  
  inline char statistical_basis::health_report(int l) const
  {    
    //---- the health is defined only for color 8 objects -----
    const particle *p = _M_outgoing-1;
    if(!(p[l].color.is_octet())) return '-';
        
    //---- it is a color 8, find the heath status -----
    int rprev = l, lnext;
    do {
      if((lnext = p[rprev].color.next) == l) return 'f';
      if((rprev = p[lnext].colorP.prev) == l) return 'h';
    } while(true);
  }

  inline void statistical_basis::make_health_report()
  {    
    particle *p = _M_outgoing-1;
    for(int l = _M_start-p; l < _M_finish-p; ++l) 
    {
      //---- the health is defined only for color 8 objects -----
      if(!(p[l].color.is_octet())) p[l].health = '-';
      
      //---- it is a color 8, find the heath status -----
      else {
        int rprev = l, lnext;
        do {
          if((lnext = p[rprev].color.next) == l) { p[l].health = 'f'; break;}
          if((rprev = p[lnext].colorP.prev) == l) { p[l].health = 'h'; break;}
        } while(true);
      }
    }
  }
 
  
  inline std::ostream& operator<<(std::ostream& os, const statistical_basis::particle::colorlink& c)
  {
    if(c.is_quark()) os<<"[Q, "<<c.next<<")";
    else if(c.is_antiquark()) os<<"("<<c.prev<<", A]";
    else if(c.is_gluon()) os<<"("<<c.prev<<","<<c.next<<")";
    else os<<"(-,-)";
    
    return os;
  }
  
   
}

#endif
