/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __qcd_color_overlap_u3_h__
#define __qcd_color_overlap_u3_h__ 1

//   std includes

//   project headers
#include <bits/qcd-color.h>
#include <bits/nlo-statistical-basis.h>




namespace duct {
  
  class color_overlap_u3
  {
  private:
    //---- class for representing color graphs ----
    class graph 
    {
    public:
      //---- typedefs ----
      typedef statistical_basis::particle::colorlink colorlink;
      
      //----- class for the color connections -----
      struct parton 
      {
        /*  Multiplicity of the parton insertion:
         *    => multi == flase inactive removed from the list
         *    => multi == true still in the list
         */
        bool active;
        
        //----- color connections -----
        graph::colorlink color, colorP;
      };
      
      //----- default constructor -----
      graph() :  _M_start(nullptr), _M_outgoing(nullptr), _M_finish(nullptr),
      _M_end_of_storage(nullptr), _M_active(0u), _M_nc_pow(0) {}
      
      //----- constructor -----
      graph(unsigned int l, unsigned int m) { 
        this->_M_allocate(l, m);
        _M_active = 0u; _M_nc_pow = 0;
      }
      
      //----- create from statistical basis -----
      graph(const statistical_basis&);
      graph& operator=(const statistical_basis&);
      
      //---- copy -----
      graph(const graph& p)
      {
        if(p._M_start) {
          this->_M_allocate(p._M_outgoing-p._M_start, p._M_finish-p._M_outgoing);
          std::copy(p._M_start, p._M_finish, _M_start);
          _M_active = p._M_active;
          _M_nc_pow = p._M_nc_pow;
        } else {
          _M_start = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
          _M_active = 0u; _M_nc_pow = 0;
        }
      }
      
      graph& operator=(const graph& p) 
      {
        if(this != &p && p._M_start) {
          this->resize(p._M_outgoing-p._M_start, p._M_finish-p._M_outgoing);
          std::copy(p._M_start, p._M_finish, _M_start);
          _M_active = p._M_active;
          _M_nc_pow = p._M_nc_pow;
        }
        
        return *this;
      }
      
      //----- move ----
      graph(graph&& p) noexcept {
        _M_start = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
        _M_active = 0u; _M_nc_pow = 0;
        this->swap(p);
      }
      
      graph& operator=(graph&& p) noexcept {
        if(this != &p) this->swap(p);
        return *this;
      }
      
      //----- destructor -----
      ~graph() {
        if(_M_start) delete [] _M_start;
      }
      
      //----- Access to the partons -----
      parton& operator[](int k) {
        return *(_M_outgoing + k-1);
      }
      
      const parton& operator[](int k) const {
        return *(_M_outgoing + k-1);
      }
            
      //--- structural informations ----
      unsigned int size() const { return _M_finish-_M_start;}
      unsigned int capacity() const { return _M_end_of_storage-_M_start;}
      bool is_finished() const { return !_M_active;}
      
      //----- this applys all the simple rules -----
      void reduce();

      //----- rules to simplify the graps -----
      void ruleG();
      void ruleQ();
      void ruleA();
      void ruleT();
      void ruleS();
      void ruleF();
    
      //----- clean the whole graph -----
      void clear() {
        for(auto pi = _M_start; pi < _M_finish; ++pi)
          pi->active = false;
        _M_active = 0U;
      }
      
      //----- resize the array -----
      void resize(unsigned int l, unsigned int m) {
        if((unsigned int)(_M_end_of_storage-_M_start) >= l+m) {
          _M_outgoing = _M_start + l; 
          _M_finish = _M_outgoing + m;
        } else {
          this->_M_deallocate();
          this->_M_allocate(l,m);
        }
      }
      
      //----- swap two color_graph -----
      void swap(graph& p) noexcept {
        std::swap(_M_start, p._M_start);
        std::swap(_M_outgoing, p._M_outgoing);
        std::swap(_M_finish, p._M_finish);
        std::swap(_M_end_of_storage, p._M_end_of_storage);
        std::swap(_M_active, p._M_active);
        std::swap(_M_nc_pow, p._M_nc_pow);
      }
 
      //------ print the graph -----
      std::ostream& print(std::ostream&) const; 

      /*   remove a gluon    */
      void remove_gluon(int r)
      {
        parton *p = _M_outgoing-1;
        
        p[p[r].color.prev].color.next = p[r].color.next;
        p[p[r].color.next].color.prev = p[r].color.prev;
        
        p[p[r].colorP.prev].colorP.next = p[r].colorP.next;
        p[p[r].colorP.next].colorP.prev = p[r].colorP.prev;
        
        p[r].active = false;
        --_M_active;
      }
      
      //----- private members -----
      parton *_M_start, *_M_outgoing, *_M_finish, *_M_end_of_storage;
      unsigned int _M_active;
      int _M_nc_pow;
      
      //---- allocate memory -----
      void _M_allocate(unsigned int l, unsigned int m) {
        unsigned int np = l+m;
        _M_start = new parton[np];
        if(!_M_start) throw "color_overlap_u3::grap : Unable to allocate memory.";
        
        _M_finish = _M_end_of_storage = _M_start + np;
        _M_outgoing = _M_start + l; 
      };
         
      //----- deallocate memory -----
      void _M_deallocate() {
        if(_M_start) delete [] _M_start;
        _M_start = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
      }
    } _M_graph;
    
  public:
    //----- calculate the weight of a color configuration -----
    double operator()(const statistical_basis&);
  };
   
}


#endif
