/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __NLO_QCD_COUPLING_H__
#define __NLO_QCD_COUPLING_H__ 1

//   std includes
#include <cmath>
#include <iostream>

//   project headers
#include "bits/qcd-color.h"



namespace duct {
  
  /*  Simple MSbar implementation of the QCD coupling
   *  with the PDG values of the masses and as(Mz).
   */
  double __qcd_coupling_msbar(double);

  
  /*  The QCD coupling implementation in the hower code.
   *  Here we need more than a simple function. We have 
   *  linear spline for the alpha_s and for the tau variables.
   */
  template<unsigned int _Np>
  struct __qcd_coupling
  {
    explicit __qcd_coupling(double q2min = 0.8, double q2max = 2.0e+8)
    : __qcd_coupling(__qcd_coupling_msbar, q2min, q2max) {}

    template<class _Func>
    explicit __qcd_coupling(const _Func&, double = 0.8, double = 2.0e+8);
    
    /*   Returns the alphas value at the given scale   */
    double operator()(double) const;
    
    /*   Returns the MC alphas value at the given scale
     *     ask = as/(1-as*K)
     */
    double ask(double) const;
    
    /*   Logarithmic integrals of alphas and its inverse   */
    double tau(double) const;
    double tau_inverse(double) const;

    /*   Logarithmic integrals of alphas and its inverse with some NLO correction (Kg term)  */
    double tauk(double) const;
    double tauk_inverse(double) const;
    
    /*   Returns the evolution range   */
    double q2min() const { return q2n;}
    double q2max() const { return s;}
    
    /*   I/O operations   */
    void write(std::ostream& os) const {
      os.write(reinterpret_cast<const char *>(this), sizeof(__qcd_coupling<_Np>));
    }

    void read(std::istream& is) {
      is.read(reinterpret_cast<char *>(this), sizeof(__qcd_coupling<_Np>));
    }
    
  private:
    /*   interpolating alphas function   */
    struct __gridpoint {
      double t, as, ask, tau, tauk;
    } grid[_Np+1];
    
    double q2n, s, dt;
  };
  
  
  template<unsigned int _Np> template<class _Func>
  __qcd_coupling<_Np>::__qcd_coupling(const _Func& __as, double __q2n, double __s)
  : q2n(__q2n), s(__s), dt(std::log(__s/__q2n)/_Np)
  {
    double kg = Kg(5U);
    
    for(unsigned k = 0; k <= _Np; k++) {
      grid[k].t  = k*dt;
      grid[k].as = __as(s*std::exp(-grid[k].t));
      grid[k].ask = grid[k].as/(1.0-grid[k].as*kg);
    }
    
    /*    calculating the tau spline   */
    grid[0].tau = grid[0].tauk = 0.0;
    
    for(unsigned k = 0; k < _Np; k++) {
      grid[k+1].tau  = grid[k].tau  + 0.5*dt*(grid[k].as  + grid[k+1].as);
      grid[k+1].tauk = grid[k].tauk + 0.5*dt*(grid[k].ask + grid[k+1].ask);
    }
  }
  
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::operator()(double mu2) const
  {
    double tin = std::log(s/mu2);
    int k = static_cast<int>(tin/dt);
    
    /*  extrapolating by freezing the basepoints  */
    if(k < 0) k = 0;
    if(k >= (int) _Np) k = (int) _Np-1;
        
    /*   interpolating the function  */
    const __gridpoint& g1 = grid[k];
    const __gridpoint& g2 = grid[k+1];
        
    return (g1.as*(g2.t-tin) + g2.as*(tin-g1.t))/dt;
  }
  
  
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::ask(double mu2) const
  {
    double tin = std::log(s/mu2);
    int k = static_cast<int>(tin/dt);
    
    /*  extrapolating by freezing the basepoints  */
    if(k < 0) k = 0;
    if(k >= (int) _Np) k = (int) _Np-1;
    
    /*   interpolating the function  */
    const __gridpoint& g1 = grid[k];
    const __gridpoint& g2 = grid[k+1];
    
    return (g1.ask*(g2.t-tin) + g2.ask*(tin-g1.t))/dt;
  }
  
  
 
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::tau(double mu2) const
  {
    double tin = std::log(s/mu2);
    int k = static_cast<int>(tin/dt);
    
    /*  extrapolating by freezing the basepoints  */
    if(k < 0) return 0.0;
    if(k >= (int) _Np) k = (int) _Np-1;
    
    /*   interpolating the function  */
    const __gridpoint& g1 = grid[k];
    const __gridpoint& g2 = grid[k+1];
    
    double da = (g2.as-g1.as);
    double ca = (g1.as*g2.t-g2.as*g1.t);
    
    return g1.tau + (tin-g1.t)*(0.5*da*(tin+g1.t) + ca)/dt;
  }
  
  
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::tau_inverse(double tt) const
  {
    /*   if it is out of range return something reasonable  */
    if(tt <= 0.0) return s;
    
    /*   findig the gridpoint  */
    int lo = 0, hi = _Np, mi;
    
    if(tt >= grid[_Np].tau)
      lo = (int) _Np-1;
    else {
      while(lo != (mi = (lo+hi)/2)) {
        if(grid[mi].tau <= tt) lo = mi;
        else hi = mi;
      }
    }
    
    /*   interpolation the inverse function   */
    const __gridpoint& g1 = grid[lo];
    double d = (grid[lo+1].as-g1.as)/dt;
    
    return s/std::exp(g1.t + (std::sqrt(g1.as*g1.as+2.0*d*(tt-g1.tau))-g1.as)/d);
  }

  
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::tauk(double mu2) const
  {
    double tin = std::log(s/mu2);
    int k = static_cast<int>(tin/dt);
    
    /*  extrapolating by freezing the basepoints  */
    if(k < 0) return 0.0;
    if(k >= (int) _Np) k = (int) _Np-1;
    
    /*   interpolating the function  */
    const __gridpoint& g1 = grid[k];
    const __gridpoint& g2 = grid[k+1];
    
    double da = (g2.ask-g1.ask);
    double ca = (g1.ask*g2.t-g2.ask*g1.t);
    
    return g1.tauk + (tin-g1.t)*(0.5*da*(tin+g1.t) + ca)/dt;
  }
  
  
  template<unsigned int _Np>
  double __qcd_coupling<_Np>::tauk_inverse(double tt) const
  {
    /*   if it is out of range return something reasonable  */
    if(tt <= 0.0) return s;
    
    /*   findig the gridpoint  */
    int lo = 0, hi = _Np, mi;
    
    if(tt >= grid[_Np].tauk)
      lo = (int) _Np-1;
    else {
      while(lo != (mi = (lo+hi)/2)) {
        if(grid[mi].tauk <= tt) lo = mi;
        else hi = mi;
      }
    }
    
    /*   interpolation the inverse function   */
    const __gridpoint& g1 = grid[lo];
    double d = (grid[lo+1].ask-g1.ask)/dt;
    
    return s/std::exp(g1.t + (std::sqrt(g1.ask*g1.ask+2.0*d*(tt-g1.tauk))-g1.ask)/d);
  }
  
  
  /*   The default qcd_coupling is based on a 50 points grid.*/
  typedef __qcd_coupling<50u> qcd_coupling;
  
}




namespace duct {
  
  struct qcd_hadron
  { 
    /*  virtual destructor   */
    virtual ~qcd_hadron() = default;
    
    /*  the pdf fucntion */
    virtual double parton(int, double, double) const = 0;
    
    /* Some structural informations those are important 
     * in the shower evolution.
     *
     * start of the evolution (in GeV^2)
     */
    virtual double q2min() const = 0;
    
    /*  threshold information (in GeV^2) */
    virtual double thrs(int) const = 0;
    
    /*  the threshold can depend on eta (in GeV^2) */
    virtual double thrs(int, double) const = 0; 

    /*  smallest x value  */
    virtual double xmin() const = 0;
    
    /*  largest x value  */
    virtual double xmax(int, double) const = 0;
  };
  
  
  /*  Anti-hadron adaptor. It turns a hadron pdf to antihadron pdf.
   */
  class qcd_antihadron : public qcd_hadron
  {
  public:
    //----- it can be constructed from a hadron -----
    explicit qcd_antihadron(const qcd_hadron *__pdf = nullptr)
    : _M_pdf(__pdf) {}
    
    //---- copy ----
    qcd_antihadron(const qcd_antihadron&) = default;
    qcd_antihadron& operator=(const qcd_antihadron&) = default;
    
    //----- move ----
    qcd_antihadron(qcd_antihadron&&) = default;
    qcd_antihadron& operator=(qcd_antihadron&&) = default;
    
    //----- destructor -----
    ~qcd_antihadron() = default;
    
    //----- assignment -----
    qcd_antihadron& operator=(const qcd_hadron *__pdf) {
      _M_pdf = __pdf;
      return *this;
    }
    
    //----- pdf function -----
    double parton(int fl, double x, double q2) const {
      return _M_pdf->parton(-fl, x, q2);
    }
    
    /* Some structural informations those are important 
     * in the shower evolution.
     *
     * start of the evolution (in GeV^2)
     */
    double q2min() const { 
      return _M_pdf->q2min();
    }
    
    /*  threshold information (in GeV^2) */
    double thrs(int fl) const { 
      return _M_pdf->thrs(fl);
    }
    
    /*  the threshold can depend on eta (in GeV^2) */
    double thrs(int fl, double eta) const { 
      return _M_pdf->thrs(fl, eta);
    }

    /*  largest x value  */
    double xmax(int fl, double q2) const {
      return _M_pdf->xmax(fl, q2);
    }
    
    /*  smallest x value  */
    double xmin() const{
      return _M_pdf->xmin();
    }

  private:
    const qcd_hadron *_M_pdf;
  };

  
  /*  Neutron adaptor. It turns a proton pdf to neutron pdf.
   */
  class qcd_neutron : public qcd_hadron
  {
  public:
    //----- it can be constructed from a hadron -----
    explicit qcd_neutron(const qcd_hadron *__pdf = nullptr)
    : _M_pdf(__pdf) {}
    
    //---- copy ----
    qcd_neutron(const qcd_neutron&) = default;
    qcd_neutron& operator=(const qcd_neutron&) = default;
    
    //----- move ----
    qcd_neutron(qcd_neutron&&) = default;
    qcd_neutron& operator=(qcd_neutron&&) = default;
    
    //----- destructor -----
    ~qcd_neutron() = default;
    
    //----- assignment -----
    qcd_neutron& operator=(const qcd_hadron *__pdf) {
      _M_pdf = __pdf;
      return *this;
    }
    
    //----- pdf function -----
    double parton(int fl, double x, double q2) const
    {
      switch(fl) {
        case -2: fl = -1; break;
        case -1: fl = -2; break;
        case  1: fl =  2; break;
        case  2: fl =  1; break;
      }
      
      return _M_pdf->parton(fl, x, q2);
    }
 
    /* Some structural informations those are important
     * in the shower evolution.
     *
     * start of the evolution (in GeV^2)
     */
    double q2min() const {
      return _M_pdf->q2min();
    }
    
    /*  threshold information (in GeV^2) */
    double thrs(int fl) const {
      return _M_pdf->thrs(fl);
    }
    
    /*  the threshold can depend on eta (in GeV^2) */
    double thrs(int fl, double eta) const {
      return _M_pdf->thrs(fl, eta);
    }
    
    /*  largest x value  */
    double xmax(int fl, double q2) const {
      return _M_pdf->xmax(fl, q2);
    }
    
    /*  smallest x value  */
    double xmin() const{
      return _M_pdf->xmin();
    }
    
  private:
    const qcd_hadron *_M_pdf;
  };
  
}


#endif