/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/11/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __qcd_dglap_h__
#define __qcd_dglap_h__ 1

//   std includes
#include <iostream>

//   project headers
#include <bits/hep-matrix.h>
#include <bits/qcd-coupling.h>
#include <bits/nlo-model.h>


namespace duct {
  
  
  class qcd_dglap : public qcd_hadron
  {
  private:
    /*    container for the pdf array   */
    struct array
    {
      /*    constructors and assignment  */
      array(double x = 0.0) { for(auto& p : _M_elem) p = x;}
      array& operator=(double x) { return (*this = array(x));}
      array& operator*=(double x) { for(auto& p : _M_elem) p *= x; return *this;}
      array& operator/=(double x) { for(auto& p : _M_elem) p /= x; return *this;}
      
      /*    more assignments    */
      array& operator+=(const array& x) {
        const double *c = x._M_elem; for(int k = 0; k < 9; k++) _M_elem[k] += c[k];
        return *this;
      }
      
      array& operator-=(const array& x) {
        const double *c = x._M_elem; for(int k = 0; k < 9; k++) _M_elem[k] -= c[k];
        return *this;
      }
      
      array& operator*=(const array& x) {
        const double *c = x._M_elem; for(int k = 0; k < 9; k++) _M_elem[k] *= c[k];
        return *this;
      }
      
      /*   element access   */
      const double& operator[](unsigned int i) const { return _M_elem[i];}
      double& operator[](unsigned int i) { return _M_elem[i];}
      
      //----- Arithmetic operators -----
      array operator-() const { return array(*this) *= -1.0;}
      array operator*(double b) const { return array(*this) *= b;}
      array operator/(double b) const { return array(*this) /= b;}
      array operator+(const array& b) const { return array(*this) += b;}
      array operator-(const array& b) const { return array(*this) -= b;}
      array operator*(const array& b) const { return array(*this) *= b;}
      
      double _M_elem[9];
    };
    
    //----- container for the splitting kernels -----
    struct kernel
    {
      /*   constructors   */
      explicit kernel(double x = 1.0)  {
        for(auto& p : _M_elem) p = 0.0;
        _M_elem[0] = _M_elem[5] = _M_elem[10] = x;
        _M_elem[15] = _M_elem[16] = x;
      }
      
      /*   assignment   */
      kernel& operator=(double x) { return (*this = kernel(x));}
      kernel& operator*=(double x) { for(auto& p : _M_elem) p *= x; return *this;}
      kernel& operator/=(double x) { for(auto& p : _M_elem) p /= x; return *this;}
      
      /*   more assignments   */
      kernel& operator+=(const kernel& x) {
        const double *c = x._M_elem; for(int k = 0; k < 17; k++) _M_elem[k] += c[k];
        return *this;
      }
      
      kernel& operator-=(const kernel& x) {
        const double *c = x._M_elem; for(int k = 0; k < 17; k++) _M_elem[k] -= c[k];
        return *this;
      }
      
      //----- element access -----
      const double* operator[](unsigned int i) const { return _M_elem + 4*i;}
      double* operator[](unsigned int i) { return _M_elem+4*i;}
      
      //----- Arithmetic operators -----
      array operator*(const array& b) const;
      kernel operator*(const kernel& b) const;
      kernel operator-() const { return kernel(*this) *= -1.0;}
      kernel operator*(double b) const { return kernel(*this) *= b;}
      kernel operator/(double b) const { return kernel(*this) /= b;}
      kernel operator+(const kernel& b) const { return kernel(*this) += b;}
      kernel operator-(const kernel& b) const { return kernel(*this) -= b;}

      double _M_elem[17];
    };
    
  public:
    //----- switches to sellect evolution code -----
    enum mode : int { msbar = 0, lambda, kT, lambdaT};
    
    /*    default constructor create an empty object   */
    qcd_dglap() = default;
    
    template<typename _Func>
    qcd_dglap(mode, double, double, double, const qcd_coupling&, _Func);

    template<typename _Func>
    qcd_dglap(mode, double, double, double, qcd_coupling&&, _Func);
    
    template<typename _Func>
    qcd_dglap(double q2min, double m2c, double m2b, const qcd_coupling& as, _Func func)
    : qcd_dglap(msbar, q2min, m2c, m2b, as, func) {}
    
    template<typename _Func>
    qcd_dglap(double q2min, double m2c, double m2b, qcd_coupling&& as, _Func func)
    : qcd_dglap(msbar, q2min, m2c, m2b, std::move(as), func) {}
    
  
    template<typename _Func>
    qcd_dglap(mode type, double q2min, _Func func, const model *mdl)
    : qcd_dglap(type, q2min, mdl->mass2(4), mdl->mass2(5), mdl->alpha_qcd, func) {}
    
    template<typename _Func>
    qcd_dglap(double q2min, _Func func, const model *mdl)
    : qcd_dglap(msbar, q2min, mdl->mass2(4), mdl->mass2(5), mdl->alpha_qcd, func) {}
    
    
    //----- calculate the pdf for a flavor -----
    /*  flavor:
     *      10 => gluon
     *      11 => singlet = u+ubar + d+dbar + s+sbar
     *      12 => c+cbar
     *      13 => b+bbar
     *      14 => d-dbar
     *      15 => u-ubar
     *      16 => s-sbar
     *      17 => T3 = u+ubar - (d+dbar)
     *      18 => T8 = u+ubar + d+dbar - 2*(s+sbar)
     *      19 => Singlet = siglet + (c+cbar) + (b+bbar)
     */
    double parton(int, double, double) const;
    double momentum(int fl, double q2) const;

    double operator()(int fl, double x, double q2) const {
      return this->parton(fl, x, q2);
    }
    
    /* Start of the evolution (in GeV^2)  */
    double q2min() const {
      return _M_q2min;
    }
    
    /*  Threshold information (in GeV^2) */
    double thrs(int fl) const {
      if(_M_type == lambda || _M_type == msbar) {
        switch(fl) {
          case 6: case -6: return 1.0e+100;
          case 5: case -5: case 13: return _M_m2b; break;
          case 4: case -4: case 12: return _M_m2c; break;
        }
      }
      
      return 0.0;
    }
    
    /*  x dependent thresholds (in GeV^2)   */
    double thrs(int, double) const;
    
    /*  Smallest x value in the interpolation table */
    double xmin() const {
      return _M_x[0];
    }
    
    /*  Largest x value where the pdf is still defined
     *  in the massive evolution it is not trivial.
     */
    double xmax(int fl, double q2) const {
      if(_M_type == lambda)
        switch(fl) {
          case -4: case 4: case 12: return 1.0-_M_m2c/q2; break;
          case -5: case 5: case 13: return 1.0-_M_m2b/q2; break;
        }
      
      return 1.0;
    }
    
    /*    binary read and write   */
    void read(std::istream&);
    void write(std::ostream&) const;
    
  private:
    /*   Type of the evolution can be massless or massive   */
    mode _M_type;
    
    /*   Strong coupling   */
    qcd_coupling _M_alpha_qcd;

    /*   threshold scales (copy from the _M_mdl table)    */
    double _M_q2min, _M_m2c, _M_m2b;
    
    /*   grids in x and t    */
    std::vector<double> _M_x, _M_t;
    
    /*   the interpolation table   */
    matrix<array> _M_h;

    /*   generate the pdf table   */
    template<typename _Func> void _M_evolve(_Func);
    
    /*   private functions to set up the grid */
    void _M_xgrid();
    void _M_tgrid();
    
    /*   private functions to do the interpolation   */
    double _M_parton(int, unsigned, unsigned) const;
    
    /*   function to calculate the weight    */
    void _M_weights_msbar(unsigned int, unsigned int, unsigned int, kernel&);
    void _M_weights_lambda(unsigned int, unsigned int, unsigned int, kernel&);
    void _M_weights_kT(unsigned int, unsigned int, unsigned int, kernel&);
    void _M_weights_lmdT(unsigned int, unsigned int, unsigned int, kernel&);
   
    /*     functions for calculating the weights (massless LO kernels)    */
    static double _M_s1qq(double, double);
    static double _M_s1qg(double, double);
    static double _M_s1gq(double, double);
    static double _M_s1gg(double, double);
    
    static double _M_s2qq(double, double);
    static double _M_s2qg(double, double);
    static double _M_s2gq(double, double);
    static double _M_s2gg(double, double);
    
    static double _M_s3qq(double);
    static double _M_s3gg(double);
    
    /* Some extra arithmetic operators are declared as friends.
     * It compiles and I hope it is OK with C++11.
     */
    friend kernel inverse(const kernel&);
    friend array operator*(double, const array&);
    friend kernel operator*(double, const kernel&);
  };
  
  
  template<typename _Func> inline
  qcd_dglap::qcd_dglap(mode type, double q2min, double m2c, double m2b, const qcd_coupling& as, _Func func)
  : _M_type(type), _M_alpha_qcd(as), _M_q2min(q2min), _M_m2c(m2c), _M_m2b(m2b) { this->_M_evolve(func); }
  
  template<typename _Func> inline
  qcd_dglap::qcd_dglap(mode type, double q2min, double m2c, double m2b, qcd_coupling&& as, _Func func)
  : _M_type(type), _M_alpha_qcd(std::move(as)), _M_q2min(q2min), _M_m2c(m2c), _M_m2b(m2b) { this->_M_evolve(func); }
  
  
  template<typename _Func>
  void qcd_dglap::_M_evolve(_Func func)
  {
    //----- setting up the t and x grids -----
    this -> _M_tgrid();
    this -> _M_xgrid();
    
    //----- select the function for calculating the weights -----
    void (qcd_dglap::*weights)(unsigned int, unsigned int, unsigned int, kernel&) = nullptr;
    
    switch(_M_type) {
      case msbar:   weights = &qcd_dglap::_M_weights_msbar; break;
      case lambda:  weights = &qcd_dglap::_M_weights_lambda; break;
      case lambdaT: weights = &qcd_dglap::_M_weights_lmdT; break;
      case kT:      weights = &qcd_dglap::_M_weights_kT; break;
    }
    
    /*    upper indices on the x,t grids   */
    unsigned int nx = _M_x.size()-1;
    unsigned int nt = _M_t.size();
    
    /*    resizing the matricies    */
    _M_h = matrix<array>(nx+1, nt);
    
    /*    at x = 1 the pdf and its derivative is zero  */
    matrix<array> _M_dh(nx+1, nt);
    
    /*    the function values at the fitting scale   */
    double __f[11], dp, up, sp;
    double *f = __f+5;
    
    for(unsigned k = 0; k < nx; k++) {
      array& h0 = _M_h[k][0];
      
      func(_M_x[k], f);
      
      dp = f[-1]+f[1];
      up = f[-2]+f[2];
      sp = f[-3]+f[3];
      
      //---- sinlet ----
      h0[0] = f[0];
      h0[1] = dp+up+sp;
      h0[2] = f[-4]+f[4];
      h0[3] = f[-5]+f[5];
      
      //---- valence ----
      h0[4] = f[1]-f[-1];
      h0[5] = f[2]-f[-2];
      h0[6] = f[3]-f[-3];
      
      //---- T3 and T8 ----
      h0[7] = up-dp;
      h0[8] = dp+up-2.0*sp;
    }
    
    //----- the t derivatives at fitting scale -----
    kernel w;
    
    for(unsigned k = 0; k < nx; k++) {
      _M_dh[k][0] = 0.0;
      
      for(unsigned l = k; l < nx; l++) {
        (this->*weights)(k, l, 0, w);
        _M_dh[k][0] += w*_M_h[l][0];
      }
    }
    
    //----- now we solve the equation iteratively ------
    array S;
    
    for(unsigned int it = 1; it < nt; it++) {
      double dt = 0.5*(_M_t[it] - _M_t[it-1]);
      
      for(int k = nx-1; k >= 0; --k) {
        //----- calculate S -----
        S = 0.0;
        for(unsigned l = k+1; l < nx; l++) {
          (this->*weights)(k, l, it, w);
          S += w*_M_h[l][it];
        }
        
        //----- calculate w matrix -----
        (this->*weights)(k, k, it, w);
        
        _M_dh[k][it] = inverse((kernel(1.0) -= dt*w))*(w*(_M_h[k][it-1]+dt*_M_dh[k][it-1])+S);
        _M_h[k][it] = _M_h[k][it-1] + dt*(_M_dh[k][it]+_M_dh[k][it-1]);
      }
    }
  }
  
  
  /*  Some operator what we need to do arithmetics with
   *  object with type array.
   */
  inline qcd_dglap::array operator*(double b, const qcd_dglap::array& a) {
    return qcd_dglap::array(a) *= b;
  }
  
  inline qcd_dglap::kernel operator*(double b, const qcd_dglap::kernel& a) {
    return qcd_dglap::kernel(a) *= b;
  }
 
  
  /*  Simple classes to select the dglap mode for the various shower ordering.
   *  (In C++14 we will use variable template.)
   */
  template<class _Ordering>
  struct __dglap_mode {
    static const qcd_dglap::mode value = qcd_dglap::msbar;
  };
  
  template<>
  struct __dglap_mode<ordering::lambda> {
    static const qcd_dglap::mode value = qcd_dglap::lambda;
  };
  
  template<>
  struct __dglap_mode<ordering::lambdaT> {
    static const qcd_dglap::mode value = qcd_dglap::lambdaT;
  };
  
  template<>
  struct __dglap_mode<ordering::kT> {
    static const qcd_dglap::mode value = qcd_dglap::kT;
  };
  
}


#endif
