/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   std includes

//   project headers
#include "bits/qcd-color-overlap-u3.h"


namespace duct {

  std::ostream& color_overlap_u3::graph::print(std::ostream& os) const
  {
    int idx = _M_start-_M_outgoing + 1;
    os<<"+------------------------\n";
    os<<"| nc_pow = "<<_M_nc_pow<<"\n";
    os<<"| active = "<<_M_active<<"\n";
    os<<"+---- qcd partons -----\n";
    if(_M_active) {
      for(auto pi = _M_start; pi < _M_finish; ++pi) 
        if(pi->active) {
          os<<"| "<<idx++
          <<"  "<<pi->color
          <<"  "<<pi->colorP
          <<std::endl;
        } else idx++;
    }
    os<<"+----\n";
    
    
    return os;
  }
 
  
  
  color_overlap_u3::graph::graph(const statistical_basis& p)
  {
    if(p.size()) {
      int np = p.number_of<qcd_out>();
      int ni = p.number_of<incomings>();
      
      //---- count the qcd incomings -----
      switch(ni) {
        case 2: if(!p[-1].color.is_qcd()) --ni;
        case 1: if(!p[ 0].color.is_qcd()) --ni;
      }
      
      //---- allcate memory ---- 
      this->_M_allocate((unsigned int) ni, (unsigned int) np);
      parton *q = _M_outgoing-1;
      
      //---- copy the color connections ----
      _M_active = ni+np;
      _M_nc_pow = 0;
      
      for(int i = 1-ni; i <= np; ++i) {
        q[i].color = p[i].color;
        q[i].colorP = p[i].colorP;
        q[i].active = true;
        
        if(q[i].color.is_gluon() || q[i].color.is_quark()) --_M_nc_pow;
      }
    }
    else {
      _M_start = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
      _M_nc_pow = 0;
    }
  }

  
  color_overlap_u3::graph& color_overlap_u3::graph::operator=(const statistical_basis& p)
  {
    if(p.size()) {
      int np = p.number_of<qcd_out>();
      int ni = p.number_of<incomings>();
      
      //---- count the qcd incomings -----
      switch(ni) {
        case 2: if(!p[-1].color.is_qcd()) --ni;
        case 1: if(!p[ 0].color.is_qcd()) --ni;
      }
      
      //---- allcate memory ----
      this->resize((unsigned int) ni, (unsigned int) np);
      parton *q = _M_outgoing-1;
      
      //---- copy the color connections ----
      _M_active = ni+np;
      _M_nc_pow = 0;
      
      for(int i = 1-ni; i <= np; ++i) {
        q[i].color = p[i].color;
        q[i].colorP = p[i].colorP;
        q[i].active = true;
        
        if(q[i].color.is_gluon() || q[i].color.is_quark()) --_M_nc_pow;
      }
    }
    
    return *this;
  }
  
  
  void color_overlap_u3::graph::ruleG()
  {
    /*  Here we try to find all the paralell and crossed gluons and combine them into 
     *  a multi-gluon.
     */
    unsigned int active;
    parton *p = _M_outgoing-1, *pi;
    
    do {
      pi = _M_start;
      active = _M_active;
      
    again:
      while(pi < _M_finish && (!pi->active || !pi->color.is_gluon())) pi++;
      
      if(pi < _M_finish) {
        int n, j = pi-p;
        
      againP:
        //----- combine two paralell multi gluons -----
        if(p[n = p[j].color.next].color.is_gluon() && n == p[j].colorP.next) {
          //----- removes closed loop squres with one multigluon -----
          if(n == j) {
            ++_M_nc_pow;
            p[j].active = false;
            --_M_active;
            
            pi++;
            goto again;
          }
          
          //----- otherwise combine the two gluons ----
          this->remove_gluon(n);
          ++_M_nc_pow;
          
          //----- repeat the paralell gluon check -----
          goto againP;
        }
        
        //----- remove two crossed gluons -----
        if(p[n = p[j].color.next].color.is_gluon() && n == p[j].colorP.prev) {
          if(p[n].color.next == j) ++_M_nc_pow;
          if(p[n].colorP.prev == j) ++_M_nc_pow;
          
          this->remove_gluon(n);
          this->remove_gluon(j);
        }
        
        //----- get another gluon -----
        pi++;
        goto again;
      }
    } while(active != _M_active);
  }
  
  
  void color_overlap_u3::graph::ruleQ()
  {
    /*  This routine tries to identify every gluon or multi-gluon those are paralell 
     *  with a quark line. If the quark is paralell with the anti-quark then we remove both 
     *  of them and increase the power of the Nc factor by 1.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  again:
    while(pi < _M_finish && (!(pi->active) || !pi->color.is_quark())) pi++;
    
    if(pi < _M_finish) {
      int n, r, l, j = pi-_M_outgoing+1;
      
      while(true) {
        //------remove gluon that is paralell to the quark -----
        if(p[j].color.next == (n = p[j].colorP.next)) {
          if(p[n].color.is_antiquark()) {
            //----- remove the whole fermion loop -----
            ++_M_nc_pow;
            p[j].active = p[n].active = false;
            _M_active -= 2;
            break;
          }
          else {
            //----- remove the gluon -----
            this->remove_gluon(n);
            ++_M_nc_pow;
          }
        }
     
        //----- now we try to find and remove two crossed gluons (not symmetric) -----
        else if((p[l=p[j].color.next].color.is_gluon() && p[r=p[j].colorP.next].colorP.is_gluon() && p[r].colorP.next == l)
                || (p[l=p[j].color.next].color.is_gluon() && p[r=p[j].colorP.next].colorP.is_gluon() && p[l].color.next == r))
        {
          this->remove_gluon(l);
          this->remove_gluon(r);
        }
        
        //----- nothing to do -----
        else break;
      }
      
      pi++;
      goto again;
    }
  }
  
  
  void color_overlap_u3::graph::ruleA()
  {
    /*  This routine tries to identify every gluon or multi-gluon those are paralell 
     *  with a anti-quark line or two or thee gluons are crossed next to a anti-quark 
     *  line and remove them. If the quark is paralell with the anti-quark then we 
     *  remove both of them and increase the power of the Nc factor by 1.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  again:
    while(pi < _M_finish && (!(pi->active) || !pi->color.is_antiquark())) pi++;
    
    if(pi < _M_finish) {
      int r, l, j = pi-_M_outgoing+1;
      
      while(true) {
        //------remove gluon that is paralell to the anit-quark -----
        if(p[j].color.prev == (r = p[j].colorP.prev)) {
          if(p[r].color.is_quark()) {
            //----- remove the whole fermion loop -----
            ++_M_nc_pow;
            p[j].active = p[r].active = false;
            _M_active -= 2;
            break;
          }
          else {
            //----- remove the gluon -----
            this->remove_gluon(r);
            ++_M_nc_pow;
          }
        }
        
        //----- now we try to find and remove two crossed gluons (not symmetric) -----
        else if((p[l=p[j].color.prev].color.is_gluon() && p[r=p[j].colorP.prev].colorP.is_gluon() && p[r].colorP.prev == l)
                || (p[l=p[j].color.prev].color.is_gluon() && p[r=p[j].colorP.prev].colorP.is_gluon() && p[l].color.prev == r))
        {
          this->remove_gluon(l);
          this->remove_gluon(r);
        }
        
        //----- nothing to do -----
        else break;
      }
      
      pi++;
      goto again;
    }
  }
  
  
  void color_overlap_u3::graph::ruleT()
  {
    /*  In the next step we try to find all the trivial open strings in the {c'}_m 
     *  color configuration. This step can create closed loop, so don't be surprised 
     *  if you see self connected gluons.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  againT1:
    while(pi < _M_finish && (!(pi->active) || !(pi->colorP.is_quark() && p[pi->colorP.next].colorP.is_antiquark()))) pi++;
    
    if(pi < _M_finish) {
      int q = pi-_M_outgoing+1, a = pi->colorP.next;
      int t = pi->color.next, ng = 0, g = 0;
      
      while(!p[t].active || !p[t].color.is_antiquark()) {
        if(p[t].color.is_gluon()) { ++ng; g = t;}
        t = p[t].color.next;
      }
      
      //----- tadpole graph ----
      if(t == a && ng == 1) {
        p[q].active = p[a].active = p[g].active = false;
        _M_active -= 3;
        
        if(p[g].colorP.next != g) {
          p[p[g].colorP.prev].colorP.next = p[g].colorP.next;
          p[p[g].colorP.next].colorP.prev = p[g].colorP.prev;
        } else ++_M_nc_pow;
      } else if(t == a && ng == 0) {
        ++_M_nc_pow;
        p[q].active = p[a].active = false;
        _M_active -= 2;
      }
      else {
        p[p[a].color.prev].color.next = p[q].color.next;
        p[p[q].color.next].color.prev = p[a].color.prev;
        
        p[q].active = p[a].active = false;
        _M_active -= 2;
      }
      
      pi++;
      goto againT1;
    }
    
    /*  In the next step we try to find all the trivial open strings in the {c}_m 
     *  color configuration. 
     */
    pi = _M_start;
    
  againT2:
    while(pi < _M_finish && (!(pi->active) || !(pi->color.is_quark() && p[pi->color.next].color.is_antiquark()))) pi++;
    
    if(pi < _M_finish) {
      int q = pi-_M_outgoing+1, a = pi->color.next;
      int t = pi->colorP.next, ng = 0, g = 0;
      
      while(!p[t].active || !p[t].colorP.is_antiquark()) {
        if(p[t].color.is_gluon()) { ++ng; g = t;}
        t = p[t].colorP.next;
      }
      
      //----- tadpole graph  ----
      if(t == a && ng == 1) {
        p[q].active = p[a].active = p[g].active = false;
        _M_active -= 3;
        
        if(p[g].color.next != g) {
          p[p[g].color.prev].color.next = p[g].color.next;
          p[p[g].color.next].color.prev = p[g].color.prev;
        } else ++_M_nc_pow;
      }
      else if(t == a && ng == 0) {
        ++_M_nc_pow;
        p[q].active = p[a].active = false;
        _M_active -= 2;
      }
      else {
        p[p[a].colorP.prev].colorP.next = p[q].colorP.next;
        p[p[q].colorP.next].colorP.prev = p[a].colorP.prev;
        
        p[q].active = p[a].active = false;
        _M_active -= 2;
      }
      
      pi++;
      goto againT2;
    }
  }
  
  
  void color_overlap_u3::graph::ruleS()
  {
    /*  In the next step we try to find all the tadpole like 
     *  configurations and remove them on the {c'}_m side.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  againS1:
    while(pi < _M_finish && (!(pi->active) || !(pi->colorP.is_gluon() && pi->colorP.next == pi-p))) pi++;
    
    if(pi < _M_finish) {
      int g = pi-p, n = pi->color.next;

      //----- tadpole graph ----
      if(n != g) {
        p[p[g].color.prev].color.next = p[g].color.next;
        p[p[g].color.next].color.prev = p[g].color.prev;
      } else ++_M_nc_pow;
      
      p[g].active = false;
      --_M_active;
      
      ++pi;
      goto againS1;
    }
    
    /* Tadpoles on the {c}_m side.
     */
    pi = _M_start;
    
  againS2:
    while(pi < _M_finish && (!(pi->active) || !(pi->color.is_gluon() && pi->color.next == pi-p))) pi++;
    
    if(pi < _M_finish) {
      int g = pi-p, n = pi->colorP.next;
      
      //----- tadpole graph  ----
      if(n != g) {
        p[p[g].colorP.prev].colorP.next = p[g].colorP.next;
        p[p[g].colorP.next].colorP.prev = p[g].colorP.prev;
      } else ++_M_nc_pow;
      
      p[g].active = false;
      --_M_active;
      
      ++pi;
      goto againS2;
    }
  }
  
  
  void color_overlap_u3::graph::reduce()
  {
    //----- applying the simple rules ----
    unsigned int active;
    
    do {
      active = _M_active;
      this->ruleG(); if(_M_active == 0) return;
      this->ruleS(); if(_M_active == 0) return;
      this->ruleQ(); if(_M_active == 0) return;
      this->ruleA(); if(_M_active == 0) return;
      this->ruleT(); if(_M_active == 0) return;
    } while(active != _M_active);
  }
    
  void color_overlap_u3::graph::ruleF()
  {
    /*  This rule is the desperate rule. It implements the Fierz identity.
     *  We pick a gluon and remove it acording to the fierz identity. In the 
     *  current graph the gluon gets removed and a new q-qbar pair is introduced.
     */
    parton *p, *pi, *pm;

  again:
    p = _M_outgoing-1; 
    pi = pm = _M_start;
    
    //----- find an empty slot -----
    while(pm < _M_finish && pm->active) pm++;
  
    if(pm == _M_finish) {
      //----- if there is reserved space, use it ----- 
      if(_M_finish < _M_end_of_storage) ++_M_finish;

      //----- otherwise allocate some memory ------
      else {
        unsigned int np = _M_finish-_M_start;
        unsigned int ni = _M_outgoing-_M_start;
        parton *start;
        
        //----- allocating memory -----
        if(!(start = new parton[np+10])) 
          throw "color_overlap_u3::grap : Unable to allocate memory.";
        
        //----- copy data -----
        std::copy(_M_start, _M_finish, start);
        
        //----- destroy the old data -----
        if(_M_start) delete [] _M_start;
       
        _M_start = start;
        _M_end_of_storage = (_M_finish = _M_start + np) + 10;
        _M_outgoing = _M_start + ni; 
        
        goto again;
      }
    }
    
    //----- find a gluon -----
    while(pi < _M_finish && (!(pi->active) || !pi->color.is_gluon())) pi++;
    
    //----- applying the fierz identity ----
    if(pi < _M_finish) {
      int l = pi-p, m = pm-p; 

      //----- the first term of the Fierz identity -----
      p[m].color = colorlink::antiquark(p[l].color.prev);
      p[p[l].color.prev].color.next = m;
      p[l].color.prev = colorlink::quarkend;
 
      p[m].colorP = colorlink::antiquark(p[l].colorP.prev);
      p[p[l].colorP.prev].colorP.next = m;
      p[l].colorP.prev = colorlink::quarkend;
      
      p[m].active = p[l].active = true;
      ++_M_active;
    }
    
    //----- just to make sure we do everything correctly -----
    else {
      this->print(std::cerr);
      throw "color_overlap_u3::graph::ruleF(...) : This message shouldn't appear...looks we have toubble.";
    }
  }
  
    
  double color_overlap_u3::operator()(const statistical_basis& p)
  {
    _M_graph = p;
    
    do {
      _M_graph.reduce();
      if(_M_graph._M_active == 0) break;
      _M_graph.ruleF();
    } while(true);
    
    return std::pow(Nc, _M_graph._M_nc_pow);
  }
  
}


