/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   std includes

//   project headers
#include "bits/qcd-color-overlap.h"


namespace duct {

  std::ostream& color_overlap::graph::print(std::ostream& os) const
  {
    int idx = _M_start-_M_outgoing + 1;
    os<<"+------------------------\n";
    os<<"| cf_pow = "<<_M_cf_pow<<"\n";
    os<<"| nc_pow = "<<_M_nc_pow<<"\n";
    os<<"| weight = "<<_M_wgt<<"\n";
    os<<"| active = "<<_M_active<<"\n";
    os<<"+---- qcd partons -----\n";
    if(_M_active) {
      for(auto pi = _M_start; pi < _M_finish; ++pi) 
        if(pi->multi) {
          os<<"| "<<idx++
          <<"  "<<pi->multi
          <<"  "<<pi->color
          <<"  "<<pi->colorP
          <<std::endl;
        } else idx++;
    }
    os<<"+----\n";
    
    
    return os;
  }
 
  
  
  color_overlap::graph::graph(const statistical_basis& p) 
  {
    if(p.size()) {
      int np = p.number_of<qcd_out>();
      int ni = p.number_of<incomings>();
      
      //---- count the qcd incomings -----
      switch(ni) {
        case 2: if(!p[-1].color.is_qcd()) --ni;
        case 1: if(!p[ 0].color.is_qcd()) --ni;
      }
      
      //---- allcate memory ---- 
      this->_M_allocate((unsigned int) ni, (unsigned int) np);
      parton *q = _M_outgoing-1;
      
      //---- copy the color connections ----
      _M_wgt = 1.0; 
      _M_active = ni+np; 
      _M_cf_pow = _M_nc_pow = 0;
      
      for(int i = 1-ni; i <= np; ++i) {
        q[i].color = p[i].color;
        q[i].colorP = p[i].colorP;
        q[i].multi = 1U;
        
        if(q[i].color.is_gluon()) {
          --_M_cf_pow;
          q[i].A = Nc/Na;
          q[i].B = -1.0/Na;
        } else if(q[i].color.is_quark()) {
          --_M_nc_pow;
          q[i].A = q[i].B = 0.0;
        } else q[i].A = q[i].B = 0.0;
      }
    }
    else {
      _M_start = _M_outgoing = _M_finish = _M_end_of_storage = nullptr;
      _M_active = _M_cf_pow = _M_nc_pow = 0; _M_wgt = 1.0;
    }
  }

  color_overlap::graph& color_overlap::graph::operator=(const statistical_basis& p) 
  {
    if(p.size()) {
      int np = p.number_of<qcd_out>();
      int ni = p.number_of<incomings>();
      
      //---- count the qcd incomings -----
      switch(ni) {
        case 2: if(!p[-1].color.is_qcd()) --ni;
        case 1: if(!p[ 0].color.is_qcd()) --ni;
      }
      
      //---- allcate memory ---- 
      this->resize((unsigned int) ni, (unsigned int) np);
      parton *q = _M_outgoing-1;
      
      //---- copy the color connections ----
      _M_wgt = 1.0;
      _M_active = ni+np; 
      _M_cf_pow = _M_nc_pow = 0;
      
      for(int i = 1-ni; i <= np; ++i) {
        q[i].color = p[i].color;
        q[i].colorP = p[i].colorP;
        q[i].multi = 1U;
        
        if(q[i].color.is_gluon()) {
          --_M_cf_pow;
          q[i].A = Nc/Na;
          q[i].B = -1.0/Na;
        } else if(q[i].color.is_quark()) {
          --_M_nc_pow;
          q[i].A = q[i].B = 0.0;
        } else q[i].A = q[i].B = 0.0;
      }
    }
    
    return *this;
  }
  
  void color_overlap::graph::ruleG()
  {
    /*  Here we try to find all the paralell and crossed gluons and combine them into 
     *  a multi-gluon.
     */
    unsigned int active;
    parton *p = _M_outgoing-1, *pi;
    
    do {
      pi = _M_start;
      active = _M_active;
      
    again:
      while(pi < _M_finish && (!pi->multi || !pi->color.is_gluon())) pi++;
      
      if(pi < _M_finish) {
        int n, j = pi-p;
        
      againP:
        //----- combine two paralell multi gluons -----
        if(p[n = p[j].color.next].color.is_gluon() && n == p[j].colorP.next) {
          //----- removes closed loop squres with one multigluon -----
          if(n == j) {
            _M_cf_pow += p[j].multi;
            _M_wgt *= Nc*p[j].A + Nc2*p[j].B;
            p[j].multi = 0U;
            --_M_active;
            
            pi++;
            goto again;
          }
          
          //----- otherwise combine the two gluons ----
          p[p[n].color.next].color.prev = p[p[n].colorP.next].colorP.prev = j;
          p[j].color.next = p[n].color.next;
          p[j].colorP.next = p[n].colorP.next;
          
          p[j].A = Nc*p[j].A*p[n].A + p[j].A*p[n].B + p[j].B*p[n].A;
          p[j].B *= p[n].B;
          
          p[j].multi += p[n].multi;
          p[n].multi = 0U;
          --_M_active;
          
          //----- repeat the paralell gluon check -----
          goto againP;
        }
        
        //----- combine two crossed multi gluons -----
        if(p[n = p[j].color.next].color.is_gluon() && n == p[j].colorP.prev) {
          p[j].color.next = p[n].color.next;
          p[p[n].color.next].color.prev = j;
          p[j].colorP.prev = p[n].colorP.prev;
          p[p[n].colorP.prev].colorP.next = j;
          
          double A = p[j].A*p[n].B + p[j].B*p[n].A;
          double B = p[j].A*p[n].A + p[j].B*p[n].B;
          
          p[j].A = A;
          p[j].B = B;
          
          p[j].multi += p[n].multi;
          p[n].multi = 0U;
          --_M_active;
          
          goto againP;
        }
        
        //----- get another gluon -----
        pi++;
        goto again;
      }
    } while(active != _M_active);
  }
  
  void color_overlap::graph::ruleQ()
  {
    /*  This routine tries to identify every gluon or multi-gluon those are paralell 
     *  with a quark line. If the quark is paralell with the anti-quark then we remove both 
     *  of them and increase the power of the Nc factor by 1.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  again:
    while(pi < _M_finish && (!(pi->multi) || !pi->color.is_quark())) pi++;
    
    if(pi < _M_finish) {
      int n, r, l, j = pi-_M_outgoing+1;
      
      while(true) {
        //------remove gluon that is paralell to the quark -----
        if(p[j].color.next == (n = p[j].colorP.next)) {
          if(p[n].color.is_antiquark()) {
            //----- remove the whole fermion loop -----
            _M_nc_pow += 1;
            p[j].multi = p[n].multi = 0;
            _M_active -= 2;
            break;
          }
          else {
            //----- remove the gluon -----
            p[p[n].color.next].color.prev = p[p[n].colorP.next].colorP.prev = j;
            p[j].color.next = p[n].color.next;
            p[j].colorP.next = p[n].colorP.next;
           
            _M_wgt *= Nc*p[n].A + p[n].B;
            _M_cf_pow += p[n].multi;
            p[n].multi = 0U;
            _M_active -= 1;
          } 
        }
     
        //----- now we try to find and remove two crossed gluons (not symmetric) -----
        else if(p[l=p[j].color.next].color.is_gluon()
                && p[r=p[j].colorP.next].colorP.is_gluon()
                && p[r].colorP.next == l) 
        {
          p[j].color.next = p[l].color.next;
          p[p[l].color.next].color.prev = j;
          p[r].colorP.next = p[l].colorP.next;
          p[p[l].colorP.next].colorP.prev = r;
          
          double A = p[l].B*p[r].A;
          double B = p[l].B*p[r].B + p[l].A*(p[r].A+Nc*p[r].B);
          
          p[r].A = A;
          p[r].B = B;
          
          _M_cf_pow += p[l].multi;
          p[l].multi = 0;
          _M_active -= 1;
        }
        else if(p[l=p[j].color.next].color.is_gluon()
                && p[r=p[j].colorP.next].colorP.is_gluon()
                && p[l].color.next == r) 
        {
          p[j].colorP.next = p[r].colorP.next;
          p[p[r].colorP.next].colorP.prev = j;
          p[l].color.next = p[r].color.next;
          p[p[r].color.next].color.prev = l;

          double A = p[r].B*p[l].A;
          double B = p[r].B*p[l].B + p[r].A*(p[l].A+Nc*p[l].B);
          
          p[l].A = A;
          p[l].B = B;
         
          _M_cf_pow += p[r].multi;
          p[r].multi = 0;
          _M_active -= 1;
        }
        
        //----- nothing to do -----
        else break;
      }
      
      pi++;
      goto again;
    }
  }
  
  void color_overlap::graph::ruleA()
  {
    /*  This routine tries to identify every gluon or multi-gluon those are paralell 
     *  with a anti-quark line or two or thee gluons are crossed next to a anti-quark 
     *  line and remove them. If the quark is paralell with the anti-quark then we 
     *  remove both of them and increase the power of the Nc factor by 1.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  again:
    while(pi < _M_finish && (!(pi->multi) || !pi->color.is_antiquark())) pi++;
    
    if(pi < _M_finish) {
      int r, l, j = pi-_M_outgoing+1;
      
      while(true) {
        //------remove gluon that is paralell to the anit-quark -----
        if(p[j].color.prev == (r = p[j].colorP.prev)) {
          if(p[r].color.is_quark()) {
            //----- remove the whole fermion loop -----
            _M_nc_pow += 1;
            p[j].multi = p[r].multi = 0;
            _M_active -= 2;
            break;
          }
          else {
            //----- remove the gluon -----
            p[p[r].color.prev].color.next = p[p[r].colorP.prev].colorP.next = j;
            p[j].color.prev = p[r].color.prev;
            p[j].colorP.prev = p[r].colorP.prev;
            
            _M_wgt *= Nc*p[r].A + p[r].B;
            _M_cf_pow += p[r].multi;
            p[r].multi = 0U;
            _M_active -= 1;
          } 
        }
        
        //----- now we try to find and remove two crossed gluons (not symmetric) -----
        else if(p[l=p[j].color.prev].color.is_gluon()
                && p[r=p[j].colorP.prev].colorP.is_gluon()
                && p[r].colorP.prev == l) 
        {
          p[j].color.prev = p[l].color.prev;
          p[p[l].color.prev].color.next = j;

          p[p[l].colorP.prev].colorP.next = r;
          p[r].colorP.prev = p[l].colorP.prev;
 
          double A = p[l].B*p[r].A;
          double B = p[l].B*p[r].B + p[l].A*(p[r].A+Nc*p[r].B);
          
          p[r].A = A;
          p[r].B = B;
          
          _M_cf_pow += p[l].multi;
          p[l].multi = 0;
          _M_active -= 1;
       }   
        else if(p[l=p[j].color.prev].color.is_gluon()
                && p[r=p[j].colorP.prev].colorP.is_gluon()
                && p[l].color.prev == r) 
        {
          p[l].color.prev = p[r].color.prev;
          p[p[r].color.prev].color.next = l;
          p[j].colorP.prev = p[r].colorP.prev;
          p[p[r].colorP.prev].colorP.next = j;
          
          double A = p[r].B*p[l].A;
          double B = p[r].B*p[l].B + p[r].A*(p[l].A+Nc*p[l].B);
          
          p[l].A = A;
          p[l].B = B;

          _M_cf_pow += p[r].multi;
          p[r].multi = 0;
          _M_active -= 1;
       }   

        //----- nothing to do -----
        else break;
      }
      
      pi++;
      goto again;
    }
  }
    
  void color_overlap::graph::ruleT()
  {
    /*  In the next step we try to find all the trivial open strings in the {c'}_m 
     *  color configuration. This step can create closed loop, so don't be surprised 
     *  if you see self connected gluons.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  againT1:
    while(pi < _M_finish && (!(pi->multi) || !(pi->colorP.is_quark() && p[pi->colorP.next].colorP.is_antiquark()))) pi++;
    
    if(pi < _M_finish) {
      int q = pi-_M_outgoing+1, a = pi->colorP.next;
      int t = pi->color.next, ng = 0, g = 0;
      
      while(!p[t].multi || !p[t].color.is_antiquark()) {
        if(p[t].color.is_gluon()) { ++ng; g = t;}
        t = p[t].color.next;
      }
      
      //----- tadpole graph ==> it is zero ----
      if(t == a && ng == 1 && p[g].multi == 1) {
        _M_wgt = 0.0;
        _M_active = 0;
       return;
      } 
      //----- tadpole with multigluon -----
      else if(t == a && ng == 1) {
        p[q].multi = p[a].multi = 0;

        if(p[g].colorP.next != g) {
          p[p[g].colorP.prev].colorP.next = p[g].colorP.next;
          p[p[g].colorP.next].colorP.prev = p[g].colorP.prev;
        } else _M_nc_pow += 1;
        
        _M_wgt *= p[g].A + Nc*p[g].B;
        _M_cf_pow += p[g].multi;
        p[g].multi = 0;
        _M_active -= 3;
      } else if(t == a && ng == 0) {
        _M_nc_pow += 1;
        p[q].multi = p[a].multi = 0;      
        _M_active -= 2;
      }
      else {
        p[p[a].color.prev].color.next = p[q].color.next;
        p[p[q].color.next].color.prev = p[a].color.prev;
        
        p[q].multi = p[a].multi = 0;
        _M_active -= 2;
      }
            
      pi++;
      goto againT1;
    }
    
    /*  In the next step we try to find all the trivial open strings in the {c}_m 
     *  color configuration. 
     */
    pi = _M_start;
    
  againT2:
    while(pi < _M_finish && (!(pi->multi) || !(pi->color.is_quark() && p[pi->color.next].color.is_antiquark()))) pi++;
    
    if(pi < _M_finish) {
      int q = pi-_M_outgoing+1, a = pi->color.next;
      int t = pi->colorP.next, ng = 0, g = 0;
      
      while(!p[t].multi || !p[t].colorP.is_antiquark()) { 
        if(p[t].color.is_gluon()) { ++ng; g = t;}
        t = p[t].colorP.next;
      }
      
      //----- tadpole graph ==> it is zero ----
      if(t == a && ng == 1 && p[g].multi == 1) {
        _M_wgt = 0.0;
        _M_active = 0;
        return;
      } 
      //----- tadpole with multigluon -----
      else if(t == a && ng == 1) {
        p[q].multi = p[a].multi = 0;

        if(p[g].color.next != g) {
          p[p[g].color.prev].color.next = p[g].color.next;
          p[p[g].color.next].color.prev = p[g].color.prev;
        } else _M_nc_pow += 1;
        
        _M_wgt *= p[g].A + Nc*p[g].B;
        _M_cf_pow += p[g].multi;
        p[g].multi = 0;
        _M_active -= 3;
      }
      else if(t == a && ng == 0) {
        _M_nc_pow += 1;
        p[q].multi = p[a].multi = 0;      
        _M_active -= 2;
      }
      else {
        p[p[a].colorP.prev].colorP.next = p[q].colorP.next;
        p[p[q].colorP.next].colorP.prev = p[a].colorP.prev;
        
        p[q].multi = p[a].multi = 0;
        _M_active -= 2;
      }
      
      pi++;
      goto againT2;
    }
  }
    
  void color_overlap::graph::ruleS()
  {
    /*  In the next step we try to find all the tadpole like 
     *  configurations and remove them on the {c'}_m side.
     */
    parton *p = _M_outgoing-1, *pi = _M_start;
    
  againS1:
    while(pi < _M_finish && (!(pi->multi) || !(pi->colorP.is_gluon() && pi->colorP.next == pi-p))) pi++;
    
    if(pi < _M_finish) {
      int g = pi-p, n = pi->color.next;

      //----- tadpole graph ==> it is zero ----
      if(p[g].multi == 1) {
        _M_wgt = 0.0;
        _M_active = 0;
        return;
      } 
      //----- tadpole with multigluon -----
      else {
        if(n != g) {
          p[p[g].color.prev].color.next = p[g].color.next;
          p[p[g].color.next].color.prev = p[g].color.prev;
        } else _M_nc_pow += 1;
        
        _M_wgt *= p[g].A + Nc*p[g].B;
        _M_cf_pow += p[g].multi;
        p[g].multi = 0;
        _M_active -= 1;
      }
    
      ++pi;
      goto againS1;
    }
    
    /* Tadpoles on the {c}_m side.
     */
    pi = _M_start;
    
  againS2:
    while(pi < _M_finish && (!(pi->multi) || !(pi->color.is_gluon() && pi->color.next == pi-p))) pi++;
    
    if(pi < _M_finish) {
      int g = pi-p, n = pi->colorP.next;
      
      //----- tadpole graph ==> it is zero ----
      if(p[g].multi == 1) {
        _M_wgt = 0.0;
        _M_active = 0;
        return;
      } 
      //----- tadpole with multigluon -----
      else {
        if(n != g) {
          p[p[g].colorP.prev].colorP.next = p[g].colorP.next;
          p[p[g].colorP.next].colorP.prev = p[g].colorP.prev;
        } else _M_nc_pow += 1;
        
        _M_wgt *= p[g].A + Nc*p[g].B;
        _M_cf_pow += p[g].multi;
        p[g].multi = 0;
        _M_active -= 1;
      }
      
      ++pi;
      goto againS2;
    }
  } 
        
  void color_overlap::graph::reduce()
  {
    //----- applying the simple rules ----
    unsigned int active;
    
    do {
      active = _M_active;
      this->ruleG(); if(_M_active == 0) return;
      this->ruleS(); if(_M_active == 0) return;
      this->ruleQ(); if(_M_active == 0) return;
      this->ruleA(); if(_M_active == 0) return;
      this->ruleT(); if(_M_active == 0) return;
    } while(active != _M_active);
  }
    
  void color_overlap::graph::ruleF(color_overlap::graph& g)
  {
    /*  This rule is the desperate rule. It implements the Fierz identity.
     *  We pick a gluon and remove it acording to the fierz identity. In the 
     *  current graph the gluon gets removed and a new q-qbar pair is introduced.
     *  the subleading part of the Fierz identity is stored in the garaph g and 
     *  here the gluon is removed.
     */
    parton *p, *pi, *pm;

  again:
    p = _M_outgoing-1; 
    pi = pm = _M_start;
    
    //----- find an empty slot -----
    while(pm < _M_finish && pm->multi) pm++;
  
    if(pm == _M_finish) {
      //----- if there is reserved space, use it ----- 
      if(_M_finish < _M_end_of_storage) ++_M_finish;

      //----- otherwise allocate some memory ------
      else {
        unsigned int np = _M_finish-_M_start;
        unsigned int ni = _M_outgoing-_M_start;
        parton *start;
        
        //----- allocating memory -----
        if(!(start = new parton[np+10])) 
          throw "color_overlap::grap : Unable to allocate memory.";
        
        //----- copy data -----
        std::copy(_M_start, _M_finish, start);
        
        //----- destroy the old data -----
        if(_M_start) delete [] _M_start;
       
        _M_start = start;
        _M_end_of_storage = (_M_finish = _M_start + np) + 10;
        _M_outgoing = _M_start + ni; 
        
        goto again;
      }
    }
    
    //----- find a gluon -----
    while(pi < _M_finish && (!(pi->multi) || !pi->color.is_gluon())) pi++;
    
    //----- applying the fierz identity ----
    if(pi < _M_finish) {
      int l = pi-p, m = pm-p; 
           
      //----- the second term in the Fierz identity -----
      parton *q = g._M_outgoing-1;
     
      if(q[l].color.next == l) g._M_nc_pow += 1;
      if(q[l].colorP.next == l) g._M_nc_pow += 1;
            
      q[q[l].color.prev].color.next = q[l].color.next;
      q[q[l].color.next].color.prev = q[l].color.prev;
   
      q[q[l].colorP.prev].colorP.next = q[l].colorP.next;
      q[q[l].colorP.next].colorP.prev = q[l].colorP.prev;
      
      g._M_wgt *= q[l].B;
      g._M_cf_pow += q[l].multi;
      g._M_active -= 1;
      q[l].multi = 0U;

      //----- the first term of the Fierz identity -----
      p[m].color = colorlink::antiquark(p[l].color.prev);
      p[p[l].color.prev].color.next = m;
      p[l].color.prev = colorlink::quarkend;
 
      p[m].colorP = colorlink::antiquark(p[l].colorP.prev);
      p[p[l].colorP.prev].colorP.next = m;
      p[l].colorP.prev = colorlink::quarkend;
      
      _M_wgt *= p[l].A;
      _M_cf_pow += p[l].multi;
      _M_active += 1;
      p[m].multi = p[l].multi = 1U;
    }
    
    //----- just to make sure we do everything correctly -----
    else {
      this->print(std::cerr);
      throw "color_overlap::graph::ruleF(...) : This message shouldn't appear...looks we have toubble.";
    }
  }
  
    
  double color_overlap::operator()(const statistical_basis& p)
  {
    //----- if the list is empty then we start a new one -----
    if(_M_graphs.empty()) {
      _M_graphs.push_back(graph(p));
      _M_begin = _M_graphs.begin();
      _M_end = _M_graphs.end();
    } 
    //----- use the previous list -----
    else {
      _M_end = _M_begin = _M_graphs.begin();
      *(_M_end++) = p;
    }
    
    //---- iterate the rules to reduce the graph to a number ------
    auto current = _M_begin;
   
    do {
      current->reduce();
           
      if(current->is_finished()) ++current;
      else {
        //----- making space for the new graph ------
        decltype(current) next;
               
        if(_M_end == _M_graphs.end()) {
          _M_graphs.push_back(*current);
          next = _M_end = _M_graphs.end();
          --next;
        } else *(next = _M_end++) = *current;
        
        //----- applying the Fierz identity -----
        current->ruleF(*next);
      }
    } while(current != _M_end);
    
    //----- calculate weight -----
    auto gi = _M_begin;
    int cfpow = gi->_M_cf_pow;
    int ncpow = gi->_M_nc_pow;
    double weight = gi->_M_wgt;
    
    while(++gi != _M_end) {
      gi->_M_cf_pow -= cfpow;
      gi->_M_nc_pow -= ncpow;
      gi->_M_wgt *= std::pow(Cf, gi->_M_cf_pow);
      weight += (gi->_M_wgt *= std::pow(Nc, gi->_M_nc_pow));
    }
    
    return weight*std::pow(Cf, cfpow)*std::pow(Nc, ncpow);
  }
  
}


