/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/11/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */
#include <cmath>


/*   project headers   */
#include "bits/mc-default-pdfs.h"
#include "bits/qcd-color.h"


namespace duct {
  
  /*   base functions of the mstw parametrization   */
  static double __mstw2008lo_ffunc(double *p, double x) {
    return p[0]*std::pow(x, p[1])*std::pow(1.0-x, p[2])*(1.0+p[3]*std::sqrt(x)+p[4]*x);
  }
  
  static double __mstw2008lo_gfunc(double *p, double x) {
    return p[0]*std::pow(x, p[1])*std::pow(1.0-x, p[2])*(1.0+p[3]*x+p[4]*x*x);
  }
  
  
  /*   this is the proton pdf funtion at the fitting scale (MSTW 2008 LO)   */
  void __mstw2008lo(double x, double *f)
  {
    static double p[7][5] = {
      {1.4335, 0.45232, 3.0409, -2.3737, 8.9924},
      {5.0903, 0.71978, 5.1244, -4.3645, 7.4730},
      {0.59964, -0.16276, 8.8801, -2.9012, 16.865},
      {8.9413, 1.876, 10.8801, 8.4703, -36.507},
      {0.0012216, -0.83657, 2.3882, -38.997, 1445.5},
      {0.10302, -0.16276, 13.242, -2.9012, 16.865},
      {-0.011523, 0.2, 10.285, -57.42506, 0.0}
    };
    
    double uv = __mstw2008lo_ffunc(p[0], x);
    double dv = __mstw2008lo_ffunc(p[1], x);
    double S  = __mstw2008lo_ffunc(p[2], x);
    double D  = __mstw2008lo_gfunc(p[3], x);
    double g  = __mstw2008lo_ffunc(p[4], x);
    double sp = __mstw2008lo_ffunc(p[5], x);
    double sv = __mstw2008lo_gfunc(p[6], x);
    
    f[-3] = 0.5*std::abs(sp-sv);
    f[-2] = 0.5*std::abs((0.5*(S-sp)-D));
    f[-1] = 0.5*std::abs((0.5*(S-sp)+D));
    f[0] = std::abs(g);
    f[1] = std::abs(dv+f[-1]);
    f[2] = std::abs(uv+f[-2]);
    f[3] = 0.5*std::abs(sp+sv);
    f[4]=f[-4]=f[5]=f[-5]=0.0;
  }

  /*  The MSTW2008 LO was fitted with LO coupling.
   *  alpha_s(Mz) = 0.130087
   */
  double __mstw2008lo_as(double q2)
  {
    constexpr double thr[7] = {0.0, 0.0, 0.0, 0.0, 1.69, 20.25, 30625};
    constexpr double lmd[7] = {0.063036, 0.063036, 0.063036, 0.063036, 0.0478824, 0.0280364, 0.00745832};
 
    unsigned nf = 6;
    while(q2 < thr[nf] && nf > 3) --nf;
    
    //  computing the LO alpha_s
    double b0 = beta0(nf);
    double t = 0.5*b0*log(q2/lmd[nf]);
    
    return 1.0/t;
  }
  
  
  
  
  static double __herapdf_func(const double *p, double x)
  {
    return p[0]*std::pow(x, p[1])*std::pow(1.0-x, p[2])*(1.0+p[3]*x+p[4]*x*x);
  }
  
  /*   this is the proton pdf funtion at the fitting scale (HERAPDF LO) 
   *   Q0 = 1GeV, alpha_s(Mz) = 0.118306 ==> as*(1+as/(2pi) K) = 0.126
   */
  void __herapdflo(double x, double *f)
  {
    constexpr double p[7][5] = {
      {2.9276,    0.6129,    4.4679,    0.0000,   12.9392},
      {1.6881,    0.6198,    3.4010,    0.0000,    0.0000},
      {0.0891,   -0.1853,    2.4870,    0.0000,    0.0000},
      {0.0891,   -0.1853,    4.8372,    0.0000,    0.0000},
      {2.5543,    0.0138,    4.4811,   -0.4657,    0.0000},
      {0.1309,   -0.1853,   16.0672,    0.0000,    0.0000}
    };
    
    f[-3] = __herapdf_func(p[5], x);
    f[-2] = __herapdf_func(p[2], x);
    f[-1] = __herapdf_func(p[3], x);
    f[ 0] = __herapdf_func(p[4], x);
    f[ 1] = __herapdf_func(p[1], x) + f[-1];
    f[ 2] = __herapdf_func(p[0], x) + f[-2];
    f[ 3] = f[-3];
    f[4]=f[-4]=f[5]=f[-5]=0.0;
  }
  
 
  /*  HERAPDF LO was fitted with LO coupling.
   *  alpha_s(Mz) = 0.118306 ==> as*(1+as/(2pi) K) = 0.126
   */
  double __herapdflo_as(double q2)
  {
    constexpr double thr[7] = {0.0, 0.0, 0.0, 0.0, 1.69, 20.25, 30625};
    constexpr double lmd[7]  = {0.0216472, 0.0216472, 0.0216472, 0.0216472, 0.0150957, 0.00799479, 0.00188725};
    
    unsigned nf = 6;
    while(q2 < thr[nf] && nf > 3) --nf;
    
    //  computing the LO alpha_s
    double b0 = beta0(nf);
    double t = 0.5*b0*log(q2/lmd[nf]);
    
    return 1.0/t;
  }
  
  
  /*   this is the proton pdf funtion at the fitting scale (HERAPDF LO)
   *   Q0 = 1GeV, alpha_s(Mz) = 0.121839 ==> as*(1+as/(2pi) K) = 0.13
   */
  void __herapdflo_hi(double x, double *f)
  {
    constexpr double p[7][5] = {
      {2.9784,    0.6278,    4.3263,    0.0000,   12.7383},
      {1.8555,    0.6584,    3.2857,    0.0000,    0.0000},
      {0.0872,   -0.1920,    2.6728,    0.0000,    0.0000},
      {0.0872,   -0.1920,    4.8723,    0.0000,    0.0000},
      {3.0061,    0.1195,    4.3692,   -0.4241,    0.0000},
      {0.1435,   -0.1920,   17.3391,    0.0000,    0.0000}
    };
    
    f[-3] = __herapdf_func(p[5], x);
    f[-2] = __herapdf_func(p[2], x);
    f[-1] = __herapdf_func(p[3], x);
    f[ 0] = __herapdf_func(p[4], x);
    f[ 1] = __herapdf_func(p[1], x) + f[-1];
    f[ 2] = __herapdf_func(p[0], x) + f[-2];
    f[ 3] = f[-3];
    f[4]=f[-4]=f[5]=f[-5]=0.0;
  }
  
  /*  HERAPDF LO was fitted with LO coupling.
   *  alpha_s(Mz) = 0.121839 ==> as*(1+as/(2pi) K) = 0.13
   */
  double __herapdflo_hi_as(double q2)
  {
    constexpr double thr[7] = {0.0, 0.0, 0.0, 0.0, 1.69, 20.25, 30625};
    constexpr double lmd[7]  = {0.0304811, 0.0304811, 0.0304811, 0.0304811, 0.021846, 0.0119477, 0.00293037};
    
    unsigned nf = 6;
    while(q2 < thr[nf] && nf > 3) --nf;
    
    //  computing the LO alpha_s
    double b0 = beta0(nf);
    double t = 0.5*b0*log(q2/lmd[nf]);
    
    return 1.0/t;
  }
  

  
}
