/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/29/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>
#include <algorithm>

//   project headers
#include "bits/hep-dist.h"
#include "bits/nlo-cxx-support.h"


namespace duct {
  namespace distpoint {
    
    
    std::vector<float>
    spacing<float>::linear(unsigned int n, float xmin, float xmax)
    {
      std::vector<float> p(n+1, xmin);
      float dx = (xmax-xmin)/n;
      
      for(unsigned i = 1; i <= n; i++)
        p[i] += i*dx;
      
      return  p;
    }
    
    std::vector<float>
    spacing<float>::logarithmic(unsigned int n, float xmin, float xmax)
    {
      std::vector<float> p(n+1, xmin);
      
      for(unsigned i = 1; i <= n; i++)
        p[i] *= std::pow(xmax/xmin, static_cast<float>(i)/n);
      
      return  p;
    }
    
    
    
    std::vector<double>
    spacing<double>::linear(unsigned int n, double xmin, double xmax)
    {
      std::vector<double> p(n+1, xmin);
      double dx = (xmax-xmin)/n;
      
      for(unsigned i = 1; i <= n; i++)
        p[i] += i*dx;
      
      return  p;
    }
    
    std::vector<double>
    spacing<double>::logarithmic(unsigned int n, double xmin, double xmax)
    {
      std::vector<double> p(n+1, xmin);
      
      for(unsigned i = 1; i <= n; i++)
        p[i] *= std::pow(xmax/xmin, static_cast<double>(i)/n);
      
      return  p;
    }
    
    
    std::vector<long double>
    spacing<long double>::linear(unsigned int n, long double xmin, long double xmax)
    {
      std::vector<long double> p(n+1, xmin);
      double dx = (xmax-xmin)/n;
      
      for(unsigned i = 1; i <= n; i++)
        p[i] += i*dx;
      
      return  p;
    }
    
    std::vector<long double>
    spacing<long double>::logarithmic(unsigned int n, long double xmin, long double xmax)
    {
      std::vector<long double> p(n+1, xmin);
      
      for(unsigned i = 1; i <= n; i++)
        p[i] *= std::pow(xmax/xmin, static_cast<long double>(i)/n);
      
      return  p;
    }
    
    
    
    std::vector<hist1d>
    spacing<hist1d>::linear(unsigned int n, double xmin, double xmax)
    {
      double dx = (xmax-xmin)/n;
      std::vector<hist1d> p(n, {xmin, xmin+dx});
      
      for(unsigned i = 1; i < n; i++) {
        p[i].xmin += i*dx;
        p[i].xmax += i*dx;
      }
      
      return  p;
    }
    
    std::vector<hist1d>
    spacing<hist1d>::logarithmic(unsigned int n, double xmin, double xmax)
    {
      double dx = std::pow(xmax/xmin, 0.5/n);
      std::vector<hist1d> p(n, {xmin, xmin*dx*dx});
      
      for(unsigned i = 1; i < n; i++) {
        p[i].xmin *= std::pow(dx, 2*i);
        p[i].xmax *= std::pow(dx, 2*i);
      }
      
      return  p;
    }
    
    std::vector<hist1d>
    spacing<hist1d>::manual(std::initializer_list<double> __m)
    {
      std::vector<double> m(__m);
      std::vector<hist1d> p;
      
      std::sort(m.begin(), m.end());
      
      for(auto i = m.begin()+1; i < m.end(); i++)
        p.push_back({*(i-1), *i});
      
      return  p;
    }
    
   
    
    std::vector<hist2d> spacing<hist2d>::
    linear(unsigned int nx, double xmin, double xmax, unsigned int ny, double ymin, double ymax)
    {
      std::vector<hist1d> hx = spacing<hist1d>::linear(nx, xmin, xmax);
      std::vector<hist1d> hy = spacing<hist1d>::linear(ny, ymin, ymax);
      std::vector<hist2d> p(nx*ny);
      unsigned idx;
      
      for(unsigned i = 0; i < nx; i++)
        for(unsigned j = 0; j < ny; j++) {
          idx = ny*i+j;
          p[idx].xmin = hx[i].xmin;
          p[idx].xmax = hx[i].xmax;
          
          p[idx].ymin = hy[j].xmin;
          p[idx].ymax = hy[j].xmax;
        }
      
      return  p;
    }
    
    std::vector<hist2d> spacing<hist2d>::
    logarithmic(unsigned int nx, double xmin, double xmax, unsigned int ny, double ymin, double ymax)
    {
      std::vector<hist1d> hx = spacing<hist1d>::logarithmic(nx, xmin, xmax);
      std::vector<hist1d> hy = spacing<hist1d>::logarithmic(ny, ymin, ymax);
      std::vector<hist2d> p(nx*ny);
      unsigned idx;
      
      for(unsigned i = 0; i < nx; i++)
        for(unsigned j = 0; j < ny; j++) {
          idx = ny*i+j;
          p[idx].xmin = hx[i].xmin;
          p[idx].xmax = hx[i].xmax;
          
          p[idx].ymin = hy[j].xmin;
          p[idx].ymax = hy[j].xmax;
        }
      
      return  p;
    }
    
    std::vector<hist2d> spacing<hist2d>::
    loglinear(unsigned int nx, double xmin, double xmax, unsigned int ny, double ymin, double ymax)
    {
      std::vector<hist1d> hx = spacing<hist1d>::logarithmic(nx, xmin, xmax);
      std::vector<hist1d> hy = spacing<hist1d>::linear(ny, ymin, ymax);
      std::vector<hist2d> p(nx*ny);
      unsigned idx;
      
      for(unsigned i = 0; i < nx; i++)
        for(unsigned j = 0; j < ny; j++) {
          idx = ny*i+j;
          p[idx].xmin = hx[i].xmin;
          p[idx].xmax = hx[i].xmax;
          
          p[idx].ymin = hy[j].xmin;
          p[idx].ymax = hy[j].xmax;
        }
      
      return  p;
    }
    
    std::vector<hist2d> spacing<hist2d>::
    linearlog(unsigned int nx, double xmin, double xmax, unsigned int ny, double ymin, double ymax)
    {
      std::vector<hist1d> hx = spacing<hist1d>::linear(nx, xmin, xmax);
      std::vector<hist1d> hy = spacing<hist1d>::logarithmic(ny, ymin, ymax);
      std::vector<hist2d> p(nx*ny);
      unsigned idx;
      
      for(unsigned i = 0; i < nx; i++)
        for(unsigned j = 0; j < ny; j++) {
          idx = ny*i+j;
          p[idx].xmin = hx[i].xmin;
          p[idx].xmax = hx[i].xmax;
          
          p[idx].ymin = hy[j].xmin;
          p[idx].ymax = hy[j].xmax;
        }
      
      return  p;
    }
  
  }
  
  
  
  std::ostream& write_bin(std::ostream& os, const ltxfigure& ltx)
  {
    write_bin(os, ltx.axis);
    
    unsigned s = ltx.addplot.size();
    os.write(reinterpret_cast<const char *>(&s), sizeof(unsigned));
    for(const std::string& pl : ltx.addplot) write_bin(os, pl);
    
    return os;
  }
  
  std::istream& read_bin(std::istream& is, ltxfigure& ltx)
  {
    read_bin(is, ltx.axis);
    ltx.addplot.clear();
    
    unsigned s;
    std::string tmp;
    
    is.read(reinterpret_cast<char *>(&s), sizeof(unsigned));
    
    for(unsigned k = 0; k < s; ++k) {
      read_bin(is, tmp);
      ltx.addplot.push_back(tmp);
    }
    
    return is;
  }
  
  
  std::ostream& write_txt(std::ostream& os, const ltxfigure& ltx)
  {
    write_txt(os, ltx.axis);
    
    os<<"  "<<ltx.addplot.size();
    for(const std::string& pl : ltx.addplot) {
      os<<" "; write_txt(os, pl);
    }
    
    return os;
  }
  
  std::istream& read_txt(std::istream& is, ltxfigure& ltx)
  {
    read_txt(is, ltx.axis);
    ltx.addplot.clear();
    
    unsigned s;
    std::string tmp;
    
    is>>s;
    
    for(unsigned k = 0; k < s; ++k) {
      read_txt(is, tmp);
      ltx.addplot.push_back(tmp);
    }
    
    return is;
  }

  
  std::ostream& operator<<(std::ostream& os, const ltxfigure& ltx)
  {
    os<<"\\begin{tikzpicture}\n"
    <<  "\\begin{axis}[width=12cm, height=10cm, grid=both,\n"<<ltx.axis<<"]\n\n";
    
    /*   adding plots    */
    for(auto const& pl : ltx.addplot)
      os<<pl<<"\n";
    
    /*   end of everything   */
    os<<"\n\\end{axis}\n"
    <<  "\\end{tikzpicture}\n";
    
    return os;
  }

  
  
  std::ostream& write_bin(std::ostream& os, const ltxplot& ltx)
  {
    write_bin(os, ltx.head.title);
    
    unsigned s = ltx.figures.size();
    os.write(reinterpret_cast<const char *>(&s), sizeof(unsigned));
    for(const ltxfigure& pl : ltx.figures) write_bin(os, pl);
    
    return os;
  }
  
  std::istream& read_bin(std::istream& is, ltxplot& ltx)
  {
    read_bin(is, ltx.head.title);

    unsigned s;
    ltxfigure tmp;
    
    is.read(reinterpret_cast<char *>(&s), sizeof(unsigned));
    
    for(unsigned k = 0; k < s; ++k) {
      read_bin(is, tmp);
      ltx.figures.push_back(tmp);
    }
    
    return is;
  }
  
  
  std::ostream& write_txt(std::ostream& os, const ltxplot& ltx)
  {
    write_txt(os, ltx.head.title);
    
    os<<"  "<<ltx.figures.size();
    for(const ltxfigure& pl : ltx.figures) {
      os<<" "; write_txt(os, pl);
    }
    
    return os;
  }
  
  std::istream& read_txt(std::istream& is, ltxplot& ltx)
  {
    read_txt(is, ltx.head.title);
    
    unsigned s;
    ltxfigure tmp;
    
    is>>s;
    
    for(unsigned k = 0; k < s; ++k) {
      read_txt(is, tmp);
      ltx.figures.push_back(tmp);
    }
    
    return is;
  }
  
  
  std::ostream& operator<<(std::ostream& os, const ltxplot::ltxhead& h)
  {
    os<<"\\documentclass[11pt]{article}\n"
    <<  "\\usepackage{pgfplots}\n"
    <<  "\\usepackage{pgfplotstable}\n"
    <<  "\\pgfplotsset{compat=newest}\n"
    <<  "\\pagestyle{empty}\n\n"
    
    <<  "\\def\\title{"<<h.title<<"}\n\n";
    
    return os;
  }
  
}
