/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes

//   project headers
#include "bits/mc-engine-kT.h"



namespace duct {
  

  double splitting_kernel<fin, ordering::kT>::y_mapping::
  refresh(const splitting_kernel<fin, ordering::kT> *engine)
  {
    /*   aliases  */
    const double &ymax = engine->_M_pars.ymax, &pQ = engine->_M_pars.pQ, &bl = engine->_M_pars.bl;
    const model *mdl = engine->_M_mdl;
    
    /*   Setting up the interpolation spline */
    this->spline.assign([=](double tau) -> double {
      double mT = mdl->alpha_qcd.tauk_inverse(tau)/pQ;
      double uh = 0.5+std::sqrt(0.25-mT*(ymax+bl)/(ymax*ymax));
      return std::log(uh/(1.0-uh));
    }, mdl->alpha_qcd.tauk(0.25*ymax*ymax/(ymax+bl)*pQ), mdl->alpha_qcd.tauk(engine->_M_pars.mT*pQ));
    
    return this->spline.integral();
  }
  
  
  double splitting_kernel<fin, ordering::kT>::_M_funcFI(double kappa, double chi) const
  {
    /*   aliases  */
    const double &ymax = _M_pars.ymax, &pQ = _M_pars.pQ, &bl = _M_pars.bl;
    
    /*   Transforming the variable kappa to y.  */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double mT = _M_mdl->alpha_qcd.tauk_inverse(tau)/pQ;
    
    /*   Transforming the variable chi to z.  */
    double u = 0.5, ujac = 1.0, retval = 0.0;
    z_mapping zmap;
    
    zmap.refresh(&_M_pars, mT);
    zmap.mapping_chi2u(chi, u, ujac);
    
    /*   Calculate the splitting function of the gluon emission process.   */
    int spec = (_M_pars.spec == 'F' ? 1 : 0);
    double y = 0.5*(mT+std::sqrt(mT*(mT+4.0*u*(1.0-u)*bl)))/(u*(1.0-u));
    
    _M_prob_proc[0] = retval = (y+bl)/(y+2.0*bl)*_M_func_emit(_M_mdl->spin(_M_pars.fl), spec, _M_pars.rl, ymax, _M_pars.xi, _M_pars.mu, y, u);
    
    /*   For gluon emitter we have to consider the g-->f+f~ processes  */
    std::memset(_M_prob_proc+1, 0, 6*sizeof(double));

    if(_M_pars.fl == 0) {
      for(unsigned int i = 1; i <= 6; i++) {
        double bi = _M_mdl->mass2(i)/pQ;
        
        if((ymax-4.0*bi)*u*(1.0-u) >= mT) {
          y = mT/(u*(1.0-u)) + 4.0*bi;
          retval += (_M_prob_proc[i] = y/(y-4.0*bi)*__mc_splitting_functions::gluon_splitting_fin(ymax, bi, y, u));
        }
      }
    }
    
    /*   adjusting the probabilities   */
    for(unsigned int i = 0; i <= 6; i++)
      _M_prob_proc[i] /= retval;
    
    /*   Jacobians:
     *     - jacobian of the u mapping ==> ujac
     *     - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     */
    retval *= ujac;
    retval *= _M_y_mapping.jacobian(tau);
    
    return retval;
  };
  
  
  double splitting_kernel<fin, ordering::kT>::_M_funcD(double kappa, double chi) const
  {
    /*   aliases  */
    const double &ymax = _M_pars.ymax, &pQ = _M_pars.pQ, &bl = _M_pars.bl;
    
    /*   Transforming the variable kappa to y.  */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double mT = _M_mdl->alpha_qcd.tauk_inverse(tau)/pQ;
    
    /*   Transforming the variable chi to z.  */
    double u = 0.5, ujac = 1.0, retval = 0.0;
    z_mapping zmap;
    
    zmap.refresh(&_M_pars, mT);
    zmap.mapping_chi2u(chi, u, ujac);
    
    /*   Calculate the splitting function of the gluon emission process.   */
    double y = 0.5*(mT+std::sqrt(mT*(mT+4.0*u*(1.0-u)*bl)))/(u*(1.0-u));
    _M_prob_proc[0] = retval = (y+bl)/(y+2.0*bl)*__mc_splitting_functions::emitting_gluon_dcy(_M_mdl->spin(_M_pars.fl), _M_pars.rl, ymax, y, u);
    
    /*   For gluon emitter we have to consider the g-->f+f~ processes  */
    std::memset(_M_prob_proc+1, 0, 6*sizeof(double));
    
    if(_M_pars.fl == 0) {
      for(unsigned int i = 1; i <= 6; i++) {
        double bi = _M_mdl->mass2(i)/pQ;
        
        if((ymax-4.0*bi)*u*(1.0-u) >= mT) {
          y = mT/(u*(1.0-u)) + 4.0*bi;
          retval += (_M_prob_proc[i] = y/(y-4.0*bi)*__mc_splitting_functions::gluon_splitting_fin(ymax, bi, y, u));
        }
      }
    }
    
    /*   adjusting the probabilities   */
    for(unsigned int i = 0; i <= 6; i++)
      _M_prob_proc[i] /= retval;
    
    /*   Jacobians:
     *     - jacobian of the u mapping ==> ujac
     *     - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     */
    retval *= ujac;
    retval *= _M_y_mapping.jacobian(tau);
    
    return retval;
  };
  
  double splitting_kernel<fin, ordering::kT>::_M_funcS(double kappa, double chi) const
  {
    /*   aliases  */
    const double &ymax = _M_pars.ymax, &pQ = _M_pars.pQ, &bl = _M_pars.bl;
    
    /*   Transforming the variable kappa to y.  */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double mT = _M_mdl->alpha_qcd.tauk_inverse(tau)/pQ;
    
    /*   Transforming the variable chi to z.  */
    double u = 0.5, ujac = 1.0, retval = 0.0;
    z_mapping zmap;
    
    zmap.refresh(&_M_pars, mT);
    zmap.mapping_chi2u(chi, u, ujac);
    
    /*   Calculate the splitting function of the gluon emission process.   */
    double y = 0.5*(mT+std::sqrt(mT*(mT+4.0*u*(1.0-u)*bl)))/(u*(1.0-u));
    _M_prob_proc[0] = retval = (y+bl)/(y+2.0*bl)*_M_func_emit(_M_mdl->spin(_M_pars.fl), ymax, y, u);
    
    /*   For gluon emitter we have to consider the g-->f+f~ processes  */
    std::memset(_M_prob_proc+1, 0, 6*sizeof(double));
    
    if(_M_pars.fl == 0) {
      for(unsigned int i = 1; i <= 6; i++) {
        double bi = _M_mdl->mass2(i)/pQ;
        
        if((ymax-4.0*bi)*u*(1.0-u) >= mT) {
          y = mT/(u*(1.0-u)) + 4.0*bi;
          retval += (_M_prob_proc[i] = y/(y-4.0*bi)*__mc_splitting_functions::gluon_splitting_fin(ymax, bi, y, u));
        }
      }
    }
    
    /*   adjusting the probabilities   */
    for(unsigned int i = 0; i <= 6; i++)
      _M_prob_proc[i] /= retval;
    
    /*   Jacobians:
     *     - jacobian of the u mapping ==> ujac
     *     - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     */
    retval *= ujac;
    retval *= _M_y_mapping.jacobian(tau);
    
    return retval;
  };
  

  /*  Selecting vertex  */
  splitting_variables<lo_shower>::zeta_flavor
  splitting_kernel<fin, ordering::kT>::vertex(double rvtx) const
  {
    //----- gluon emitter -----
    if(_M_pars.fl == 0) {
      int ic = 0;
      double cprob = _M_prob_proc[0];
      while(cprob < rvtx) cprob += _M_prob_proc[++ic];
      if(ic) return {ic, -ic};
    }
    
    //---- quark emitter -----
    return {_M_pars.fl, 0};
  }
  
  
  bool splitting_prob<fin, ordering::kT>::
  _M_parameters(const statistical_basis& p, int emitter, int spectator, const lorentzvector<double>& qnull) const
  {
    /*  Some simple information   */
    const model *mdl = _M_split.get_model();
    bool is_decay = (p.number_of<incomings>() == 1U && spectator == 0);
    int fl = p[emitter].flavor;
    
    /*   Total incoming momentum  */
    lorentzvector<double> Q = p.momentum_sum<incomings>();
    double pQ = 2.0*(p[emitter].momentum*Q), Q2 = Q.mag2();
    
    /*  Standard parameters: al, bl, rl, ymax, ymin, ynull    */
    double al = Q2/pQ, bl = mdl->mass2(fl)/pQ;
    double rl = std::sqrt(1.0-4.0*al*bl);
    double tmp = (1.0-bl)/(std::sqrt(al)+std::sqrt(std::abs(al+bl-1.0)));
    double ymax = tmp*tmp-bl;
    
    this->_M_mTmax = 0.25*ymax*ymax/(ymax+bl);
    this->_M_mTmin = std::max(_M_q2np, _M_k2T_veto)/pQ;
  
    /*  The splitting is not possible. No phase space left
     *  in this channel to make a splitting.
     */
    if(_M_mTmax < _M_mTmin) return false;
    
    /*  Variable to be able to map the shower time to kappa   */
    this->_M_mTnull = qnull.mag2()/pQ;
    
    /*  When the spectator is the incoming in a decay process then it is simple. */
    if(is_decay) {
      _M_bind_jac = _M_split.bindD(fl, _M_q2np, rl, ymax, pQ);
      _M_bounding_kappa_chi = &(_M_split_approx.bindD(fl, ymax, pQ));
      
      return true;
    }
    
    /*   In the special case when the spectator set consists of one massless parton.   */
    double m2k = mdl->mass2(p[spectator].flavor);
    
    if(p.number_of<outgoings>() == 2u && m2k < 1e-15) {
      _M_bind_jac = _M_split.bindS(fl, _M_q2np, ymax, pQ);
      _M_bounding_kappa_chi = &(_M_split_approx.bindS(fl, pQ));
      
      return true;
    }
    
    /*   Otherwise the splitting depends on the spectator partons coordinates   */
    lorentzvector<double> pl = (2.0*al/(1.0+rl))*p[emitter].momentum;
    lorentzvector<double> nl = Q - pl; pl -= ((1.0-rl)/(1.0+rl))*Q;
    
    double kQ = (Q*p[spectator].momentum)*2.0*rl/(1.0+rl);
    double xi = (pl*p[spectator].momentum)/kQ, mu = m2k*(pl*nl)/(kQ*kQ);
    
    if(spectator > 0) {
      _M_bind_jac = _M_split.bindF(fl, _M_q2np, rl, ymax, pQ, xi, mu);
      _M_bounding_kappa_chi = &(_M_split_approx.bindF(fl, ymax, pQ, xi, mu));
    } else {
      _M_bind_jac = _M_split.bindI(fl, _M_q2np, rl, ymax, pQ, xi, mu);
      _M_bounding_kappa_chi = &(_M_split_approx.bindI(fl, ymax, pQ, xi, mu));
    }
    
    return true;
  }
  
  
  bool splitting_prob<fin, ordering::kT>::
  _M_make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
                double& tmT, int emitter, int spectator, double tmU, double tmL, double cfac) const
  {
    //----- veto procedure -----
    double lrho, chi;
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*    adjusting to the available kinematical region   */
    double mL = _M_mTnull*std::exp(-tmU), mU = _M_mTnull*std::exp(-tmL);
    
    if(mL < _M_mTmin) mL = _M_mTmin;
    if(mU > _M_mTmax) mU = _M_mTmax;
    if(mL >= mU) return false;
    
    /*   the integral bounderies in variable kappa   */
    double kappaU = _M_split.mapping_mT2k(mL);
    double kappaT = _M_split.mapping_mT2k(mU);
    
    do {
      //------ generate a random number ------
      lrho = -std::log(rng(rne));
      lrho /= cfac;
      lrho /= _M_bind_jac;
      
      //------ no chance to emit in this channel ------
      if(this->_M_primitive_kappa(kappaU, kappaT) < lrho) return false;
      
      //------ otherwise found the new trial time -----
      kappaT = __mc_helper::newton_root_finder_kappa(this, kappaT, lrho);
      
      //------ generates a trial u -----
      chi = __mc_helper::newton_root_finder_chi(this, kappaT, rng(rne));
      
      /*   calculate the hit-or-miss ratio   */
      ratio = _M_split(kappaT, chi)/(*_M_bounding_kappa_chi)({kappaT, chi});
    } while(rng(rne) > ratio);
    
    
    /*  selecting vertex   */
    vars->zeta = _M_split.vertex(rng(rne));
    vars->emitter = emitter;
    vars->spectator = (vars->zeta.key2 != 0 ? emitter : spectator);
    
    /*   getting y and z variables  */
    double mT = _M_split.mapping_k2mT(kappaT);
    
    vars->z = _M_split.mapping_chi2u(mT, chi);
    vars->y = _M_split.getting_y(vars->zeta.key2, mT, vars->z);
    
    /*  selecting the azimuthal angle  */
    vars->phi = _M_split.azimuthal_angle(vars->emitter, vars->spectator, rng(rne));
    
    /*   transforming variable y to shower time   */
    tmT = std::log(_M_mTnull/mT);
    
    return true;
  }
  

  bool splitting_prob<fin, ordering::kT>::
  _M_make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
                double& tmU, double tmL, const statistical_basis& p, const lorentzvector<double>& qnull) const
  {
    bool is_possible = false;
    unsigned int np = p.number_of<qcd_out>();
    std::pair<int, double> spectator[4];
    double tmT, ratioT;
    
    for(int emitter = 1; emitter <= (int) np; emitter++)
    {
      //----- possible spectators ----
      unsigned int ns = __mc_helper::counting_color_connections(p, emitter, spectator);
      
      //----- make trial emission with every spectator -----
      for(auto is = spectator; is < spectator+ns; ++is) {
        //----- setting up the probability distribution ------
        if(!_M_parameters(p, emitter, is->first, qnull)) continue;
      
        //---- make a trial emission -----
        if(!_M_make_trial(rne, vars, ratioT, tmT, emitter, is->first, tmU, tmL, is->second)) continue;

        //----- splitting is possible ----
        tmU = tmT;
        ratio = ratioT;
        is_possible = true;
      }
    }
    
    return is_possible;
  }
  

}




