/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/14/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <algorithm>

//   project headers
#include "bits/mc-engine-kT.h"



namespace duct {
  
  void splitting_kernel<ini, ordering::kT>::z_mapping::
  refresh(const params *pars, double mT)
  {
    /*   aliases  */
    const double &eta = pars->eta, &ma = pars->ma, &mb = pars->mb;
    
    //----- z limits -----
    is_gluon = (engine->_M_pars.fl ? false : true);
    
    double zmax = std::pow(std::sqrt(mT) + std::sqrt(mT+1.0+ma+mb+ma*mb),2)-ma-mb;
    zmax = (ma*mb > 1e-12 ? 2.0/(zmax+std:sqrt(zmax*zmax-4.0*ma*mb)) : 1.0/zmax);
    
    //----- refresing the y dependent variables -----
    double ozmax = 1.0-zmax;
    oeta = 1.0-eta;
    
    if(is_gluon) {
      ozmax /= zmax;
      oeta  /= eta;
    }
    
    J = std::log(oeta/ozmax);
  }
  

  
  double splitting_kernel<ini, ordering::lambda>::y_mapping::
  refresh(const splitting_kernel<ini, ordering::lambda> *engine)
  {
    /*   aliases  */
    const int &fl = engine->_M_pars.fl;
    const double &eta = engine->_M_pars.eta, &Q2 = engine->_M_pars.Q2;
    const double &ma = engine->_M_pars.ma, &mb = engine->_M_pars.mb, &mT = engine->_M_pars.mT;
    const model *mdl = engine->_M_mdl;
    const qcd_hadron *pdf = mdl->pdf(hadron(engine->_M_pars.hdr));
    
    /*  Trivial cut. It is effective when both incomming partons are massive  */
    if(1.0 < ma*mb) return 0.0;
    
    /*  Transforming the variable kappa to y. This is like in the final state
     *  related to the primitive function of the log integral of alpha_s.
     */
    double tmp1 = 1.0-eta*(1.0+ma*mb), tmp2 = 2.0*eta*(1.0+eta*mb);
    if(tmp1*tmp1 <= 2.0*(ma+mT)*tmp2) return 0.0;
    
    double stmp = std::sqrt(tmp1*tmp1 - 2.0*(ma+mT)*tmp2);
    double yl = (tmp1-stmp)/tmp2, yu = (tmp1+stmp)/tmp2;
    double q2T = std::max(pdf->q2min(), pdf->thrs(fl, eta));
    
    yl = std::max(yl, q2T/Q2);
    if(yl >= yu) return 0.0;
    
    /*   Setting up the interpolation spline */
    this->spline.assign([=](double tau) -> double {
      double y = mdl->alpha_qcd.tauk_inverse(tau)/Q2;
      
      //----- z limits -----
      double zmax = 1.0+y+ma*mb;
      if(mb > 1e-12) zmax = 0.5*(zmax+std::sqrt(zmax*zmax+4.0*mb*(y-ma-mT)));
      zmax = (1.0-(ma+mT)/y)/zmax;
      
      //----- refresing the y dependent variables -----
      double ozmax = 1.0-zmax, oeta = 1.0-eta;
      double as = mdl->alpha_qcd(y*Q2);
      double asb = 0.5*beta0(5U)*as;
      
      if(fl == 0) {
        ozmax /= zmax;
        oeta  /= eta;
      }
      
      //---- contribution from the z jacobian -----
      double retval = std::log((1.0+asb*std::log(oeta))/(1.0+asb*std::log(ozmax)))/asb;
      
      //---- try to map out some pdf dependence ----
      double pdfden = pdf->parton(-fl, eta, y*Q2);
      
      if(fl == 0) retval *= 1.0 + pdf->parton(19, eta, y*Q2)/pdfden;
      else retval *= (pdf->parton(-fl, eta/zmax, y*Q2) + pdf->parton(0, eta/zmax, y*Q2))/pdfden;
      
      //---- for heavy quark we have to remove the singularity of the pdf ratio -----
      if(std::abs(fl) > 3) retval *= mdl->alpha_qcd.tauk(q2T)-tau;
      
      //---- in the gluon case we try to map out the large y suppression ----
      else if(fl == 0) retval /= (1.0+y)*(1.0+y);
      
      return retval;
    }, mdl->alpha_qcd.tauk(yu*Q2), mdl->alpha_qcd.tauk(yl*Q2));
    
    
    /*  Taking care of the heavy flavor threshold.   */
    if(std::abs(fl) > 3) {
      q2T = Q2*(tmp1-std::sqrt(tmp1*tmp1 - 2.0*ma*tmp2))/tmp2;
      kappaT = this->mapping_tau2k(mdl->alpha_qcd.tauk(q2T));
      kappaF = std::min(0.95, this->mapping_tau2k(mdl->alpha_qcd.tauk(1.6*q2T)));
    } else kappaT = kappaF = 2.0;
    
    return this->spline.integral();
  }
  
  
  double splitting_kernel<ini, ordering::lambda>::y_mapping::
  refresh_hq(const splitting_kernel<ini, ordering::lambda> *engine)
  {
    /*   aliases  */
    const int &fl = engine->_M_pars.fl;
    const double &eta = engine->_M_pars.eta, &Q2 = engine->_M_pars.Q2;
    const double &ma = engine->_M_pars.ma, &mb = engine->_M_pars.mb, &mT = engine->_M_pars.mT;
    const model *mdl = engine->_M_mdl;
    const qcd_hadron *pdf = mdl->pdf(hadron(engine->_M_pars.hdr));
    
    /*  Trivial cut. It is effective when both incomming partons are massive  */
    if(1.0 < ma*mb) return 0.0;
    
    /*  Transforming the variable kappa to y. This is like in the final state
     *  related to the primitive function of the log integral of alpha_s.
     */
    double tmp1 = 1.0-eta*(1.0+ma*mb), tmp2 = 2.0*eta*(1.0+eta*mb);
    if(tmp1*tmp1 <= 2.0*(ma+mT)*tmp2) return 0.0;
    
    double stmp = std::sqrt(tmp1*tmp1 - 2.0*(ma+mT)*tmp2);
    double yl = (tmp1-stmp)/tmp2, yu = (tmp1+stmp)/tmp2;
    double q2T = std::max(pdf->q2min(), pdf->thrs(fl, eta));
    
    yl = std::max(yl, q2T/Q2);
    if(yl >= yu) return 0.0;
    
    /*   Setting up the interpolation spline */
    this->spline.assign([=](double tau) -> double {
      double y = mdl->alpha_qcd.tauk_inverse(tau)/Q2;
      
      //----- z limits -----
      double zmin = 1.0+y+ma*mb;
      if(mb > 1e-12) zmin = 0.5*(zmin+std::sqrt(zmin*zmin+4.0*mb*(y-ma-mT)));
      zmin = (1.0-(ma+mT)/y)/zmin;
      
      double zmax = 1.0+y;
      if(mb > 1e-12) zmax = 0.5*(zmax+std::sqrt(zmax*zmax+4.0*mb*(y-mT)));
      zmax = (1.0-mT/y)/zmax;
      
      //----- refresing the y dependent variables -----
      double as = mdl->alpha_qcd(y*Q2);
      double asb = 0.5*beta0(5U)*as;
      
      //---- contribution from the z jacobian -----
      return std::log((1.0+asb*std::log(1.0-zmin))/(1.0+asb*std::log(1.0-zmax)))/asb;
    }, mdl->alpha_qcd.tauk(yu*Q2), mdl->alpha_qcd.tauk(yl*Q2));
    
    
    /*  In the gluon emission term we don't have to worry about the threshold.   */
    kappaT = kappaF = 2.0;
    
    return this->spline.integral();
  }
  
  
  double splitting_kernel<ini, ordering::lambda>::_M_gluon(double kappa, double chi, double k2Tmin, double k2Tmax) const
  {
    /*   Safety cut  */
    double fac = 1.0;
    
    if(kappa > 0.98) {
      fac = std::pow((1.0-kappa)/0.02, 0.25);
      kappa = 0.98;
    }
    
    /*   aliases  */
    const double &eta = _M_pars.eta, &Q2 = _M_pars.Q2;
    const double &mb = _M_pars.mb, &mT = _M_pars.mT;
    const qcd_hadron *pdf = _M_mdl->pdf(hadron(_M_pars.hdr));

    /*   Transforming the variable kappa to y.  */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double y = _M_mdl->alpha_qcd.tauk_inverse(tau)/Q2;
    double m2F = y*Q2;
    
    /*   Transforming the variable chi to z.  */
    double z = eta, zjac = 1.0;
    z_mapping zmap;
    
    zmap.refresh(this, y);
    zmap.mapping_chi2z(chi, z, zjac);
    
    /*   Calculate the splitting functions with transverse momentum veto  */
    double *__prob = _M_prob_proc + 5, retval = 0.0, wgq;
    double k2T = (1.0 - z*(1.0 + y*(1.0+z*mb)))*y*Q2;
    
    if(k2Tmin < k2T && k2T < k2Tmax)
    {
      /*   Contribution of the gluon emission  */
      __prob[0] = this->_M_func_emit(2, _M_pars.spec, 0.0, mb, _M_pars.xi, _M_pars.xT, y, z);
      __prob[0] *= pdf->parton(0, eta/z, m2F);
      retval = __prob[0];
      
      /*   Contribution of the light quarks   */
      wgq = __mc_splitting_functions::splitting_gluon_quark_ini(0.0, mb, y, z);
      
      for(int i = 1; i < 4; i++) {
        retval += (__prob[ i] = pdf->parton(-i, eta/z, m2F)*wgq);
        retval += (__prob[-i] = pdf->parton( i, eta/z, m2F)*wgq);
      }
    } else {
      for(int i = -3; i < 4; i++)
        __prob[i] = 0.0;
    }
    
    /*  Contribution of the charm and bottom quarks  */
    for(int i = 4; i < 6; i++)
    {
      double ma = _M_mdl->mass2(i)/Q2;
      double xa = 1.0/z - 1.0 - y*(1.0+z*mb) - z*ma*mb, xn = 1.0-z*z*ma*mb;
      double xb = z*(y + ma + z*ma*(1.0+y-z*ma*mb));
      k2T = xa*xb/(xn*xn)-ma;
      
      if(std::max(mT, k2Tmin/Q2) < k2T && k2T < k2Tmax/Q2) {
        wgq = __mc_splitting_functions::splitting_gluon_quark_ini(ma, mb, y, z);
        retval += (__prob[ i] = pdf->parton(-i, eta/z, m2F)*wgq);
        retval += (__prob[-i] = pdf->parton( i, eta/z, m2F)*wgq);
      } else __prob[i] = __prob[-i] = 0.0;
    }
    
    /*   return zero if it was vetod   */
    if(retval < 1e-19) return 0.0;
    
    /*   adjusting the probabilities   */
    for(int i = -5; i <= 5; i++)
      __prob[i] /= retval;
    
    /*   Jacobians:
     *     - jacobian of the z mapping ==> zjac
     *     - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     *     - denom. of the pdf ratio
     */
    retval *= zjac;
    retval *= _M_y_mapping.jacobian(tau);
    retval *= z/pdf->parton(0, eta, m2F);
    
    return retval*fac;
  }
  
  
  double splitting_kernel<ini, ordering::lambda>::_M_quark(double kappa, double chi, double k2Tmin, double k2Tmax) const
  {
    /*   aliases  */
    const int &fl = _M_pars.fl;
    const double &eta = _M_pars.eta, &Q2 = _M_pars.Q2;
    const double &ma = _M_pars.ma, &mb = _M_pars.mb, &mT = _M_pars.mT;
    const qcd_hadron *pdf = _M_mdl->pdf(hadron(_M_pars.hdr));
  
    /*   Transforming the variable kappa to y.   */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double y = _M_mdl->alpha_qcd.tauk_inverse(tau)/Q2;
    double m2F = y*Q2;
    
    /*   Transforming the variable chi to z.  */
    double z = eta, zjac = 1.0;
    z_mapping zmap;
    
    zmap.refresh(this, y);
    zmap.mapping_chi2z(chi, z, zjac);
    
    /*   Calculate the splitting function of the gluon emission process.   */
    double *__prob = _M_prob_proc + 5;
    double xa = 1.0/z-1.0 - y*(1.0+z*mb) - (1.0-z)*ma*mb, xn = 1.0-z*z*ma*mb;
    double xb = z*(y*(1.0+z*ma) - ma*(1.0-z)*(1.0-z*ma*mb));
    double k2T = xa*xb/(xn*xn);
    
    /*  probability of the gluon emission channel   */
    for(int i = -5; i <= 5; i++) __prob[i] = 0.0;
    
    if(std::max(mT, k2Tmin/Q2) < k2T && k2T < k2Tmax/Q2 && xb > 0.0) {
      __prob[fl] = this->_M_func_emit(1, _M_pars.spec, ma, mb, _M_pars.xi, _M_pars.xT, y, z);
      __prob[fl] *= pdf->parton(-fl, eta/z, m2F);
    }
    
    /*   contribution of the gluon splitting   */
    k2T = (1.0 - z*(1.0 + y*(1.0+z*mb) + ma*mb))*y - ma;
    
    if(k2Tmin/Q2 < k2T && k2T < k2Tmax/Q2) {
      __prob[0] = __mc_splitting_functions::splitting_quark_antiquark_ini(ma, mb, y, z);
      __prob[0] *= pdf->parton(0, eta/z, m2F);
    }
    
    /*   adjusting the probabilities   */
    double retval = __prob[0] + __prob[fl];
    
    /*   return zero if it was vetod   */
    if(retval < 1e-19) return 0.0;
    
    __prob[0]  /= retval;
    __prob[fl] /= retval;
    
    /*  Jacobians:
     *   - jacobian of the z mapping ==> zjac
     *   - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     *   - denom. of the pdf ratio
     */
    retval *= zjac;
    retval *= _M_y_mapping.jacobian(tau);
    retval *= z/pdf->parton(-fl, eta, m2F);
    
    return retval;
  }
  
  
  double splitting_kernel<ini, ordering::lambda>::_M_quark_hq(double kappa, double chi, double k2Tmin, double k2Tmax) const
  {
    /*   Safety cut  */
    double fac = 1.0;
    
    if(kappa > 0.95) {
      fac = std::pow((1.0-kappa)/0.05, 0.25);
      kappa = 0.95;
    }
    
    /*   aliases  */
    const int &fl = _M_pars.fl;
    const double &eta = _M_pars.eta, &Q2 = _M_pars.Q2;
    const double &ma = _M_pars.ma, &mb = _M_pars.mb, &mT = _M_pars.mT;
    const qcd_hadron *pdf = _M_mdl->pdf(hadron(_M_pars.hdr));
    
    /*   Transforming the variable kappa to y.   */
    double tau = _M_y_mapping.mapping_k2tau(kappa);
    double y = _M_mdl->alpha_qcd.tauk_inverse(tau)/Q2;
    double m2F = y*Q2;
    
    /*   Transforming the variable chi to z.  */
    double z = eta, zjac = 1.0;
    z_mapping zmap;
    
    zmap.refresh_hq(this, y);
    zmap.mapping_chi2z(chi, z, zjac);
    
    /*   Calculate the splitting function of the gluon emission process.   */
    double retval = 0.0;
    double xa = 1.0/z-1.0 - y*(1.0+z*mb) - (1.0-z)*ma*mb, xn = 1.0-z*z*ma*mb;
    double xb = z*(y*(1.0+z*ma) - ma*(1.0-z)*(1.0-z*ma*mb));
    double k2T = xa*xb/(xn*xn);
    
    if(std::max(mT, k2Tmin/Q2) < k2T && k2T < k2Tmax/Q2 && xb > 0.0) {
      retval = this->_M_func_emit(1, _M_pars.spec, ma, mb, _M_pars.xi, _M_pars.xT, y, z);
      retval *= pdf->parton(-fl, eta/z, m2F);
    }
    
    /*   return zero if it was vetod   */
    if(retval < 1e-19) return 0.0;
    
    /*  Jacobians:
     *   - jacobian of the z mapping ==> zjac
     *   - jacobian of the tau mapping ==> _M_y_mapping.jacobian(tau)
     *   - denom. of the pdf ratio
     */
    retval *= zjac;
    retval *= _M_y_mapping.jacobian(tau);
    retval *= z/pdf->parton(-fl, eta, m2F);
    
    return retval*fac;
  }
  
  
  
  splitting_variables<lo_shower>::zeta_flavor
  splitting_kernel<ini, ordering::lambda>::vertex(double rvtx) const
  {
    //----- selecting vertex -----
    const int &fl = _M_pars.fl;
    const double *prob = this->_M_prob_proc + 5;
    
    //----- gluon emitter -----
    if(fl == 0) {
      int ic = -5;
      double cprob = prob[ic];
      
      while(cprob < rvtx)
        cprob += prob[++ic];
      
      return {ic, -ic};
    }
    //----- quark emitter ------
    else {
      if(rvtx < prob[fl]) return {fl, 0};
      else return {0, fl};
    }
  }


 
  bool splitting_prob<ini, ordering::lambda>::
  _M_parameters(const event<dis>& q, bool hq, int spectator, const lorentzvector<double>& qnull) const
  {
    int hdr = q[hadron(0)].key, fl = q[0].flavor;
    const model *mdl = _M_split.get_model();
    
    double s = q[-1].momentum*q[hadron(0)].momentum;
    double eta = (q[0].momentum*q[hadron(-1)].momentum)/s;
    double Q2 = 2.0*eta*s;
        
    /*    the allowed kinematical range in variable y   */
    {
      double ma = mdl->mass2(fl)/Q2, mT = std::max(_M_q2np, _M_k2T_veto_min)/Q2;
      const qcd_hadron *pdf = mdl->pdf(hadron(hdr));
      
      /*  Transforming the variable kappa to y. This is like in the final state
       *  related to the primitive function of the log integral of alpha_s.
       */
      double tmp1 = 1.0-eta;
      if(tmp1*tmp1 <= 4.0*(ma+mT)*eta) return false;
      
      double stmp = std::sqrt(tmp1*tmp1 - 4.0*(ma+mT)*eta);
      double yl = 0.5*(tmp1-stmp)/eta, q2T = std::max(pdf->q2min(), pdf->thrs(fl, eta));
      
      this->_M_ymin = std::max(yl, q2T/Q2);
      this->_M_ymax = 0.5*(tmp1+stmp)/eta;
      
      if(_M_ymin >= _M_ymax) return false;
    }

    /*  Variable to be able to map the shower time to kappa   */
    this->_M_ynull = q[hadron(0)].momentum*qnull/s;
    this->_M_isheavy = hq;
    
    /*  Spectator related parameters  */
    double ska = eta*(q[spectator].momentum*q[hadron(0)].momentum);
    double skb = q[spectator].momentum*q[-1].momentum;
    double m2k = mdl->mass2(q[spectator].flavor);
    
    double xi = skb/(ska+skb);
    double xT = std::sqrt(std::abs(1.0-m2k*Q2/(4.0*ska*skb)));
    
    /*  Bind the exact splitting function to the parameters. This also
     *  checks if the splitting is possible and sets up the mappings.
     */
    if(hq) _M_bind_jac = _M_split.bindFH(hdr, fl, _M_q2np, 0.0, eta, Q2, xi, xT);
    else _M_bind_jac = _M_split.bindF(hdr, fl, _M_q2np, 0.0, eta, Q2, xi, xT);
    
    if(_M_bind_jac < 1.0e-13) return false;
    
    /*  We need the bounding function at fixed parameters. */
    auto iska = _M_split_approx.find(hdr);
    
    if(iska != _M_split_approx.end()) {
      /*    Pure gluon emission from heavy quark   */
      if(hq) this->_M_bounding_kappa_chi = &(iska->second.bindFH(fl, eta, Q2, xi, xT));
      
      /*    Main term with gluon splitting   */
      else {
        this->_M_bounding_kappa_chi = &(iska->second.bindF(fl, eta, Q2, xi, xT));
        
        /*  In the heavy quark case the bounding function is freezed at kappaF.
         *  It helps on the numerical stability.
         */
        if(std::abs(_M_split.emitter_flavor()) > 3) {
          double kappaF = _M_split.kappaF();
          _M_funcF_hq = _M_bounding_kappa_chi->average({{kappaF,kappaF},{0.0,1.0}});
        }
      }
    }
    
    return true;
  }
  

  bool splitting_prob<ini, ordering::lambda>::
  _M_parameters(const event<hhc>& q, bool hq, int ia, int spectator, const lorentzvector<double>& qnull) const
  {
    //----- some trivial checks ----
    if(ia < -1 || ia > 0)
      throw "shower_engine_virt_ini::parameters(...) : emitter is not an incoming parton";
    
    int ib = (ia == -1 ? 0 : -1);
    int hdr = q[hadron(ia)].key, fl = q[ia].flavor;
    const model *mdl = _M_split.get_model();
    
    double s = q[hadron(-1)].momentum*q[hadron(0)].momentum;
    double eta = (q[ia].momentum*q[hadron(ib)].momentum)/s;
    double etb = (q[ib].momentum*q[hadron(ia)].momentum)/s;
    double Q2 = 2.0*eta*etb*s, m2b = mdl->mass2(q[ib].flavor);    
    
    /*    the allowed kinematical range in variable y   */
    {
      double ma = mdl->mass2(fl)/Q2, mb = m2b/Q2, mT = std::max(_M_q2np, _M_k2T_veto_min)/Q2;
      const qcd_hadron *pdf = mdl->pdf(hadron(hdr));

      /*  Trivial cut. It is effective when both incomming partons are massive  */
      if(1.0 < ma*mb) return 0.0;
      
      /*  Transforming the variable kappa to y. This is like in the final state
       *  related to the primitive function of the log integral of alpha_s.
       */
      double tmp1 = 1.0-eta*(1.0+ma*mb), tmp2 = 2.0*eta*(1.0+eta*mb);
      if(tmp1*tmp1 <= 2.0*(ma+mT)*tmp2) return false;
      
      double stmp = std::sqrt(tmp1*tmp1 - 2.0*(ma+mT)*tmp2);
      double yl = (tmp1-stmp)/tmp2, q2T = std::max(pdf->q2min(), pdf->thrs(fl, eta));
      
      this->_M_ymin = std::max(yl, q2T/Q2);
      this->_M_ymax = (tmp1+stmp)/tmp2;
      
      if(_M_ymin >= _M_ymax) return false;
    }

    /*  Variable to be able to map the shower time to kappa   */
    this->_M_ynull = (q[hadron(ia)].momentum*qnull)/(etb*s);
    this->_M_isheavy = hq;

    /*  Spectator related parameters  */
    double xi = m2b/(Q2+m2b), xT = 0.0;
    
    if(spectator > 0) {
      double ska = eta*(q[spectator].momentum*q[hadron(ia)].momentum);
      double skb = etb*(q[spectator].momentum*q[hadron(ib)].momentum);
      double m2k = mdl->mass2(q[spectator].flavor);
      
      xi = skb/(ska+skb);
      xT = std::sqrt(std::abs(1.0-m2k*Q2/(4.0*ska*skb)));
    }
    
    
    /*  Bind the exact splitting function to the parameters. This also
     *  checks if the splitting is possible and sets up the mappings.
     */
    
    /*     Pure gluon emission from the heavy quark   */
    if(hq) {
      if(spectator > 0) _M_bind_jac = _M_split.bindFH(hdr, fl, _M_q2np, m2b, eta, Q2, xi, xT);
      else _M_bind_jac = _M_split.bindIH(hdr, fl, _M_q2np, m2b, eta, Q2);
    }
    /*    Main term with gluon splitting  */
    else {
      if(spectator > 0) _M_bind_jac = _M_split.bindF(hdr, fl, _M_q2np, m2b, eta, Q2, xi, xT);
      else _M_bind_jac = _M_split.bindI(hdr, fl, _M_q2np, m2b, eta, Q2);
    }
    
    if(_M_bind_jac < 1.0e-13) return false;
    
    /*  We need the bounding function at fixed parameters. */
    auto iska = _M_split_approx.find(hdr);
    
    if(iska != _M_split_approx.end()) {
      /*   Pure gluon emission heavy quark   */
      if(hq) {
        if(spectator < 1) this->_M_bounding_kappa_chi = &(iska->second.bindIH(fl, eta, Q2));
        else this->_M_bounding_kappa_chi = &(iska->second.bindFH(fl, eta, Q2, xi, xT));
      }
      /*   Main contribution with gluon spitting    */
      else {
        if(spectator < 1) this->_M_bounding_kappa_chi = &(iska->second.bindI(fl, eta, Q2));
        else this->_M_bounding_kappa_chi = &(iska->second.bindF(fl, eta, Q2, xi, xT));
        
        /*  In the heavy quark case the bounding function is freezed at kappaF.
         *  It helps on the numerical stability.
         */
        if(std::abs(_M_split.emitter_flavor()) > 3) {
          double kappaF = _M_split.kappaF();
          _M_funcF_hq = _M_bounding_kappa_chi->average({{kappaF,kappaF},{0.0,1.0}});
        }
      }
    }
    
    return true;
  }
  
 
  double splitting_prob<ini, ordering::lambda>::_M_primitive_kappa(double kappaU, double kappaL) const
  {
    /*  trivial case  */
    if(kappaL >= kappaU) return 0.0;
    
    /*  For heavy flavor we have a singularity at kappaT   */
    if(!_M_isheavy && std::abs(_M_split.emitter_flavor()) > 3) {
      double retval = 0.0, kappaT = _M_split.kappaT();
      double kappaF = _M_split.kappaF();
      double kappaX = std::min(kappaU, kappaF);
      double kappaY = std::max(kappaL, kappaF);
     
      if(kappaL < kappaX)
        retval = _M_bounding_kappa_chi->integral({{kappaL,kappaX},{0.0,1.0}},
                                                 [=](const std::pair<double,double> *v) -> double {
                                                   return std::log((kappaT-v[0].first)/(kappaT-v[0].second))*(v[1].second-v[1].first);
                                                 });
      
      if(kappaY < kappaU)
        retval += _M_funcF_hq*std::log((kappaT-kappaY)/(kappaT-kappaU));
      
      return retval;
    }
    
    /*   For light partons it is simple   */
    return _M_bounding_kappa_chi->integral({{kappaL,kappaU},{0.0,1.0}});
  }
  
  
  bool splitting_prob<ini, ordering::lambda>::
  _M_make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
                double& tmT, int emitter, int spectator, double tmU, double tmL, double cfac) const
  {
    //----- veto procedure -----
    bool is_forced = false;
    double lrho, chi;
    std::uniform_real_distribution<double> rng(0.0,1.0);
    
    /*    adjusting to the available kinematical region   */
    double yL = _M_ynull*std::exp(-tmU), yU = _M_ynull*std::exp(-tmL);
    
    if(yL < _M_ymin) yL = _M_ymin;
    if(yU > _M_ymax) yU = _M_ymax;
    if(yL >= yU) return false;
    
    /*   the integral bounderies in variable kappa   */
    double kappaU = _M_split.mapping_y2k(yL);
    double kappaT = _M_split.mapping_y2k(yU);
    
    do {
      //------ generate a random number ------
      lrho = -std::log(rng(rne));
      lrho /= cfac;
      lrho /= _M_bind_jac;
      
      /*   Check if the splitting is possible.  */
      if(kappaU < _M_split.kappaT() && this->_M_primitive_kappa(kappaU, kappaT) < lrho) return false;
      
      //------ otherwise found the new trial time -----
      kappaT = __mc_helper::newton_root_finder_kappa(this, kappaT, lrho);
      
      //------ generates a trial u -----
      chi = __mc_helper::newton_root_finder_chi(this, kappaT, rng(rne));
      
      /*   calculate the hit-or-miss ratio   */
      ratio = _M_function_veto(kappaT, chi)/(*_M_bounding_kappa_chi)({kappaT, chi});
      
      /*   check if we have to do forced splitting for heavy flavor   */
      if(!_M_isheavy && std::abs(_M_split.emitter_flavor()) > 3 && kappaT > _M_split.kappaF()) {
        ratio = 0.999999999;
        is_forced = true;
        break;
      }
    } while(rng(rne) > ratio);
    
    
    /*   getting y and z variables  */
    vars->y = _M_split.mapping_k2y(kappaT);
    vars->z = _M_split.mapping_chi2z(vars->y, chi);
    
    /*  selecting vertex   */
    vars->emitter = emitter;

    if(_M_isheavy) vars->zeta = {_M_split.emitter_flavor(), 0};
    else if(is_forced) vars->zeta = {0, _M_split.emitter_flavor()};
    else vars->zeta = _M_split.vertex(rng(rne));
    
    vars->spectator = (vars->zeta.key2 != 0 ? emitter : spectator);
    
    /*  selecting the azimuthal angle  */
    vars->phi = _M_split.azimuthal_angle(vars->emitter, vars->spectator, rng(rne));
    
    /*   transforming variable y to shower time   */
    tmT = std::log(_M_ynull/vars->y);
    
    return true;
  }
  

  template<> bool splitting_prob<ini, ordering::lambda>::
  make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
             double& tmU, double tmL, const event<dis>& p, const lorentzvector<double>& qnull) const
  {
    bool is_possible = false;
    std::pair<int, double> spectator[4];
    double tmT, ratioT;
    
    //----- possible spectators ----
    unsigned int ns = __mc_helper::counting_color_connections(p, 0, spectator);
    
    //----- make trial emission with every spectator -----
    for(auto is = spectator; is < spectator+ns; ++is) {
      //----- setting up the probability distribution ------
      if(!_M_parameters(p, false, is->first, qnull)) continue;
      
      //---- make a trial emission -----
      if(!_M_make_trial(rne, vars, ratioT, tmT, 0, is->first, tmU, tmL, is->second)) continue;
      
      //----- splitting is possible ----
      tmU = tmT;
      ratio = ratioT;
      is_possible = true;
    }
    
    /*   For heavy qaurks we have extra term   */
    if(std::abs(p[0].flavor) > 3) {
      for(auto is = spectator; is < spectator+ns; ++is) {
        //----- setting up the probability distribution ------
        if(!_M_parameters(p, true, is->first, qnull)) continue;
        
        //---- make a trial emission -----
        if(!_M_make_trial(rne, vars, ratioT, tmT, 0, is->first, tmU, tmL, is->second)) continue;
        
        //----- splitting is possible ----
        tmU = tmT;
        ratio = ratioT;
        is_possible = true;
      }
    }
    
    return is_possible;
  }
  
  
  template<> bool splitting_prob<ini, ordering::lambda>::
  make_trial(random_engine_type& rne, splitting_variables<lo_shower> *vars, double& ratio,
             double& tmU, double tmL, const event<hhc>& p, const lorentzvector<double>& qnull) const
  {
    bool is_possible = false;
    std::pair<int, double> spectator[4];
    double tmT, ratioT;
    
    for(int emitter : {-1, 0})
    {
      //----- possible spectators ----
      unsigned int ns = __mc_helper::counting_color_connections(p, emitter, spectator);
      
      //----- make trial emission with every spectator -----
      for(auto is = spectator; is < spectator+ns; ++is) {
        //----- setting up the probability distribution ------
        if(!_M_parameters(p, false, emitter, is->first, qnull)) continue;
        
        //---- make a trial emission -----
        if(!_M_make_trial(rne, vars, ratioT, tmT, emitter, is->first, tmU, tmL, is->second)) continue;
        
        //----- splitting is possible ----
        tmU = tmT;
        ratio = ratioT;
        is_possible = true;
      }

      /*   For heavy qaurks we have an extra term   */
      if(std::abs(p[emitter].flavor) > 3) {
        for(auto is = spectator; is < spectator+ns; ++is) {
          //----- setting up the probability distribution ------
          if(!_M_parameters(p, true, emitter, is->first, qnull)) continue;
          
          //---- make a trial emission -----
          if(!_M_make_trial(rne, vars, ratioT, tmT, emitter, is->first, tmU, tmL, is->second)) continue;
          
          //----- splitting is possible ----
          tmU = tmT;
          ratio = ratioT;
          is_possible = true;
        }
      }
    }
    
    return is_possible;
  }
  
  
  void splitting_prob<ini, ordering::lambda>::
  initialize(const char *filename, int hdr, double q2min, double s, unsigned int nspar, unsigned int nsvar)
  {
    auto iend = _M_split_approx.end();
    
    /*   if it is not yet initialized then we have to   */
    if(_M_split_approx.find(hdr) == iend) {
      
      /*   neither the anti hadron is initialized   */
      if(_M_split_approx.find(-hdr) == iend) {
        _M_split_approx[hdr] = splitting_kernel_approx<ini>(_M_q2np, q2min, s);
        _M_split_approx[hdr].initialize(filename, _M_split, hdr, nspar, nsvar);
      }
      
      /*   the anti-hadron is initialized  */
      else {
        _M_split_approx[hdr] = _M_split_approx[-hdr];
        _M_split_approx[hdr].anti();
      }
      
    }
  }
  
  
  
}


