/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   project includes   */
#include "bits/hep-lorentzvector.h"


namespace duct {

  void lorentzvector<float>::boost(const float& bx, const float& by, const float& bz) 
  {
    float b2 = bx*bx + by*by + bz*bz;
    float gamma = 1.0f/std::sqrt(1.0f - b2);
    float bp = bx*_M_x + by*_M_y + bz*_M_z;
    float gamma2 = (b2 > 0.0f ? (gamma - 1.0f)/b2 : 0.0f);
	
    _M_x = _M_x + gamma2*bp*bx + gamma*bx*_M_t;
    _M_y = _M_y + gamma2*bp*by + gamma*by*_M_t;
    _M_z = _M_z + gamma2*bp*bz + gamma*bz*_M_t;
    _M_t = gamma*(_M_t + bp);
  }
  
  void lorentzvector<double>::boost(const double& bx, const double& by, const double& bz) 
  {
    double b2 = bx*bx + by*by + bz*bz;
    double gamma = 1.0/std::sqrt(1.0 - b2);
    double bp = bx*_M_x + by*_M_y + bz*_M_z;
    double gamma2 = (b2 > 0.0 ? (gamma - 1.0)/b2 : 0.0);
	
    _M_x = _M_x + gamma2*bp*bx + gamma*bx*_M_t;
    _M_y = _M_y + gamma2*bp*by + gamma*by*_M_t;
    _M_z = _M_z + gamma2*bp*bz + gamma*bz*_M_t;
    _M_t = gamma*(_M_t + bp);
  }

  void lorentzvector<long double>::boost(const long double& bx, const long double& by, const long double& bz) 
  {
    long double b2 = bx*bx + by*by + bz*bz;
    long double gamma = 1.0L/std::sqrt(1.0L - b2);
    long double bp = bx*_M_x + by*_M_y + bz*_M_z;
    long double gamma2 = (b2 > 0.0L ? (gamma - 1.0L)/b2 : 0.0L);
	
    _M_x = _M_x + gamma2*bp*bx + gamma*bx*_M_t;
    _M_y = _M_y + gamma2*bp*by + gamma*by*_M_t;
    _M_z = _M_z + gamma2*bp*bz + gamma*bz*_M_t;
    _M_t = gamma*(_M_t + bp);
  }
  
  
  
  lorentzvector<float> 
  lorentzvector<float>::transverse(const lorentzvector<float>& na, const lorentzvector<float>& nb) const 
  {
    float na2 = na.mag2(), nb2 = nb.mag2(), sab = na*nb;
    float sab2 = sab*sab, xa = na*(*this), xb = nb*(*this);
    lorentzvector<float> retval(*this);
    
    retval *= -1.0;
    retval += sab2/(sab2-na2*nb2)*(na*(xb/sab - xa*nb2/sab2) + nb*(xa/sab - xb*na2/sab2));
    
    return retval;
  }
  
  lorentzvector<double> 
  lorentzvector<double>::transverse(const lorentzvector<double>& na, const lorentzvector<double>& nb) const 
  {
    double na2 = na.mag2(), nb2 = nb.mag2(), sab = na*nb;
    double sab2 = sab*sab, xa = na*(*this), xb = nb*(*this);
    lorentzvector<double> retval(*this);
    
    retval *= -1.0;
    retval += sab2/(sab2-na2*nb2)*(na*(xb/sab - xa*nb2/sab2) + nb*(xa/sab - xb*na2/sab2));
    
    return retval;
  }
  
  lorentzvector<long double> 
  lorentzvector<long double>::transverse(const lorentzvector<long double>& na, const lorentzvector<long double>& nb) const 
  {
    long double na2 = na.mag2(), nb2 = nb.mag2(), sab = na*nb;
    long double sab2 = sab*sab, xa = na*(*this), xb = nb*(*this);
    lorentzvector<long double> retval(*this);
    
    retval *= -1.0;
    retval += sab2/(sab2-na2*nb2)*(na*(xb/sab - xa*nb2/sab2) + nb*(xa/sab - xb*na2/sab2));
    
    return retval;
  }
  
  
}   //  namespace nlo
