/*  Copyright (C) 2010-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 10/5/10.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   standard c++ includes
#include <limits>

//   nlojet++ includes
#include "bits/nlo-mapping.h"



namespace duct {
  
  //========> Some helper fuctions to compute the mappings <=====
  namespace __mc_helper {
    
    /*  Standard lorentz transformation. We use this one in the final 
     *  state splitting. This is the general transformation, for the special
     *  case we have to do something more elaborate.
     */
    void statndard_boost(const lorentzvector<double>& hK, const lorentzvector<double>& K, 
                         statistical_basis::iterator pl, statistical_basis::iterator k1, statistical_basis::iterator kn, 
                         statistical_basis::iterator p1, statistical_basis::iterator pn) 
    {
      //----- do the boost when the spectator sector is not a single massless particle -----
      lorentzvector<double> KK = hK+K;
      double  K2 = K.mag2(), KK2 = KK.mag2();
      double alpha = 0.5*std::sqrt(KK2/K2);
      
      diracalgebra<double> lambda(alpha);
      lambda -= commutator(diracalgebra<double>(K), diracalgebra<double>(hK*(alpha/KK2)));
      
      //----- boosting the non-qcd momenta -----
      for(auto ki = k1; ki < kn; ki++) {
        ki->momentum += hK*(2.0*(K*ki->momentum)/K2) - KK*(2.0*(KK*ki->momentum)/KK2);
        ki->plus = lambda*(ki->plus);
        ki->minus = lambda*(ki->minus);
      }
      
      //----- boosting the qcd momenta -----
      for(auto pi = p1; pi < pn; pi++) 
        if(pi != pl) { 
          pi->momentum += hK*(2.0*(K*pi->momentum)/K2) - KK*(2.0*(KK*pi->momentum)/KK2);
          pi->plus = lambda*(pi->plus);
          pi->minus = lambda*(pi->minus);
        }
    }
    
    
    /*  Lorentz transformation, that takes care about the special case when the 
     *  spectator sector is massles.
     */
    void statndard_boost_fin(const lorentzvector<double>& hK, const lorentzvector<double>& K, 
                             statistical_basis::iterator pl, statistical_basis::iterator k1, statistical_basis::iterator kn, 
                             statistical_basis::iterator p1, statistical_basis::iterator pn)
    {
      unsigned int nnqcd = kn-k1, nqcd = pn-p1;
      
      //----- when the spectator sector consists of one particle the boost is trivial -----
      if(nnqcd + nqcd == 2U && pl >= p1) {
        statistical_basis::iterator pk;
        
        if(nnqcd == 1U) pk = k1;
        else pk = (p1 == pl ? p1+1 : p1);
        
        double m2k = pk->momentum.mag2(), Ek = pk->momentum.T();
        
        if(m2k/(Ek*Ek) < 1.0e-10) {
          double fac = std::sqrt((pl->momentum*hK)/(pk->momentum*K));
          
          pk->momentum = hK;
          pk->plus *= fac;
          pk->minus *= fac;
          return;
        }
      }
      
      //----- do the boost when the spectator sector is not a single massless particle -----
      statndard_boost(hK, K, pl, k1, kn, p1, pn);
    }    
    
    /*  Lorentz transformation in the initial state. One could use the standard 
     *  transformation but this one is a little bit better for the initial state. 
     */    
    void boost_ini(const lorentzvector<double>& na, const lorentzvector<double>& nb, double omega, 
              const lorentzvector<double>& vT, statistical_basis::iterator k1, statistical_basis::iterator kn,
              statistical_basis::iterator p1, statistical_basis::iterator pn)
    {
      double s = na*nb, vT2 = vT.mag2();
      double rs = std::sqrt(2.0*s), so = std::sqrt(omega);
      double alpha, beta, twors = 2.0/rs;
      lorentzvector<double> pT;
      
      //----- spinor representation of the Lorentz transformetion
      diracalgebra<double> slasha((so-1.0/so)/rs*na), slashb(nb/rs), slashT(vT);
      diracalgebra<double> lambda(1.0/so), transverse(1.0);
      
      lambda += slasha*slashb;
      transverse -= slashb*slashT;
      lambda *= transverse; 
     
      //----- boosting the non-qcd momenta -----
      for(auto ki = k1; ki < kn; ki++) {
        alpha = ki->momentum*nb/s;
        beta = ki->momentum*na/s;
        pT = ki->momentum - alpha*na - beta*nb;
        
        beta -= alpha*vT2 + twors*(vT*pT);
        beta /= omega;
        pT += alpha*rs*vT;
        alpha *= omega;
        
        ki->momentum = alpha*na + beta*nb + pT;
        ki->plus = lambda*(ki->plus);
        ki->minus = lambda*(ki->minus);
      }
      
      //----- boosting the qcd momenta -----
      for(auto pi = p1; pi < pn; pi++) {
        alpha = pi->momentum*nb/s;
        beta = pi->momentum*na/s;
        pT = pi->momentum - alpha*na - beta*nb;
        
        beta -= alpha*vT2 + twors*(vT*pT);
        beta /= omega;
        pT += alpha*rs*vT;
        alpha *= omega;
        
        pi->momentum = alpha*na + beta*nb + pT;
        pi->plus = lambda*(pi->plus);
        pi->minus = lambda*(pi->minus);
      }
    }
    
    
#define QUARK statistical_basis::particle::colorlink::quarkend
#define ANTI_QUARK statistical_basis::particle::colorlink::antiquarkend
    
    //        Elementary color operators
    void color_a_dagger(int sigma, int emitter, statistical_basis& p)
    {
      int spectator, m = p.number_of<qcd_out>();
      
      switch(sigma) {
        case 1:
          spectator = p[emitter].color.next;
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = spectator;
          p[spectator].color.prev = m;
          break;
        case -1:
          spectator = p[emitter].color.prev;
          p[spectator].color.next = m;
          p[m].color.prev = spectator;
          p[m].color.next = emitter;
          p[emitter].color.prev = m;
          break;
        case 2:
          spectator = p[emitter].color.prev;
          p[emitter].color.prev  = QUARK;
          p[spectator].color.next  = m;
          p[m].color.prev = spectator;
          p[m].color.next = ANTI_QUARK;
          break;
        case -2:
          p[p[emitter].color.prev].color.next = p[emitter].color.next;
          p[p[emitter].color.next].color.prev = p[emitter].color.prev;
          p[emitter].color.prev = QUARK;
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = ANTI_QUARK;
          break;
        case 3:
          spectator = p[emitter].color.next;
          p[emitter].color.next = ANTI_QUARK;
          p[spectator].color.prev = m;
          p[m].color.next = spectator;
          p[m].color.prev = QUARK;
          break;
        case -3:
          p[p[emitter].color.prev].color.next = p[emitter].color.next;
          p[p[emitter].color.next].color.prev = p[emitter].color.prev;
          p[emitter].color.next = ANTI_QUARK;
          p[emitter].color.prev  = m;
          p[m].color.next = emitter;
          p[m].color.prev = QUARK;
          break;
        case 4:
          p[emitter].color.prev  = m;
          p[m].color.next = emitter;
          p[m].color.prev = QUARK;
          break;
        case -4:
          p[emitter].color.next  = m;
          p[m].color.prev = emitter;
          p[m].color.next = ANTI_QUARK;
          break;
        default:
          //   do nothing
          break;
      } 
    }
    
    void color_a(int sigma, int emitter, statistical_basis& p)
    {
      int spectator, m = p.number_of<qcd_out>();
      
      switch(sigma) {
        case 1:
          spectator = p[emitter].colorP.next;
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = spectator;
          p[spectator].colorP.prev = m;
          break;
        case -1:
          spectator = p[emitter].colorP.prev;
          p[spectator].colorP.next = m;
          p[m].colorP.prev = spectator;
          p[m].colorP.next = emitter;
          p[emitter].colorP.prev = m;
          break;
        case 2:
          spectator = p[emitter].colorP.prev;
          p[emitter].colorP.prev  = QUARK;
          p[spectator].colorP.next  = m;
          p[m].colorP.prev = spectator;
          p[m].colorP.next = ANTI_QUARK;
          break;
        case -2:
          p[p[emitter].colorP.prev].colorP.next = p[emitter].colorP.next;
          p[p[emitter].colorP.next].colorP.prev = p[emitter].colorP.prev;
          p[emitter].colorP.prev = QUARK;
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = ANTI_QUARK;
          break;
        case 3:
          spectator = p[emitter].colorP.next;
          p[emitter].colorP.next = ANTI_QUARK;
          p[spectator].colorP.prev = m;
          p[m].colorP.next = spectator;
          p[m].colorP.prev = QUARK;
          break;
        case -3:
          p[p[emitter].colorP.prev].colorP.next = p[emitter].colorP.next;
          p[p[emitter].colorP.next].colorP.prev = p[emitter].colorP.prev;
          p[emitter].colorP.next = ANTI_QUARK;
          p[emitter].colorP.prev  = m;
          p[m].colorP.next = emitter;
          p[m].colorP.prev = QUARK;
          break;
        case 4:
          p[emitter].colorP.prev  = m;
          p[m].colorP.next = emitter;
          p[m].colorP.prev = QUARK;
          break;
        case -4:
          p[emitter].colorP.next  = m;
          p[m].colorP.prev = emitter;
          p[m].colorP.next = ANTI_QUARK;
          break;
        default:
          //  do nothing
          break;
      } 
    }

    
    /*  Momentum mapping in the final state shower.
     */
    void momentum_mapping_fin(statistical_basis& p, const statistical_basis& q, const splitting_variables<lo_shower>& vars)
    {
      //----- number of partons & particle table -----
      unsigned int np = q.number_of<qcd_out>();
      unsigned int nnqcd = q.number_of<non_qcd>();
      unsigned int ninc = q.number_of<incomings>();
      const model *mdl = q.get_model();
      
      //----- copying the spectator momenta -----
      {
        //----- resize if it is needed ---
        if(p.number_of<non_qcd>() != nnqcd || p.number_of<incomings>() != ninc || p.number_of<qcd_out>() != np+1) 
          p.resize(nnqcd, ninc, np+1);
        
        //----- matching the particle tables -----
        p.set_model(mdl);
        p.color_suppression_index = q.color_suppression_index; 
      
        //----- copying the state q into p ---- 
        statistical_basis::iterator pi = p.begin();
        statistical_basis::const_iterator qi = q.begin();
        
        for(; qi < q.end(); ++pi, ++qi) *pi = *qi;
      }
      
      //----- iterators -----
      statistical_basis::iterator pb = p.begin();
      statistical_basis::iterator p1 = pb+nnqcd+ninc;
      statistical_basis::iterator pn = p1+np, pl = p1+vars.emitter-1;
      
      //----- total incoming momentum -----
      lorentzvector<double> Q = p.momentum_sum<incomings>();
      double pQ = 2.0*(pl->momentum*Q), Q2 = Q.mag2();
      
      //----- al, bl and rl ------
      double al = Q2/pQ, bl = mdl->mass2(pl->flavor)/pQ;
      double rl = std::sqrt(std::abs(1.0-4.0*al*bl)), tmp = (1.0-bl)/(std::sqrt(al)+std::sqrt(std::abs(al+bl-1.0)));
      double ymax = tmp*tmp-bl;
      double b1 = mdl->mass2(vars.zeta.key1)/pQ, b2 = mdl->mass2(vars.zeta.key2)/pQ;
      
      //----- y dependent variables ------
      double y = vars.y;
      double lmd = std::sqrt(std::abs((1.0-y/ymax)*(rl*rl-ymax*y)));
      double x0 = (1.0+y-lmd)/(1.0+y+lmd), Fg = (1.0+y + lmd)*(1.0+rl)/(4.0*al*rl);
      double zp = 1.0, zm  = 0.0;
      
      if(b2 == 0.0) zm = b1/(y+bl);
      else if(b1 == b2) {
        zp = 0.5 + 0.5*std::sqrt(std::abs(1.0-4.0*b1/(y+bl)));
        zm = 1.0-zp;
      } else {
        double yl12 = y+bl+b1-b2, yl = y+bl;
        double sr = std::sqrt(std::abs(yl12*yl12-4.0*yl*b1));
        zp = 0.5*(yl12+sr)/yl;
        zm = 0.5*(yl12-sr)/yl;
      }	  
      
      //----- momenta nl and Pl ----
      lorentzvector<double> nl = Q - (2.0*al/(1.0+rl))*pl->momentum;
      lorentzvector<double> kl = 2.0*al/(1.0+rl)*pl->momentum - (1.0-rl)/(1.0+rl)*Q;
      lorentzvector<double> Pl = Fg*(kl + x0*nl);
      
      //----- boost the spectators -----
      __mc_helper::statndard_boost_fin(Q-Pl, Q-pl->momentum, pl, pb, pb+nnqcd, p1, pn); 
      
      //----- the transverse momentum in the c.m. of Pl -----
      double z = (zp-zm)*vars.z + zm;
      double mT = std::sqrt(std::abs((z-zm)*(zp-z)*x0*2.0*(kl*nl)));
      lorentzvector<double> kT(mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0, 0.0);
        
      //------ boosting the transverse momentum back to the original frame ------
      lorentzvector<double> ql = kl, qk = p[vars.spectator].momentum;
      threevector<double> bvec = Pl.boostVector();
      
      ql.boost(-bvec); 
      double rPhi = ql.phi(), rTheta = ql.theta();
      
      if(vars.spectator != vars.emitter) {
        qk.boost(-bvec);
        qk.rotateZ(-rPhi);
        qk.rotateY(-rTheta);
        kT.rotateZ(qk.phi());
      }
      
      kT.rotateY(rTheta);
      kT.rotateZ(rPhi);
      kT.boost(bvec);
      
      //----- the daughter momenta -----
      pl->flavor = vars.zeta.key1;
      pl->momentum = Fg*(z*kl + (zp+zm-z)*x0*nl - kT);
      pl->plus = diracspinor<double,ket>(pl->momentum, 1);
      pl->minus = diracspinor<double,ket>(pl->momentum, -1);
      
      pn->flavor = vars.zeta.key2;
      pn->momentum = Pl-pl->momentum;
      pn->plus = diracspinor<double,ket>(pn->momentum, 1);
      pn->minus = diracspinor<double,ket>(pn->momentum, -1);
    }
    
    
    /*  Momentum mapping in the initial state shower.
     */
    void momentum_mapping_ini(statistical_basis& p, const lorentzvector<double>& pA, const lorentzvector<double>& pB, 
                              const statistical_basis& q, const splitting_variables<lo_shower>& vars)
    {
      //----- number of partons & particle table -----
      unsigned int np = q.number_of<qcd_out>();
      unsigned int nnqcd = q.number_of<non_qcd>();
      unsigned int ninc = q.number_of<incomings>();
      const model *mdl = q.get_model();
      
      //----- copying the spectator momenta -----
      {
        //----- resize if it is needed ---
        if(p.number_of<non_qcd>() != nnqcd || p.number_of<incomings>() != ninc || p.number_of<qcd_out>() != np+1) 
          p.resize(nnqcd, ninc, np+1);
        
        //----- matching the particle tables -----
        p.set_model(mdl);
        p.color_suppression_index = q.color_suppression_index; 
              
        //----- copying the state q into p ---- 
        statistical_basis::iterator pi = p.begin();
        statistical_basis::const_iterator qi = q.begin();
        
        for(; qi < q.end(); ++pi, ++qi) *pi = *qi;
      }
      
      //----- iterators -----
      statistical_basis::iterator pa = p.begin()+nnqcd;
      statistical_basis::iterator pb = pa+1, p1 = pa+2, pn = p.end()-1;
      
      //----- incoming momentum ----
      double s = pA*pB;
      double eta = (pa->momentum*pB)/s, etb = (pb->momentum*pA)/s;
      lorentzvector<double> na = eta*pA, nb = etb*pB;
      if(vars.emitter == 0) { std::swap(pa,pb); std::swap(na,nb); std::swap(eta,etb);}
      
      //----- parameters -----
      double xa, xb, mT, Q2 = 2.0*eta*etb*s;
      double ma = mdl->mass2(pa->flavor)/Q2;
      double mb = mdl->mass2(pb->flavor)/Q2;
      double ha = mdl->mass2(vars.zeta.key1)/Q2;
      double h1 = mdl->mass2(vars.zeta.key2)/Q2;
      double z = vars.z, y = vars.y;

      //----- calculates the coordinates of the Sudakov parametrization -----
      if(ma+mb+ha+h1 < 1e-11) {
        xa = 1.0/z-1.0-y;
        xb = z*y;
        mT = std::sqrt(xa*xb);
      } else if(mb < 1e-12) {
        xa = 1.0/z-1.0-y;
        xb = z*(y+h1-ma+z*ha*(1.0+y));
        mT = std::sqrt(xa*xb-h1);
      } else {
        double xn = 1.0-z*z*ha*mb;
        
        xa = (1.0/z-1.0 - y*(1.0+z*mb) - (1.0-z)*ma*mb - z*h1*mb)/xn;
        xb = z*(y+h1-ma + z*ha*(1.0+y) + z*ha*mb*(ma-z*ha))/xn;
        mT = std::sqrt(xa*xb-h1);
      }
      
      //----- Now we have every parameters we can define the momenta -----
      lorentzvector<double> kT(mT*std::cos(vars.phi), mT*std::sin(vars.phi), 0.0, 0.0);
      
      //------ azimuthal angle is measured from the color connected parton -----      
      if(vars.spectator != vars.emitter) kT.rotateZ(p[vars.spectator].momentum.phi());
      
      //----- the daughter momenta -----
      pa->flavor = vars.zeta.key1;
      pa->momentum = na/z + (ha*z)*nb;
      pa->plus = diracspinor<double,ket>(pa->momentum, 1);
      pa->minus = diracspinor<double,ket>(pa->momentum, -1);

      pn->flavor = vars.zeta.key2;
      pn->momentum = xa*na + xb*nb + std::sqrt(Q2)*kT;
      pn->plus = diracspinor<double,ket>(pn->momentum, 1);
      pn->minus = diracspinor<double,ket>(pn->momentum, -1);

      //----- boost parameters -----
      double omega = (1.0/z+mb-xa)/(1.0+mb);
      kT /= -1.0-mb;
            
      //----- do the Lorentz boost -----
      __mc_helper::boost_ini(na, nb, omega, kT, p.begin(), p.begin()+nnqcd, p1, pn);      
    }
  }    //  namespace __mc_helper
  
  
  template<>
  void mapping(event<decay>& p, const event<decay>& q, const splitting_variables<lo_shower>& vars) 
  {
    if(vars.emitter > 0) __mc_helper::momentum_mapping_fin(p, q, vars);
    else throw "mapping(event<decay>&,....) : Unable to do the mapping, emitter is not a QCD parton.";
    
    //----- assigning the new color state ----
    __mc_helper::color_a_dagger(vars.sigma, vars.emitter, p);
    __mc_helper::color_a(vars.sigmaP, vars.emitter, p);
    
    //----- issuing the health report for the new state ----
    p.make_health_report();

    //----- healthy gluon, crossed configuration -----
    if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
      p.color_suppression_index += 2U;
    
    //------ gluon splits to quark-antiquark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2)) 
      p.color_suppression_index += 2U;
  }
  
  
  template<>
  void mapping(event<epa>& p, const event<epa>& q, const splitting_variables<lo_shower>& vars) 
  {
    if(vars.emitter > 0) __mc_helper::momentum_mapping_fin(p, q, vars);
    else throw "mapping(event<epa>&,...) : Unable to do the mapping, emitter is not a QCD parton.";
    
    //----- assigning the new color state ----
    __mc_helper::color_a_dagger(vars.sigma, vars.emitter, p);
    __mc_helper::color_a(vars.sigmaP, vars.emitter, p);
    
    //----- issuing the health report for the new state ----
    p.make_health_report();
    
    //----- healthy gluon, crossed configuration -----
    if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
      p.color_suppression_index += 2U;
    
    //------ gluon splits to quark-antiquark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2)) 
      p.color_suppression_index += 2U;
  }
  
  
  template<>
  void mapping(event<dis>& p, const event<dis>& q, const splitting_variables<lo_shower>& vars) 
  {
    if(vars.emitter > 0) 
      __mc_helper::momentum_mapping_fin(p, q, vars);
    else if(vars.emitter == 0)
      __mc_helper::momentum_mapping_ini(p, q[-1].momentum, q[hadron(0)].momentum, q, vars);
    else throw "mapping(event<dis>&,...) : Unable to do the mapping, emitter is not a QCD parton.";
    
    //----- copying the hadron -----
    p[hadron(0)] = q[hadron(0)];
    
    //----- assigning the new color state ----
    __mc_helper::color_a_dagger(vars.sigma, vars.emitter, p);
    __mc_helper::color_a(vars.sigmaP, vars.emitter, p);
    
    //----- issuing the health report for the new state ----
    p.make_health_report();
    
    //----- healthy gluon, crossed configuration -----
    if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
      p.color_suppression_index += 2U;
    
    //------ gluon splits to quark-antiquark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2)) 
      p.color_suppression_index += 2U;

    //------ gluon splits to antiquark-quark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3)) 
      p.color_suppression_index += 2U;
  }
  
  
  template<>
  void mapping(event<hhc>& p, const event<hhc>& q, const splitting_variables<lo_shower>& vars) 
  {
    if(vars.emitter > 0) 
      __mc_helper::momentum_mapping_fin(p, q, vars);
    else if(vars.emitter == -1 || vars.emitter == 0)
      __mc_helper::momentum_mapping_ini(p, q[hadron(-1)].momentum, q[hadron(0)].momentum, q, vars);
    else throw "mapping(event<hhc>&,...) : Unable to do the mapping, emitter is not a QCD parton.";
    
    //----- copying the hadron -----
    p[hadron(-1)] = q[hadron(-1)];
    p[hadron(0)] = q[hadron(0)];
    
    //----- assigning the new color state ----
    __mc_helper::color_a_dagger(vars.sigma, vars.emitter, p);
    __mc_helper::color_a(vars.sigmaP, vars.emitter, p);
    
    //----- issuing the health report for the new state ----
    p.make_health_report();
    
    //----- healthy gluon, crossed configuration -----
    if(q[vars.emitter].health == 'h' && std::abs(vars.sigma) == 1 && vars.sigma == -vars.sigmaP)
      p.color_suppression_index += 2U;
    
    //------ gluon splits to quark-antiquark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -2 || vars.sigmaP == -2)) 
      p.color_suppression_index += 2U;

    //------ gluon splits to antiquark-quark ----
    if(q[vars.emitter].health == 'h' && (vars.sigma == -3 || vars.sigmaP == -3)) 
      p.color_suppression_index += 2U;
  }
  
  
}

