/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 7/4/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <climits>
#include <fstream>

//   project headers
#include "bits/nlo-model.h"


namespace duct {
  
  static const particle_info __particle_table_three_light_quarks[] = {
    {-6,      "-t",      175.0,        -3, 1},
    {-5,      "-b",        4.75,       -3, 1},
    {-4,      "-c",        1.4,        -3, 1},
    {-3,      "-s",        0.0,        -3, 1},
    {-2,      "-u",        0.0,        -3, 1},
    {-1,      "-d",        0.0,        -3, 1},
    { 0,      "gluon",     0.0,         8, 2},
    { 1,      "d",         0.0,         3, 1},
    { 2,      "u",         0.0,         3, 1},
    { 3,      "s",         0.0,         3, 1},
    { 4,      "c",         1.4,         3, 1},
    { 5,      "b",         4.75,        3, 1},
    { 6,      "t",       175.0,         3, 1},
 
    { 11,     "e-",        0.0,         0, 1},
    {-11,     "e+",        0.0,         0, 1},
    { 12,     "nu_e",      0.0,         0, 1},
    {-12,     "-nu_e",     0.0,         0, 1},

    { 13,     "mu-",       0.0,         0, 1},
    {-13,     "mu+",       0.0,         0, 1},
    { 14,     "nu_mu",     0.0,         0, 1},
    {-14,     "-nu_mu",    0.0,         0, 1},

    { 15,     "tau-",      1.77682,     0, 1},
    {-15,     "tau+",      1.77682,     0, 1},
    { 16,     "nu_tau",    0.0,         0, 1},
    {-16,     "-nu_tau",   0.0,         0, 1},
    
    { 21,     "gluon",     0.0,         8, 2},
    { 22,     "photon",    0.0,         0, 2},
    { 23,     "Z0",       91.182,       0, 2},
    { 24,     "W+",       80.4,         0, 2},
    {-24,     "W-",       80.4,         0, 2},
    { 25,     "h0",      125.0,         0, 0},
    {INT_MAX, "",          0.0,         0, 0}
  };
 
  static const hadron_info __hadron_table_some_light_hadron[] = {
    { 2212, "proton", nullptr},
    {-2212, "anti-proton", nullptr},
    { 2112, "neutron", nullptr},
    {INT_MAX, "", nullptr}
  };
  
  
  model::model()
  : model(__particle_table_three_light_quarks, __hadron_table_some_light_hadron) {}
  
  
  model::model(const particle_info *pi, const hadron_info *hi)
  {
    if(pi != nullptr)
      while(pi->key != INT_MAX) {
        particle_table[pi->key] = *pi;
        ++pi;
      }
    
    if(hi != nullptr)
      while(hi->key != INT_MAX) {
        hadron_table[hi->key] = *hi;
        ++hi;
      }
  }
  
  
  void model::write(std::ostream& os) const
  {
    /*   write out the sizes of the tables   */
    unsigned np = particle_table.size(), nh = hadron_table.size();

    os.write(reinterpret_cast<const char *>(&np), sizeof(unsigned));
    os.write(reinterpret_cast<const char *>(&nh), sizeof(unsigned));
    
    /*   writting out the particle table   */
    for(const auto& p : particle_table) {
      os.write(reinterpret_cast<const char *>(&(p.second.key)), sizeof(int));
      os.write(reinterpret_cast<const char *>(&(p.second.mass)), sizeof(double));
      os.write(reinterpret_cast<const char *>(&(p.second.mass2)), sizeof(double));
      os.write(reinterpret_cast<const char *>(&(p.second.color)), sizeof(int));
      os.write(reinterpret_cast<const char *>(&(p.second.spin)), sizeof(int));
      os<<p.second.name<<'\n';
    }
    
    /*   writting out the particle table   */
    for(const auto& p : hadron_table) {
      os.write(reinterpret_cast<const char *>(&(p.second.key)), sizeof(int));
      os<<p.second.name<<'\n';
    }
    
    /*   writting out the alpha_qcd   */
    alpha_qcd.write(os);
  }
  
  
  void model::read(std::istream& is)
  {
    /*   cleaning up   */
    this->particle_table.clear();
    this->hadron_table.clear();
    
    /*   reading the size of tables   */
    unsigned np = 0u, nh = 0u;

    is.read(reinterpret_cast<char *>(&np), sizeof(unsigned));
    is.read(reinterpret_cast<char *>(&nh), sizeof(unsigned));
    
    /*   reading back the particle table   */
    particle_info pi;
    
    for(unsigned k = 0; k < np; k++) {
      /*   reading the particle properties   */
      is.read(reinterpret_cast<char *>(&(pi.key)), sizeof(int));
      is.read(reinterpret_cast<char *>(&(pi.mass)), sizeof(double));
      is.read(reinterpret_cast<char *>(&(pi.mass2)), sizeof(double));
      is.read(reinterpret_cast<char *>(&(pi.color)), sizeof(int));
      is.read(reinterpret_cast<char *>(&(pi.spin)), sizeof(int));
      std::getline(is, pi.name);
      
      /*   inserting it to the table   */
      particle_table[pi.key] = pi;
    }
    
    /*   reading back the hadron table   */
    hadron_info hi;
    
    for(unsigned k = 0; k < nh; k++) {
      /*   reading the hadron properties   */
      is.read(reinterpret_cast<char *>(&(hi.key)), sizeof(int));
      std::getline(is, hi.name);
      
      /*   inserting it to the table   */
      hadron_table[hi.key] = hi;
    }
    
    /*   reading back the alpha_qcd   */
    alpha_qcd.read(is);
  }
  
  
  
  double model::alpha_qed(double Q2) const
  {
    double rpigg, alem = 1.0/137.04;
    double alem3pi = alem/(3.0*M_PI);
    
    if(Q2 <= 2e-6) rpigg = 0.0;
    else if(Q2 < 0.09) rpigg = alem3pi*(13.4916 + std::log(Q2)) + 0.00835*std::log(1.0+Q2);
    else if(Q2 < 9.0) rpigg = alem3pi*(16.3200 + 2.0*std::log(Q2)) + 0.00238*std::log(1.0+3.927*Q2);
    else if (Q2 < 10000.0) rpigg = alem3pi*(13.4955 + 3.0*std::log(Q2)) + 0.00165 + 0.00299*std::log(1.0 + Q2);
    else rpigg = alem3pi*(13.4955 + 3.0*std::log(Q2)) + 0.00221 + 0.00293*std::log(1.0 + Q2);
    
    return alem/(1.0-rpigg);
  }

}




