/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 11/8/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>

//   project headers
#include "bits/qcd-dglap.h"


namespace duct {
  
  namespace __dglap_helper {
    
    static std::pair<double,double> prob_kT_qq(double z, double xi)
    {
      double oz = 1.0-z, rxi = std::sqrt(xi), oxi = 1.0+xi;
      double loz = z+1e-15 > 1.0 ? 0.0 : 2.0*std::log(oz);
      
      double r1 = -1.0/(1.0 + oz*oz*xi) - std::atan(-oz*rxi)/(rxi*oxi) - loz
      + std::log(z)/oxi + 0.5*(1.0+2.0*xi)*std::log(1.0 + oz*oz*xi)/oxi;
      
      double rz = -z/(1.0 + oz*oz*xi) - std::atan(-oz*rxi)/rxi - loz
      - 0.5*(1.0-2.0*xi)*std::log(1.0 + oz*oz*xi)/xi;
      
      return {Cf*r1, Cf*rz};
    }
    
    static std::pair<double,double> prob_kT_gq(double z, double xi)
    {
      double rxi = std::sqrt(xi);
      
      double r1 = -2.0/z - (1.0+z*xi)/(1.0+z*z*xi) + (1.0-3.0*xi)*std::atan(z*rxi)/rxi
      - 2.0*std::log(z) + std::log(1.0+z*z*xi);
      
      double rz =  (1.0-z)/(1.0+z*z*xi) - std::atan(z*rxi)/rxi + 2.0*std::log(z)
      + 0.5*(1.0-2.0*xi)*std::log(1.0+z*z*xi)/xi;
      
      return {Cf*r1, Cf*rz};
    }
    
    static std::pair<double,double> prob_kT_qg(double z, double xi)
    {
      double oxi = 1.0+xi;
      double r1 = (-2.0*z + z*z + oxi*std::log(z))/(oxi*oxi);
      double rz = z*(oxi - z + 2.0*z*z/3.0)/(oxi*oxi);
      
      return {0.5*r1, 0.5*rz};
    }
    
    
    
    inline double s1_kT_qq(double u, double v, double xi)
    {
      auto pu = prob_kT_qq(u, xi);
      auto pv = prob_kT_qq(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double  s2_kT_qq(double u, double v, double xi)
    {
      auto pu = prob_kT_qq(u, xi);
      auto pv = prob_kT_qq(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    inline double s1_kT_gq(double u, double v, double xi)
    {
      auto pu = prob_kT_gq(u, xi);
      auto pv = prob_kT_gq(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double  s2_kT_gq(double u, double v, double xi)
    {
      auto pu = prob_kT_gq(u, xi);
      auto pv = prob_kT_gq(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    inline double s1_kT_qg(double u, double v, double xi)
    {
      auto pu = prob_kT_qg(u, xi);
      auto pv = prob_kT_qg(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double s2_kT_qg(double u, double v, double xi)
    {
      auto pu = prob_kT_qg(u, xi);
      auto pv = prob_kT_qg(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    static double s3_kT_qq(double u, double xi)
    {
      double ou = 1.0-u, rxi = std::sqrt(xi), oxi = 1.0+xi;
      double lou = std::log(1.0-u), loxi = std::log(oxi);
      
      return Cf*(2.0*lou + loxi*(1.0-0.5/xi) + std::atan(rxi)/rxi - (ou+xi)*std::atan(ou*rxi)/(ou*oxi*rxi)
                 + u*std::log(u)/(ou*oxi) + (1.0-ou*xi*(1.0+2.0*xi))*std::log(1.0+ou*ou*xi)/(2.0*ou*oxi*xi));
    }
    
  }   //   namespace __dglap_helper
    
    
  void qcd_dglap::_M_weights_kT(unsigned int i, unsigned int k, unsigned int it, qcd_dglap::kernel& res)
  {
    using namespace __dglap_helper;
    
    /*   calculate the scale   */
    double q2 = _M_alpha_qcd.tauk_inverse(_M_t.back()-_M_t[it]);
    double xic = _M_m2c/q2, xib = _M_m2b/q2;
    double sp = _M_x[i]/_M_x[k+1];
    
    //----- set all elements of the kernel to zero -----
    res = kernel(0.0);
    
    if(i < k) {
      double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
      
      //----- gluon-gluon vertex -----
      res[0][0] = _M_s1gg(sp, s) + _M_s2gg(s, sm);
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, s) + _M_s2gq(s, sm);
      res[0][2] = s1_kT_gq(sp, s, xic) + s2_kT_gq(s, sm, xic);
      res[0][3] = s1_kT_gq(sp, s, xib) + s2_kT_gq(s, sm, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*(_M_s1qg(sp, s) + _M_s2qg(s, sm));
      res[2][0] = 2.0*(s1_kT_qg(sp, s, xic) + s2_kT_qg(s, sm, xic));
      res[3][0] = 2.0*(s1_kT_qg(sp, s, xib) + s2_kT_qg(s, sm, xib));
      
      //----- quark-quark vertices -----
      res[1][1] = _M_s1qq(sp, s) + _M_s2qq(s, sm);
      res[2][2] = s1_kT_qq(sp, s, xic) + s2_kT_qq(s, sm, xic);
      res[3][3] = s1_kT_qq(sp, s, xib) + s2_kT_qq(s, sm, xib);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- diagonal elements -----
    else if(i == k) {
      //----- gluon-gluon vertex -----
      double nf = 3.0;
      nf += (1.0+1.5*xic)/(1.0+xic)/(1.0+xic);
      nf += (1.0+1.5*xib)/(1.0+xib)/(1.0+xib);
      
      res[0][0] = _M_s3gg(sp) + 11.0*Ca/6.0 - nf/3.0;
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, 1.0);
      res[0][2] = s1_kT_gq(sp, 1.0, xic);
      res[0][3] = s1_kT_gq(sp, 1.0, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*_M_s1qg(sp, 1.0);
      res[2][0] = 2.0*s1_kT_qg(sp, 1.0, xic);
      res[3][0] = 2.0*s1_kT_qg(sp, 1.0, xib);
      
      //----- quark-quark vertices -----
      res[1][1] = _M_s3qq(sp);
      res[2][2] = s3_kT_qq(sp, xic);
      res[3][3] = s3_kT_qq(sp, xib);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- we shouldn't call this function with k < i -----
    else throw "pdf::weights_kT(...) : We have some trouble...";
  }

}
