/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/6/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>

//   project headers
#include "bits/qcd-dglap.h"


namespace duct {
  
  namespace __dglap_helper {
    
    
    static double s1_lmd_qq(double u, double v, double xi)
    {
      double u1 = std::max(u, 1.0-1.0/xi);
      if(u1 > v) return 0.0;
      
      return Cf*v*(0.5*((u1-v)*(2.0-2.0*u+u1+v+2.0*(u1-2.0*u+v)*xi))
                   -2.0*(1.0-u)*std::log((1.0-v)/(1.0-u1)) + u*std::log(u1/v))/(v-u);
    }
    
    static double s2_lmd_qq(double u, double v, double xi)
    {
      double u1 = std::max(u, 1.0-1.0/xi);
      if(u1 > v) return 0.0;
      
      if(v+1e-15 > 1.0) return 0.5*Cf*u*((u1-v)*(2.0+u1-v+2.0*xi*(u1-v))+2.0*v*std::log(u1/v))/(u-v);
      return 0.5*Cf*u*((u1-v)*(2.0+u1-v+2.0*xi*(u1-v))+4.0*(1.0-v)*std::log((1.0-u1)/(1.0-v))+2.0*v*std::log(u1/v))/(u-v);
    }
    
    static double s3_lmd_qq(double u, double xi)
    {
      double u1 = std::max(u, 1.0-1.0/xi);
      double ou1 = 1.0-u1, ou = 1.0-u;
      return Cf*(2.0*std::log(ou1) + u/ou*std::log(u1) - ou1/ou*(1.0 + 0.5*(1.0-2.0*u+u1)*(1.0+2.0*xi)));
    }
    
    
    static double s1_lmd_qg(double u, double v, double xi)
    {
      double v1 = v < 1.0-xi ? v : 1.0-xi;
      if(u > 1.0-xi) return 0.0;
      return 0.5*v*((u-v1)*(u*u-3.0 + v1*(3.0-2.0*v1-3.0*xi) + u*(v1-3.0+3.0*xi))/3.0 - u*std::log(v1/u))/(v-u);
    }
    
    static double s2_lmd_qg(double u, double v, double xi)
    {
      double v1 = v < 1.0-xi ? v : 1.0-xi;
      if(u > 1.0-xi) return 0.0;
      return 0.5*u*((u-v1)*(3.0+2.0*u*u+v1*(2.0*v1-3.0+3.0*xi)-3.0*v*(v1+2.0*xi-2.0)
                            + u*(2*v1-3.0*(1.0+v-xi)))/3.0 + v*std::log(v1/u))/(v-u);
    }
    
    
    static double s1_lmd_gq(double u, double v, double xi)
    {
      double u1 = std::max(u, 2.0*xi/(1.0+std::sqrt(1.0+4.0*xi)));
      if(u1 > v) return 0.0;
      
      return Cf*((u1-v)*(4.0*u1*v + u*(4.0+2.0*u1*v*(1.0-2.0*xi))-u1*v*(u1+v)*(1.0-2.0*xi))
                 + 4.0*(1.0+u)*u1*v*std::log(v/u1))/(2.0*u1*(v-u));
    }
    
    static double s2_lmd_gq(double u, double v, double xi)
    {
      double u1 = std::max(u, 2.0*xi/(1.0+std::sqrt(1.0+4.0*xi)));
      if(u1 > v) return 0.0;
      
      return Cf*(u*((v-u1)*(4.0+u1*u1*(2.0*xi-1.0) + u1*(4.0+v-2.0*v*xi))
                    -4.0*u1*(1.0+v)*std::log(v/u1)))/(2.0*u1*(v-u));
    }
    
  }  //   namespace __dglap_helper


  
  void qcd_dglap::_M_weights_lambda(unsigned int i, unsigned int k, unsigned int it, qcd_dglap::kernel& res)
  {
    using namespace __dglap_helper;
    
    /*   calculate the scale   */
    double q2 = _M_alpha_qcd.tauk_inverse(_M_t.back()-_M_t[it]);
    double xic = _M_m2c/q2, xib = _M_m2b/q2;
    double sp = _M_x[i]/_M_x[k+1];
    
    //----- set all elements of the kernel to zero -----
    res = kernel(0.0);
    
    if(i < k) {
      double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
      
      //----- gluon-gluon vertex -----
      res[0][0] = _M_s1gg(sp, s) + _M_s2gg(s, sm);
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, s) + _M_s2gq(s, sm);
      res[0][2] = s1_lmd_gq(sp, s, xic) + s2_lmd_gq(s, sm, xic);
      res[0][3] = s1_lmd_gq(sp, s, xib) + s2_lmd_gq(s, sm, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*(_M_s1qg(sp, s) + _M_s2qg(s, sm));
      res[2][0] = 2.0*(s1_lmd_qg(sp, s, xic) + s2_lmd_qg(s, sm, xic));
      res[3][0] = 2.0*(s1_lmd_qg(sp, s, xib) + s2_lmd_qg(s, sm, xib));
      
      //----- quark-quark vertices -----
      res[1][1] = _M_s1qq(sp, s) + _M_s2qq(s, sm);
      res[2][2] = s1_lmd_qq(sp, s, xic) + s2_lmd_qq(s, sm, xic);
      res[3][3] = s1_lmd_qq(sp, s, xib) + s2_lmd_qq(s, sm, xib);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- diagonal elements -----
    else if(i == k) {
      //----- gluon-gluon vertex -----
      double nf = 3.0;
      if(1.0 > 4.0*xic) nf += (1.0+2.0*xic)*std::sqrt(1.0-4.0*xic);
      if(1.0 > 4.0*xib) nf += (1.0+2.0*xib)*std::sqrt(1.0-4.0*xib);
      
      res[0][0] = _M_s3gg(sp) + 11.0*Ca/6.0 - nf/3.0;
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, 1.0);
      res[0][2] = s1_lmd_gq(sp, 1.0, xic);
      res[0][3] = s1_lmd_gq(sp, 1.0, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*_M_s1qg(sp, 1.0);
      res[2][0] = 2.0*s1_lmd_qg(sp, 1.0, xic);
      res[3][0] = 2.0*s1_lmd_qg(sp, 1.0, xib);
      
      //----- quark-quark vertices -----
      double oxic = 1.0+xic, oxib = 1.0+xib;
      double lxc = std::log(oxic), lxb = std::log(oxib);
      
      res[1][1] = _M_s3qq(sp);
      res[2][2] = s3_lmd_qq(sp, xic) + (2.0-0.5/(oxic*oxic) + 2.0*lxc)*Cf;
      res[3][3] = s3_lmd_qq(sp, xib) + (2.0-0.5/(oxib*oxib) + 2.0*lxb)*Cf;
            
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- we shouldn't call this function with k < i -----
    else throw "pdf::weights_lambda(...) : We have some trouble...";
  }
  

  


}


