/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/6/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>
#include <complex>

//   project headers
#include "bits/qcd-dglap.h"


namespace duct {
  
  
  namespace __dglap_helper {
    
    
    
    static std::pair<double,double> prob_lmdT_qq(double z, double xi)
    {
      double oz = 1.0-z, oxi = 1.0+xi;
      double loz = z+1e-15 > 1.0 ? 0.0 : 2.0*std::log(oz);
      
      double r1 = -2.0/(1.0+xi*oz) - loz + (1.0/(xi*oxi)+2.0)*std::log(1.0+xi*oz) + std::log(z)/oxi;
      double rz = -oz/xi - 2.0*oxi/(xi*(1.0+xi*oz)) - loz + (1.0/(xi*xi)+2.0)*std::log(1.0+oz*xi);
      
      return {Cf*r1, Cf*rz};
    }
    
    static std::pair<double,double> prob_lmdT_gq(double z, double xi)
    {
      /*   at xi = 0.25 the expressions below are not well defined   */
      if(xi == 0.25) xi = 0.2500001;
      
      /*   calculate the primitiv functions   */
      double xi2 = xi*xi, z2 = z*z, loz = std::log(1.0-z+xi*z*z);
      double rxi = std::sqrt(std::abs(1.0-4.0*xi)), atg;
      
      double r1 = -2.0/z + z - 2.0*std::log(z);
      double rz = -2.0*z + 0.5*z*z + 2.0*std::log(z);
      
      if(xi > 0.00001) {
        if(xi < 0.25) {
          atg = 0.5*std::log((1.0-2.0*xi*z/(1.0+rxi))/(1.0-0.5*z*(1.0+rxi)))/rxi;
        } else atg = (std::atan(1.0/rxi) - std::atan((1.0-2.0*xi*z)/rxi))/rxi;
        
        r1 += -z - (0.5/xi-1.0)*loz - (1.0/xi - 8.0 + xi*(20.0-24.0*xi))*atg/(1.0-4.0*xi)
        - 2.0*z*xi*((1.0-z)*(1.0-2.0*xi)+xi*z)/((1.0-4.0*xi)*(1.0-z+xi*z2));
        
        rz += -loz*(1.0-xi+2.0*xi2)/(2.0*xi2)
        - atg*(1.0/xi2-7.0/xi+(14.0-12.0*xi))/(1.0-4.0*xi)
        -0.5*z*(2.0*(1.0-z)/xi - 4.0*xi2*z*(2.0-4.0*z+z2) - 8.0+9.0*z+z2
                +xi*(4.0*(1.0-z)-(8.0-z)*z2))/((1.0-4.0*xi)*(1.0-z+xi*z2));
      }
      
      return {Cf*r1, Cf*rz};
    }
    
    static std::pair<double,double> prob_lmdT_pqg(double z, double xi)
    {
      double oxi = 1.0+xi, xi2 = xi*xi, z2 = z*z;
      double loz = std::log(oxi-z), xi3 = 2.0*xi2*xi;
      
      double r1 = xi3/(oxi-z) - z*(2.0-4.0*xi-z) + std::log(z)/oxi + xi*(1.0/oxi + 6.0*xi)*loz;
      double rz = xi3*oxi/(oxi-z) + z + 6.0*xi2*z - (1.0-2.0*xi-2.0*z/3.0)*z2 + xi*(1.0+6.0*xi+8.0*xi2)*loz;
      
      return {0.5*r1, 0.5*rz};
    }
  
  
    inline double s1_lmdT_qq(double u, double v, double xi)
    {
      auto pu = prob_lmdT_qq(u, xi);
      auto pv = prob_lmdT_qq(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double s2_lmdT_qq(double u, double v, double xi)
    {
      auto pu = prob_lmdT_qq(u, xi);
      auto pv = prob_lmdT_qq(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    inline double s1_lmdT_gq(double u, double v, double xi)
    {
      auto pu = prob_lmdT_gq(u, xi);
      auto pv = prob_lmdT_gq(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double s2_lmdT_gq(double u, double v, double xi)
    {
      auto pu = prob_lmdT_gq(u, xi);
      auto pv = prob_lmdT_gq(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    inline double s1_lmdT_qg(double u, double v, double xi)
    {
      auto pu = prob_lmdT_pqg(u, xi);
      auto pv = prob_lmdT_pqg(v, xi);
      
      return v*(pv.second-pu.second - u*(pv.first-pu.first))/(v-u);
    }
    
    inline double s2_lmdT_qg(double u, double v, double xi)
    {
      auto pu = prob_lmdT_pqg(u, xi);
      auto pv = prob_lmdT_pqg(v, xi);
      
      return u*(v*(pv.first-pu.first) - (pv.second-pu.second))/(v-u);
    }
    
    
    static double s3_lmdT_qq(double u, double xi)
    {
      double ou = 1.0-u, xi2 = xi*xi, oxi = 1.0+xi;
    
      return Cf*(1.0/xi-2.0 + 2.0*std::log(ou) + u*std::log(u)/(ou*oxi)
                 - (1.0 + ou*xi*(1.0+2.0*oxi*xi))*std::log(1.0+xi*ou)/(ou*oxi*xi2));
    }
    
    static double gamma_lmdT_q(double xi)
    {
      double oxi = 1.0-xi;
      return Cf*(3.0 - xi*(4.0 - (5.0-4.0*xi)*xi) - 2.0*xi*(1.0-2.0*(3.0-xi)*xi)*std::log(xi))/(2.0*oxi*oxi*oxi);
    }
    
    static double nf_lmdT(double xi)
    {
      double rxi = std::sqrt(1.0+4.0*xi);
      double acth = std::log((rxi+1.0)/(rxi-1.0))/rxi;
      
      return (1.0 + 10.0*xi*(1.0+3.0*xi) - 3.0*xi*(1.0+10.0*xi*(1.0+2.0*xi))*acth)/(1.0+4.0*xi);
    }
    
  }    //    namespace __dglap_helper

  
  void qcd_dglap::_M_weights_lmdT(unsigned int i, unsigned int k, unsigned int it, qcd_dglap::kernel& res)
  {
    using namespace __dglap_helper;
    
    /*   calculate the scale   */
    double q2 = _M_alpha_qcd.tauk_inverse(_M_t.back()-_M_t[it]);
    double xic = _M_m2c/q2, xib = _M_m2b/q2;
    double sp = _M_x[i]/_M_x[k+1];
    
    //----- set all elements of the kernel to zero -----
    res = kernel(0.0);
    
    if(i < k) {
      double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
      
      //----- gluon-gluon vertex -----
      res[0][0] = _M_s1gg(sp, s) + _M_s2gg(s, sm);
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, s) + _M_s2gq(s, sm);
      res[0][2] = s1_lmdT_gq(sp, s, xic) + s2_lmdT_gq(s, sm, xic);
      res[0][3] = s1_lmdT_gq(sp, s, xib) + s2_lmdT_gq(s, sm, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*(_M_s1qg(sp, s) + _M_s2qg(s, sm));
      res[2][0] = 2.0*(s1_lmdT_qg(sp, s, xic) + s2_lmdT_qg(s, sm, xic));
      res[3][0] = 2.0*(s1_lmdT_qg(sp, s, xib) + s2_lmdT_qg(s, sm, xib));
      
      //----- quark-quark vertices -----
      res[1][1] = _M_s1qq(sp, s) + _M_s2qq(s, sm);
      res[2][2] = s1_lmdT_qq(sp, s, xic) + s2_lmdT_qq(s, sm, xic);
      res[3][3] = s1_lmdT_qq(sp, s, xib) + s2_lmdT_qq(s, sm, xib);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- diagonal elements -----
    else if(i == k) {
      //----- gluon-gluon vertex -----
      res[0][0] = _M_s3gg(sp) + 11.0*Ca/6.0 - (3.0 + nf_lmdT(xic) + nf_lmdT(xib))/3.0;
      
      //----- gluon-quark vertices -----
      res[0][1] = _M_s1gq(sp, 1.0);
      res[0][2] = s1_lmdT_gq(sp, 1.0, xic);
      res[0][3] = s1_lmdT_gq(sp, 1.0, xib);
      
      //----- quark-gluon vertices -----
      res[1][0] = 6.0*_M_s1qg(sp, 1.0);
      res[2][0] = 2.0*s1_lmdT_qg(sp, 1.0, xic);
      res[3][0] = 2.0*s1_lmdT_qg(sp, 1.0, xib);
      
      //----- quark-quark vertices -----
      res[1][1] = _M_s3qq(sp);
      res[2][2] = s3_lmdT_qq(sp, xic) + gamma_lmdT_q(xic);
      res[3][3] = s3_lmdT_qq(sp, xib) + gamma_lmdT_q(xib);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- we shouldn't call this function with k < i -----
    else throw "pdf::weights_lmdT(...) : We have some trouble...";
  }
  
  
  
}
