/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/11/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>
#include <set>

//   project headers
#include "bits/qcd-dglap.h"


namespace duct {
  
  qcd_dglap::array qcd_dglap::kernel::operator*(const qcd_dglap::array& b) const
  {
    qcd_dglap::array res;
    
    //---- singlet ----
    for(unsigned int i = 0U; i < 4U; i++)
      for(unsigned int j = 0U; j < 4U; j++)
        res[i] += _M_elem[4U*i+j]*b[j];
    
    //----- non-singlet part ----
    for(unsigned int i = 4U; i < 9U; i++)
      res[i] = _M_elem[16]*b[i];
    
    return res;
  }
  
  qcd_dglap::kernel qcd_dglap::kernel::operator*(const qcd_dglap::kernel& b) const
  {
    qcd_dglap::kernel res(0.0);
    
    //----- singlet -----
    for(unsigned int i = 0U; i < 4U; i++)
      for(unsigned int j = 0U; j < 4U; j++)
        for(unsigned int k = 0U; k < 4U; k++)
          res._M_elem[4U*i+j] += _M_elem[4U*i+k]*b._M_elem[4U*k+j];
    
    //----- non-singlet ----
    res._M_elem[16] = _M_elem[16]*b._M_elem[16];
    
    return res;
  }
  
  qcd_dglap::kernel inverse(const qcd_dglap::kernel& k)
  {
    const double *a = k[0];
    double det = a[0]*a[10]*a[15]*a[5]-a[1]*a[10]*a[15]*a[4]-a[10]*a[12]*a[3]*a[5]-a[15]*a[2]*a[5]*a[8];
    qcd_dglap::kernel res;
    
    res[0][0] = a[10]*a[15]*a[5];
    res[0][1] = -a[1]*a[10]*a[15];
    res[0][2] = -a[15]*a[2]*a[5];
    res[0][3] = -a[10]*a[3]*a[5];
    
    res[1][0] = -a[10]*a[15]*a[4];
    res[1][1] = a[0]*a[10]*a[15]-a[10]*a[12]*a[3]-a[15]*a[2]*a[8];
    res[1][2] = a[15]*a[2]*a[4];
    res[1][3] = a[10]*a[3]*a[4];
    
    res[2][0] = -a[15]*a[5]*a[8];
    res[2][1] = a[1]*a[15]*a[8];
    res[2][2] = -a[1]*a[15]*a[4]+a[0]*a[15]*a[5]-a[12]*a[3]*a[5];
    res[2][3] = a[3]*a[5]*a[8];
    
    res[3][0] = -a[10]*a[12]*a[5];
    res[3][1] = a[1]*a[10]*a[12];
    res[3][2] = a[12]*a[2]*a[5];
    res[3][3] = -a[1]*a[10]*a[4]+a[0]*a[10]*a[5]-a[2]*a[5]*a[8];
    
    res /= det;
    res[4][0] = 1.0/a[16];
    
    return res;
  }
  

  template<typename RandomIterator>
  static inline unsigned int __mstw_locate(RandomIterator xb, RandomIterator xe, double x)
  {
    auto n = xe-xb;
    int ju = n, jm, jl = 0;
    
    while(ju-jl > 1) {
      jm=(ju+jl)/2;
      if(x >= xb[jm]) jl = jm;
      else ju = jm;
    }
    
    return static_cast<unsigned>(jl == n-1 ? jl-1 : jl);
  }
  
  double qcd_dglap::thrs(int fl, double x) const
  {
    switch(_M_type) {
      case msbar:
        switch(fl) {
          case -4: case 4: case 12: return _M_m2c; break;
          case -5: case 5: case 13: return _M_m2b; break;
          default: return 0.0;
        }
        break;
      case lambda:
        switch(fl) {
          case -4: case 4: case 12: return x < 1.0 ? _M_m2c/(1.0-x) : 1e+100; break;
          case -5: case 5: case 13: return x < 1.0 ? _M_m2b/(1.0-x) : 1e+100; break;
        }
        break;
      default:
        return 0.0;
    }
    
    return 0.0;
  }
  

  double qcd_dglap::parton(int fl, double x, double q2) const
  {
    //----- trivial case -----
    double xmx = xmax(fl, q2), thr2 = thrs(fl, x);
    if(std::abs(fl) == 6 || x >= xmx || q2 <= thr2) return 0.0;
    
    //---- finding the x, t interval ----
    double t = _M_t.back()-_M_alpha_qcd.tauk(q2);
    unsigned l = __mstw_locate(_M_x.begin(), _M_x.end(), x);
    unsigned k = __mstw_locate(_M_t.begin(), _M_t.end(), t);
    
    //----- the of the function in the corners of the interpolation square -----
    double f[] = {
      _M_parton(fl,l,k), _M_parton(fl,l,k+1),
      _M_parton(fl,l+1,k), _M_parton(fl,l+1,k+1)
    };
    
    //----- normalized variables -----
    double xx = (_M_x[l+1]-x)/(_M_x[l+1]-_M_x[l]);
    double tt = (_M_t[k+1]-t)/(_M_t[k+1]-_M_t[k]);
    
    /*   When every point is well defined
     *   we intepolate the log of the function.
     */
    if(f[0] > 0.0 && f[1] > 0.0 && f[2] > 0.0 && f[3] > 0.0) {
      //----- interpolation in variable t -----
      double h1 = std::log(f[0])*tt + std::log(f[1])*(1.0-tt);
      double h2 = std::log(f[2])*tt + std::log(f[3])*(1.0-tt);
      
      //----- interpolation in variable x -----
      return std::exp(h1*xx + h2*(1.0-xx));
    }
    
    //----- interpolation in variable t -----
    double h1 = f[0]*tt + f[1]*(1.0-tt);
    double h2 = f[2]*tt + f[3]*(1.0-tt);
    
    //----- interpolation in variable x -----
    return h1*xx + h2*(1.0-xx);
  }

  double qcd_dglap::momentum(int fl, double q2) const
  {
    //----- trivial case -----
    if(std::abs(fl) == 6 || q2 <= thrs(fl)) return 0.0;
    
    //---- finding the x, t interval ----
    double t = _M_t.back()-_M_alpha_qcd.tauk(q2);
    unsigned k = __mstw_locate(_M_t.begin(), _M_t.end(), t);
    
    //----- the of the function in the corners of the interpolation square -----
    double h1 = 0.0, h2 = 0.0, xi1, xi2;
    
    for(unsigned i = 1; i < _M_x.size(); i++) {
      xi1 = std::log(_M_x[i-1]);
      xi2 = std::log(_M_x[i]);
      
      h1 += 0.5*(xi2-xi1)*(_M_x[i-1]*_M_parton(fl,i-1,k) + _M_x[i]*_M_parton(fl,i,k));
      h2 += 0.5*(xi2-xi1)*(_M_x[i-1]*_M_parton(fl,i-1,k+1) + _M_x[i]*_M_parton(fl,i,k+1));
    }
    
    //----- normalized variables -----
    double tt = (_M_t[k+1]-t)/(_M_t[k+1]-_M_t[k]);
    
    //----- interpolation in variable x -----
    return h1*tt + h2*(1.0-tt);
  }
  
  
  void qcd_dglap::write(std::ostream& os) const
  {
    unsigned int nx = _M_x.size(), nt = _M_t.size();
    
    os.write((const char *) &_M_type, sizeof(mode));
    os.write((const char *) &_M_q2min, sizeof(double));
    os.write((const char *) &_M_m2c, sizeof(double));
    os.write((const char *) &_M_m2b, sizeof(double));
    os.write((const char *) &nx, sizeof(unsigned int));
    os.write((const char *) &nt, sizeof(unsigned int));
    
    os.write((const char *) _M_x.data(), nx*sizeof(double));
    os.write((const char *) _M_t.data(), nt*sizeof(double));
    
    for(const array& x : _M_h) os.write((const char *) x._M_elem, 9U*sizeof(double));
    
    _M_alpha_qcd.write(os);
  }
  
  void qcd_dglap::read(std::istream& os)
  {
    unsigned int nx, nt;
    
    os.read((char *) &_M_type, sizeof(mode));
    os.read((char *) &_M_q2min, sizeof(double));
    os.read((char *) &_M_m2c, sizeof(double));
    os.read((char *) &_M_m2b, sizeof(double));
    os.read((char *) &nx, sizeof(unsigned int));
    os.read((char *) &nt, sizeof(unsigned int));
    
    _M_x.resize(nx);
    _M_t.resize(nt);
    
    os.read((char *) _M_x.data(), nx*sizeof(double));
    os.read((char *) _M_t.data(), nt*sizeof(double));
    
    _M_h.resize(nx, nt);
        
    for(array& x : _M_h) os.read((char *) x._M_elem, 9U*sizeof(double));

    _M_alpha_qcd.read(os);
  }
  

  double qcd_dglap::_M_parton(int fl, unsigned ix, unsigned it) const
  {
    const array& h = _M_h[ix][it];
    
    if(fl == 20) return h[0] + h[1] + h[2] + h[3];
    if(fl == 19) return h[1] + h[2] + h[3];
    if(fl >= 10) return h[fl-10];
    
    switch(fl) {
        //----- b and c quarks -----
      case -5: case 5: return 0.5*h[3]; break;
      case -4: case 4: return 0.5*h[2]; break;
        
        //----- strange quark -----
      case -3: return (h[1] - h[8] - 3.0*h[6])/6.0; break;
      case  3: return (h[1] - h[8] + 3.0*h[6])/6.0; break;
        
        //----- up quark -----
      case -2: return (2.0*h[1] + h[8] + 3.0*h[7] - 6.0*h[5])/12.0; break;
      case  2: return (2.0*h[1] + h[8] + 3.0*h[7] + 6.0*h[5])/12.0; break;
        
        //----- down quark -----
      case -1: return (2.0*h[1] + h[8] - 3.0*h[7] - 6.0*h[4])/12.0; break;
      case  1: return (2.0*h[1] + h[8] - 3.0*h[7] + 6.0*h[4])/12.0; break;
        
        //----- gluon -----
      case  0: return h[0]; break;
    }
    
    return 0.0;
  }

  
  void qcd_dglap::_M_xgrid()
  {
    /*  These parameters are hard wired. I don't think
     *  that the users should change them. We can set up
     *  a large grid over a big interval and that should
     *  be OK for every processes.
     */
    constexpr double dx = 0.0075;
    constexpr double xmin = 1e-8, xmax = 0.9999;
    
    //----- Don't change beyond this point! -----
    double lx = std::log(xmin), lox = std::log(1.0-xmax);
    double dlx = -dx*lx, dlox = dx*lox;
    
    unsigned int k = 0U;
    double xk, x1 = xmin;
    std::set<double> xx({1.0});
    
    while((xk = xmin*std::exp((k++)*dlx))-x1 < dx)
      xx.insert(x1 = xk);
    
    k = 0U;
    double x2 = xmax;
    while(x2-(xk = 1.0-std::exp(lox-(k++)*dlox)) < dx)
      xx.insert(x2 = xk);
    
    unsigned int nx = 1 + (x2-x1)/dx;
    double dmx = (x2-x1)/nx;
    
    for(k = 1; k < nx; k++)
      xx.insert(dmx*k+x1);
    
    //------ copy the basepoint to the _M_x array -----
    _M_x.assign(xx.cbegin(), xx.cend());
  }
  

  void qcd_dglap::_M_tgrid()
  {
    /*  These parameters are hard wired. I don't think
     *  that the users should change them. We can set up
     *  a large grid over a big interval and that should
     *  be OK for every processes.
     */
    constexpr unsigned int nt = 200U;
    
    //----- Don't change beyond this point! -----
    const qcd_coupling& as = _M_alpha_qcd;
    
    double tmax = as.tauk(_M_q2min);
    double dt = tmax/nt, tc = tmax-as.tauk(_M_m2c+1e-12), tb = tmax-as.tauk(_M_m2b+1e-12);
    
    /*    setting up the t grid   */
    std::set<double> tt({0.0, tc, tb, tmax});
    auto ti = tt.find(0.0), tl = tt.find(tmax);
    double t1, t2, dtn;
    unsigned int n;
    
    _M_t.reserve(nt+20u);
    _M_t.resize(0u);
    
    while(ti != tl) {
      t1 = *ti;
      t2 = *(++ti);
      
      n = 1u + static_cast<unsigned>((t2-t1)/dt);
      if(n < 10u) n = 10u;
      dtn = (t2-t1)/n;
      
      for(unsigned k = 0; k < n; k++)
        _M_t.push_back(t1+dtn*k);
    }
    
    _M_t.push_back(tmax);
  }
  

  void qcd_dglap::_M_weights_msbar(unsigned int i, unsigned int k, unsigned int it, qcd_dglap::kernel& res)
  {
    /*   calculate the scale   */
    double q2 = _M_alpha_qcd.tauk_inverse(_M_t.back()-_M_t[it]);
    double sp = _M_x[i]/_M_x[k+1];
    
    //----- set all elements of the kernel to zero -----
    res = kernel(0.0);
    
    if(i < k) {
      double s = _M_x[i]/_M_x[k], sm = _M_x[i]/_M_x[k-1];
      
      //----- gluon-gluon vertex -----
      res[0][0] = _M_s1gg(sp, s) + _M_s2gg(s, sm);
      
      //----- gluon-quark vertices -----
      res[0][1] = res[0][2] = res[0][3] = _M_s1gq(sp, s) + _M_s2gq(s, sm);
      
      //----- quark-gluon vertices -----
      double sqg = _M_s1qg(sp, s) + _M_s2qg(s, sm);
      res[1][0] = 6.0*sqg;
      
      if(q2 > _M_m2c) res[2][0] = 2.0*sqg;
      if(q2 > _M_m2b) res[3][0] = 2.0*sqg;
      
      //----- quark-quark vertices -----
      res[1][1] = res[2][2] = res[3][3] = _M_s1qq(sp, s) + _M_s2qq(s, sm);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- diagonal elements -----
    else if(i == k) {
      //----- gluon-gluon vertex -----
      unsigned int nf = 3U;
      if(q2 > _M_m2c) nf++;
      if(q2 > _M_m2b) nf++;
      
      res[0][0] = _M_s3gg(sp) + Gg(nf);
      
      //----- gluon-quark vertices -----
      res[0][1] = res[0][2] = res[0][3] = _M_s1gq(sp, 1.0);
      
      //----- quark-gluon vertices -----
      double sqg = _M_s1qg(sp, 1.0);
      res[1][0] = 6.0*sqg;
      
      if(q2 > _M_m2c) res[2][0] = 2.0*sqg;
      if(q2 > _M_m2b) res[3][0] = 2.0*sqg;
      
      //----- quark-quark vertices -----
      res[1][1] = res[2][2] = res[3][3] = _M_s3qq(sp);
      
      //----- non-singlet -----
      res[4][0] = res[1][1];
    }
    
    //----- we shouldn't call this function with k < i -----
    else throw "pdf::weights_msbar(...) : We have some trouble...";
  }
  
  
  
  double qcd_dglap::_M_s1qq(double u, double v) {
    double vu = v-u, ov = 1.0-v, ou = 1.0-u;
    return Cf*(2.0*v*ou/vu*std::log(ou/ov)-u*v/vu*std::log(v/u)-0.5*v*(2.0+vu));
  }
  
  double qcd_dglap::_M_s2qq(double u, double v) {
    double vu = v-u, ov = 1.0-v, ou = 1.0-u;
    if(v+1e-15 > 1.0) return Cf*(u*v/vu*std::log(v/u)+0.5*u*(2.0-vu));
    return Cf*(u*v/vu*std::log(v/u)-2.0*u*ov/vu*std::log(ou/ov)+0.5*u*(2.0-vu));
  }
  
  double qcd_dglap::_M_s3qq(double u) {
    double ou = 1.0-u;
    return Cf*(u/ou*std::log(u)+2.0*std::log(ou)+0.5*u);
  }
  
  
  double qcd_dglap::_M_s1qg(double u, double v) {
    double vu = v-u;
    return 0.5*(v*(1.0-vu)+v*vu*(u+2.0*v)/3.0-u*v/(v-u)*std::log(v/u));
  }
  
  double qcd_dglap::_M_s2qg(double u, double v) {
    return 0.5*((u*u*(2.0*u-3.0*(1.0+v))+u*(3.0+6.0*v)+v*(v*v-3.0-3.0*v))/3.0+v*std::log(v/u))*u/(v-u);
  }
  
  
  double qcd_dglap::_M_s1gq(double u, double v) {
    double vu = v-u;
    return Cf*(2.0*v*(1.0+u)/vu*std::log(v/u)-0.5*(4.0+v*(4.0-vu)));
  }
  
  double qcd_dglap::_M_s2gq(double u, double v) {
    double vu = v-u;
    return Cf*(2.0*u*(1.0+v)/vu*std::log(u/v)+0.5*(4.0+u*(4.0+vu)));
  }
  
  
  double qcd_dglap::_M_s1gg(double u, double v) {
    double vu = v-u, ov = 1.0-v, ou = 1.0-u;
    return 2.0*Ca*(v/vu*(ou*std::log(ou/ov)+(1.0+u)*std::log(v/u)) - 1.0 + v*(u*(u-3.0+v)+(3.0-2.0*v)*v-12.0)/6.0);
  }
  
  double qcd_dglap::_M_s2gg(double u, double v) {
    double vu = v-u, ov = 1.0-v, ou = 1.0-u;
    if(v+1e-14 > 1.0) return 2.0*Ca*(u/vu*((1.0+v)*std::log(u/v)) + 1.0 + u*(12.0+ u*(2.0*u-v-3.0)-(v-3.0)*v)/6.0);
    return 2.0*Ca*(u/vu*(ov*std::log(ov/ou)+(1.0+v)*std::log(u/v)) + 1.0 + u*(12.0+ u*(2.0*u-v-3.0)-(v-3.0)*v)/6.0);
  }
  
  double qcd_dglap::_M_s3gg(double u) {
    double ou = 1.0-u;
    return 2.0*Ca*(std::log(ou)-(1.0+u)/ou*std::log(u) + (u*(u-2.0)-17.0)/6.0);
  }
  
}

