/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/13/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

//   std includes
#include <cmath>

//   project headers
#include "bits/qcd-color.h"
#include "bits/mc-splitting-functions.h"


#define __two_pi 6.28318530717958647696
#define __one_pi 3.14159265358979323848



namespace duct {
  
  /*   Spliting functions in the final state 
   */
  namespace __mc_splitting_functions
  {
    
    void emitting_gluon_fin::
    caching(int sl, int spectator, double rl, double ymax, double xi, double mu, double y, double u)
    {
      //------ basic papameters ------
      double rym = rl*rl+(1.0+ymax)*(1.0+ymax)-1.0;
      double al = rym/(4.0*ymax), bl = (1.0-rl*rl)*ymax/rym;
      
      //------ check the type of the spectator -----
      isSpec = (spectator > 0 ? 1 : 2);
 
      double oy = 1.0+y, zm = bl != 0.0 ? bl/(y+bl) : 0.0;
      double lmd = al+bl < 1.00000001 ? rl - y : std::sqrt(oy*oy-4.0*al*(y+bl));
      double x0 = (oy-lmd)/(oy+lmd);
            
      //----- we do lorentz transformation if the spectator is in the final state -----
      if(isSpec == 1) {
        double omega = (2.0*al-1.0-y-lmd)/(2.0*al-1.0-rl);
        double den = xi/omega + (1.0-xi)*omega;
        
        xi /= omega*den;
        mu /= den*den;
      }
      
      //----- basic variables -----
      D = x0 + zm + u*(1.0-zm)*(1.0-x0);
      B = x0/D;
      D *= mu;
      A = 2.0*(xi+(1.0-xi)*x0) + D;
      C = xi*(1.0-u) + (1.0-xi)*u*x0;
      S = std::sqrt(std::abs((2.0*xi*(1.0-xi) - mu)*2.0*u*(1.0-u)*x0));
      
      //----- caching more variables -----
      I1 = (A + 2.0*B)/std::sqrt(std::abs((B+C)*(B+C)-S*S));
      I2 = D > 0.0 ? D/std::sqrt(std::abs(C*C-S*S)) : 0.0;
      R1 = std::sqrt((B+C+S)/(B+C-S));
      R2 = std::sqrt((C+S)/(C-S));
          
      //----- caching the pure collinear contributions -----
      switch(sl) {
        case 1: {
          double orl = 1.0+rl, rg = lmd+y-rl, ozm = 1.0-zm;
          double h = (rg*rg + 2.0*y*orl)/(lmd*orl), gam = ozm*ozm*(oy+lmd)/orl;
          
          //----- the collinear part when the emitter is a fermion -----
          Hcoll = h*(-1.0 + (1.0 + x0 - x0/(1.0-u+x0*u))/(1.0-u+x0*u)) + gam*(1.0-u);
        } break;
        case 2: {
          double Dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          
          //----- the collinear part when the emitter is a massless vector (gluon) -----
          Hcoll = u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
        }	break;
        default: Hcoll = 0.0; break;
      }
      
      //----- overall factors ----
      F = lmd/rl;
    }
    
    void emitting_gluon_fin::caching(int sl, double ymax, double y, double u)
    {
      //------ basic papameters ------
      double bl = 0.5*(1.0-ymax);
      double oy = 1.0+y, zm = bl/(y+bl), lmd = ymax - y;
      double rl = ymax, x0 = (oy-lmd)/(oy+lmd);
            
      //---- In this case xi = 1 and mu = 0 and the lorentz transformation doesn't change them. ----
      isSpec = 0;
      
      A = 2.0;
      B = x0/(x0 + zm + u*(1.0-zm)*(1.0-x0));
      C = 1.0-u;
      D = S = 0.0;
      I1 = (A + 2.0*B)/(B + C) - 2.0;
      I2 = 0.0; R1 = R2 = 1.0;
    
      //----- caching the pure collinear contributions -----
      switch(sl) {
        case 1: {
          double orl = 1.0+rl, rg = lmd+y-rl, ozm = 1.0-zm;
          double h = (rg*rg + 2.0*y*orl)/(lmd*orl), gam = ozm*ozm*(oy+lmd)/orl;
          
          //----- the collinear part when the emitter is a fermion -----
          Hcoll = h*(-1.0 + (1.0 + x0 - x0/(1.0-u+x0*u))/(1.0-u+x0*u)) + gam*(1.0-u);
        } break;
        
        case 2: {
          double Dlt = 2.0*x0/((x0+u*(1.0-x0))*(1.0-u*(1.0-x0)));
          
          //----- the collinear part when the emitter is a massless vector (gluon) -----
          Hcoll = u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
        } break;
      
        default: Hcoll = 0.0; break;
      }
      
      //----- overall factors ----
      F = lmd/rl;
    }
    
    
    double emitting_gluon_fin::function_phi(double phi) const
    {
      if(isSpec == 0) return F*(I1 + Hcoll);
      
      double cphi = std::cos(__two_pi*phi);
      double ret = (A + 2.0*B)/(B+C - S*cphi) - 2.0;
      if(D > 0.0) ret -= D/(C - S*cphi);
      
      return F*(ret + Hcoll);
    }
    
    double emitting_gluon_fin::primitive_phi(double phi) const
    {
      if(isSpec == 0) return F*(I1 + Hcoll)*phi;
      
      //----- if phi is 2pi then we have simpler expression -----
      phi *= __two_pi;
      
      if(std::abs(1.0-phi/__two_pi) < 1e-13 || phi >= __two_pi)
        return F*((I1 - I2 - 2.0) + Hcoll);
      
      //----- phi is smaller than 2pi -----
      double ctg = std::tan(0.5*phi), one = phi <= __one_pi ? 0.0 : __one_pi;
      double ret = I1*(one + std::atan(R1*ctg)) - phi;
      if(D > 1e-13) ret -= I2*(one + std::atan(R2*ctg));
      
      return F*(ret + 0.5*Hcoll*phi)/__one_pi;
    }
    
    
    double emitting_gluon_dcy(int sl, double rl, double ymax, double y, double u)
    {
      //------ basic papameters ------
      double rym = rl*rl+(1.0+ymax)*(1.0+ymax)-1.0;
      double al = rym/(4.0*ymax), bl = (1.0-rl*rl)*ymax/rym, oy = 1.0+y;
      
      double lmd = al+bl <= 1.0 ? rl - y : std::sqrt(oy*oy-4.0*al*(y+bl));
      double x0 = (oy-lmd)/(oy+lmd);
      
      //----- soft-collinear term -----
      double u1 = 1.0-u+x0*u;
      double Fc = 0.0, Fsc = 2.0*((1.0+x0)/u1 - 1.0 - x0/(u1*u1));
      
      //----- pure collinear contributions -----
      switch(sl) {
        case 1: {
          double oy = 1.0+y, ozm = bl != 0.0 ? y/(y+bl) : 1.0;
          double orl = 1.0+rl, rg = lmd+y-rl;
          double H2 = (rg*rg + 2.0*y*orl)/(lmd*orl);
          double H3 = 0.5*ozm*ozm*(oy+lmd)/orl;
          
          Fc = H2*(-1.0 + (1.0 + x0 - x0/u1)/u1) + H3*2.0*(1.0-u);
        } break;
          
        case 2: {
          double Dlt = 2.0*x0/(u1*(x0+u*(1.0-x0)));
          Fc = u*(1.0-u)*(1.0 - 0.5*Dlt*(2.0-Dlt));
        } break;
      }
      
      return lmd/rl*(Fsc + Fc);
    }
   
    
    double gluon_splitting_fin(double ymax, double b, double y, double u)
    {
      if(y > 4.0*b) {
        double zpm = (b > 0.0 ? std::sqrt(1.0-4.0*b/y) : 1.0);
        
        //----- splitting function of g --> f + f~ -----
        double lmd = std::sqrt((1.0-y/ymax)*(1.0-ymax*y));
        double weight = lmd*(1.0 - zpm*zpm*2.0*u*(1.0-u))*zpm;
        
        /* Since in the emiiting_gluon_fin there is no color factor
         * we drop it also here. At the end it has to be multiplied
         * by the gluon charge, Ca. Thus the net color factors is Tr/Ca.
         */
        weight *= 0.5/Ca;
      
        return weight;
      }
      
      return 0.0;
    }
  
  }
  
  
  /*   Spliting functions in the initial state
   */
  namespace __mc_splitting_functions {
    
    
    void emitting_gluon_ini::
    caching(int sl, int spectator, double ma, double mb, double xi, double xT, double y, double z)
    {
      //----- xa and xb coordinates of the emitted patons -----
      double xn = 1.0-ma*mb*z*z, saQ = 1.0/z+2.0*ma+ma*mb*z;
      double xa = (1.0/z-1.0 - y*(1.0+mb*z) - (1.0-z)*ma*mb)/xn;
      double xb = z*(y*(1.0+ma*z) - ma*(1.0-z)*(1.0-ma*mb*z))/xn;
      
      //----- special case when the spectator is incoming parton  -----
      if(spectator < 1) {
        this->isSpec = true;
        this->Cp = this->Cm = xa+mb*xb;
        this->Bp = this->Bm = 1.0/z+ma*mb*z;
        this->Ap = this->Am = Cp*saQ + y*(1.0/z+2.0*mb+ma*mb*z);        
        
        this->w1 = 2.0*saQ;
        this->w2 = 2.0*mb*saQ*y;
        this->w3 = 2.0*ma*saQ/y;
      }
      //----- when the spectator is an outgoing parton then we have to work harder -----
      else {
        //----- boost parameters -----
        double kT = std::sqrt(xa*xb);
        double omega = (1.0/z+mb-xa)/(1.0+mb), vT = kT/(1.0+mb);
        
        //----- coordinates of the spectator after the boost ----
        constexpr double sqrt2 = 1.4142135623730950488;
        double sxi = std::sqrt(2.0*xi*(1.0-xi));
        double alpha = omega*xi;
        double betap = (1.0-xi + (xi*vT+sxi*sqrt2*xT)*vT)/omega;
        double betam = (1.0-xi + (xi*vT-sxi*sqrt2*xT)*vT)/omega;
                
        //----- cached variables -----
        this->isSpec = false;
        this->Cp = alpha*xb + betap*xa + (sxi*xT + sqrt2*xi*vT)*sqrt2*kT;
        this->Cm = alpha*xb + betam*xa - (sxi*xT - sqrt2*xi*vT)*sqrt2*kT;
        
        this->Bp = betap/z + ma*z*alpha;
        this->Bm = betam/z + ma*z*alpha;
        
        this->Ap = Cp*saQ + y*((1.0+z*ma)*alpha+(1.0/z+mb)*betap);
        this->Am = Cm*saQ + y*((1.0+z*ma)*alpha+(1.0/z+mb)*betam);
        
        this->w1 = 2.0*saQ;
        this->w2 = sxi*sxi*(1.0-xT*xT)*saQ*y;
        this->w3 = 2.0*ma*saQ/y;
        
        this->rAp = std::sqrt(Ap);
        this->rAm = std::sqrt(Am);
        this->rCp = std::sqrt(Cp);
        this->rCm = std::sqrt(Cm);
        
        double omb = 1.0+mb, ysaQ = saQ+y, ombz = 1.0+mb*z;
        
        this->R2 = 1.0/(ysaQ*ombz);
        this->R1 = xi*(ma*ombz-y)/omb + R2*(saQ*omb*(1.0-xi) - xi*y*(1.0+ma));
        this->R5 = -(omb + (ma-mb)*xi)*y/ysaQ;
        this->R3 = 1.0/R5;
        this->R4 = R3/ysaQ;
        this->R6 = 1.0/ysaQ;
      }
      
      //----- pure collinear part -----
      switch(sl) {
          //----- quark emitter -----
        case 1: {
          this->Hcoll = xa;
          
          if(ma*mb > 1e-13) {
            double oz = 1.0-z;
            double den = oz*(1.0-z*ma*mb);
            double Feik = 2.0*((1.0+z*ma + z*(ma-y*(1.0+z*ma)/den)*(1.0+z*mb))/den - ma/y);
            this->Hcoll += oz*oz*ma*mb*Feik/xn;
          }
        } break;
          
          //----- gluon emitter -----
        case 2: {
          double za = xa*z;
          double oz = 1.0-z, oza = 1.0-za;
          this->Hcoll = 2.0*za*(za*za/oza + (oz+za)/oz)/oza;
        } break;
          
        default: this->Hcoll = 0.0; break;
      }
    
      //----- overall factors ----
      double fac = (1.0-ma*mb)/xn;
            
      this->w1 *= fac;
      this->w2 *= fac;
      this->w3 *= fac;
      this->Hcoll *= fac;
    }
    
    double emitting_gluon_ini::function_phi(double phi) const
    {
      //----- in the special case it is simple -----
      if(isSpec) return w1*Bp/Ap - w2/(Ap*Cp) - w3*Cp/Ap + Hcoll;
      
      //----- general case -----
      double c = std::cos(__two_pi*phi);
      double gp = 0.5*(1.0+c), gm = 0.5*(1.0-c);
      double retval = w1*(Bp*gp+Bm*gm)/(Ap*gp+Am*gm);
      
      if(w2 > 1e-13) retval -= w2/(Ap*gp+Am*gm)/(Cp*gp+Cm*gm);
      if(w3 > 1e-13) retval -= w3*(Cp*gp+Cm*gm)/(Ap*gp+Am*gm);
      
      return retval + Hcoll;
    }
    
    double emitting_gluon_ini::primitive_phi(double phi) const
    {
      //----- with initial state spectator it is simple ----
      if(isSpec) return (w1*Bp/Ap - w2/(Ap*Cp) - w3*Cp/Ap + Hcoll)*phi;
      
      //----- the full integral -----
      if(phi == 1.0) {
        double retval = w1*(R1/(rAp*rAm) + R2) + Hcoll;
       
        if(w2 > 1e-13) retval -= w2*(R3/(rAp*rAm) - R4/(rCp*rCm));
        if(w3 > 1e-13) retval -= w3*(R5/(rAp*rAm) + R6);
        
        return retval;
      }
      
      //----- the primitive function in the general case -----
      phi *= __two_pi;
      
      double t = std::tan(0.5*phi), one = (phi <= __one_pi ? 0.0 : __one_pi);
      double iA = (one + std::atan(rAm*t/rAp))/(rAp*rAm);
      double retval = w1*(R1*iA + R2*0.5*phi) + 0.5*Hcoll*phi;
      
      if(w2 > 1e-13) {
        double iC = (one + std::atan(rCm*t/rCp))/(rCp*rCm);
        retval -= w2*(R3*iA - R4*iC);
      }
      
      if(w3 > 1e-13) retval -= w3*(R5*iA + R6*0.5*phi);
      
      return retval/__one_pi;
    }
    

    double splitting_gluon_quark_ini(double ma, double mb, double y, double z)
    {
      //----- coordinates -----
      double xn = 1.0, xa, xb;
      
      if(ma < 1e-14) {
        xa = 1.0/z-1.0 - y*(1.0+mb*z);
        xb = y*z;
      } else {
        xn = 1.0-ma*mb*z*z;
        xa = (1.0/z-1.0 - y*(1.0+mb*z) - ma*mb*z)/xn;
        xb = z*(y + ma + ma*z*(1.0+y-ma*mb*z))/xn;
      }
      
      //----- splitting function -----
      double w = 1.0-xa*z;
      double retval = (1.0 + 2.0*(xa*xb-ma)/(y*w*w))/xn;
      
      /*  The color factor is Tr and we have a factor 
       *  from the color averaging in the initial state, Na/Nc
       */
      retval *= 0.5*Na/Nc;
    
      /* Since in the emiiting_gluon_ini there is no color factor
       * we drop it also here. At the end it has to be multiplied
       * by the gluon charge, Ca.
       */
      retval /= Ca;
      
      return retval;
    }
    

    double splitting_quark_antiquark_ini(double ma, double mb, double y, double z)
    {
      double zxa = 1.0-z*(1.0+y*(1.0+mb*z)+ma*mb);
      double retval = (1.0-ma*mb)*(1.0-2.0*(1.0-zxa)*(zxa-ma/y))/z;

      /*  The color factor is Cf and the we have a factor
       *  from the color averaging in the initial state, Nc/Na
       *  Since in the emiiting_gluon_ini there is no color factor
       *  we drop it also here. At the end it has to be multiplied
       *  by the gluon charge, Cf.
       */
      retval *= Nc/Na;
      
      return retval;
    }
    
    
    
    
  }
  
  
}

