/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 5/3/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */

//   std includes
#include <cmath>
#include <limits>
#include <complex>

//   project headers
#include "proc-dyjets.h"
#include "bits/qcd-color.h"


namespace duct {
  
  //----- FIXME: these constats should be gotten from the model description file -----
  constexpr double Mz = 91.182, Gz = 2.4952, s2w = 0.231;
  constexpr double swcw = 0.84294484;   // 2.0*std::sqrt(s2w*(1.0-s2w));
  constexpr double M2z = Mz*Mz;
  
  //----- electroweak couplings ------
  constexpr double Ql = -1.0, Vl = (-0.5 + 2.0*s2w)/swcw, Al = -0.5/swcw;
  constexpr double Qu = 2.0/3.0, Vu = (0.5 - 4.0*s2w/3.0)/swcw, Au = 0.5/swcw;
  constexpr double Qd = -1.0/3.0, Vd = (-0.5 + 2.0*s2w/3.0)/swcw, Ad = -0.5/swcw;

  
  
  dyjets::dyjets(const model *mdl, double Ecm, double mL, double mH, int hdr1, int hdr2,
                 std::initializer_list<int> lpts, std::initializer_list<int> partons)
  : _M_mdl(mdl), _M_Ecm(Ecm), _M_lpts(lpts), _M_partons(partons)
  {
    /*   incoming hadrons   */
    incoming_hadrons<hhc>::hadronA = hdr1;
    incoming_hadrons<hhc>::hadronB = hdr2;
    
    /*    the pdf functions   */
    _M_pdf1 = _M_mdl->pdf(hadron(hdr1));
    _M_pdf2 = _M_mdl->pdf(hadron(hdr2));

    /*    tau limits   */
    double s = _M_Ecm*_M_Ecm;
    double tauL = mL*mL/s, tauH = mH*mH/s;
    
    /*   coupling part of the normalization    */
    double Ng = (Qu*Qu+Qd*Qd)*Ql*Ql;
    double Nz = (std::pow((Vu+Vl)*(Au+Al), 2)+std::pow((Vd+Vl)*(Ad+Al), 2))*std::log(s/M2z);
   
    /*   the full normalization   */
    double pL = -Mz/Gz*std::atan((M2z-s*tauL)/(Mz*Gz));
    double pH = -Mz/Gz*std::atan((M2z-s*tauH)/(Mz*Gz));
    double lL = std::pow(std::log(tauL),2);
    double lH = std::pow(std::log(tauH),2);
    
    double Ngz = 0.5*(lL-lH)*Ng + (pH-pL)*Nz;
    double pg = 0.5*(lL-lH)*Ng/Ngz;
    
    /*   the tau probability density    */
    this->_M_rhotau = [=](double tau) -> double {
      double pp = s*tau - M2z;
      return (Nz*s*M2z/(pp*pp + M2z*Gz*Gz) - Ng*std::log(tau)/tau)/Ngz;
    };
    
    /*   generating tau   */
    this->_M_generate_tau = [=](double rho, double chi) -> double {
      if(rho < pg) return std::exp(-std::sqrt(chi*lH + (1.0-chi)*lL));
      return Mz/s*(Mz - Gz*std::tan(-Gz/Mz*(pL+chi*(pH-pL))));
    };
  }

  
  double dyjets::_M_select_flavors(double tau, double eta, double x, double rho)
  {
    double norm = 0.0, __fa[11], __fb[11];
    double *fa = __fa+5, *fb = __fb+5, m2F = _M_Ecm*_M_Ecm*tau;
    
    /*   precalculating the pdfs  */
    for(int i : _M_partons) {
      fa[ i] = _M_pdf1->parton( i, eta, m2F);
      fb[-i] = _M_pdf2->parton(-i, tau/eta, m2F);
    }

    /*   calculating the matrix element square with the pdfs  */
    unsigned ip = 0u;
    struct { int fq, fl; double w;} proc[50];
    
    for(int fl : _M_lpts) {
      /*   processes q + qbar -> l + lbar and
       *   processes qbar + q -> l + lbar   
       */
      for(int iq : _M_partons)
        if(iq != 0 && -6 < iq && iq < 6) {
          proc[ip].fq = -iq;
          proc[ip].fl =  fl;
          norm += (proc[ip++].w = fa[iq]*fb[-iq]*_M_process_qqll(iq > 0 ? x : 1.0-x, tau, iq, fl));
        }
    }
    
    //----- selecting subprocess ----
    double cproc = proc[0].w;
    
    ip = 0U;
    while(cproc < rho*norm)
      cproc += proc[++ip].w;
 
    /*   Assigning the flavours   */
    this->state[-3].flavor = -proc[ip].fl;
    this->state[-2].flavor =  proc[ip].fl;
    this->state[-1].flavor =  proc[ip].fq;
    this->state[ 0].flavor = -proc[ip].fq;

    return norm/tau;   //  We have a 1/tau because the pdfs are the eta*f(eta,mu2) functions.
  }

  
  double dyjets::_M_process_qqll(double x, double tau, int fq, int fl) const
  {
    /*    masses invariants    */
    double s = _M_Ecm*_M_Ecm, m2q = _M_mdl->mass2(fq), m2l = _M_mdl->mass2(fl);
    double nq = m2q/s;
    double shat = (tau+nq)*(1.0+nq/tau)*s, sab = s*(tau*tau+nq*nq)/tau;
    double sa1 = (x*tau + (1.0-x)*nq)*shat/(tau+nq);
    
    /*  trivial cuts  */
    if(x*(1.0-x) <= m2l/shat || tau < nq) return 0.0;
      
    /*    Z propagator factor   */
    std::complex<double> chi = shat/std::complex<double>(shat-M2z, Mz*Gz);
    std::complex<double> A, B, C, D;
    
    /*    calulate EW coupling factor   */
    switch(std::abs(fq)) {
      //----- down-like quarks -----
      case 1: case 3: case 5: {
        A = Ql*Qd + Vd*Vl*chi;
        B = Vd*Al*chi;
        C = Ad*Vl*chi;
        D = Ad*Al*chi;
      } break;
        
      //------ up-like quarks -----
      case 2: case 4: case 6: {
        A = Ql*Qu + Vu*Vl*chi;
        B = Vu*Al*chi;
        C = Au*Vl*chi;
        D = Au*Al*chi;
      } break;
    }
    
    /*   derived dot products   */
    double s12 = shat-2.0*m2l, sa2 = shat-sa1;
    double X1 = sa1*sa1+sa2*sa2, X2 = 2.0*m2q*s12, X3 = 2.0*m2l*sab, X4 = 8.0*m2q*m2l;
    
    return (std::norm(A)*(X1 + X2 + X3 + X4) + std::norm(B)*(X1 + X2 - X3 - X4)
            + std::norm(C)*(X1 - X2 + X3 - X4) + std::norm(D)*(X1 - X2 - X3 + X4)
            + 2.0*std::real(A*std::conj(D)+B*std::conj(C))*(sa2*sa2 - sa1*sa1))/(shat*shat);
  }
  
  
  shower_stage<hhc>& dyjets::operator()(random_engine_type& rne)
  {
    std::uniform_real_distribution<double> rng(0.0,1.0);
  
    /*   resizing the state, setting the model  */
    this->state.resize(2,0);
    this->state.set_model(_M_mdl);

    /*    incoming hadrons   */
    this->state[hadron(-1)].momentum = {0.0,0.0, 0.5*_M_Ecm, 0.5*_M_Ecm};
    this->state[hadron( 0)].momentum = {0.0,0.0,-0.5*_M_Ecm, 0.5*_M_Ecm};
    this->state[hadron(-1)].key = incoming_hadrons<hhc>::hadronA;
    this->state[hadron( 0)].key = incoming_hadrons<hhc>::hadronB;

    /*   selecting tau, eta and x   */
    double tau = _M_generate_tau(rng(rne), rng(rne));
    double eta = tau*std::pow(1.0/tau, rng(rne));
    double etb = tau/eta, x = rng(rne);
    
    /*   selecting the flavor   */
    this->weight = _M_select_flavors(tau, eta, x, rng(rne));
    this->weight /= _M_rhotau(tau);
    this->weight *= -std::log(tau)/tau;
    
    /*   For the outgoing momenta we use Sudakov paremetrization.   */
    double s = _M_Ecm*_M_Ecm;
    double nq = _M_mdl->mass2(this->state[0].flavor)/s;
    double m2l = _M_mdl->mass2(this->state[-2].flavor);
    double shat = s*(tau+nq)*(tau+nq)/tau;
    double mT = std::sqrt(x*(1.0-x)*shat - m2l);
    
    //----- + and - momenta ------
    lorentzvector<double> na = this->state[hadron(-1)].momentum*((tau+nq)/etb);
    lorentzvector<double> nb = this->state[hadron( 0)].momentum*((tau+nq)/eta);
    
    //----- transverse momentum ------
    double phi = 2.0*M_PI*rng(rne);
    lorentzvector<double> kT(mT*std::cos(phi), mT*std::sin(phi), 0.0, 0.0);
    
    this->state[-3].momentum =      x *na + (1.0-x)*nb - kT;
    this->state[-2].momentum = (1.0-x)*na +      x *nb + kT;
    this->state[-1].momentum = (tau*na + nq*nb)/(tau+nq); 
    this->state[ 0].momentum = (tau*nb + nq*na)/(tau+nq);

    //----- heath and spinors -----
    for(int i = -3; i <= 0; ++i) {
      this->state[i].health = '-';
      this->state[i].plus = diracspinor<double,ket>(this->state[i].momentum, 1);
      this->state[i].minus = diracspinor<double,ket>(this->state[i].momentum, -1);
    }
    
    //---- color connections -----
    typedef statistical_basis::particle::colorlink colorlink;
    this->state[-3].color = this->state[-3].colorP = colorlink::dead();
    this->state[-2].color = this->state[-2].colorP = colorlink::dead();
    this->state[-1].color = this->state[-1].colorP = this->state[-1].flavor > 0 ? colorlink::quark( 0) : colorlink::antiquark( 0);
    this->state[ 0].color = this->state[ 0].colorP = this->state[ 0].flavor > 0 ? colorlink::quark(-1) : colorlink::antiquark(-1);
    
    //----- color suppression of the state -----
    this->state.color_suppression_index = 0U;

    /*   We always start with SU(Nc) states.   */
    this->is_color_u3 = false;

    //----- shower time ----
    this->time = 0.0;
    
    //---- reference momentum to define shower time -----
    this->qnull = this->state[-1].momentum + this->state[0].momentum;
    
    /*   maximum allowed tranverse momentum in the initial and final state splittings    */
    this->max_k2T_initial_state = std::numeric_limits<double>::max();
    this->max_k2T_final_state = std::numeric_limits<double>::max();
    
    /*  This is maximum allowed virtuality in the initial and final state splitting.   */
    this->max_q2_final_state = this->max_q2_initial_state = std::numeric_limits<double>::max();
   
    /*   alpha_qed coupling   */
    //double aqed = _M_mdl->alpha_qed(shat);
    
    double alpha_MZ = 1.0/127.944;  // From PDG Chapter 10, page 4
    double betamod = 20.0/9.0/M_PI;   // 1/4/pi * (1 + 1 + 1 + 12/9 + 12/9 + 1/3 + 1/3 + 1/3)
    double aqed = alpha_MZ/(1.0 - alpha_MZ*betamod*std::log(shat/M2z));

    //----- adjusting the weight -----
    this->weight *= 2.0*M_PI*aqed*aqed/(s*Nc);

    return *this;
  }
  
}


