/*  Copyright (C) 2012-2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/23/12.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __proc_hccjets_h__
#define __proc_hccjets_h__ 1

//   std includes

//   project headers
#include <bits/mc-emission.h>
#include <bits/qcd-coupling.h>


namespace duct {
  
  class hhcjets : public hard_state<hhc>
  {
  public:
    //----- constructor -----
    hhcjets(const model *, double, std::initializer_list<double>, int = 2212, int = 2212);
    
    /*   function to clone the objects   */
    hard_state<hhc> * clone() const {
      return new hhcjets(*this);
    }
    
    /*   generates the hard event   */
    shower_stage<hhc>& operator()(random_engine_type&);
    
    /*   set the renormalization and hadronization scale factor   */
    void scales(double xR, double xF) { _M_xR = xR; _M_xF = xF; }
    
  private:
    /*   pointer to the model  */
    const model *_M_mdl;
    
    /*   ponter to the pdfs   */
    const qcd_hadron *_M_pdf1, *_M_pdf2;
    
    //----- genreate the event from an array of random numbers ------
    double _M_generate_event(random_engine_type&);
    
    //----- select the flavor configuration ------
    double _M_select_flavors(double, double, double, double);
    
    //----- private members ------
    double _M_s, _M_xR, _M_xF;
    
    /*   p2T ranges  */
    std::vector<double> _M_p2T;
    mutable double _M_p2Tmin, _M_p2Tmax;
    
    //----- structure for represnting the subprocesses -----
    struct subprocess {
      int fl[4];
      double (hhcjets::*weight_function)(double, double, const int *) const;
      double (hhcjets::*select_color)(double);
    };
    
    //------ table of the possible subprocesses -----
    static const subprocess _M_proc_tbl[250];
    
    //----- cached variables to communicate between the private member functions -----
    mutable double _M_nu[7], _M_fa[11], _M_fb[11];
    mutable const subprocess *_M_active;

    //----- tables of color supression indices ----
    static const unsigned int _M_color_suppr_idx_4g[6][6];
    static const unsigned int _M_color_suppr_idx_2q2g[2][2];
    static const unsigned int _M_color_suppr_idx_4q[2][2];
    static const unsigned int _M_color_suppr_idx_4qC[2][2];
    
    /*  Private functions help to compute the weight 
     *  of the subprocesses and select color states.
     */
    
    //----- transverse momentum  -----
    double _M_transverse_momentum(double tau, double x, const int *fl) const {
      return _M_s*(tau + _M_mdl->mass2(fl[0])/_M_s)*(tau + _M_mdl->mass2(fl[1])/_M_s)/tau*x*(1.0-x)
      - _M_mdl->mass2(fl[3])*x - _M_mdl->mass2(fl[2])*(1.0-x);
    }

    /*   generating the x variable   */
    void _M_generate_x(double&, double&, random_engine_type&, double, double) const;
    
    //----- weight functions of the possible subprocesses -----
    double _M_process_gggg(double, double, const int *) const;
    double _M_process_ggqQ(double, double, const int *) const;
    double _M_process_gqgq(double, double, const int *) const;
    double _M_process_qggq(double, double, const int *) const;
    double _M_process_qQgg(double, double, const int *) const;
    double _M_process_qQrR(double, double, const int *) const;
    double _M_process_qrqr(double, double, const int *) const;
    double _M_processA_qQqQ(double, double, const int *) const;
    double _M_processB_qQqQ(double, double, const int *) const;
    double _M_processC_qQqQ(double, double, const int *) const;
    double _M_processA_qqqq(double, double, const int *) const;
    double _M_processB_qqqq(double, double, const int *) const;
    double _M_processC_qqqq(double, double, const int *) const;
    
    //----- functions for selecting color configurations -----
    double _M_select_color_gggg(double);
    double _M_select_color_ggqQ(double);
    double _M_select_color_gqgq(double);
    double _M_select_color_gQgQ(double);
    double _M_select_color_qggq(double);
    double _M_select_color_QggQ(double);
    double _M_select_color_qQgg(double);
    double _M_select_color_Qqgg(double);
    double _M_select_color_qQrR(double);
    double _M_select_color_QqRr(double);
    double _M_select_color_qrqr(double);
    double _M_select_color_QRQR(double);
    double _M_select_color_qRqR(double);
    double _M_select_color_QrQr(double);
    
    double _M_select_colorA_qQqQ(double);
    double _M_select_colorB_qQqQ(double);
    double _M_select_colorC_qQqQ(double);
    double _M_select_colorA_QqQq(double);
    double _M_select_colorB_QqQq(double);
    double _M_select_colorC_QqQq(double);
    double _M_select_colorA_qqqq(double);
    double _M_select_colorB_qqqq(double);
    double _M_select_colorC_qqqq(double);
    double _M_select_colorA_QQQQ(double);
    double _M_select_colorB_QQQQ(double);
    double _M_select_colorC_QQQQ(double);
  };
  
}
#endif
