/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


//   std includes
#include <cstdlib>
#include <cstring>
#include <string>
#include <iostream>
#include <getopt.h>
#include <fstream>
#include <list>
#include <unistd.h>

//   project headers
#include "bits/nlo-c-support.h"
#include "main-mod.h"


//----- print out the help -----
void main_mod_help()
{
  std::cout
  <<"Usage: deductor --module my-module --cflags=\"opt1 opt2 ...\"\n"
  <<"                                   --cxxflags=\"opt1 opt2 ...\"\n"
  <<"                                   --ldflags=\"opt1 opt2 ...\"\n"
  <<"                                   source1 source2 ...\n"
  <<"                                   -Lpath1 -llibrary1 -Lpath2 -llibrary2 ...\n"
  
  <<"\nOptioins of the --module mode:\n"
  <<"  my-module                        The name of the module. It will create the\n"
  <<"                                   loadable module file my-module.so.\n"
  <<"  --cflags=\"opt1 opt2 ...\"         C compiler options\n"
  <<"  --cxxflags=\"opt1 opt2 ...\"       C++ compiler options\n"
  <<"  --ldflags=\"opt1 opt2 ...\"        linker options\n"
  <<"  source1 source2 ...              C or C++ source files. Recognized extensions\n"
  <<"                                   are .c, .cc, .c++, .cxx, .cpp, .C\n"
  <<"  -Lpath                           library search path\n"
  <<"  -llibrary                        link the liblibrary.so (e.g.: -lfastjet)\n"
  <<"--------------------------------------------------------------------------------\n"
  <<std::endl;
}


inline bool check_ext(const char *str, const char *ext) {
  return std::strcmp(str+std::strlen(str)-std::strlen(ext), ext) ? false : true;
}



int main_mod(int argc, char **argv)
{
  //----- long options -----
  option longopt[] = {
    {"cflags",   required_argument, nullptr, 0},
    {"cxxflags", required_argument, nullptr, 0},
    {"ldflags",  required_argument, nullptr, 0},
     {nullptr, 0, nullptr, 0}
  };
  
  //------ default options -----
  const char *module = argv[optind-1];
  std::string cflags = "", cxxflags = "", ldflags = "";
  std::list<std::string> libs, lpath;
  
  
  //------ argument parsing -----
  while(true) {
    int optidx = 0;
    int c = getopt_long(argc, argv, "L:l:", longopt, &optidx);
    if(c == -1) break;
    
    switch(c) {
        case 'L': lpath.push_back(optarg); break;
        case 'l': libs.push_back(optarg); break;
        case 0:
        switch(optidx) {
          case 0: cflags += optarg; cflags += " "; break;
          case 1: cxxflags += optarg; cxxflags += " "; break;
          case 2: ldflags += optarg; ldflags += " ";  break;
        }
        break;
        case '?': break;
      default: throw; break;
    }
  }
  
  /*   making the list of sources and object files   */
  std::list<std::string> sources;
  
  for(int k = optind; k < argc; k++)
    if(check_ext(argv[k], ".c")   || check_ext(argv[k], ".cc")  ||
       check_ext(argv[k], ".cpp") || check_ext(argv[k], ".cxx") ||
       check_ext(argv[k], ".c++") || check_ext(argv[k], ".C"))
      sources.push_back(argv[k]);
  
  
  /*   adding deductor libraries   */
  libs.push_back("deductor");
  libs.push_back("deductor-proc");
  
  /*   creating the module build directory  */
  std::string moddir = std::string(module) + ".bundle";
  
  if(::__nlo_helper_mkdir(moddir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP| S_IROTH | S_IXOTH)) {
    std::cerr<<"Unable to make directory : "<<moddir.c_str()<<std::endl;
    throw "Unable to make the output directory";
  }
  
  /*   writing out the makefile    */
  std::string fname = moddir+"/CMakeLists.txt";
  std::fstream makefile(fname.c_str(), std::ios::out);
  
  makefile<<"cmake_minimum_required(VERSION 2.6)\n"
  <<"set(CMAKE_C_COMPILER "<<"\""<<__CC__<<"\")\n"
  <<"set(CMAKE_CXX_COMPILER "<<"\""<<__CXX__<<"\")\n"
  <<"project(\"module\")\n\n"
  
  <<"set(CMAKE_INSTALL_PREFIX ..)\n"
  <<"set(CMAKE_C_FLAGS \"${CMAKE_C_FLAGS} "<<__CFLAGS__<<" "<<cflags<<"\")\n"
  <<"set(CMAKE_CXX_FLAGS \"${CMAKE_CXX_FLAGS} "<<__CXXFLAGS__<<" "<<cxxflags<<"\")\n"
  <<"set(CMAKE_MODULE_LINKER_FLAGS \"${CMAKE_MODULE_LINKER_FLAGS} "<<ldflags<<"\")\n\n"
  
  <<"include_directories(\""<<__INCLUDEDIR__<<"\")\n"
  <<"link_directories(\""<<__LIBDIR__<<"\" ";
  for(const auto& path : lpath) makefile<<"\""<<path<<"\" ";
  makefile<<")\n\n";
  
  makefile<<"add_library(module MODULE\n";
  for(const auto& src : sources)
    makefile<<(src[0] == '/' ? "" : "../")<<src<<"\n";
  makefile<<")\n\n";
  
  makefile<<"target_link_libraries(module\n";
  for(const auto& lib : libs) makefile<<lib<<"\n";
  makefile<<")\n\n";
  
  makefile<<"set_target_properties(module PROPERTIES PREFIX \"\")\n"
  <<std::endl;
  
  makefile.flush();
  makefile.close();
  
  
  /*   change into the bundle   */
  ::chdir(moddir.c_str());
  
  /*   rum cmake and make   */
  std::string cmd = __CMAKE__;
  cmd += " -G \"Unix Makefiles\" .; ";
  cmd += __MAKE__;
  cmd += " all";
  
  std::system(cmd.c_str());
  
  
  return 0;
}
